Package: futurize
Version: 1.0.0
Title: Parallelize Common Functions via One Magic Function
Description: The futurize() function turns sequential map-reduce functions such as base::lapply(), purrr::map(), 'foreach::foreach() %do% { ... }' into concurrent alternatives, providing you with a simple, straightforward path to scalable parallel computing via the 'future' ecosystem <doi:10.32614/RJ-2021-048>. By combining this transpiler function with R's native pipe operator, you have a convenient way for speeding up iterative computations with minimal refactoring, e.g. 'lapply(xs, fcn) |> futurize()', 'purrr::map(xs, fcn) |> futurize()', and 'foreach::foreach(x = xs) %do% { fcn(x) } |> futurize()'. Other map-reduce packages that can be "futurized" are 'BiocParallel', 'plyr', 'crossmap', 'pbapply' packages. There is also support for a growing set of domain-specific packages on CRAN (e.g. 'boot', 'caret', 'DiceKriging', 'ez', 'fgsea', 'fwb', 'gamlss', 'glmmTMB', 'glmnet', 'kernelshap', 'lme4', 'metafor', 'mgcv', 'modelsummary', 'parameters', 'partykit', 'pls', 'pvclust', 'riskRegression', 'rugarch', 'sandwich', 'seriation', 'shapr', 'Sim.DiffProc', 'SimDesign', 'stars', 'strucchange', 'SuperLearner', 'tm', 'TSP', and 'vegan') and on Bioconductor (e.g. 'DESeq2', 'GenomicAlignments', 'GSVA', 'Rsamtools', 'scater', 'scuttle', 'SingleCellExperiment', and 'sva').
Authors@R: c(
   person("Henrik", "Bengtsson", role=c("aut", "cre", "cph"),
          email = "henrikb@braju.com",
          comment = c(ORCID = "0000-0002-7579-5165")))
License: Apache License (>= 2)
URL: https://futurize.futureverse.org,
        https://github.com/futureverse/futurize
BugReports: https://github.com/futureverse/futurize/issues
Depends: R (>= 4.1.0), future (>= 1.69.0)
Imports: utils
Suggests: methods, future.apply (>= 1.20.2), foreach, doFuture (>=
        1.2.1), purrr, furrr, crossmap, plyr, pbapply, BiocParallel,
        boot, survival, caret, randomForest, DESeq2, DiceKriging, ez,
        fgsea, GenomicAlignments, Rsamtools, fwb, gamlss, glmmTMB,
        glmnet, GSVA, kernelshap, lme4, metafor, mgcv, modelsummary,
        parameters (>= 0.29.1), partykit, pls, pvclust, riskRegression,
        rugarch, sandwich, scater, scuttle, seriation, shapr,
        Sim.DiffProc, SimDesign, SingleCellExperiment, stars,
        strucchange, SuperLearner, sva, tm, vegan, tools, commonmark,
        base64enc
VignetteBuilder: futurize
Language: en-US
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-12 14:06:47 UTC; hb
Author: Henrik Bengtsson [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-7579-5165>)
Maintainer: Henrik Bengtsson <henrikb@braju.com>
Repository: CRAN
Date/Publication: 2026-06-12 15:00:02 UTC
Built: R 4.5.2; ; 2026-06-12 15:31:20 UTC; unix

Package: futurize
Version: 1.0.0
Title: Parallelize Common Functions via One Magic Function
Description: The futurize() function turns sequential map-reduce functions such as base::lapply(), purrr::map(), 'foreach::foreach() %do% { ... }' into concurrent alternatives, providing you with a simple, straightforward path to scalable parallel computing via the 'future' ecosystem <doi:10.32614/RJ-2021-048>. By combining this transpiler function with R's native pipe operator, you have a convenient way for speeding up iterative computations with minimal refactoring, e.g. 'lapply(xs, fcn) |> futurize()', 'purrr::map(xs, fcn) |> futurize()', and 'foreach::foreach(x = xs) %do% { fcn(x) } |> futurize()'. Other map-reduce packages that can be "futurized" are 'BiocParallel', 'plyr', 'crossmap', 'pbapply' packages. There is also support for a growing set of domain-specific packages on CRAN (e.g. 'boot', 'caret', 'DiceKriging', 'ez', 'fgsea', 'fwb', 'gamlss', 'glmmTMB', 'glmnet', 'kernelshap', 'lme4', 'metafor', 'mgcv', 'modelsummary', 'parameters', 'partykit', 'pls', 'pvclust', 'riskRegression', 'rugarch', 'sandwich', 'seriation', 'shapr', 'Sim.DiffProc', 'SimDesign', 'stars', 'strucchange', 'SuperLearner', 'tm', 'TSP', and 'vegan') and on Bioconductor (e.g. 'DESeq2', 'GenomicAlignments', 'GSVA', 'Rsamtools', 'scater', 'scuttle', 'SingleCellExperiment', and 'sva').
Authors@R: c(
   person("Henrik", "Bengtsson", role=c("aut", "cre", "cph"),
          email = "henrikb@braju.com",
          comment = c(ORCID = "0000-0002-7579-5165")))
License: Apache License (>= 2)
URL: https://futurize.futureverse.org,
        https://github.com/futureverse/futurize
BugReports: https://github.com/futureverse/futurize/issues
Depends: R (>= 4.1.0), future (>= 1.69.0)
Imports: utils
Suggests: methods, future.apply (>= 1.20.2), foreach, doFuture (>=
        1.2.1), purrr, furrr, crossmap, plyr, pbapply, BiocParallel,
        boot, survival, caret, randomForest, DESeq2, DiceKriging, ez,
        fgsea, GenomicAlignments, Rsamtools, fwb, gamlss, glmmTMB,
        glmnet, GSVA, kernelshap, lme4, metafor, mgcv, modelsummary,
        parameters (>= 0.29.1), partykit, pls, pvclust, riskRegression,
        rugarch, sandwich, scater, scuttle, seriation, shapr,
        Sim.DiffProc, SimDesign, SingleCellExperiment, stars,
        strucchange, SuperLearner, sva, tm, vegan, tools, commonmark,
        base64enc
VignetteBuilder: futurize
Language: en-US
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-12 14:06:47 UTC; hb
Author: Henrik Bengtsson [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-7579-5165>)
Maintainer: Henrik Bengtsson <henrikb@braju.com>
Repository: CRAN
Date/Publication: 2026-06-12 15:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(print,transpiled_call)
export(futurize)
export(futurize_options)
export(futurize_supported_functions)
export(futurize_supported_packages)
export(fz)
importFrom(future,plan)
importFrom(utils,capture.output)
importFrom(utils,getS3method)
importFrom(utils,globalVariables)
importFrom(utils,packageName)
importFrom(utils,packageVersion)

<div id="badges"><!-- pkgdown markup -->
<a href="https://CRAN.R-project.org/web/checks/check_results_futurize.html"><img border="0" src="https://www.r-pkg.org/badges/version/futurize" alt="CRAN check status"/></a> <a href="https://github.com/futureverse/futurize/actions?query=workflow%3AR-CMD-check"><img border="0" src="https://github.com/futureverse/futurize/actions/workflows/R-CMD-check.yaml/badge.svg?branch=develop" alt="R CMD check status"/></a>     <a href="https://app.codecov.io/gh/futureverse/futurize"><img border="0" src="https://codecov.io/gh/futureverse/futurize/branch/develop/graph/badge.svg" alt="Coverage Status"/></a> 
</div>

# futurize: Parallelize Common Functions via One Magic Function <img border="0" src="man/figures/futurize-logo.png" style="width: 120px; margin: 2ex;" alt="The 'futurize' hexlogo" align="right"/>

## TL;DR 

The **futurize** package makes it extremely simple to parallelize your
existing map-reduce calls, but also a growing set of domain-specific
calls.  All you need to know is that there is a single function called
`futurize()` that will take care of everything, e.g.

```r
y <- lapply(x, fcn) |> futurize()
y <- map(x, fcn) |> futurize()
b <- boot(city, ratio, R = 999) |> futurize()
```

The `futurize()` function parallelizes via **[futureverse]**, meaning
your code can take advantage of any **[supported future backends]**,
whether it be parallelization on your local computer, across multiple
computers, in the cloud, or on a high-performance compute (HPC) cluster.
The **futurize** package has only one hard dependency - the
**[future]** package. All other dependencies are optional "buy-in"
dependencies as shown in the below tables.

In addition to getting access to all future-based parallel backends,
by using `futurize()` you also get access to all the benefits that
come with **futureverse**, including **structured concurrency**. For
example, it ensures that remaining parallel tasks are cancelled if
there is an error or an interrupt. Also, if the function you
parallelize outputs messages and warnings, they will be relayed from
the parallel worker to your main R session, just as you get when
running sequentially. This is particularly useful when troubleshooting
or debugging.

Using **futurize** comes with a zero risk buy-in. If there is ever a
parallel universe where `futurize()` suddenly stops working, setting
`futurize <- identical` avoids rewrites while make all code to run
sequentially.


## Supported map-reduce packages

The **futurize** package supports transpilation of functions from
multiple packages. The tables below summarize the supported map-reduce
(Table 1) and domain-specific (Tables 2 and 3) functions,
respectively.  To programmatically see which packages are currently
supported, use:

```r
futurize_supported_packages()
```

To see which functions are supported for a specific package, use:

```r
futurize_supported_functions("caret")
```

| Package            | Functions                                                                                                          | Requires                 |
|--------------------|--------------------------------------------------------------------------------------------------------------------|--------------------------|
| **base**           | `lapply()`, `sapply()`, `tapply()`, `vapply()`, `mapply()`, `.mapply()`, `Map()`, `eapply()`, `apply()`, `by()`, `replicate()`, `Filter()` | **[future.apply]** |
| **stats**          | `kernapply()`                                                                                  | **[future.apply]** |
| **[purrr]**        | `map()` and variants, `map2()` and variants, `pmap()` and variants, `imap()` and variants, `modify()`, `modify_if()`, `modify_at()`, `map_if()`, `map_at()` | **[furrr]** |
| **[crossmap]**     | `xmap()` and variants, `xwalk()`, `map_vec()`, `map2_vec()`, `pmap_vec()`, `imap_vec()`        | -                  |
| **[foreach]**      | `%do%`, e.g. `foreach() %do% { }`, `times() %do% { }`                                          | **[doFuture]**     | 
| **[plyr]**         | `aaply()` and variants, `ddply()` and variants, `llply()` and variants, `mlply()` and variants | **[doFuture]**     | 
| **[pbapply]**      | `pblapply()`, `pbsapply()` and variants, `pbby()`, `pbreplicate()` and `pbwalk()`              | **[future.apply]** |
| **[BiocParallel]** | `bplapply()`, `bpmapply()`, `bpvec()`, `bpiterate()`, `bpaggregate()`                          | **[doFuture]**     | 

_Table 1: Map-reduce functions currently supported by `futurize()` for parallel transpilation._

Here are some examples:

```r
library(futurize)
plan(multisession)

xs <- 1:10
ys <- lapply(xs, sqrt) |> futurize()

xs <- 1:10
ys <- purrr::map(xs, sqrt) |> futurize()

xs <- 1:10
ys <- crossmap::xmap_dbl(xs, ~ .y * .x) |> futurize()

library(foreach)
xs <- 1:10
ys <- foreach(x = xs) %do% { sqrt(x) } |> futurize()

xs <- 1:10
ys <- plyr::llply(xs, sqrt) |> futurize()

xs <- 1:10
ys <- pbapply::pblapply(xs, sqrt) |> futurize()

xs <- 1:10
ys <- BiocParallel::bplapply(xs, sqrt) |> futurize()
```

and

```r
ys <- replicate(3, rnorm(1)) |> futurize()

y <- by(warpbreaks, warpbreaks[,"tension"],
        function(x) lm(breaks ~ wool, data = x)) |> futurize()

xs <- EuStockMarkets[, 1:2]
k <- kernel("daniell", 50)
xs_smooth <- stats::kernapply(xs, k = k) |> futurize()
```


## Supported domain-specific packages

You can also futurize calls from a growing set of domain-specific CRAN
and Bioconductor packages that have optional built-in support for
parallelization.

### CRAN packages with support for futurize

| Package                    | Functions                                                                    | Requires           |
|----------------------------|------------------------------------------------------------------------------|--------------------|
| **[boot]**                 | `boot()`, `censboot()`, `tsboot()`                                           | -                  |
| **[caret]**                | `bag()`, `gafs()`, `nearZeroVar()`, `rfe()`, `safs()`, `sbf()`, `train()`    | **[doFuture]**     |
| **[DiceKriging]**          | `km()`                                                                       | **[doFuture]**     |
| **[ez]**                   | `ezBoot()`, `ezPerm()`, `ezPlot2()`                                          | **[doFuture]**     |
| **[fwb]**                  | `fwb()`, `vcovFWB()`                                                         | -                  |
| **[gamlss]**               | `add1All()`, `add1TGD()`, `drop1All()`, `drop1TGD()`, `gamlssCV()`           | -                  |
| **[glmmTMB]**              | `profile()` for 'glmmTMB'                                                    | -                  |
| **[glmnet]**               | `cv.glmnet()`                                                                | **[doFuture]**     |
| **[kernelshap]**           | `kernelshap()`, `permshap()`                                                 | **[doFuture]**     |
| **[lme4]**                 | `allFit()`, `bootMer()`, `influence()` and `profile()` for 'merMod'          | -                  |
| **[metafor]**              | `profile()`, `rstudent()`, `cooks.distance()`, `dfbetas()` for 'rma'         | -                  |
| **[mgcv]**                 | `bam()`, `predict()` for 'bam'                                               | -                  |
| **[modelsummary]**         | `modelsummary()`, `msummary()`, `modelplot()`                                | **[future.apply]** |
| **[parameters]**           | `bootstrap_model()`, `bootstrap_parameters()`                                | -                  |
| **[partykit]**             | `cforest()`, `ctree_control()`, `mob_control()`, `varimp()` for 'cforest'    | **[future.apply]** |
| **[pls]**                  | `mvr()`, `plsr()`, `pcr()`, `cppls()`, `crossval()`                          | -                  |
| **[pvclust]**              | `pvclust()`                                                                  | -                  |
| **[riskRegression]**       | `Score()` for 'list'                                                         | **[doFuture]**     |
| **[rugarch]**              | `arfimacv()`, `arfimadistribution()`, `arfimaroll()`, `autoarfima()`, `multifilter()`, `multifit()`, `multiforecast()`, `ugarchboot()`, `ugarchdistribution()`, `ugarchroll()` | -                  |
| **[sandwich]**             | `vcovBS()`, `vcovJK()`                                                       | **[future.apply]** |
| **[seriation]**            | `seriate_best()`, `seriate_rep()`                                            | **[doFuture]**     |
| **[shapr]**                | `explain()`, `explain_forecast()`                                            | -                  |
| **[Sim.DiffProc]**         | `MCM.sde()`                                                                  | -                  |
| **[SimDesign]**            | `runSimulation()`, `runArraySimulation()`                                    | -                  |
| **[stars]**                | `st_apply()`                                                                 | **[future.apply]** |
| **[strucchange]**          | `breakpoints()` for 'formula'                                                | **[doFuture]**     |
| **[SuperLearner]**         | `CV.SuperLearner()`                                                          | -                  |
| **[tm]**                   | `TermDocumentMatrix()`, `tm_index()`, `tm_map()`                             | -                  |
| **[TSP]**                  | `solve_TSP()`                                                                | **[doFuture]**     |
| **[vegan]**                | `adonis()`, `adonis2()`, `anova()` for 'cca', `anosim()`, `cascadeKM()`, `estaccumR()`, `mantel()`, `mantel.partial()`, `metaMDSiter()`, `mrpp()`, `oecosimu()`, `ordiareatest()`, `permutest()` for 'betadisper', and 'cca'  | -                  |

_Table 2: CRAN packages with domain-specific functions currently
supported by `futurize()` for parallel transpilation._

Here are some examples:

```r
ratio <- function(d, w) sum(d$x * w)/sum(d$u * w)
b <- boot::boot(boot::city, ratio, R = 999) |> futurize()

ctrl <- caret::trainControl(method = "cv", number = 10)
model <- caret::train(Species ~ ., data = iris, method = "rf", trControl = ctrl) |> futurize()

rt <- ez::ezBoot(data = ANT, dv = rt, wid = subnum, within = .(cue, flank), between = group) |> futurize()

f <- fwb::fwb(boot::city, ratio, R = 999) |> futurize()

m <- DiceKriging::km(~., design = design, response = response, multistart = 8L) |> futurize()

cv <- gamlss::gamlssCV(y ~ pb(x), data = abdom, K.fold = 10) |> futurize()

cv <- glmnet::cv.glmnet(x, y) |> futurize()

ks <- kernelshap::kernelshap(model, X = x_explain, bg_X = bg_X) |> futurize()

m <- lme4::allFit(models) |> futurize()

fit <- metafor::rma(yi, vi)
pr <- profile(fit) |> futurize()

b <- mgcv::bam(y ~ s(x0, bs = bs) + s(x1, bs = bs), data = dat) |> futurize()

fit <- parameters::bootstrap_model(model, iterations = 1000) |> futurize()

cf <- partykit::cforest(dist ~ speed, data = cars) |> futurize()

m <- pls::plsr(density ~ NIR, ncomp = 10, data = yarn, validation = "CV") |> futurize()

fit <- pvclust::pvclust(mtcars, nboot = 1000) |> futurize()

v <- sandwich::vcovBS(fm) |> futurize()

sc <- riskRegression::Score(list("CSC" = fit), data = d,
  formula = Hist(time, event) ~ 1, times = 5, B = 100,
  split.method = "bootcv") |> futurize()

roll <- rugarch::ugarchroll(spec, sp500ret, n.start = 1000, 
  refit.window = "moving", refit.every = 100) |> futurize()

result <- shapr::explain(model, x_explain, x_train, approach = "empirical", phi0 = phi0) |> futurize()


o <- seriation::seriate_best(d_supreme) |> futurize()

res <- Sim.DiffProc::MCM.sde(model, statistic = stat, R = 100) |> futurize()

res <- SimDesign::runSimulation(Design, replications = 1000,
  generate = Generate, analyse = Analyse, summarise = Summarise) |> futurize()

s <- stars::st_as_stars(matrix(1:20, nrow = 5, ncol = 4))
res <- stars::st_apply(s, MARGIN = 1, FUN = mean) |> futurize()

bp <- strucchange::breakpoints(Nile ~ 1) |> futurize()

res <- SuperLearner::CV.SuperLearner(Y, X, SL.library = SL.library) |> futurize()

m <- tm::tm_map(crude, content_transformer(tolower)) |> futurize()

tour <- TSP::solve_TSP(USCA50, method = "nn", rep = 10) |> futurize()

md <- vegan::mrpp(dune, Management) |> futurize()
```


### Bioconductor packages with support for futurize

| Package                    | Functions                                                                 | Requires           |
|----------------------------|---------------------------------------------------------------------------|--------------------|
| **[DESeq2]**               | `DESeq()`, `lfcShrink()`, `results()`                                     | **[doFuture]**     |
| **[fgsea]**                | `fgsea()`, `fgseaMultilevel()`, `fgseaSimple()`, `fgseaLabel()`, `geseca()`, `gesecaSimple()`, `collapsePathwaysGeseca()` | **[doFuture]**     |
| **[GenomicAlignments]**    | `summarizeOverlaps()`                                                     | **[doFuture]**     |
| **[GSVA]**                 | `gsva()`, `gsvaRanks()`, `gsvaScores()`, `spatCor()`                      | **[doFuture]**     |
| **[Rsamtools]**            | `countBam()`, `scanBam()`                                                 | **[doFuture]**     |
| **[scater]**               | `calculatePCA()`, `calculateTSNE()`, `calculateUMAP()`, `runPCA()`, `runTSNE()`, `runUMAP()`, `runColDataPCA()`, `nexprs()`, `getVarianceExplained()`, `plotRLE()` | **[doFuture]** |
| **[scuttle]**              | `calculateAverage()`, `logNormCounts()`, `normalizeCounts()`, `perCellQCMetrics()`, `perFeatureQCMetrics()`, `addPerCellQCMetrics()`, `addPerFeatureQCMetrics()`, `addPerCellQC()`, `addPerFeatureQC()`, `numDetectedAcrossCells()`, `numDetectedAcrossFeatures()`, `sumCountsAcrossCells()`, `sumCountsAcrossFeatures()`, `summarizeAssayByGroup()`, `aggregateAcrossCells()`, `aggregateAcrossFeatures()`, `librarySizeFactors()`, `computeLibraryFactors()`, `geometricSizeFactors()`, `computeGeometricFactors()`, `medianSizeFactors()`, `computeMedianFactors()`, `pooledSizeFactors()`, `computePooledFactors()`, `fitLinearModel()` | **[doFuture]** |
| **[SingleCellExperiment]** | `applySCE()`                                                              | **[doFuture]**     |
| **[sva]**                  | `ComBat()`, `read.degradation.matrix()`                                   | **[doFuture]**     |

_Table 3: Bioconductor packages with domain-specific functions
currently supported by `futurize()` for parallel transpilation._


Here are some examples:

```r
dds <- DESeq2::DESeq(dds) |> futurize()

res <- fgsea::fgsea(pathways, stats) |> futurize()

se <- GenomicAlignments::summarizeOverlaps(features, bam_files) |> futurize()

es <- GSVA::gsva(GSVA::gsvaParam(expr, geneSets)) |> futurize()

counts <- Rsamtools::countBam(bamViews) |> futurize()

sce <- scater::runPCA(sce) |> futurize()

qc <- scuttle::perFeatureQCMetrics(sce) |> futurize()

result <- SingleCellExperiment::applySCE(sce, scuttle::perFeatureQCMetrics) |> futurize()
  
adjusted <- sva::ComBat(dat = dat, batch = batch) |> futurize()
```



[futureverse]: https://www.futureverse.org
[future]: https://future.futureverse.org
[future.apply]: https://future.apply.futureverse.org
[furrr]: https://furrr.futureverse.org
[doFuture]: https://doFuture.futureverse.org
[TSP]: https://cran.r-project.org/package=TSP
[boot]: https://cran.r-project.org/package=boot
[caret]: https://cran.r-project.org/package=caret
[crossmap]: https://cran.r-project.org/package=crossmap
[DESeq2]: https://bioconductor.org/packages/DESeq2/
[DiceKriging]: https://cran.r-project.org/package=DiceKriging
[ez]: https://cran.r-project.org/package=ez
[fgsea]: https://bioconductor.org/packages/fgsea/
[foreach]: https://cran.r-project.org/package=foreach
[fwb]: https://ngreifer.github.io/fwb/
[gamlss]: https://cran.r-project.org/package=gamlss
[glmmTMB]: https://cran.r-project.org/package=glmmTMB
[glmnet]: https://cran.r-project.org/package=glmnet
[GenomicAlignments]: https://bioconductor.org/packages/GenomicAlignments/
[GSVA]: https://bioconductor.org/packages/GSVA/
[kernelshap]: https://cran.r-project.org/package=kernelshap
[lme4]: https://cran.r-project.org/package=lme4
[metafor]: https://cran.r-project.org/package=metafor
[mgcv]: https://cran.r-project.org/package=mgcv
[modelsummary]: https://cran.r-project.org/package=modelsummary
[parameters]: https://cran.r-project.org/package=parameters
[partykit]: https://cran.r-project.org/package=partykit
[pbapply]: https://cran.r-project.org/package=pbapply
[plyr]: https://cran.r-project.org/package=plyr
[pls]: https://cran.r-project.org/package=pls
[pvclust]: https://cran.r-project.org/package=pvclust
[purrr]: https://cran.r-project.org/package=purrr
[riskRegression]: https://cran.r-project.org/package=riskRegression
[rugarch]: https://cran.r-project.org/package=rugarch
[sandwich]: https://cran.r-project.org/package=sandwich
[Rsamtools]: https://bioconductor.org/packages/Rsamtools/
[scater]: https://bioconductor.org/packages/scater/
[scuttle]: https://bioconductor.org/packages/scuttle/
[SingleCellExperiment]: https://bioconductor.org/packages/SingleCellExperiment/
[seriation]: https://cran.r-project.org/package=seriation
[shapr]: https://cran.r-project.org/package=shapr
[Sim.DiffProc]: https://cran.r-project.org/package=Sim.DiffProc
[SimDesign]: https://cran.r-project.org/package=SimDesign
[stars]: https://cran.r-project.org/package=stars
[strucchange]: https://cran.r-project.org/package=strucchange
[sva]: https://bioconductor.org/packages/sva/
[SuperLearner]: https://cran.r-project.org/package=SuperLearner
[tm]: https://cran.r-project.org/package=tm
[vegan]: https://cran.r-project.org/package=vegan
[BiocParallel]: https://bioconductor.org/packages/BiocParallel/
[supported future backends]: https://www.futureverse.org/backends.html

library(futurize)
options(futurize.debug = TRUE)

# --------------------------------------------------------------------
# futurize()
# --------------------------------------------------------------------
message("futurize(NA):")
res <- futurize(NA)
print(res)
stopifnot(isTRUE(res))

message("futurize(FALSE):")
res <- futurize(FALSE)
print(res)
stopifnot(isTRUE(res))

message("futurize(TRUE):")
res <- futurize(TRUE)
print(res)
stopifnot(isFALSE(res))

if (requireNamespace("future.apply", quietly = TRUE)) {
  message("futurize(when = FALSE):")
  y_truth <- lapply(1:3, identity)
  y <- lapply(1:3, identity) |> futurize(when = FALSE)
  stopifnot(identical(y, y_truth))
  y <- lapply(1:3, identity) |> futurize(when = TRUE)
  stopifnot(identical(y, y_truth))
  expr <- lapply(1:3, identity) |> futurize(when = FALSE, eval = FALSE)
  print(expr)
}

## Cannot futurize non-calls
res <- tryCatch(base::pi |> futurize::futurize(), error = identity)
print(res)
stopifnot(inherits(res, "error"))

## Cannot futurize non-calls
res <- tryCatch(quote(1 + 2) |> futurize::futurize(), error = identity)
print(res)
stopifnot(inherits(res, "error"))

## Cannot futurize non-existing functions
res <- tryCatch(futurize:::unknown |> futurize::futurize(), error = identity)
print(res)
stopifnot(inherits(res, "error"))

## Cannot futurize non-existing infix operators
res <- tryCatch(futurize:::`%unknown%` |> futurize::futurize(), error = identity)
print(res)
stopifnot(inherits(res, "error"))

## Cannot futurize non-supported functions
res <- tryCatch(futurize:::futurize_supported_packages() |> futurize::futurize(), error = identity)
print(res)
stopifnot(inherits(res, "error"))

## Cannot futurize private functions
res <- tryCatch(futurize:::import_future() |> futurize::futurize(), error = identity)
print(res)
stopifnot(inherits(res, "error"))


# --------------------------------------------------------------------
# futurize_options()
# --------------------------------------------------------------------
message("*** futurize_options() - specified tracking")
opts <- futurize_options(seed = TRUE)
stopifnot("seed" %in% attr(opts, "specified"))
stopifnot(!("globals" %in% attr(opts, "specified")))

opts <- futurize_options(seed = TRUE, globals = FALSE, packages = "foo",
                         stdout = FALSE, conditions = "warning",
                         scheduling = 2.0, chunk_size = 10L)
specified <- attr(opts, "specified")
stopifnot(all(c("seed", "globals", "packages", "stdout", "conditions",
                "scheduling", "chunk_size") %in% specified))

opts <- futurize_options(extra_opt = 42)
stopifnot("extra_opt" %in% attr(opts, "specified"))


# --------------------------------------------------------------------
# futurize_supported_packages() and futurize_supported_functions()
# --------------------------------------------------------------------
pkgs <- futurize_supported_packages()
print(pkgs)

## Assert that there are not clashes between supported packages
pkgs <- futurize_supported_packages()
for (pkg in rep(pkgs, times = 2L)) {
  cat(sprintf("Package %s:\n", pkg))
  ## futurize_supported_functions() fail if required packages
  ## are not supported
  fcns <- tryCatch({
    futurize::futurize_supported_functions(pkg)
  }, error = identity)
  print(fcns)
}

#' @tags pkg-future.apply
#' @tags future_Filter
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)

message("*** future_Filter() ...")

is_even <- function(x) { x %% 2 == 0 }
x <- sample.int(100, size = 1000, replace = TRUE)

y_truth <- x[vapply(x, FUN.VALUE = NA, FUN = is_even)]
str(y_truth)

for (strategy in supportedStrategies()) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)

  y <- Filter(is_even, x)
  str(y)

  stopifnot(identical(y, y_truth))
  
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
}

plan(sequential)

message("*** future_Filter() ... DONE")

} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags pkg-boot
if (requireNamespace("boot") && getRversion() >= "4.4.0") {
library(futurize)
library(boot)
options(future.rng.onMisuse = "error")

all_equal <- function(a, b, ...) {
  a$call <- b$call <- NULL
  all.equal(a, b, ...)
}

plan(multisession)

#------------------------------------------------------------------
# boot()
#------------------------------------------------------------------
message("boot() ...")
## Adopted from example("boot", package = "boot")

ratio <- function(d, w) {
  sum(d$x * w)/sum(d$u * w)
}

set.seed(42)
b_truth <- boot(city, ratio, R = 99L, stype = "w")
print(b_truth)

set.seed(42)
b <- boot(city, ratio, R = 99L, stype = "w") |> futurize_and_verify()
print(b)

stopifnot(all_equal(b, b_truth))

message("boot() ... done")

#------------------------------------------------------------------
# censboot()
#------------------------------------------------------------------
message("censboot() ...")
## Adopted from example("censboot", package = "boot")

if (requireNamespace("survival", quietly = TRUE)) {
  ## From example("censboot", package = "boot")
  library(survival)
  data(aml, package = "boot") # not the version in 'survival'
  
  aml.fun <- function(data) {
    surv <- survfit(Surv(time, cens) ~ group, data = data)
    out <- NULL
    st <- 1
    for (s in seq_along(surv$strata)) {
       inds <- st:(st + surv$strata[s] - 1)
       md <- min(surv$time[inds[1-surv$surv[inds] >= 0.5]])
       st <- st + surv$strata[s]
       out <- c(out, md)
    }
    out
  }

  R <- 100L
  
  set.seed(42)
  b_truth <- censboot(aml, aml.fun, R = R, strata = aml$group)
  print(b_truth)
  
  set.seed(42)
  b <- censboot(aml, aml.fun, R = R, strata = aml$group) |> futurize_and_verify()
  print(b)

  b_truth$call <- NULL
  b$call <- NULL
  stopifnot(all.equal(b, b_truth))
}

message("censboot() ... done")

#------------------------------------------------------------------
# tsboot()
#------------------------------------------------------------------
message("tsboot() ...")
## Adopted from example("tsboot", package = "boot")

## stats::ar()
if (requireNamespace("stats", quietly = TRUE)) {

  lynx.fun <- function(tsb) {
    ar.fit <- ar(tsb, order.max = 25)
    c(ar.fit$order, mean(tsb), tsb)
  }

  R <- 99L
  
  set.seed(42)
  b_truth <- tsboot(log(lynx), lynx.fun, R = R, l = 20, sim = "geom")
  str(b_truth)
  
  set.seed(42)
  b <- tsboot(log(lynx), lynx.fun, R = R, l = 20, sim = "geom") |> futurize_and_verify()
  str(b)

  b_truth$call <- NULL
  b$call <- NULL
  stopifnot(all.equal(b, b_truth))
}

message("tsboot() ... done")

plan(sequential)
} ## if (requireNamespace("boot"))

#' @tags pkg-future.apply
#' @tags future_lapply future_mapply
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)



message("*** future_*apply() and 'future.stdout' ...")

options(future.debug = TRUE)

truth <- list()

out <- utils::capture.output({
  y <- lapply(1:0, FUN = function(x) {
    print(x)
  })
})
truth[["lapply"]] <- list(value = y, stdout = out)

out <- utils::capture.output({
  y <- mapply(1:0, 0:1, FUN = function(x, y) {
    print(list(x = x, y = y))
  })
})
truth[["mapply"]] <- list(value = y, stdout = out)

for (cores in 1:availCores) {
  message(sprintf("  - Testing with %d cores ...", cores))
  options(mc.cores = cores)
  
  for (strategy in supportedStrategies(cores)) {
    message(sprintf("* plan('%s') ...", strategy))
    plan(strategy)

    for (fun in names(truth)) {
      for (stdout in c(FALSE, TRUE, NA)) {
        message(sprintf("* future_%s(x, ..., future.stdout = %s) ...", fun, stdout))
  
        out <- utils::capture.output({
          if (fun == "lapply") {
            y <- lapply(1:0, FUN = function(x) {
              Sys.sleep(x / 20)  ## make futures resolve out of order
	      print(x)
            }) |> futurize_and_verify(stdout = stdout)
	  } else if (fun == "mapply") {
            y <- mapply(1:0, 0:1, FUN = function(x, y) {
              Sys.sleep(x / 20)  ## make futures resolve out of order
              print(list(x = x, y = y))
            }) |> futurize_and_verify(stdout = stdout)
	  }
        })
        stopifnot(identical(y, truth[[fun]]$value))
  
        if (isTRUE(stdout)) {
          stopifnot(identical(out, truth[[fun]]$stdout))
        } else if (is.na(stdout)) {
        } else {
          stopifnot(nchar(out) == 0)
        }
      
        message(sprintf("* future_%s(x, ..., future.stdout = %s) ... DONE", fun, stdout))
      } ## for (stdout ...)
    } ## for (fun ...)
    
    message(sprintf("* plan('%s') ... DONE", strategy))
  }
  
  message(sprintf("  - Testing with %d cores ... DONE", cores))
}

message("*** future_*apply() and 'future.stdout' ... DONE")



} ## if (requireNamespace("future.apply", quietly = TRUE))

library(futurize)

message("make_options_for_makeClusterFuture()")
opts <- futurize:::make_options_for_makeClusterFuture(options = list())
str(opts)
opts <- futurize:::make_options_for_makeClusterFuture(options = list(), defaults = list(packages = character(0L), stdout = TRUE))
str(opts)

message("*** make_options_for_doFuture()")
if (requireNamespace("doFuture", quietly = TRUE)) {
  opts <- futurize_options(chunk_size = 10L)
  result <- futurize:::make_options_for_doFuture(opts, wrap = FALSE)
  stopifnot("chunk.size" %in% names(result))
  stopifnot(!("chunk_size" %in% names(result)))
}

## Assert that future options are properly named
options <- list(seed = TRUE)
attr(options, "specified") <- "seed"
doFuture_options <- futurize:::make_options_for_doFuture(options, wrap = TRUE)
print(doFuture_options)
stopifnot(
  length(doFuture_options) == 1L,
  names(doFuture_options) == ".options.future"
)
opts <- doFuture_options[[".options.future"]]
stopifnot(
  length(opts) == 1L,
  names(opts) == "seed"
)

#' @tags skip_on_cran  ## (20s) to limit total check time
if (requireNamespace("GSVA") && requireNamespace("doFuture")) {
library(futurize)
library(GSVA)
print(sessionInfo())

options(future.rng.onMisuse = "error")

plan(multisession)

## Create simple test data
set.seed(42)
n_genes <- 120L    # > 100 for gsva() to parallelize
n_samples <- 120L  # > 100 for gsva() to parallelize
expr <- matrix(rnorm(n_genes * n_samples), nrow = n_genes, ncol = n_samples)
rownames(expr) <- paste0("gene", seq_len(n_genes))
colnames(expr) <- paste0("sample", seq_len(n_samples))

geneSets <- list(
  geneSet1 = paste0("gene", sample(n_genes, 10L)),
  geneSet2 = paste0("gene", sample(n_genes, 15L)),
  geneSet3 = paste0("gene", sample(n_genes, 20L))
)

## gsva() in GSVA (< 2.4) relies on BiocParallel::bpiterate(),
## which does _not_ support DoparParam by design.
if ("gsva" %in% futurize_supported_functions("GSVA")) {
  ## ---------------------------------------------------------
  ## gsva() with gsvaParam()
  ## ---------------------------------------------------------
  param <- gsvaParam(expr, geneSets)
  
  set.seed(42)
  result_truth <- gsva(param, verbose = FALSE)
  str(result_truth)
  
  set.seed(42)
  result <- gsva(param, verbose = FALSE) |> futurize_and_verify()
  str(result)
  stopifnot(all.equal(result, result_truth))
  
  set.seed(42)
  result2 <- GSVA::gsva(param, verbose = FALSE) |> futurize_and_verify()
  stopifnot(all.equal(result2, result_truth))


  ## ---------------------------------------------------------
  ## gsva() with ssgseaParam()
  ## ---------------------------------------------------------
  param_ssgsea <- ssgseaParam(expr, geneSets)
  
  set.seed(42)
  result_truth <- gsva(param_ssgsea, verbose = FALSE)
  str(result_truth)
  
  set.seed(42)
  result <- gsva(param_ssgsea, verbose = FALSE) |> futurize_and_verify()
  str(result)
  stopifnot(all.equal(result, result_truth))
  
  
  ## ---------------------------------------------------------
  ## gsva() with plageParam()
  ## ---------------------------------------------------------
  param_plage <- plageParam(expr, geneSets)
  
  set.seed(42)
  result_truth <- gsva(param_plage, verbose = FALSE)
  str(result_truth)
  
  set.seed(42)
  result <- gsva(param_plage, verbose = FALSE) |> futurize_and_verify()
  str(result)
  stopifnot(all.equal(result, result_truth))
} ## if ("gsva" %in% futurize_supported_functions("GSVA"))  

plan(sequential)
} ## if (requireNamespace("GSVA") && ...)

#' @tags pkg-vegan
if (requireNamespace("vegan") && getRversion() >= "4.4.0") {

all_equal_ignore_call <- function(a, b, ...) {
  attr(a, "heading") <- attr(b, "heading") <- NULL
  a$call <- b$call <- NULL
  a <- lapply(a, FUN = function(x) if ("call" %in% names(x)) x$call <- NULL)
  b <- lapply(b, FUN = function(x) if ("call" %in% names(x)) x$call <- NULL)
  if (is.list(a)) {
    a$control <- b$control <- NULL
  }
  if (!is.null(a$perm)) {
    a$perm <- as.vector(a$perm)
    b$perm <- as.vector(a$perm)
    names(a$perm) <- names(b$perm) <- NULL
  }
  res <- all.equal(a, b, ...)
  if (!isTRUE(res)) {
    str(list(a = a, b = b))
    print(res)
    str(res)
  }
  res
}

library(futurize)
library(vegan)
options(future.rng.onMisuse = "error")

plan(multisession)

data(dune)
data(dune.env)

message("*** mrpp()")

## Adopted from example("mrpp", package = "vegan")
set.seed(42)
res_truth <- mrpp(dune, dune.env$Management, permutations = 9L)
print(res_truth)

set.seed(42)
res <- mrpp(dune, dune.env$Management, permutations = 9L) |> futurize_and_verify()
print(res)

stopifnot(all_equal_ignore_call(res, res_truth))


message("*** adonis2()")

## Adopted from example("adonis2", package = "vegan")
set.seed(42)
res_truth <- adonis2(dune ~ Management, data = dune.env, permutations = 9L)
print(res_truth)

set.seed(42)
res <- adonis2(dune ~ Management, data = dune.env, permutations = 9L) |> futurize_and_verify()
print(res)

stopifnot(all_equal_ignore_call(res, res_truth))


message("*** anova() on cca object - S3 method anova.cca()")

## Adopted from example("anova.cca", package = "vegan")
ord <- cca(dune ~ A1 + Management, data = dune.env)

set.seed(42)
res_truth <- anova(ord, permutations = 9L)
print(res_truth)

set.seed(42)
res <- anova(ord, permutations = 9L) |> futurize_and_verify()
print(res)

stopifnot(all_equal_ignore_call(res, res_truth))


message("*** anosim()")

## Adopted from example("anosim", package = "vegan")
set.seed(42)
res_truth <- anosim(dune, dune.env$Management, permutations = 9L)
print(res_truth)

set.seed(42)
res <- anosim(dune, dune.env$Management, permutations = 9L) |> futurize_and_verify()
print(res)

stopifnot(all_equal_ignore_call(res, res_truth))


message("*** mantel()")

## Adopted from example("mantel", package = "vegan")
veg.dist <- vegdist(dune)
env.dist <- dist(dune.env[, "A1", drop = FALSE])

set.seed(42)
res_truth <- mantel(veg.dist, env.dist, permutations = 9L)
print(res_truth)

set.seed(42)
res <- mantel(veg.dist, env.dist, permutations = 9L) |> futurize_and_verify()
print(res)

stopifnot(all_equal_ignore_call(res, res_truth))


message("*** mantel.partial()")

## Adopted from example("mantel.partial", package = "vegan")
xdis <- vegdist(dune)
ydis <- dist(dune.env$A1)
zdis <- xdis + ydis
set.seed(42)
res_truth <- mantel.partial(xdis, ydis, zdis, permutations = 9L)
print(res_truth)

set.seed(42)
res <- mantel.partial(xdis, ydis, zdis, permutations = 9L) |> futurize_and_verify()
print(res)

stopifnot(all_equal_ignore_call(res, res_truth))


message("*** cascadeKM()")

## Adopted from example("cascadeKM", package = "vegan")
set.seed(42)
res_truth <- cascadeKM(dune, inf.gr = 2, sup.gr = 3, iter = 9L)
#print(res_truth)

set.seed(42)
res <- cascadeKM(dune, inf.gr = 2, sup.gr = 3, iter = 9L) |> futurize_and_verify()
#print(res)

## NOTE: cascadeKM() is not numerically reproducible
res_truth$partition <- res$partition <- NULL
res_truth$size <- sort(res_truth$size)
res$size <- sort(res$size)
stopifnot(all_equal_ignore_call(res, res_truth))


message("*** estaccumR()")

## Adopted from example("estaccumR", package = "vegan")
set.seed(42)
res_truth <- estaccumR(dune, permutations = 9L)
print(res_truth)

set.seed(42)
res <- estaccumR(dune, permutations = 9L) |> futurize_and_verify()
print(res)

stopifnot(all_equal_ignore_call(res, res_truth))


message("*** oecosimu()")
data(sipoo)

## Adopted from example("oecosimu", package = "vegan")
set.seed(42)
res_truth <- oecosimu(sipoo, nestedchecker, "r0")
print(res_truth)

set.seed(42)
res <- oecosimu(sipoo, nestedchecker, "r0") |> futurize_and_verify()
print(res)

stopifnot(all_equal_ignore_call(res, res_truth))

if (FALSE) {
  ## Skip for now, because of
  ## https://github.com/HenrikBengtsson/future-ideas/issues/1052
  message("*** ordiareatest()")
   
  ## Adopted from example("ordiareatest", package = "vegan")
  ord <- rda(dune)
  
  set.seed(42)
  res_truth <- ordiareatest(ord, dune.env$Management, permutations = 9)
  print(res_truth)
   
  set.seed(42)
  res <- ordiareatest(ord, dune.env$Management, permutations = 9) |> futurize_and_verify()
  print(res)
   
  stopifnot(all_equal_ignore_call(res, res_truth))
}

## vegan 2.8.0: simper() ignores argument 'parallel' with warning
## "argument 'parallel' is not used (yet)"
if (FALSE) {
  message("*** simper()")
  
  ## Adopted from example("simper", package = "vegan")
  set.seed(42)
  res_truth <- simper(dune, dune.env$Management, permutations = 9)
  print(res_truth)
  
  set.seed(42)
  res <- simper(dune, dune.env$Management, permutations = 9) |> futurize_and_verify()
  print(res)
  
  stopifnot(all_equal_ignore_call(res, res_truth))
}

plan(sequential)
} ## if (requireNamespace("vegan"))

## Local functions for test scripts
printf <- function(...) cat(sprintf(...))
mstr <- function(...) message(paste(capture.output(str(...)), collapse = "\n"))

## Call futurize() + validates that at least one future was created
futurize_and_verify <- function(expr, substitute = TRUE, ..., when = TRUE, eval = TRUE, envir = parent.frame()) {
  if (substitute) expr <- substitute(expr)
  if (!when || !eval) {
    res <- futurize::futurize(expr, substitute = FALSE, ..., when = when, eval = eval, envir = envir)
    return(res)
  }
  
  backend <- future::plan("backend")
  counters <- backend[["counters"]]
  
  res <- futurize::futurize(expr, substitute = FALSE, ..., when = when, eval = eval, envir = envir)
  
  delta <- backend[["counters"]] - counters

  ## Assert that at least one future was created
  if (delta[["created"]] == 0L) {
    e <- simpleError(sprintf("futurize() test did not result in futures being created"))
    class(e) <- c("FuturizeTestAssertionError", class(e))
    stop(e)
  }

  res
}

#' @tags pkg-foreach
if (requireNamespace("foreach") && requireNamespace("doFuture")) {
library(futurize)
library(foreach)
options(future.rng.onMisuse = "error")

plan(multisession)

message("%do%")
if (! "covr" %in% loadedNamespaces()) {
  ## NOTE: This basic foreach() call produces "Error in frameTypes(env) :
  ## namespace found within global environments" when checked with 'covr'
  y_truth <- foreach(x = 1:3, .combine = c) %do% sqrt(x)
  y_truth_2 <- foreach(x = 1:2) %:% foreach(y = 3L) %do% c(x,y)
} else {
  y_truth <- sqrt(1:3)
  y_truth_2 <- list(list(c(1L, 3L)), list(2:3))
}
str(list(y_truth = y_truth, y_truth_2 = y_truth_2))

message("foreach(...) %do% |> futurize()")
y <- foreach(x = 1:3, .combine = c) %do% {
  print(x)
  sqrt(x)
} |> futurize_and_verify()
print(y)
stopifnot(identical(y, y_truth))

message("foreach(...) %:% foreach(...) %do% |> futurize()")
y_2 <- foreach(x = 1:2) %:% foreach(y = 3L) %do% c(x,y) |> futurize_and_verify()
str(y_2)
stopifnot(identical(y_2, y_truth_2))

out <- utils::capture.output({
  y <- foreach(x = 1:3, .combine = c) %do% {
    print(x)
    sqrt(x)
  } |> futurize_and_verify(stdout = FALSE)
})
print(out)
stopifnot(
  identical(out, character(0L)),
  identical(y, y_truth)
)

message("Test with RNG:")
y <- local({
  opts <- options(future.rng.onMisuse = "error")
  on.exit(options(opts))
  foreach(x = 1:3, .combine = c) %do% {
    dummy <- sample.int(2L)
    sqrt(x)
  } |> futurize_and_verify(seed = TRUE)
})
print(y)
stopifnot(identical(y, y_truth))

message("times(...) %do% |> futurize()")
y <- times(1L) %do% { 42L } |> futurize_and_verify()
print(y)
stopifnot(identical(y, 42L))

message("Non-supported %dopar% and %dofuture%")
res <- tryCatch({ foreach(x = 1) %dopar% x |> futurize_and_verify() }, FuturizeTestAssertionError = stop, error = identity)
print(res)
stopifnot(inherits(res, "error"))

res <- tryCatch({ foreach(x = 1) %dofuture% x |> futurize_and_verify() }, FuturizeTestAssertionError = stop, error = identity)
print(res)
stopifnot(inherits(res, "error"))

message("Special case: Zero futurize() options")
y <- foreach(x = 1) %do% identity(x) |> futurize_and_verify(options = list())

plan(sequential)

message("Special case: Zero futurize() options")
y <- times(1L) %do% { 42L } |> futurize_and_verify(options = list())
print(y)
stopifnot(identical(y, 42L))

message("Internals")
opts <- futurize:::make_options_for_doFuture(options = list(), defaults = list(stdout = TRUE))
str(opts)

} ## if (requireNamespace("foreach") && requireNamespace("doFuture"))

#' @tags pkg-future.apply
#' @tags future_mapply
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)


library(tools) ## toTitleCase()

message("*** future_mapply() - globals ...")

#plan(cluster, workers = "localhost")
plan(sequential)

options(future.debug = FALSE)
a <- 1
b <- 2

globals_set <- list(
  A = FALSE,
  B = TRUE,
  C = c("a", "b"),
  D = list(a = 1, b = 2)
)

x <- list(1)

for (name in names(globals_set)) {
  globals <- globals_set[[name]]
  message("Globals set ", sQuote(name))
  str(globals)
  
  y_truth <- tryCatch({
    mapply(function(x) {
      median(c(x, a, b))
    }, x)
  }, FuturizeTestAssertionError = stop, error = identity)

  y1 <- tryCatch({
    mapply(function(x) {
      median(c(x, a, b))
    }, x) |> futurize_and_verify()
  }, FuturizeTestAssertionError = stop, error = identity)
  print(y1)
  stopifnot(identical(y1, y_truth))

  y2 <- tryCatch({
    mapply(function(x) {
      median(c(x, a, b))
    }, x) |> futurize_and_verify(globals = globals, packages = "utils")
  }, FuturizeTestAssertionError = stop, error = identity)
  print(y2)
  stopifnot(identical(y2, y_truth))
}

message("*** future_mapply() - globals ... DONE")



## Test adopted from http://stackoverflow.com/questions/42561088/nested-do-call-within-a-foreach-dopar-environment-cant-find-function-passed-w

message("*** future_mapply() - tricky globals ...")

my_add <- function(a, b) a + b

call_my_add <- function(a, b) {
  do.call(my_add, args = list(a = a, b = b))
}

call_my_add_caller <- function(a, b, FUN = call_my_add) {
  do.call(FUN, args = list(a = a, b = b))
}

main <- function(x = 1:2, caller = call_my_add_caller,
                 args = list(FUN = call_my_add)) {
  results <- mapply(function(i) {
    do.call(caller, args = c(list(a = i, b = i + 1L), args))
  }, x) |> futurize_and_verify()
  results
}

x <- list(list(1:2))
z_length <- mapply(do.call, args = x, MoreArgs = list(what = length))
fun <- function(...) sum(...)
z_fun <- mapply(do.call, args = x, MoreArgs = list(what = fun))

y0 <- NULL
for (strategy in supportedStrategies()) {
  plan(strategy)

  y <- main(1:3)
  if (is.null(y0)) y0 <- y
  stopifnot(identical(y, y0))

  message("- mapply(do.call, x, ...) |> futurize() ...")
  z <- mapply(do.call, args = x, MoreArgs = list(what = length)) |> futurize_and_verify()
  stopifnot(identical(z, z_length))
  z <- mapply(do.call, args = x, MoreArgs = list(what = fun)) |> futurize_and_verify()
  stopifnot(identical(z, z_fun))

  message("- mapply(FUN, x, ...) |> futurize() - passing arguments via '...' ...")
  ## typeof() == "list"
  obj <- data.frame(a = 1:2)
  stopifnot(typeof(obj) == "list")
  y <- mapply(function(a, b) typeof(b), 1L, MoreArgs = list(b = obj)) |> futurize_and_verify()
  stopifnot(identical(y[[1]], typeof(obj)))

  ## typeof() == "environment"
  obj <- new.env()
  stopifnot(typeof(obj) == "environment")
  y <- mapply(function(a, b) typeof(b), 1L, MoreArgs = list(b = obj)) |> futurize_and_verify()
  stopifnot(identical(y[[1]], typeof(obj)))

  ## typeof() == "S4"
  if (requireNamespace("methods")) {
    obj <- methods::getClass("MethodDefinition")
    stopifnot(typeof(obj) == "S4")
    y <- mapply(function(a, b) typeof(b), 1L, MoreArgs = list(b = obj)) |> futurize_and_verify()
    stopifnot(identical(y[[1]], typeof(obj)))
  }

  message("- mapply(FUN, X, ...) |> futurize() - 'X' containing globals ...")
  ## From https://github.com/futureverse/future.apply/issues/12
  a <- 42
  b <- 21
  X <- list(
    function(b) 2 * a,
    function() b / 2,
    function() a + b,
    function() nchar(tools::toTitleCase("hello world"))
  )
  z0 <- mapply(function(s, f) f() + s, s = seq_along(X), X)
  str(z0)
  z1 <- mapply(function(s, f) f() + s, s = seq_along(X), X) |> futurize_and_verify()
  str(z1)
  stopifnot(identical(z1, z0))
}

message("*** future_mapply() - tricky globals ... DONE")


message("*** future_mapply() - missing arguments ...")

## Here 'abc' becomes missing, i.e. missing(abc) is TRUE
foo <- function(x, abc) mapply(function(y) y, x)
y0 <- foo(1:2)
foo <- function(x, abc) mapply(function(y) y, x) |> futurize_and_verify()
y <- foo(1:2)
stopifnot(identical(y, y0))

message("*** future_mapply() - missing arguments ... DONE")


message("*** future_mapply() - false positives ...")

## Here 'abc' becomes a promise, which fails to resolve
## iff 'xyz' does not exist. (Issue #161)
suppressWarnings(rm(list = "xyz"))
foo <- function(x, abc) mapply(function(y) y, x)
y0 <- foo(1:2, abc = (xyz >= 3.14))
foo <- function(x, abc) mapply(function(y) y, x) |> futurize_and_verify()
y <- foo(1:2, abc = (xyz >= 3.14))
stopifnot(identical(y, y0))

message("*** future_mapply() - false positives ... DONE")


message("*** future_mapply() - too large ...")

X <- replicate(10L, 1:100, simplify = FALSE)
FUN <- function(x) {
  getOption("future.globals.maxSize")
}

y0 <- mapply(FUN = FUN, X)

sizes <- unclass(c(FUN = object.size(FUN), X = object.size(X)))
cat(sprintf("Baseline size of globals: %.2f KiB\n", sizes[["FUN"]] / 1024))

message("- true positive ...")
oMaxSize <- getOption("future.globals.maxSize")
options(future.globals.maxSize = 1L)
res <- tryCatch({
  y <- mapply(FUN = FUN, X) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))
res <- NULL
options(future.globals.maxSize = oMaxSize)

maxSize <- getOption("future.globals.maxSize")
y <- mapply(FUN = FUN, X) |> futurize_and_verify()
str(y)
stopifnot(all(sapply(y, FUN = identical, oMaxSize)))

message("- approximately invariant to chunk size ...")
maxSize <- sizes[["FUN"]] + sizes[["X"]] / length(X)
maxSize <- 4.0 * maxSize  ## Add a bit of leeway
if ("covr" %in% loadedNamespaces()) maxSize <- maxSize + 50e3
options(future.globals.maxSize = maxSize)

for (chunk_size in c(1L, 2L, 5L, structure(10L, ordering = "random"))) {
  y <- mapply(FUN = FUN, X) |> futurize_and_verify(chunk_size = chunk_size)
  str(y)
  stopifnot(all(unlist(y) == maxSize))
  cat(sprintf("maxSize = %g bytes\nfuture.globals.maxSize = %g bytes\n",
              maxSize, getOption("future.globals.maxSize")))
  stopifnot(getOption("future.globals.maxSize") == maxSize)
}
y <- NULL
options(future.globals.maxSize = oMaxSize)

message("*** future_mapply() - too large ... DONE")


message("*** future_mapply() - globals exceptions ...")

res <- tryCatch({
  y <- mapply(function(x) x, 1) |> futurize_and_verify(globals = 42)
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))

res <- tryCatch({
  y <- mapply(function(x) x, 1) |> futurize_and_verify(globals = list(1))
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))

res <- tryCatch({
  y <- mapply(function(x) x, 1) |> futurize_and_verify(globals = "...future.FUN")
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))

...future.elements_ii <- 42L
X <- list(function() 2 * ...future.elements_ii)
res <- tryCatch({
  y <- mapply(FUN = function(f) f(), X) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
if (! "covr" %in% loadedNamespaces()) {
  stopifnot(inherits(res, "error"))
}

message("*** future_mapply() - globals exceptions ... DONE")



} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags pkg-doFuture
#' @tags %dofuture%
#' @tags rng
#' @tags sequential multisession cluster multicore

if (requireNamespace("foreach") && requireNamespace("doFuture")) {
library(futurize)
library(foreach)

strategies <- future:::supportedStrategies()

options(future.debug = FALSE, doFuture.debug = FALSE)

message("*** %dofuture% w/ RNG ...")
print(sessionInfo())

r1_0 <- r3_0 <- NULL

for (strategy in strategies) {
  message(sprintf("- plan('%s') ...", strategy))
  plan(strategy)

  r1 <- foreach(i = 1:4) %do% {
    runif(1)
  } |> futurize(seed = 42)
  str(r1)
  if (is.null(r1_0)) r1_0 <- r1
  stopifnot(identical(r1, r1_0))
  
  r2 <- foreach(i = 1:4) %do% {
    runif(1)
  } |> futurize(seed = 42)
  str(r2)
  stopifnot(identical(r2, r1))

  set.seed(42)
  r3 <- foreach(i = 1:4) %do% {
    runif(1)
  } |> futurize(seed = TRUE)
  str(r3)
  if (is.null(r3_0)) r3_0 <- r3
  stopifnot(identical(r3, r3_0))
  
  set.seed(42)
  r4 <- foreach(i = 1:4) %do% {
    runif(1)
  } |> futurize(seed = TRUE)
  str(r4)
  stopifnot(identical(r4, r3))

  # Shutdown current plan
  plan(sequential)

  message(sprintf("- plan('%s') ... DONE", strategy))
} ## for (strategy ...)

message("*** %dofuture% w/ RNG ... DONE")

plan(sequential)
} ## if (requireNamespace("foreach") && requireNamespace("doFuture"))

#' @tags detritus-files
if (requireNamespace("pbapply") && requireNamespace("future.apply")) {
library(futurize)
library(pbapply)
options(future.rng.onMisuse = "error")

plan(multisession)

counters <- plan("backend")[["counters"]]
y <- pblapply(1:3, function(x) { print(x) }) |> futurize(stdout = FALSE)
delta <- plan("backend")[["counters"]] - counters
cat(sprintf("Futures created: %d\n", delta[["created"]]))
stopifnot(delta[["created"]] > 0L)
print(y)


xs <- list(aa = 1, bb = 1:2, cc = 1:10, dd = 1:5, .ee = -6:6)
FUN_no_rng <- function(x, na.rm = TRUE) {
  a <- 1:5
  add <- NULL
  if (length(x) == 2) add <- list(C = 42)
  median(c(a, x), na.rm = na.rm)
}

FUN_rng <- function(x, na.rm = TRUE) {
  dummy <- sample.int(10L)
  a <- 1:5
  add <- NULL
  if (length(x) == 2) add <- list(C = 42)
  median(c(a, x), na.rm = na.rm)
}

es <- as.environment(xs)


exprs <- list(
  pblapply = quote(pbapply::pblapply(xs, FUN) ),
  pblapply = quote(pblapply(xs, FUN) ),
  pbsapply = quote(pbsapply(xs, FUN) ),
  pbvapply = quote(pbvapply(xs, FUN, FUN.VALUE = NA_real_) ),
  pbeapply = quote(pbeapply(es, FUN) )
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))

  FUN <- FUN_no_rng
  truth <- eval(expr)
  counters <- plan("backend")[["counters"]]
  expr_f <- bquote(.(expr) |> futurize())
  res <- eval(expr_f)
  delta <- plan("backend")[["counters"]] - counters
  cat(sprintf("Futures created: %d\n", delta[["created"]]))
  stopifnot(delta[["created"]] > 0L)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> futurize(stdout = FALSE, conditions = character(0L)))
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> futurize(chunk_size = 1L))
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))

  message("Test with RNG:")
  FUN <- FUN_rng
  expr_f4 <- bquote(.(expr) |> futurize(seed = TRUE))
  print(expr_f4)
  res4 <- local({
    opts <- options(future.rng.onMisuse = "error")
    on.exit(options(opts))
    eval(expr_f4)
  })
  stopifnot(
    identical(res4, res),
    identical(res4, truth)
  )
}


plan(sequential)
} ## if (requireNamespace("pbapply") && requireNamespace("future.apply"))

#' @tags pkg-sandwich
if (requireNamespace("sandwich") && requireNamespace("future.apply")) {
library(futurize)
library(sandwich)
options(future.rng.onMisuse = "error")

plan(multisession, workers = 2L)

RNGkind("L'Ecuyer-CMRG")

fm <- lm(dist ~ speed, data = cars)

## Truth (sequential lapply)
## Since the non-parallel version uses a different set of RNG seeds
## than the parallel versions, we will not get numerically identical
## results
#set.seed(42)
#vcov_truth <- sandwich::vcovBS(fm, R = 5L)

## futurize()
set.seed(42)
vcov <- sandwich::vcovBS(fm, R = 5L) |> futurize()

stopifnot(
  is.matrix(vcov),
  all(dim(vcov) == c(2, 2))
)

## vcovJK()
set.seed(42)
vcov_jk <- sandwich::vcovJK(fm) |> futurize()
stopifnot(
  is.matrix(vcov_jk),
  all(dim(vcov_jk) == c(2, 2))
)

plan(sequential)
}

#!/usr/bin/env Rscript

#' Run a 'testme' Test Script
#'
#' R usage:
#' testme("<name>")
#'
#' Command-line usage:
#' tests/test-<name>.R
#'
#' Command-line usage without package re-install:
#' inst/testme/run.R --name=<test_name>
#' inst/testme/run.R <test-name.R>
#'
#' Options:
#' --package=<pkg>     The name of the package being tested
#'                     (Environment variable: `R_TESTME_PACKAGE`)
#'                     (Default: The `Package` field of the DESCRIPTION file)
#' --name=<name>       The name of the test to run, used to locate the test
#'                     script `test-<name>.R`
#'                     (Environment variable: `R_TESTME_NAME`)
#' --not-cran          Set environment variable `NOT_CRAN=true`
#' --coverage=summary  Estimate test code coverage with basic summary
#' --coverage=tally    Estimate test code coverage with full tally summary
#' --coverage=report   Estimate test code coverage with full HTML report
#' --debug             Output debug messages
#'                     (Environment variable: `R_TESTME_DEBUG`)
#'
#' Examples:
#' testme/test-abc.R
#' testme/test-abc.R --not-cran
#' tests/test-cpuLoad.R --coverage=report
#'
#' inst/testme/run.R inst/testme/test-abc.R
#' inst/testme/run.R inst/testme/test-abc.R --coverage
#'
#' Environment variables:
#' * R_TESTME_PACKAGE
#' * R_TESTME_NAME
#' * R_TESTME_PATH
#' * R_TESTME_FILTER_NAME:
#"    An R expression to be evaluated as TRUE
#' * R_TESTME_FILTER_TAGS:
#'    An R expression to be evaluated as TRUE, e.g.
#'    R_TESTME_FILTER_TAGS='"pkg-furrr" %in% tags'
#' * R_TESTME_COVERAGE
#' * R_TESTME_DEBUG
main <- function() {
  cmd_args <- commandArgs(trailingOnly = TRUE)
  
  pattern <- "--package=([[:alpha:][:alnum:]]+)"
  idx <- grep(pattern, cmd_args)
  if (length(idx) > 0L) {
    stopifnot(length(idx) == 1L)
    testme_package <- gsub(pattern, "\\1", cmd_args[idx])
    cmd_args <- cmd_args[-idx]
  } else {
    testme_package <- Sys.getenv("R_TESTME_PACKAGE", NA_character_)
    if (is.na(testme_package)) {
      desc <- read.dcf("DESCRIPTION")
      testme_package <- desc[1, "Package"]
    }
  }
  
  pattern <- "--path=([[:alpha:][:alnum:]]+)"
  idx <- grep(pattern, cmd_args)
  if (length(idx) > 0L) {
    stopifnot(length(idx) == 1L)
    path <- gsub(pattern, "\\1", cmd_args[idx])
    cmd_args <- cmd_args[-idx]
  } else {
    path <- Sys.getenv("R_TESTME_PATH", NA_character_)
    if (is.na(path)) {
      path <- file.path("inst", "testme")
    }
    if (!utils::file_test("-d", path)) {
      stop("There exist no such 'R_TESTME_PATH' folder: ", sQuote(path))
    }
  }
  
  pattern <- "--name=([[:alpha:][:alnum:]]+)"
  idx <- grep(pattern, cmd_args)
  if (length(idx) > 0L) {
    stopifnot(length(idx) == 1L)
    testme_name <- gsub(pattern, "\\1", cmd_args[idx])
    cmd_args <- cmd_args[-idx]
  } else {
    testme_name <- NULL
  }

  pattern <- "^--not-cran"
  idx <- grep(pattern, cmd_args)
  if (length(idx) > 0L) {
    cmd_args <- cmd_args[-idx]
    Sys.setenv(NOT_CRAN = "TRUE")
  }

  pattern <- "^--debug"
  idx <- grep(pattern, cmd_args)
  if (length(idx) > 0L) {
    cmd_args <- cmd_args[-idx]
    Sys.setenv(R_TESTME_DEBUG = "TRUE")
  }

  pattern <- "^--coverage(|=([[:alpha:][:alnum:]]+))$"
  idx <- grep(pattern, cmd_args)
  if (length(idx) > 0L) {
    value <- gsub(pattern, "\\2", cmd_args[idx])
    if (!nzchar(value)) {
      coverage <- "summary"
    } else {
      coverage <- match.arg(value, choices = c("none", "summary", "tally", "report"))
    }
    cmd_args <- cmd_args[-idx]
  } else {
    value <- Sys.getenv("R_TESTME_COVERAGE", "none")
    coverage <- match.arg(value, choices = c("none", "summary", "tally", "report"))
  }
  if (coverage != "none") {
    if (!utils::file_test("-f", "DESCRIPTION")) {
      stop("Current folder does not look like a package folder")
    }
  }
  
  ## Fallback for 'testme_name'?
  if (is.null(testme_name)) {
    if (length(cmd_args) > 0) {
      stopifnot(length(cmd_args) == 1L)
      file <- cmd_args[1]
      if (utils::file_test("-f", file)) {
        testme_name <- gsub("(^test-|[.]R$)", "", basename(file))
      } else {
        stop("No such file: ", file)
      }
    } else {
      testme_name <- Sys.getenv("R_TESTME_NAME", NA_character_)
      if (is.na(testme_name)) {
        stop("testme: Environment variable 'R_TESTME_NAME' is not set")
      }
    }
  } 
  
  testme_file <- file.path(path, sprintf("test-%s.R", testme_name))
  if (!utils::file_test("-f", testme_file)) {
    stop("There exist no such 'testme' file: ", sQuote(testme_file))
  }

  
  ## -----------------------------------------------------------------
  ## testme environment
  ## -----------------------------------------------------------------
  on_cran <- function() {
    not_cran <- Sys.getenv("NOT_CRAN", NA_character_)
    if (is.na(not_cran)) {
      not_cran <- FALSE
    } else {
      not_cran <- isTRUE(as.logical(not_cran))
    }
    !interactive() && !not_cran
  } ## on_cran()
  
  
  ## Get test script tags
  tags <- local({
    lines <- readLines(testme_file, warn = FALSE)
    pattern <- "^#'[[:blank:]]+@tags[[:blank:]]+"
    lines <- grep(pattern, lines, value = TRUE)
    tags <- sub(pattern, "", lines)
    tags
  })
  if (length(tags) > 0) {
    tags <- sub("[[:blank:]]*$", "", tags)
    tags <- unlist(strsplit(tags, split = "[[:blank:]]+"))
    tags <- sort(unique(tags))
  } else {
    tags <- character(0L)
  }

  debug <- isTRUE(as.logical(Sys.getenv("R_TESTME_DEBUG")))

  coverage <- match.arg(coverage, choices = c("none", "summary", "tally", "report"))

  ## Create 'testme' environment on the search() path
  testme_config <- list(
     testme = TRUE,
    package = testme_package,
       name = testme_name,
       tags = tags,
     status = "created",
      start = proc.time(),
     script = testme_file,
       path = path,
    on_cran = on_cran(),
       coverage = coverage,
      debug = debug
  )
  if ("testme" %in% search()) detach(name = "testme")
  testme <- attach(testme_config, name = "testme", warn.conflicts = FALSE)
  rm(list = c("tags", "testme_package", "testme_name", "testme_file"))
  

  ## -----------------------------------------------------------------
  ## Filters
  ## -----------------------------------------------------------------
  ## Skip on CRAN? To run these tests, set env var NOT_CRAN=true
  if ("skip_on_cran" %in% tags && on_cran()) {
    testme[["status"]] <- "skipped"
  }

  code <- Sys.getenv("R_TESTME_FILTER_NAME", NA_character_)
  if (!is.na(code)) {
    expr <- tryCatch(parse(text = code), error = identity)
    if (inherits(expr, "error")) {
      stop("Syntax error in R_TESTME_FILTER_NAME: ", sQuote(code))
    }
    
    keep <- tryCatch(eval(expr, envir = testme), error = identity)
    if (inherits(keep, "error")) {
      stop(sprintf("Evaluation of R_TESTME_FILTER_NAME=%s produced an error: %s",
           sQuote(code), conditionMessage(keep)))
    }
    if (!isTRUE(keep)) testme[["status"]] <- "skipped"
  }
  
  code <- Sys.getenv("R_TESTME_FILTER_TAGS", NA_character_)
  if (!is.na(code)) {
    expr <- tryCatch(parse(text = code), error = identity)
    if (inherits(expr, "error")) {
      stop("Syntax error in R_TESTME_FILTER_TAGS: ", sQuote(code))
    }
    keep <- tryCatch(eval(expr, envir = testme), error = identity)
    if (inherits(keep, "error")) {
      stop(sprintf("Evaluation of R_TESTME_FILTER_TAGS=%s produced an error: %s",
           sQuote(code), conditionMessage(keep)))
    }
    if (!isTRUE(keep)) testme[["status"]] <- "skipped"
  }
  
  testme_run_test(testme)
} ## main()



#' @param testme A names list
#'
testme_run_test <- function(testme) {
  message(sprintf("Test %s ...", sQuote(testme[["name"]])))
  if (testme[["debug"]]) {
    message("testme:")
    message(paste(utils::capture.output(utils::str(as.list(testme))), collapse = "\n"))
  }

  path <- testme[["path"]]

  ## Process prologue scripts, if they exist
  if (testme[["status"]] != "skipped" &&
      utils::file_test("-d", file.path(path, "_prologue"))) {
    testme[["status"]] <- "prologue"
    local({
      ## Find all prologue scripts
      files <- dir(file.path(path, "_prologue"), pattern = "*[.]R$", full.names = TRUE)
      files <- sort(files)
      testme[["prologue_scripts"]] <- files

      ## Source all prologue scripts inside the 'testme' environment
      expr <- bquote({
        files <- prologue_scripts
        if (.(testme[["debug"]])) message(sprintf("Sourcing %d prologue scripts ...", length(files)))
        for (kk in seq_along(files)) {
          file <- files[kk]
          if (.(testme[["debug"]])) message(sprintf("%02d/%02d prologue script %s", kk, length(files), sQuote(file)))
          source(file, local = TRUE)
        }
        if (.(testme[["debug"]])) message(sprintf("Sourcing %d prologue scripts ... done", length(files)))
        rm(list = c("kk", "file", "files"))
      })
      eval(expr, envir = testme)
    })
  
  #  ## In case prologue scripts overwrote some elements in 'testme'
  #  for (name in names(testme_config)) {
  #    testme[[name]] <- testme_config[[name]]
  #  }
  }
  
  
  ## Run test script
  ## Note, prologue scripts may trigger test to be skipped
  if (testme[["status"]] != "skipped") {
    if (testme[["debug"]]) message("Running test script: ", sQuote(testme[["script"]]))
    testme[["status"]] <- "failed"
    str(testme[["coverage"]])
    if (testme[["coverage"]] != "none") {
      pkg_env <- pkgload::load_all()
      cov <- covr::environment_coverage(pkg_env[["env"]], test_files = testme[["script"]])
      ## Keep source files with non-zero coverage
      tally <- covr::tally_coverage(cov)
      tally <- subset(tally, value > 0)
      cov <- cov[covr::display_name(cov) %in% unique(tally$filename)]
      testme[["test_coverage"]] <- cov
    } else {
      testme[["test_coverage"]] <- NULL
      source(testme[["script"]], echo = TRUE)
    }
    testme[["status"]] <- "success"
  }
  
  
  ## Process epilogue scripts, if they exist
  ## Note, epilogue scripts may change status or produce check errors
  if (testme[["status"]] == "success" &&
      utils::file_test("-d", file.path(path, "_epilogue"))) {
    testme[["status"]] <- "epilogue"
    local({
      ## Find all epilogue scripts
      files <- dir(file.path(path, "_epilogue"), pattern = "*[.]R$", full.names = TRUE)
      files <- sort(files)
      testme[["epilogue_scripts"]] <- files
    
      ## Source all epilogue scripts inside the 'testme' environment
      expr <- bquote({
        files <- epilogue_scripts
        if (.(testme[["debug"]])) message(sprintf("Sourcing %d epilogue scripts ...", length(files)))
        for (kk in seq_along(files)) {
          file <- files[kk]
          if (.(testme[["debug"]])) message(sprintf("%02d/%02d epilogue script %s", kk, length(files), sQuote(file)))
          source(file, local = TRUE)
        }
        if (.(testme[["debug"]])) message(sprintf("Sourcing %d epilogue scripts ... done", length(files)))
        rm(list = c("kk", "file", "files"))
      })
      eval(expr, envir = testme)
    })
    testme[["status"]] <- "success"
  }
  
  testme[["stop"]] <- proc.time()
  dt <- testme[["stop"]] - testme[["start"]]
  dt_str <- sprintf("%s=%.1gs", names(dt), dt)
  message("Test time: ", paste(dt_str, collapse = ", "))
  
  if ("testme" %in% search()) detach(name = "testme")

  cov <- testme[["test_coverage"]]
  if (!is.null(cov)) {
    message("Source files covered by the test script:")
    if (length(cov) > 0) {
      print(cov)
      if ("tally" %in% testme[["coverage"]]) {
        tally <- covr::tally_coverage(cov)
        print(tally)
      }
      if ("report" %in% testme[["coverage"]]) {
        html <- covr::report(cov, browse = FALSE)
        browseURL(html)
        Sys.sleep(5.0)
      }
    } else {
      message("* No source files were covered by this test!")
    }
  }

  message(sprintf("Test %s ... %s", sQuote(testme[["name"]]), testme[["status"]]))
} ## testme_run_test()


main()

#' @tags pkg-doFuture
#' @tags %dofuture%
#' @tags sequential multisession cluster multicore

if (requireNamespace("foreach") && requireNamespace("doFuture")) {
library(futurize)
library(foreach)

strategies <- future:::supportedStrategies()

message("*** doFuture - reproducibility ...")

res0 <- NULL

for (strategy in strategies) {
  message(sprintf("- plan('%s') ...", strategy))
  plan(strategy)

  mu <- 1.0
  sigma <- 2.0
  res <- foreach(i = 1:3) %do% {
    dnorm(i, mean = mu, sd = sigma)
  } |> futurize(packages = "stats")
  print(res)

  if (is.null(res0)) {
    res0 <- res
  } else {
    stopifnot(all.equal(res, res0))
  }

  # Shutdown current plan
  plan(sequential)

  message(sprintf("- plan('%s') ... DONE", strategy))
} ## for (strategy ...)

message("*** doFuture - reproducibility ... DONE")

plan(sequential)
} ## if (requireNamespace("foreach") && requireNamespace("doFuture"))

#' @tags pkg-future.apply
#' @tags future_tapply
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)



library("datasets") ## warpbreaks, iris

options(future.debug = FALSE)
message("*** future_tapply() ...")

for (strategy in supportedStrategies()[1]) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)

  message("- From example(tapply) ...")

  message("  - Example #1")
  library("stats")  ## rbinom()
  groups <- as.factor(stats::rbinom(32, n = 5, prob = 0.4))
  t <- table(groups)
  print(t)
  y0 <- tapply(groups, INDEX = groups, FUN = length)
  print(y0)
  y1 <- tapply(groups, INDEX = groups, FUN = length) |> futurize_and_verify()
  print(y1)
  stopifnot(all.equal(y1, y0))
  y2 <- tapply(groups, INDEX = groups, FUN = "length") |> futurize_and_verify()
  print(y2)
  stopifnot(all.equal(y2, y0))

  message("  - Example #2")
  ## contingency table from data.frame : array with named dimnames
  y0 <- tapply(warpbreaks$breaks, INDEX = warpbreaks[,-1], FUN = sum)
  print(y0)
  y1 <- tapply(warpbreaks$breaks, INDEX = warpbreaks[,-1], FUN = sum) |> futurize_and_verify()
  print(y1)
  stopifnot(all.equal(y1, y0))

  message("  - Example #3")
  y0 <- tapply(warpbreaks$breaks, warpbreaks[, 3, drop = FALSE], sum)
  print(y0)
  y1 <- tapply(warpbreaks$breaks, warpbreaks[, 3, drop = FALSE], sum) |> futurize_and_verify()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #4")
  n <- 17
  fac <- factor(rep_len(1:3, n), levels = 1:5)
  t <- table(fac)  
  y0 <- tapply(1:n, fac, sum)
  print(y0)
  y1 <- tapply(1:n, fac, sum) |> futurize_and_verify()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #5")
  if ("default" %in% names(formals(tapply))) {
    y0 <- tapply(1:n, fac, sum, default = 0) # maybe more desirable
    print(y0)
    y1 <- tapply(1:n, fac, sum, default = 0) |> futurize_and_verify() # maybe more desirable
    print(y1)
    stopifnot(all.equal(y1, y0))
  }

  message("  - Example #6")
  y0 <- tapply(1:n, fac, sum, simplify = FALSE)
  print(y0)
  y1 <- tapply(1:n, fac, sum, simplify = FALSE) |> futurize_and_verify()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #7")
  y0 <- tapply(1:n, fac, range)
  print(y0)
  y1 <- tapply(1:n, fac, range) |> futurize_and_verify()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #8")
  y0 <- tapply(1:n, fac, quantile)
  print(y0)
  y1 <- tapply(1:n, fac, quantile) |> futurize_and_verify()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #9")
  y0 <- tapply(1:n, fac, length) ## NA's
  print(y0)
  y1 <- tapply(1:n, fac, length) |> futurize_and_verify() ## NA's
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #10")
  if ("default" %in% names(formals(tapply))) {
    y0 <- tapply(1:n, fac, length, default = 0) # == table(fac)
    print(y0)
    y1 <- tapply(1:n, fac, length, default = 0) |> futurize_and_verify() # == table(fac)
    print(y1)
    stopifnot(all.equal(y1, y0))
  }
  
  message("  - Example #11")
  ## example of ... argument: find quarterly means
  y0 <- tapply(presidents, cycle(presidents), mean, na.rm = TRUE)
  print(y0)
  y1 <- tapply(presidents, cycle(presidents), mean, na.rm = TRUE) |> futurize_and_verify()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #12")
  ind <- list(c(1, 2, 2), c("A", "A", "B"))
  t <- table(ind)
  print(t)
  y0 <- tapply(1:3, ind) #-> the split vector
  print(y0)
  y1 <- tapply(1:3, ind) |> futurize::futurize() #-> the split vector
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #13")
  y0 <- tapply(1:3, ind, sum)
  print(y0)
  y1 <- tapply(1:3, ind, sum) |> futurize_and_verify()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  ## Some assertions (not held by all patch propsals):
  message("  - Example #14")
  nq <- names(quantile(1:5))
  y_truth <- c(1L, 2L, 4L)
  stopifnot(identical(tapply(1:3, ind), y_truth))
  stopifnot(identical(tapply(1:3, ind) |> futurize::futurize(), y_truth))
  
  message("  - Example #15")
  y_truth <- matrix(c(1L, 2L, NA, 3L), nrow = 2L,
                    dimnames = list(c("1", "2"), c("A", "B")))
  stopifnot(identical(tapply(1:3, ind, sum), y_truth))
  stopifnot(identical(tapply(1:3, ind, sum) |> futurize_and_verify(), y_truth))
  
  message("  - Example #16")
  y_truth <- array(list(
    `2` = structure(c(2, 5.75, 9.5, 13.25, 17), .Names = nq),
    `3` = structure(c(3, 6, 9, 12, 15), .Names = nq),
    `4` = NULL, `5` = NULL),
    dim = 4L, dimnames = list(as.character(2:5)))
  stopifnot(identical(tapply(1:n, fac, quantile)[-1], y_truth))
  stopifnot(identical((tapply(1:n, fac, quantile) |> futurize_and_verify())[-1], y_truth))

  if (getRversion() >= "4.3.0") {
    data <- iris[, c("Sepal.Length", "Sepal.Width")]
    y_truth <- tapply(data, INDEX = iris$Species, FUN = sum)
    y <- tapply(data, INDEX = iris$Species, FUN = sum) |> futurize_and_verify()
    stopifnot(identical(y, y_truth))
    
    y_truth2 <- tapply(data, INDEX = ~ iris$Species + iris$Petal.Width, FUN = sum)
    y2 <- tapply(data, INDEX = ~ iris$Species + iris$Petal.Width, FUN = sum) |> futurize_and_verify()
    stopifnot(identical(y2, y_truth2))
  }

  plan(sequential)
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
} ## for (strategy in ...) 


message("*** exceptions ...")

## Error: 'INDEX' is of length zero
res <- tryCatch({
  y <- tapply(1L, INDEX = list()) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))

## Error: total number of levels >= 2^31
res <- tryCatch({
  y <- tapply(1:216, INDEX = rep(list(1:216), times = 4L)) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))

## Error: arguments must have same length
res <- tryCatch({
  y <- tapply(1L, INDEX = list(1:2)) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))

message("*** future_tapply() ... DONE")

} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags pkg-doFuture
#' @tags %dofuture%
#' @tags detritus-files
#' @tags sequential multisession cluster multicore

if (requireNamespace("foreach") && requireNamespace("doFuture")) {
library(futurize)
library(foreach)

strategies <- future:::supportedStrategies()

message("*** doFuture - nested w/ %:% ...")

for (strategy1 in strategies) {
  for (strategy2 in strategies) {
    message(sprintf("- plan(list('%s', '%s')) ...", strategy1, strategy2))
    plan(list(a = strategy1, b = strategy2))
    nested <- plan("list")
    
    as <- 1:2
    bs <- 3:1
    x <- foreach(a = as) %:% foreach(b = bs) %do% {
      list(a = a, b = b, plan_b = future::plan("list"), plan = future::plan("next"))
    } |> futurize()

    stopifnot(length(x) == length(as))
    for (aa in seq_along(as)) {
      x_aa <- x[[aa]]
      stopifnot(length(x_aa) == length(bs))
      a <- as[aa]
      for (bb in seq_along(bs)) {
        x_aa_bb <- x_aa[[bb]]
        b <- bs[bb]
        stopifnot(
          length(x_aa_bb) == 4L,
          all(names(x_aa_bb) == c("a", "b", "plan_b", "plan")),
          x_aa_bb$a == a,
          x_aa_bb$b == b,
          length(x_aa_bb$plan_b) == length(nested[-1]),
          inherits(x_aa_bb$plan_b[[1]], strategy2),
          inherits(x_aa_bb$plan, strategy2)
        )
      }
    }

    ## Cleanup in order make sure none of these variables exist as
    ## proxies for missing globals of the name names
    rm(list = c("as", "bs", "x", "nested"))

    ## WORKAROUND: Shut down *nested* parallel workers as an attempt to avoid
    ## * checking for detritus in the temp directory ... NOTE
    ## from 'R CMD check --as-cran' when running on MS Windows. This looks
    ## like a bug in R, cf. https://bugs.r-project.org/show_bug.cgi?id=18133    
    message("- shut down nested workers")
    dummy <- foreach(ii = 1:nbrOfWorkers()) %do% plan("sequential") |> futurize()
    plan("sequential")
    
    message(sprintf("- plan(list('%s', '%s')) ... DONE", strategy1, strategy2))
  }
}

message("*** doFuture - nested w/ %:% ... DONE")

plan(sequential)
} ## if (requireNamespace("foreach") && requireNamespace("doFuture"))

#' @tags pkg-kernelshap
if (requireNamespace("kernelshap") && requireNamespace("doFuture")) {
  library(futurize)
  library(kernelshap)
  options(future.rng.onMisuse = "error")

  plan(multisession)

  ## Simple linear model
  set.seed(42)
  n <- 20L
  x_train <- data.frame(x1 = rnorm(n), x2 = rnorm(n))
  y_train <- 2 * x_train$x1 + x_train$x2 + rnorm(n)
  model <- lm(y ~ x1 + x2, data = data.frame(y = y_train, x_train))
  x_explain <- x_train[1:4, ]
  bg_X <- x_train[1:20, ]


  ## -------------------------------------------------------
  ## kernelshap()
  ## -------------------------------------------------------
  set.seed(42)
  result_truth <- kernelshap(
    model, X = x_explain, bg_X = bg_X
  )
  print(result_truth)

  set.seed(42)
  result <- kernelshap(
    model, X = x_explain, bg_X = bg_X
  ) |> futurize_and_verify()
  print(result)

  stopifnot(all.equal(result$S, result_truth$S))
  stopifnot(all.equal(result$SE, result_truth$SE))


  ## -------------------------------------------------------
  ## permshap()
  ## -------------------------------------------------------
  set.seed(42)
  result_truth2 <- permshap(
    model, X = x_explain, bg_X = bg_X
  )
  print(result_truth2)

  set.seed(42)
  result2 <- permshap(
    model, X = x_explain, bg_X = bg_X
  ) |> futurize_and_verify()
  print(result2)

  stopifnot(all.equal(result2$S, result_truth2$S))
  stopifnot(all.equal(result2$SE, result_truth2$SE))


  plan(sequential)
} ## if (requireNamespace("kernelshap") && requireNamespace("doFuture"))

#' @tags pkg-doFuture
#' @tags %dofuture%
#' @tags detritus-files
#' @tags sequential multisession cluster multicore

if (requireNamespace("foreach") && requireNamespace("doFuture")) {
library(futurize)
library(foreach)

options(
  parallelly.debug = TRUE,
  future.debug = TRUE,
  doFuture.debug = TRUE
)

strategies <- future:::supportedStrategies()

message("*** doFuture - explicitly exported globals ...")

message("- Globals manually specified as named list - also with '...' ...")

x <- 1:3
y_truth <- lapply(1:2, FUN = function(i) x[c(i, 2:3)])
str(y_truth)
  
for (strategy in strategies) {
  message(sprintf("- plan('%s') ...", strategy))
  plan(strategy)

  ## (a) automatic
  sub <- function(x, ...) {
    foreach(i = 1:2) %do% { x[c(i, ...)] } |> futurize()
  }
  y <- sub(x, 2:3)
  str(y)
  stopifnot(identical(y, y_truth))

  ## (b) explicit and with '...'
  sub <- function(x, ...) {
    foreach(i = 1:2) %do% { x[c(i, ...)] } |> futurize(globals = c("x", "..."))
  }
  y <- sub(x, 2:3)
  str(y)
  stopifnot(identical(y, y_truth))

  ## (c) with '...', but not last
  sub <- function(x, ...) {
    foreach(i = 1:2) %do% { x[c(i, ...)] } |> futurize(globals = c("...", "x"))
  }
  y <- sub(x, 2:3)
  str(y)
  stopifnot(identical(y, y_truth))
  
  ## (d) explicit, but forgotten '...'
  sub <- function(x, ...) {
    foreach(i = 1:2) %do% { x[c(i, ...)] } |> futurize(globals = c("x"))
  }
  y <- tryCatch(sub(x, 2:3), error = identity)
  str(y)
  stopifnot(inherits(y, "simpleError"))

  # Shutdown current plan
  plan(sequential)

  message(sprintf("- plan('%s') ... DONE", strategy))
} ## for (strategy ...)

message("*** doFuture - explicitly exported globals ... DONE")


message("*** doFuture - automatically finding globals ...")

## Example adopted from StackOverflow comment
## https://stackoverflow.com/a/10114192/1072091
fibonacci <- function(n) {
  if (n <= 1L) return(1L)
  return(fibonacci(n - 1L) + fibonacci(n - 2L))
}
X <- matrix(1:4, nrow = 3L, ncol = 4L)
y_truth <- foreach(rr = 1:nrow(X)) %do% {
  Vectorize(fibonacci)(X[rr, ])
}
str(y_truth)

res0 <- NULL

for (strategy in strategies) {
  message(sprintf("- plan('%s') ...", strategy))
  plan(strategy)

  y <- foreach(rr = 1:nrow(X)) %do% {
    Vectorize(fibonacci)(X[rr, ])
  } |> futurize()
  str(y)
  stopifnot(all.equal(y, y_truth))

  # Shutdown current plan
  plan(sequential)

  message(sprintf("- plan('%s') ... DONE", strategy))
} ## for (strategy ...)


message("*** doFuture - automatically finding globals ... DONE")


message("*** doFuture - automatically finding globals in 'args_list' ...")

library("tools") ## toTitleCase()

a <- 42
b <- 21
F <- list(
  function() b / 2,
  function(b) 2 * a,
  function() a + b
)
G <- list(
  function(b) 2 * a,
  function() b / 2,
  function() nchar(toTitleCase("hello world"))
)
z0 <- foreach(f = F, g = G) %do% list(f(), g())
str(z0)

for (strategy in strategies) {
  message(sprintf("- plan('%s') ...", strategy))
  plan(strategy)
  
  message("- foreach(f = X, ...) - 'f' containing globals ...")
  ## From https://github.com/futureverse/future.apply/issues/12
  z1 <- foreach(f = F, g = G) %do% list(f(), g()) |> futurize()
  str(z1)
  stopifnot(identical(z1, z0))

  # Shutdown current plan
  plan(sequential)
} ## for (strategy ...)


message("*** doFuture - automatically finding globals in 'args_list' ... DONE")

plan(sequential)
} ## if (requireNamespace("foreach") && requireNamespace("doFuture"))

#' @tags pkg-future.apply
#' @tags future_apply
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)

message("*** future_apply() ...")

z0 <- NULL

for (strategy in supportedStrategies()) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)

  message("- From example(apply) ...")
  X <- matrix(c(1:4, 1, 6:8), nrow = 2L)
  
  Y0 <- apply(X, MARGIN = 1L, FUN = table)
  Y1 <- apply(X, MARGIN = 1L, FUN = table) |> futurize()
  print(Y1)
  stopifnot(all.equal(Y1, Y0, check.attributes = FALSE)) ## FIXME

  Y2 <- apply(X, MARGIN = 1L, FUN = "table") |> futurize()
  print(Y2)
  stopifnot(identical(Y2, Y1))

  Y0 <- apply(X, MARGIN = 1L, FUN = stats::quantile)
  Y1 <- apply(X, MARGIN = 1L, FUN = stats::quantile) |> futurize()
  print(Y1)
  stopifnot(all.equal(Y1, Y0))

  x <- cbind(x1 = 3, x2 = c(4:1, 2:5))
  names(dimnames(x)) <- c("row", "col")
  x3 <- array(x, dim = c(dim(x), 3),
              dimnames = c(dimnames(x), list(C = paste0("cop.", 1:3))))
  
  y0 <- apply(x, MARGIN = 2L, FUN = identity)
  stopifnot(identical(y0, x))
  y1 <- apply(x, MARGIN = 2L, FUN = identity) |> futurize()
  print(y1)
  stopifnot(identical(y1, y0))
  
  y0 <- apply(x3, MARGIN = 2:3, FUN = identity)
  stopifnot(identical(y0, x3))
  y1 <- apply(x3, MARGIN = 2:3, FUN = identity) |> futurize()
  print(y1)
  stopifnot(identical(y1, y0))

  z <- array(1:24, dim = 2:4)
  y0 <- apply(z, MARGIN = 1:2, FUN = function(x) seq_len(max(x)))
  y1 <- apply(z, MARGIN = 1:2, FUN = function(x) seq_len(max(x))) |> futurize()
  print(y1)
  stopifnot(identical(y1, y0))

  message("- apply(X, MARGIN = <character>, ...) ...")
  X <- matrix(1:2, nrow = 2L, ncol = 1L, dimnames = list(rows = c("a", "b")))
  y0 <- apply(X, MARGIN = "rows", FUN = identity)
  y1 <- apply(X, MARGIN = "rows", FUN = identity) |> futurize()
  print(y1)
  stopifnot(identical(y1, y0))

  message("- apply(X, ...) - dim(X) > 2 ...")
  X <- array(1:12, dim = c(2, 2, 3))
  y0 <- apply(X, MARGIN = 1L, FUN = identity)
  y1 <- apply(X, MARGIN = 1L, FUN = identity) |> futurize()
  print(y1)
  stopifnot(identical(y1, y0))

  message("- apply(X, ...) - not all same names ...")
  FUN <- function(x) {
    if (x[1] == 1L) names(x) <- letters[seq_along(x)]
    x
  }
  X <- matrix(1:4, nrow = 2L, ncol = 2L)
  y0 <- apply(X, MARGIN = 1L, FUN = FUN)
  y1 <- apply(X, MARGIN = 1L, FUN = FUN) |> futurize()
  print(y1)
  stopifnot(identical(y1, y0))

  message("- example(future_apply) - reproducible RNG ...")
  z1 <- apply(X, MARGIN = 1L, FUN = sample) |>
          futurize(
            seed = 0xBEEF,
            ## Test also all other '*' arguments
            stdout     = TRUE,
            conditions = NULL,
            globals    = TRUE,
            packages   = NULL,
            scheduling = 1.0,
            chunk_size = NULL,
            label      = "future_apply-%d"
          )
  print(z1)
  if (is.null(z0)) {
    z0 <- z1
  } else {
    stopifnot(identical(z1, z0))
  }

  plan(sequential)
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
} ## for (strategy in ...) 


message("*** apply(X, ...) - prod(dim(X)) == 0 [non-parallel] ...")
X <- matrix(nrow = 0L, ncol = 2L)
y0 <- apply(X, MARGIN = 1L, FUN = identity)
y1 <- apply(X, MARGIN = 1L, FUN = identity) |> futurize()
print(y1)
stopifnot(identical(y1, y0))
  

message("*** exceptions ...")

## Error: dim(X) must have a positive length
res <- tryCatch({
  y <- apply(1L, MARGIN = 1L, FUN = identity) |> futurize()
}, error = identity)
stopifnot(inherits(res, "error"))

## Error: 'X' must have named dimnames
X <- matrix(1:2, nrow = 2L, ncol = 1L)
res <- tryCatch({
  y <- apply(X, MARGIN = "rows", FUN = identity) |> futurize()
}, error = identity)
stopifnot(inherits(res, "error"))

## Error: not all elements of 'MARGIN' are names of dimensions
X <- matrix(1:2, nrow = 2L, ncol = 1L, dimnames = list(rows = c("a", "b")))
res <- tryCatch({
  y <- apply(X, MARGIN = "cols", FUN = identity) |> futurize()
}, error = identity)
stopifnot(inherits(res, "error"))


message("*** future_apply() ... DONE")

} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags pkg-BiocParallel
if (requireNamespace("BiocParallel") && requireNamespace("doFuture")) {
library(futurize)
library(BiocParallel)
options(future.rng.onMisuse = "error")

plan(multisession)

y <- bplapply(1:3, function(x) { print(x) }) |> futurize_and_verify(stdout = FALSE)
print(y)

xs <- list(aa = 1, bb = 1:2, cc = 1:10, dd = 1:5, .ee = -6:6)
FUN_no_rng <- function(x, na.rm = TRUE) {
  a <- 1:5
  add <- NULL
  if (length(x) == 2) add <- list(C = 42)
  median(c(a, x), na.rm = na.rm)
}

FUN_rng <- function(x, na.rm = TRUE) {
  dummy <- sample.int(10L)
  a <- 1:5
  add <- NULL
  if (length(x) == 2) add <- list(C = 42)
  invisible(median(c(a, x), na.rm = na.rm))
}

es <- as.environment(xs)

exprs <- list(
  bplapply = quote(bplapply(xs, FUN) ),
  bplapply = quote(BiocParallel::bplapply(xs, FUN) ),
  bpmapply = quote(bpmapply(FUN, xs) ),
  bpmapply = quote(BiocParallel::bpmapply(FUN, xs) )
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))

  FUN <- FUN_no_rng
  truth <- eval(expr)
  expr_f <- bquote(.(expr) |> futurize_and_verify())

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> futurize_and_verify(stdout = FALSE, conditions = character(0L)))
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> futurize_and_verify(chunk_size = 1L))
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))

  message("Test with RNG:")
  FUN <- FUN_rng
  expr_f4 <- bquote(.(expr) |> futurize_and_verify(seed = TRUE))
  print(expr_f4)
  res4 <- local({
    opts <- options(future.rng.onMisuse = "error")
    on.exit(options(opts))
    eval(expr_f4)
  })
  stopifnot(
    identical(res4, res),
    identical(res4, truth)
  )
}

plan(sequential)
} ## if (requireNamespace("BiocParallel"))

get_connections <- function() {
  cons <- lapply(getAllConnections(), FUN = function(idx) {
    tryCatch({
      con <- getConnection(idx)
      as.data.frame(c(index = idx, summary(con)))
    }, error = function(e) {
      NULL
    })
  })
  do.call(rbind, cons)
}

diff_connections <- function(after, before) {
  index <- NULL ## To please R CMD check
  
  ## Nothing to do?
  if (length(before) + length(after) == 0L) {
    return(c(added = NULL, removed = NULL, replaced = NULL))
  }

  idxs <- setdiff(after[["index"]], before[["index"]])
  if (length(idxs) > 0) {
    added <- subset(after, index %in% idxs)
    after <- subset(after, ! index %in% idxs)
  } else {
    added <- NULL
  }
  
  idxs <- setdiff(before[["index"]], after[["index"]])
  if (length(idxs) > 0) {
    removed <- subset(before, index %in% idxs)
    before <- subset(before, ! index %in% idxs)
  } else {
    removed <- NULL
  }

  idxs <- intersect(before[["index"]], after[["index"]])
  if (length(idxs) > 0) {
    replaced <- list()
    for (idx in idxs) {
      before_idx <- subset(before, index == idx)
      after_idx <- subset(after, index == idx)
      if (!identical(before_idx, after_idx)) {
        for (name in colnames(after_idx)) {
          value <- after_idx[[name]]
          if (!identical(before_idx[[name]], value)) {
            value <- sprintf("%s (was %s)", value, before_idx[[name]])
            after_idx[[name]] <- value
          }
        }
        replaced <- c(replaced, list(after_idx))
      }
    }
    replaced <- do.call(rbind, replaced)
  } else {
    replaced <- NULL
  }

  list(added = added, removed = removed, replaced = replaced)
}

testme <- as.environment("testme")
testme[["testme_connections"]] <- get_connections()

#' @tags pkg-fgsea
#' @tags skip_on_cran  ## (20s) to limit total check time
if (requireNamespace("fgsea") && requireNamespace("doFuture")) {
library(futurize)
library(fgsea)
options(future.rng.onMisuse = "error")

plan(multisession)

## Create simple test data
set.seed(42)
n_genes <- 100L
stats <- rnorm(n_genes)
names(stats) <- paste0("gene", seq_len(n_genes))

pathways <- list(
  pathway1 = paste0("gene", sample(n_genes, 10L)),
  pathway2 = paste0("gene", sample(n_genes, 15L)),
  pathway3 = paste0("gene", sample(n_genes, 20L))
)

## ---------------------------------------------------------
## fgseaSimple()
## ---------------------------------------------------------
set.seed(42)
result_truth <- fgseaSimple(pathways, stats, nperm = 10000)
print(result_truth)

set.seed(42)
result <- fgseaSimple(pathways, stats, nperm = 10000) |> futurize_and_verify()
print(result)
stopifnot(all.equal(result$pval, result_truth$pval))

set.seed(42)
result2 <- fgsea::fgseaSimple(pathways, stats, nperm = 10000) |> futurize_and_verify()
stopifnot(all.equal(result2$pval, result_truth$pval))

## ---------------------------------------------------------
## fgsea() via fgseaMultilevel()
## ---------------------------------------------------------
## To test fgsea(), we need a large enough data set with
## strong enrichment signals so that fgseaMultilevel() needs
## multilevel refinement. Hence the a bit complicated
## simulation set up.
set.seed(42)
n_genes2 <- 500L
stats2 <- rnorm(n_genes2)
names(stats2) <- paste0("gene", seq_len(n_genes2))

## Give pathway genes strong positive signals
p1 <- paste0("gene", 1:15)
p2 <- paste0("gene", 31:55)
p3 <- paste0("gene", 71:105)
p4 <- paste0("gene", 121:170)
stats2[p1] <- stats2[p1] + 4
stats2[p2] <- stats2[p2] + 3
stats2[p3] <- stats2[p3] + 3
stats2[p4] <- stats2[p4] + 3

pathways2 <- list(
  pathway1 = p1,
  pathway2 = p2,
  pathway3 = p3,
  pathway4 = p4
)

set.seed(42)
result_truth <- fgsea(pathways2, stats2)
print(result_truth)

set.seed(42)
result <- fgsea(pathways2, stats2) |> futurize_and_verify()
print(result)
stopifnot(all.equal(result$pval, result_truth$pval))

set.seed(42)
result2 <- fgsea::fgsea(pathways2, stats2) |> futurize_and_verify()
stopifnot(all.equal(result2$pval, result_truth$pval))

plan(sequential)
} ## if (requireNamespace("fgsea") && ...)

#' @tags pkg-furrr
if (requireNamespace("purrr") && requireNamespace("furrr")) {
library(futurize)


library(purrr)

message("walk functions work")

x <- 1:5
out <- walk(x, ~"hello") |> futurize_and_verify()
stopifnot(identical(out, x))

y <- 6:10
out <- walk2(x, y, ~"hello") |> futurize_and_verify()
stopifnot(identical(out, x))

l <- list(x, y)
out <- pwalk(list(x, y), ~"hello") |> futurize_and_verify()
stopifnot(identical(out, l))

out <- iwalk(x, ~"hello") |> futurize_and_verify()
stopifnot(identical(out, x))

plan(sequential)
} ## if (requireNamespace("purrr") && requireNamespace("furrr"))

#' @tags pkg-future.apply
#' @tags future_apply
#' @tags globals
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)

## Test adopted from http://stackoverflow.com/questions/42561088/nested-do-call-within-a-foreach-dopar-environment-cant-find-function-passed-w

options(future.debug = FALSE)

message("*** Tricky globals requiring recursive search ...")

my_add <- function(a, b) a + b

call_my_add <- function(a, b) {
  do.call(my_add, args = list(a = a, b = b))
}

call_my_add_caller <- function(a, b, FUN = call_my_add) {
  do.call(FUN, args = list(a = a, b = b))
}

main_lapply <- function(x = 1:2, caller = call_my_add_caller,
                               args = list(FUN = call_my_add)) {
  lapply(x, FUN = function(i) {
    do.call(caller, args = c(list(a = i, b = i+1L), args))
  })
}

main_lapply_no_FUN <- function(x = 1:2, caller = call_my_add_caller,
                               args = list(FUN = call_my_add)) {
  lapply(x, FUN = function(i) {
    do.call(caller, args = list(a = i, b = i+1L))
  })
}

main_future_lapply <- function(x = 1:2, caller = call_my_add_caller,
                               args = list(FUN = call_my_add)) {
  lapply(x, FUN = function(i) {
    do.call(caller, args = c(list(a = i, b = i + 1L), args))
  }) |> futurize()
}

main_future_lapply_no_FUN <- function(x = 1:2, caller = call_my_add_caller,
                               args = list(FUN = call_my_add)) {
  lapply(x, FUN = function(i) {
    do.call(caller, args = list(a = i, b = i + 1L))
  }) |> futurize()
}

x0 <- y0 <- z0 <- NULL
for (strategy in supportedStrategies()) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  
  plan(strategy)

  z <- main_lapply()
  str(list(z = z))
  if (is.null(z0)) z0 <- z
  stopifnot(identical(z, z0))

  z2 <- main_lapply_no_FUN()
  str(list(z2 = z2))
  stopifnot(identical(z2, z0))
  
  z3 <- main_future_lapply()
  str(list(z3 = z3))
  stopifnot(identical(z3, z0))

  z4 <- main_future_lapply_no_FUN()
  str(list(z4 = z4))
  stopifnot(identical(z4, z0))
  
  message(sprintf("*** strategy = %s ... DONE", sQuote(strategy)))
}

message("*** Tricky globals requiring recursive search ... DONE")



} ## if (requireNamespace("future.apply", quietly = TRUE))

# parameters::bootstrap_model(...) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   parameters::bootstrap_model(..., parallel = "snow", ncpus = 2L, cluster = cl)
# })
#
# parameters::bootstrap_parameters(...) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   parameters::bootstrap_parameters(..., parallel = "snow", ncpus = 2L, cluster = cl)
# })
#
append_transpilers_for_parameters <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'parameters' functions requires R (>= 4.4.0)", getRversion()))
  }

  template_ignore_clusterEvalQ <- bquote_compile(
    local({
      cl <- do.call(.(CALL), args = .(OPTS))
      oopts <- options(future.ClusterFuture.clusterEvalQ = "ignore")
      on.exit(options(oopts))
      .(EXPR)
    })
  )

  transpilers <- make_package_transpilers("parameters", FUN = function(fcn, name) {
    if (name %in% c("bootstrap_model", "bootstrap_parameters")) {
      transpiler <- make_futurize_for_makeClusterFuture(
        template = template_ignore_clusterEvalQ,
        args = list(
          parallel = "snow",
          n_cpus = 2L,   ## only used for test ncpus > 1
          cluster = quote(cl)
        ), defaults = list(
          label = sprintf("fz:parameters::%s", name),
          seed = TRUE
        )
      )

      list(
        label = sprintf("parameters::%s() ~> parameters::%s(..., parallel = 'snow', cluster = cl)",  name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("parameters", "future")
}

#' @tags skip_on_cran  ## (35s) to limit total check time
if (requireNamespace("glmmTMB") && getRversion() >= "4.4.0") {
library(futurize)
library(glmmTMB)
options(future.rng.onMisuse = "error")

plan(multisession)

## Adopted from example("confint.glmmTMB", package = "glmmTMB")
m <- glmmTMB(count ~ mined + (1 | site), data = Salamanders, family = nbinom2)

message("Ordinary processing:")
pr_truth <- profile(m)
print(head(pr_truth))

message("Futurized processing:")
pr <- profile(m) |> futurize_and_verify()
print(head(pr))
message("Comparing results:")
stopifnot(all.equal(pr, pr_truth))

## Skip confint() for 'glmmTMB' until has been fixed per
## https://github.com/glmmTMB/glmmTMB/issues/1268
if (FALSE) {
  ## confint.glmmTMB() via stats::confint() S3 generic dispatch
  message("Ordinary processing:")
  ci_truth <- confint(m, method = "profile")
  print(ci_truth)
  
  message("Futurized processing:")
  ci <- confint(m, method = "profile") |> futurize_and_verify()
  print(ci)
  message("Comparing results:")
  stopifnot(all.equal(ci, ci_truth))
}

plan(sequential)
} ## if (requireNamespace("glmmTMB"))

#' Check whether function is an S3 generic function
#'
#' @param fcn A function.
#'
#' @return
#' Return TRUE if `fcn` is an S3 generic function, otherwise FALSE.
#'
#' @noRd
is_s3_generic <- local({
  has_UseMethod <- function(expr) {
    if (is.call(expr)) {
      first <- expr[[1]]
      if (is.symbol(first) && identical(first, as.symbol("UseMethod"))) {
        return(TRUE)
      }
      for (ii in seq_along(expr)) {
        if (has_UseMethod(expr[[ii]])) return(TRUE)
      }
    }
    FALSE
  } ## has_UseMethod()

  function(fcn) {
    if (is.primitive(fcn)) return(FALSE)
    bd <- body(fcn)
    if (is.null(bd)) return(FALSE)
    has_UseMethod(bd)
  }
}) ## is_s3_generic()

#' @tags pkg-fwb
if (requireNamespace("fwb") && requireNamespace("boot")) {
  library(futurize)
  library(fwb)
  city <- boot::city
  options(future.rng.onMisuse = "error")

  all_equal <- function(a, b, ...) {
    a$call <- b$call <- NULL
    all.equal(a, b, ...)
  }

  plan(multisession)

  ## Adopted from example("boot", package = "boot")
  ratio <- function(d, w) {
    sum(d$x * w)/sum(d$u * w)
  }

  set.seed(42)
  b_truth <- fwb(city, ratio, R = 19L, verbose = FALSE, simple = FALSE)
  print(b_truth)

  set.seed(42)
  b <- fwb(city, ratio, R = 19L, verbose = FALSE, simple = FALSE) |> futurize_and_verify()
  print(b)

  set.seed(42)
  RNGkind("L'Ecuyer-CMRG")
  b2 <- fwb(city, ratio, R = 19L, verbose = FALSE, simple = FALSE, cl = "future")
  print(b2)

  stopifnot(all_equal(b, b2, check.attributes = FALSE))

  #vcovFWB
  fit <- lm(u ~ x, data = city)

  set.seed(42)
  v_truth <- vcovFWB(fit, R = 19L)
  print(v_truth)

  set.seed(42)
  v <- vcovFWB(fit, R = 19L) |> futurize_and_verify()
  print(v)

  #v and v_truth should differ if parallelization is done
  stopifnot(!isTRUE(all.equal(v_truth, v)))

  set.seed(42)
  RNGkind("L'Ecuyer-CMRG")
  v2 <- vcovFWB(fit, R = 19L, cl = "future")
  print(v2)

  #v and v2 should not differ if futurize_and_verify() engages cl = "future"
  stopifnot(all.equal(v, v2))

  plan(sequential)
} ## if (requireNamespace("fwb") && requireNamespace("boot"))

#' @tags pkg-doFuture
#' @tags %dofuture%
#' @tags sequential multisession cluster multicore

if (requireNamespace("foreach") && requireNamespace("doFuture")) {
library(futurize)
library(foreach)

message("*** Options in nested parallelization ...")

options(future.debug = FALSE)
options(future.apply.debug = FALSE)
options(future.globals.maxSize = 1234000)

for (cores in 1:availCores) {
  message(sprintf("Testing with %d cores ...", cores))
  options(mc.cores = cores)
  strategies <- supportedStrategies(cores)

  for (strategy1 in strategies) {
    for (strategy2 in strategies) {
      message(sprintf("- plan('%s') ...", strategy2))
      plan(list(outer = tweak(strategy1), inner = strategy2))

      v <- foreach(x = 1:2) %do% {
        outer <- data.frame(
          label   = "outer",
          idx     = x,
          pid     = Sys.getpid(),
          maxSize = getOption("future.globals.maxSize", NA_real_)
        )
        
        inner <- foreach(x = 3:4) %do% {
          data.frame(
            label   = "inner",
            idx     = x,
            pid     = Sys.getpid(),
            maxSize = getOption("future.globals.maxSize", NA_real_))
        } |> futurize()
        inner <- do.call(rbind, inner)
        rbind(outer, inner)
      } |> futurize()
      v <- do.call(rbind, v)
      print(v)
      stopifnot(!anyNA(v$maxSize))
    } ## for (strategy2 ...)
  } ## for (strategy1 ...)
} ## for (cores in ...)

message("*** Options in nested parallelization ... done")

plan(sequential)
} ## if (requireNamespace("foreach") && requireNamespace("doFuture"))

#' @tags pkg-future.apply
#' @tags future_tapply
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)

library("datasets") ## warpbreaks, iris

options(future.debug = FALSE)
message("*** future_tapply() ...")

for (strategy in supportedStrategies()[1]) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)

  message("- From example(tapply) ...")

  message("  - Example #1")
  library("stats")  ## rbinom()
  groups <- as.factor(stats::rbinom(32, n = 5, prob = 0.4))
  t <- table(groups)
  print(t)
  y0 <- tapply(groups, INDEX = groups, FUN = length)
  print(y0)
  y1 <- tapply(groups, INDEX = groups, FUN = length) |> futurize()
  print(y1)
  stopifnot(all.equal(y1, y0))
  y2 <- tapply(groups, INDEX = groups, FUN = "length") |> futurize()
  print(y2)
  stopifnot(all.equal(y2, y0))

  message("  - Example #2")
  ## contingency table from data.frame : array with named dimnames
  y0 <- tapply(warpbreaks$breaks, INDEX = warpbreaks[,-1], FUN = sum)
  print(y0)
  y1 <- tapply(warpbreaks$breaks, INDEX = warpbreaks[,-1], FUN = sum) |> futurize()
  print(y1)
  stopifnot(all.equal(y1, y0))

  message("  - Example #3")
  y0 <- tapply(warpbreaks$breaks, warpbreaks[, 3, drop = FALSE], sum)
  print(y0)
  y1 <- tapply(warpbreaks$breaks, warpbreaks[, 3, drop = FALSE], sum) |> futurize()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #4")
  n <- 17
  fac <- factor(rep_len(1:3, n), levels = 1:5)
  t <- table(fac)  
  y0 <- tapply(1:n, fac, sum)
  print(y0)
  y1 <- tapply(1:n, fac, sum) |> futurize()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #5")
  if ("default" %in% names(formals(tapply))) {
    y0 <- tapply(1:n, fac, sum, default = 0) # maybe more desirable
    print(y0)
    y1 <- tapply(1:n, fac, sum, default = 0) |> futurize() # maybe more desirable
    print(y1)
    stopifnot(all.equal(y1, y0))
  }
  
  message("  - Example #6")
  y0 <- tapply(1:n, fac, sum, simplify = FALSE)
  print(y0)
  y1 <- tapply(1:n, fac, sum, simplify = FALSE) |> futurize()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #7")
  y0 <- tapply(1:n, fac, range)
  print(y0)
  y1 <- tapply(1:n, fac, range) |> futurize()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #8")
  y0 <- tapply(1:n, fac, quantile)
  print(y0)
  y1 <- tapply(1:n, fac, quantile) |> futurize()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #9")
  y0 <- tapply(1:n, fac, length) ## NA's
  print(y0)
  y1 <- tapply(1:n, fac, length) |> futurize() ## NA's
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #10")
  if ("default" %in% names(formals(tapply))) {
    y0 <- tapply(1:n, fac, length, default = 0) # == table(fac)
    print(y0)
    y1 <- tapply(1:n, fac, length, default = 0) |> futurize() # == table(fac)
    print(y1)
    stopifnot(all.equal(y1, y0))
  }
  
  message("  - Example #11")
  ## example of ... argument: find quarterly means
  y0 <- tapply(presidents, cycle(presidents), mean, na.rm = TRUE)
  print(y0)
  y1 <- tapply(presidents, cycle(presidents), mean, na.rm = TRUE) |> futurize()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #12")
  ind <- list(c(1, 2, 2), c("A", "A", "B"))
  t <- table(ind)
  print(t)
  y0 <- tapply(1:3, ind) #-> the split vector
  print(y0)
  y1 <- tapply(1:3, ind) |> futurize() #-> the split vector
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  message("  - Example #13")
  y0 <- tapply(1:3, ind, sum)
  print(y0)
  y1 <- tapply(1:3, ind, sum) |> futurize()
  print(y1)
  stopifnot(all.equal(y1, y0))
  
  ## Some assertions (not held by all patch propsals):
  message("  - Example #14")
  nq <- names(quantile(1:5))
  y_truth <- c(1L, 2L, 4L)
  stopifnot(identical(tapply(1:3, ind), y_truth))
  stopifnot(identical(tapply(1:3, ind) |> futurize(), y_truth))
  
  message("  - Example #15")
  y_truth <- matrix(c(1L, 2L, NA, 3L), nrow = 2L,
                    dimnames = list(c("1", "2"), c("A", "B")))
  stopifnot(identical(tapply(1:3, ind, sum), y_truth))
  stopifnot(identical(tapply(1:3, ind, sum) |> futurize(), y_truth))
  
  message("  - Example #16")
  y_truth <- array(list(
    `2` = structure(c(2, 5.75, 9.5, 13.25, 17), .Names = nq),
    `3` = structure(c(3, 6, 9, 12, 15), .Names = nq),
    `4` = NULL, `5` = NULL),
    dim = 4L, dimnames = list(as.character(2:5)))
  stopifnot(identical(tapply(1:n, fac, quantile)[-1], y_truth))
  stopifnot(identical((tapply(1:n, fac, quantile) |> futurize())[-1], y_truth))

  if (getRversion() >= "4.3.0") {
    data <- iris[, c("Sepal.Length", "Sepal.Width")]
    y_truth <- tapply(data, INDEX = iris$Species, FUN = sum)
    y <- tapply(data, INDEX = iris$Species, FUN = sum) |> futurize()
    stopifnot(identical(y, y_truth))
    
    y_truth2 <- tapply(data, INDEX = ~ iris$Species + iris$Petal.Width, FUN = sum)
    y2 <- tapply(data, INDEX = ~ iris$Species + iris$Petal.Width, FUN = sum) |> futurize()
    stopifnot(identical(y2, y_truth2))
  }

  plan(sequential)
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
} ## for (strategy in ...) 


message("*** exceptions ...")

## Error: 'INDEX' is of length zero
res <- tryCatch({
  y <- tapply(1L, INDEX = list()) |> futurize()
}, error = identity)
stopifnot(inherits(res, "error"))

## Error: total number of levels >= 2^31
res <- tryCatch({
  y <- tapply(1:216, INDEX = rep(list(1:216), times = 4L)) |> futurize()
}, error = identity)
stopifnot(inherits(res, "error"))

## Error: arguments must have same length
res <- tryCatch({
  y <- tapply(1L, INDEX = list(1:2)) |> futurize()
}, error = identity)
stopifnot(inherits(res, "error"))

message("*** future_tapply() ... DONE")

} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags skip_on_cran  ## (20s) to limit total check time
if (requireNamespace("sva") && requireNamespace("doFuture")) {
library(futurize)
library(sva)
print(sessionInfo())

options(future.rng.onMisuse = "error")

plan(multisession)

## Create simple test data for ComBat
set.seed(42)
n_genes <- 50L
n_samples <- 20L
dat <- matrix(rnorm(n_genes * n_samples), nrow = n_genes, ncol = n_samples)
rownames(dat) <- paste0("gene", seq_len(n_genes))
colnames(dat) <- paste0("sample", seq_len(n_samples))

## Add batch effect
batch <- rep(c(1, 2), each = n_samples / 2L)
dat[, batch == 2] <- dat[, batch == 2] + 2

## ---------------------------------------------------------
## ComBat()
## ---------------------------------------------------------
set.seed(42)
result_truth <- ComBat(dat = dat, batch = batch)
str(result_truth)

set.seed(42)
result <- ComBat(dat = dat, batch = batch) |> futurize_and_verify()
str(result)
stopifnot(all.equal(result, result_truth))

set.seed(42)
result2 <- sva::ComBat(dat = dat, batch = batch) |> futurize_and_verify()
stopifnot(all.equal(result2, result_truth))

## ---------------------------------------------------------
## ComBat() with mod
## ---------------------------------------------------------
## Covariate must not be confounded with batch
group <- rep(c("A", "B"), times = n_samples / 2L)
mod <- model.matrix(~ group)

set.seed(42)
result_truth <- ComBat(dat = dat, batch = batch, mod = mod)
str(result_truth)

set.seed(42)
result <- ComBat(dat = dat, batch = batch, mod = mod) |> futurize_and_verify()
str(result)
stopifnot(all.equal(result, result_truth))

plan(sequential)
} ## if (requireNamespace("sva") && ...)

#' @tags pkg-seriation
if (requireNamespace("seriation") && requireNamespace("doFuture")) {
library(futurize)
library(seriation)
options(future.rng.onMisuse = "error")

plan(multisession)

data(SupremeCourt)
d_supreme <- as.dist(SupremeCourt)

message("*** seriate_rep()")

RNGkind("L'Ecuyer-CMRG")
set.seed(42)
o_truth <- seriate_rep(d_supreme, "QAP_LS", rep = 5L)
print(o_truth)
set.seed(42)
o <- seriate_rep(d_supreme, "QAP_LS", rep = 5L) |> futurize_and_verify()
print(o)

## seriate_rep() does not allow for RNG reproducibility
o[[1]] <- o_truth[[1]]
stopifnot(all.equal(o, o_truth, tolerance = 0.01))

message("*** seriate_best()")

RNGkind("L'Ecuyer-CMRG")
set.seed(42)
o_truth <- seriate_best(d_supreme, criterion = "AR_events")
print(o_truth)
set.seed(42)
o <- seriate_best(d_supreme, criterion = "AR_events") |> futurize_and_verify()
print(o)
attr(o, "time") <- attr(o_truth, "time")
stopifnot(all.equal(o, o_truth))

plan(sequential)
} ## if (requireNamespace("seriation"))

#' Turn common R function calls into concurrent calls for parallel evaluation
#'
#' \if{html}{\figure{futurize-logo.png}{options: style='float: right;' alt='logo' width='120'}}
#'
#' @inheritParams future::future
#'
#' @param expr An \R expression, typically a function call to futurize.
#' If FALSE, then futurization is disabled, and if TRUE, it is
#' re-enabled.
#'
#' @param options,\ldots Named options, passed to [futurize_options()],
#' controlling how futures are resolved.
#' 
#' @param when If TRUE (default), the expression is futurized, otherwise not.
#'
#' @param eval If TRUE (default), the futurized expression is evaluated,
#' otherwise it is returned.
#'
#' @returns
#' Returns the value of the evaluated expression `expr`.
#'
#' If `expr` is TRUE or FALSE, then a logical is returned indicating 
#' whether futurization was previously enabled or disabled.
#'
#'
#' @section Expression unwrapping:
#' The transpilation mechanism includes logic to "unwrap" expressions
#' enclosed in constructs such as `!`, `{ }`, `( )`, `local()`, `I()`,
#' `identity()`, `invisible()`, `suppressMessages()`, `suppressWarnings()`,
#' and `with()`. The transpiler descends through wrapping
#' constructs until it finds a transpilable expression, avoiding the
#' need to place `futurize()` inside such constructs. This allows for
#' patterns like:
#'
#' ```r
#' y <- {
#'   lapply(xs, fcn)
#' } |> suppressMessages() |> futurize()
#' ```
#'
#' avoiding having to write:
#'
#' ```r
#' y <- {
#'   lapply(xs, fcn) |> futurize()
#' } |> suppressMessages()
#' ```
#'
#'
#' @section Conditional futurization:
#' It is possible to control whether futurization should take place at
#' run-time. For example,
#'
#' ```r
#' y <- lapply(xs, fun) |> futurize(when = { length(xs) >= 10 })
#' ```
#'
#' will be futurized, unless `length(xs)` is less than ten, in which case it is
#' evaluated as:
#'
#' ```r
#' y <- lapply(xs, fun)
#' ```
#'
#'
#' @section Disable and re-enable all futurization:
#' It is possible to globally disable the effect of all `futurize()` calls
#' by calling `futurize(FALSE)`. The effect is as if `futurize()` was never
#' applied. For example,
#'
#' ```r
#' futurize(FALSE)
#' y <- lapply(xs, fun) |> futurize()
#' ```
#'
#' evaluates as:
#'
#' ```r
#' y <- lapply(xs, fun)
#' ```
#'
#' To re-enable futurization, call `futurize(TRUE)`.
#' Please note that it is only the end-user that may control whether
#' futurization should be disabled and enabled. A package must _never_
#' disable or enable futurization.
#'
#'
#' @example incl/futurize.R
#'
#' @seealso
#' To see which CRAN and Bioconductor packages are supported, use
#' [futurize_supported_packages()].
#' To see which functions a specific package supports, use
#' [futurize_supported_functions()].
#'
#' @aliases fz
#' @export
futurize <- function(expr, substitute = TRUE, options = futurize_options(...), ..., when = TRUE, eval = TRUE, envir = parent.frame()) {
  if (substitute) expr <- substitute(expr)
  debug <- isTRUE(getOption("futurize.debug"))
  if (debug) {
    mdebug_push("futurize() ...")
    on.exit(mdebug_pop())
  }

  disable <- !isTRUE(getOption("futurize.enable", TRUE))
  
  transpile(expr, substitute = FALSE, options = options, when = when, eval = eval, disable = disable, type = "futurize::add-on", envir = envir, what = "futurize", debug = debug)
} ## futurize()
class(futurize) <- c("transpiler", class(futurize))


#' @export
fz <- futurize

#' @tags pkg-doFuture
#' @tags %dopar%
#' @tags sequential multisession cluster multicore

if (requireNamespace("foreach") && requireNamespace("doFuture")) {
library(futurize)
library(foreach)
`%dofuture%` <- doFuture::`%dofuture%`

strategies <- future:::supportedStrategies()

message("*** times() ...")

res0 <- NULL

for (strategy in strategies) {
  message(sprintf("- plan('%s') ...", strategy))
  plan(strategy)

  mu <- 1.0
  sigma <- 2.0
  ## NOTE: Using times() with %dopar% without proper parallel RNG will likely
  ## produce unreliable results.  Here we don't produce random numbers so it
  ## is ok, but it's a toy example because just like base::replicate(),
  ## foreach::times() is commonly used for resampling purposes.
  res <- times(3L) %do% {
    dnorm(2L, mean = mu, sd = sigma)
  } |> futurize(seed = TRUE)
  print(res)

  if (is.null(res0)) {
    res0 <- res
  } else {
    stopifnot(all.equal(res, res0))
  }

  # Shutdown current plan
  plan(sequential)
  
  message(sprintf("- plan('%s') ... DONE", strategy))
} ## for (strategy ...)

message("*** times() ... DONE")

plan(sequential)
} ## if (requireNamespace("foreach") && requireNamespace("doFuture"))

#' @tags skip_on_cran  ## to limit total check time
#' @tags pkg-ez
if (requireNamespace("ez") && getRversion() >= "4.4.0") {
library(futurize)
library(ez)
options(future.rng.onMisuse = "error")

plan(multisession)

# Read in the ANT data and subset to make it fast
data(ANT, package = "ez")
ANT <- subset(ANT, subnum %in% levels(subnum)[c(1, 2, 11, 12)])

#------------------------------------------------------------------
# ezBoot() and ezPlot2()
#------------------------------------------------------------------
message("ezBoot() and ezPlot2() ...")

set.seed(42)
rt_truth <- ezBoot(
  data = ANT,
  dv = rt,
  wid = subnum,
  within = .(cue, flank),
  between = group,
  iterations = 3L
)

set.seed(42)
rt <- ezBoot(
  data = ANT,
  dv = rt,
  wid = subnum,
  within = .(cue, flank),
  between = group,
  iterations = 3L
) |> futurize_and_verify()

stopifnot(
  identical(class(rt), class(rt_truth)),
  identical(names(rt), names(rt_truth)),
  identical(dim(rt$cells), dim(rt_truth$cells))
)

message("ezPlot2() ...")
p_truth <- ezPlot2(
  preds = rt_truth,
  x = flank,
  split = cue,
  col = group,
  do_plot = FALSE
)

p <- ezPlot2(
  preds = rt,
  x = flank,
  split = cue,
  col = group,
  do_plot = FALSE
) |> futurize_and_verify()

stopifnot(
  identical(class(p), class(p_truth)),
  identical(names(p), names(p_truth))
)

message("ezBoot() and ezPlot2() ... done")

#------------------------------------------------------------------
# ezPerm()
#------------------------------------------------------------------
message("ezPerm() ...")

# Compute some useful statistics per cell.
library(plyr)
cell_stats <- ddply(
  .data = ANT,
  .variables = .(subnum, group, cue, flank),
  .fun = function(x) {
    to_return <- data.frame(
      mrt = mean(x$rt[x$error == 0])
    )
    return(to_return)
  }
)

# Compute the grand mean RT per Ss.
gmrt <- ddply(
  .data = cell_stats,
  .variables = .(subnum, group),
  .fun = function(x) {
    to_return <- data.frame(
      mrt = mean(x$mrt)
    )
    return(to_return)
  }
)

set.seed(42)
perm_truth <- ezPerm(
  data = gmrt,
  dv = mrt,
  wid = subnum,
  between = group,
  perms = 3L
)

set.seed(42)
perm <- ezPerm(
  data = gmrt,
  dv = mrt,
  wid = subnum,
  between = group,
  perms = 3L
) |> futurize_and_verify()

# The exact p-values may differ due to different sequential vs. future RNG kinds (MT vs. L'Ecuyer-CMRG),
# but the structure, names, and dimensions should match exactly.
stopifnot(
  identical(class(perm), class(perm_truth)),
  identical(names(perm), names(perm_truth)),
  identical(dim(perm), dim(perm_truth))
)

message("ezPerm() ... done")

plan(sequential)
}

# vegan::mrpp(...) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   vegan::mrpp(..., cluster = cl)
# })
#
append_transpilers_for_vegan <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'vegan' functions requires R (>= 4.4.0)", getRversion()))
  }

  template <- bquote_compile(
    local({
      cl <- do.call(.(CALL), args = .(OPTS))
      
      ## WORKAROUND: https://github.com/vegandevs/vegan/issues/771
      base_attach <- base::attach # silence R CMD check
      base_attach(list(cl = cl), name = "vegan:patch")
      on.exit(detach("vegan:patch"))
      
      oopts <- options(future.ClusterFuture.clusterEvalQ = "error")
      on.exit(options(oopts), add = TRUE)
      .(EXPR)
    })
  )

  transpilers <- make_package_transpilers("vegan", FUN = function(fcn, name) {
    ## vegan 2.8.0: simper() ignores argument 'parallel' with warning
    ## "argument 'parallel' is not used (yet)"
    if (name %in% c("simper")) return()

    defaults <- list(label =  sprintf("fz:vegan::%s", name))
    if (name %in% ("cascadeKM")) defaults$seed <- TRUE

    if ("parallel" %in% names(formals(fcn))) {
      list(
        label = sprintf("vegan::%s() ~> vegan::%s(..., parallel = <cluster>)", name, name),
        transpiler = make_futurize_for_makeClusterFuture(
          args = list(
            parallel = quote(cl)
          ),
          defaults = defaults,
          template = if (name == "cascadeKM") template else NULL
        )
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("vegan", "future")
}

#' @tags skip_on_cran  ## to limit total check time
#' @tags pkg-gamlss
if (requireNamespace("gamlss") && requireNamespace("gamlss.data") && getRversion() >= "4.4.0") {
library(futurize)
library(gamlss)
options(future.rng.onMisuse = "ignore")

plan(multisession)

## -------------------------------------------------------------------
## gamlssCV() - k-fold cross-validation
## -------------------------------------------------------------------
## Adopted from help("gamlssCV", package = "gamlss")
data(abdom, package = "gamlss.data")

set.seed(42)
cv_truth <- gamlssCV(y ~ pb(x), data = abdom, K.fold = 5)
cat(sprintf("CV truth: %s\n", paste(cv_truth, collapse = ", ")))

set.seed(42)
cv <- gamlssCV(y ~ pb(x), data = abdom, K.fold = 5) |> futurize_and_verify()
cat(sprintf("CV futurized: %s\n", paste(cv, collapse = ", ")))

stopifnot(all.equal(cv, cv_truth))

## -------------------------------------------------------------------
## drop1All() - drop terms from model
## -------------------------------------------------------------------
set.seed(42)
m <- gamlss(y ~ pb(x) + x, data = abdom)

drop_truth <- drop1All(m, trace = FALSE)
print(drop_truth)

drop_res <- drop1All(m, trace = FALSE) |> futurize_and_verify()
print(drop_res)

plan(sequential)
} ## if (requireNamespace("gamlss"))

#' Options used by futurize
#'
#' Below are the \R options and environment variables that are used by the
#' \pkg{futurize} package and packages enhancing it.\cr
#' \cr
#' _WARNING: Note that the names and the default values of these options may
#'  change in future versions of the package.  Please use with care until
#'  further notice._
#'
#' @section Packages must not change futurize options:
#'
#' Just like for other R options, as a package developer you must _not_ change
#' any of the below `futurize.*` options.  Only the end-user should set these.
#' If you find yourself having to tweak one of the options, make sure to
#' undo your changes immediately afterward.
#'
#' @section Options for debugging:
#' \describe{
#'  \item{\option{futurize.debug}:}{(logical) If `TRUE`, extensive debug messages are generated. (Default: `FALSE`)}
#' }
#'
#' @section Options for controlling futurization:
#' \describe{
#'  \item{\option{futurize.enable}:}{(logical) If `TRUE` (default),
#'    `futurize()` transpilation will be applied, otherwise not.}
#' }
#'
#' @section Environment variables that set R options:
#' All of the above \R \option{futurize.*} options can be set by corresponding
#' environment variable \env{R_FUTURIZE_*} _when the \pkg{futurize} package is
#' loaded_. This means that those environment variables must be set before
#' the \pkg{futurize} package is loaded in order to have an effect.
#' For example, if `R_FUTURIZE_DEBUG=true`, then option
#' \option{futurize.debug} is set to `TRUE` (logical).
#'
#' @seealso
#' To set \R options or environment variables when \R starts (even before the \pkg{futurize} package is loaded), see the \link[base]{Startup} help page.  The \href{https://cran.r-project.org/package=startup}{\pkg{startup}} package provides a friendly mechanism for configuring \R's startup process.
#'
#' @aliases futurize.options 
#'
#' @aliases futurize.debug
#' @aliases futurize.enable
#'
#' @aliases R_FUTURIZE_DEBUG
#' @aliases R_FUTURIZE_ENABLE
#'
#' @name zzz-futurize.options 
NULL


update_package_option <- import_future("update_package_option")

## Set future options based on environment variables
update_package_options <- function(debug = FALSE) {
  update_package_option("futurize.enable", mode = "logical", default = TRUE, debug = debug)
}

#' @tags pkg-pvclust
if (requireNamespace("pvclust") && getRversion() >= "4.4.0") {
library(futurize)
library(pvclust)
options(future.rng.onMisuse = "error")

all_equal <- function(a, b, ...) {
  a$call <- b$call <- NULL
  all.equal(a, b, ...)
}

plan(multisession)

#------------------------------------------------------------------
# pvclust()
#------------------------------------------------------------------
message("pvclust() ...")
data(mtcars, package = "datasets")

set.seed(42)
fit_truth <- pvclust(mtcars, nboot = 10L, parallel = FALSE)
print(fit_truth)

set.seed(42)
fit <- pvclust(mtcars, nboot = 10L, parallel = FALSE) |> futurize_and_verify()
print(fit)

stopifnot(inherits(fit, "pvclust"))
stopifnot(length(fit$edges) == length(fit_truth$edges))

message("pvclust() ... done")

plan(sequential)
}

#' @tags pkg-SuperLearner
if (requireNamespace("SuperLearner") && getRversion() >= "4.4.0") {
library(futurize)
library(SuperLearner)
options(future.rng.onMisuse = "error")

plan(multisession)

#------------------------------------------------------------------
# CV.SuperLearner()
#------------------------------------------------------------------
message("CV.SuperLearner() ...")

set.seed(42)
n <- 100
p <- 5
X <- as.data.frame(matrix(rnorm(n * p), n, p))
Y <- X[, 1] + X[, 2] + rnorm(n)
SL.library <- c("SL.glm", "SL.mean")

set.seed(1)
res_truth <- CV.SuperLearner(Y = Y, X = X, V = 2, SL.library = SL.library, method = "method.NNLS")

set.seed(1)
res <- CV.SuperLearner(Y = Y, X = X, V = 2, SL.library = SL.library, method = "method.NNLS") |> futurize_and_verify()

stopifnot(
  inherits(res, "CV.SuperLearner"),
  identical(names(res), names(res_truth)),
  identical(dim(res$coef), dim(res_truth$coef)),
  identical(colnames(res$coef), colnames(res_truth$coef))
)

message("CV.SuperLearner() ... done")

plan(sequential)
}

make_options_for_doFuture <- local({
  defaults_base <- NULL

  function(options, defaults = NULL, wrap = TRUE) {
    ## Nothing to do?
    if (length(options) == 0 && length(defaults) == 0) return(options)

    if (is.null(defaults_base)) {
      ## The 'doFuture' package already imports 'future.apply'
      defaults_base <<- names(formals(future.apply::future_lapply))
    }

    if (length(defaults) > 0) {
      names <- setdiff(names(defaults), attr(options, "specified"))
      for (name in names) options[[name]] <- defaults[[name]]
    }

    ## Remap chunk_size -> chunk.size
    names(options) <- sub("^chunk_size$", "chunk.size", names(options))

    ## Remap future options for doFuture
    names <- sprintf("future.%s", names(options))

    ## Silently drop unknown future options
    keep <- intersect(defaults_base, names)
    idxs <- match(keep, table = names)
    options <- options[idxs]
    
    if (wrap) options <- list(.options.future = options)

    options
  }
})

#' @tags pkg-future.apply
if (requireNamespace("future.apply")) {
library(futurize)

plan(multisession)

truth <- future.apply::future_lapply(1, identity)

## Wrapped in { ... }
y <- { lapply(1, identity) } |> futurize_and_verify()
stopifnot(identical(y, truth))

## Wrapped in ( ... )
y <- ( lapply(1, identity) ) |> futurize_and_verify()
stopifnot(identical(y, truth))

## Wrapped in local( ... )
y <- local( lapply(1, identity) ) |> futurize_and_verify()
stopifnot(identical(y, truth))

## Wrapped in identity( ... )
y <- identity( lapply(1, identity) ) |> futurize_and_verify()
stopifnot(identical(y, truth))

## Wrapped in I( ... )
y <- I( lapply(1, identity) ) |> futurize_and_verify()
stopifnot(identical(y, I(truth)))

## Wrapped in { { ... } }
y <- { { lapply(1, identity) } } |> futurize_and_verify()
stopifnot(identical(y, truth))

## Wrapped in { ( ... ) }
y <- { ( lapply(1, identity) ) } |> futurize_and_verify()
stopifnot(identical(y, truth))

## Wrapped in ( ( ... ) )
y <- ( ( lapply(1, identity) ) ) |> futurize_and_verify()
stopifnot(identical(y, truth))

## Wrapped in local({ ... })
y <- local({ lapply(1, identity) }) |> futurize_and_verify()
stopifnot(identical(y, truth))

## Wrapped in { local({ ... }) }
y <- { local({ lapply(1, identity) }) } |> futurize_and_verify()
stopifnot(identical(y, truth))

truth2 <- list(-1, 2, -3)

## Wrapped in { x <- truth2; local({ ... }) }
y <- { x <- truth2; lapply(x, identity) } |> futurize_and_verify()
stopifnot(identical(y, truth2))

## Wrapped in { { { x <- truth2; local({ ... }) } } }
y <- { { { x <- truth2; lapply(x, identity) } } } |> futurize_and_verify()
stopifnot(identical(y, truth2))

## Wrapped in { ( { x <- truth2; local({ ... }) ) } }
y <- { ( { x <- truth2; lapply(x, identity) } ) } |> futurize_and_verify()
stopifnot(identical(y, truth2))

## Wrapped in { local({ x <- truth2; local({ ... }) }) }
y <- { local({ x <- truth2; lapply(x, identity) }) } |> futurize_and_verify()
stopifnot(identical(y, truth2))

## Wrapped in { local({ 42; x <- truth2; local({ ... }) } })
y <- { local({ 42; x <- truth2; lapply(x, identity) }) } |> futurize_and_verify()
stopifnot(identical(y, truth2))

## Wrapped in { 3.14; local({ 42; x <- truth2; local({ ... }) } })
y <- { 3.14; local({ 42; x <- truth2; lapply(x, identity) }) } |> futurize_and_verify()
stopifnot(identical(y, truth2))

## Wrapped in { 3.14; local({ 42; x <- truth2; local({ ... }) } })
y <- { 3.14; local({ 42; x <- truth2; lapply(x, identity) }) } |> futurize_and_verify()
stopifnot(identical(y, truth2))

truth3 <- !sapply(truth2, \(x) { x > 0 })
y <- !sapply(truth2, \(x) { x > 0 }) |> futurize_and_verify()
stopifnot(identical(y, truth3))

y <- { !sapply(truth2, \(x) { x > 0 }) } |> futurize_and_verify()
stopifnot(identical(y, truth3))

y <- { 42; !sapply(truth2, \(x) { x > 0 }) } |> futurize_and_verify()
stopifnot(identical(y, truth3))

## Wrapped in with(data, ...)
data <- data.frame(a = 1:3)
truth_with <- future.apply::future_lapply(data$a, identity)
y <- with(data, lapply(a, identity)) |> futurize_and_verify()
stopifnot(identical(y, truth_with))

## Wrapped in { with(data, ...) }
y <- { with(data, lapply(a, identity)) } |> futurize_and_verify()
stopifnot(identical(y, truth_with))

## Wrapped in with(data, { ... })
y <- with(data, { lapply(a, identity) }) |> futurize_and_verify()
stopifnot(identical(y, truth_with))

## Wrapped in local(..., envir = ...) - not the last element
e <- new.env()
truth4 <- local(lapply(1:3, identity), envir = e)
y <- local({ lapply(1:3, identity) }, envir = e) |> futurize_and_verify()
stopifnot(identical(y, truth4))

e <- new.env()
truth5 <- local({ lapply(1:3, identity) }, envir = e)
y <- local({ lapply(1:3, identity) }, envir = e) |> futurize_and_verify()
stopifnot(identical(y, truth5))

## Wrapped in suppressWarnings(..., classes = ...) - not the last element
truth6 <- suppressWarnings({ lapply(1:3, identity) }, classes = "warning")
y <- suppressWarnings({ lapply(1:3, identity) }, classes = "warning") |> futurize_and_verify()
stopifnot(identical(y, truth6))

plan(sequential)
} ## if (requireNamespace("future.apply"))

#' @tags skip_on_cran  ## (35s) to limit total check time
if (requireNamespace("SingleCellExperiment") && requireNamespace("scuttle") && requireNamespace("doFuture") && requireNamespace("DelayedArray")) {
library(futurize)
library(SingleCellExperiment)
library(scuttle)
library(DelayedArray)

## Use a small block size to ensure multiple blocks and thus multiple futures
setAutoBlockSize(10000)

plan(multisession)

## Create a simple SingleCellExperiment with alternative experiments
set.seed(42)
sce <- mockSCE(ncells = 20L, ngenes = 70L, nspikes = 30L)

## ---------------------------------------------------------
## applySCE() with perFeatureQCMetrics
## ---------------------------------------------------------
result_truth <- applySCE(sce, perFeatureQCMetrics)

result <- applySCE(sce, perFeatureQCMetrics) |> futurize_and_verify()
stopifnot(all.equal(result, result_truth))

result2 <- SingleCellExperiment::applySCE(sce, perFeatureQCMetrics) |> futurize_and_verify()
stopifnot(all.equal(result2, result_truth))

plan(sequential)
} ## if (requireNamespace("SingleCellExperiment") && ...)

#' @tags pkg-caret
if (requireNamespace("caret") && requireNamespace("randomForest") && requireNamespace("doFuture")) {
library(futurize)
library(caret)
options(future.rng.onMisuse = "error")

plan(multisession)

d_truth <- nearZeroVar(iris[, -5], saveMetrics = TRUE)
print(d_truth)

d <- nearZeroVar(iris[, -5], saveMetrics = TRUE) |> futurize_and_verify()
print(d)
stopifnot(all.equal(d, d_truth))

# Define training control
ctrl <- trainControl(method = "cv", number = 3L)

set.seed(1011)
model_truth <- train(Species ~ ., data = iris, method = "rf", trControl = ctrl)
print(model_truth)

set.seed(1011)
model <- train(Species ~ ., data = iris, method = "rf", trControl = ctrl) |> futurize_and_verify()
print(model)

## Cannot really compare results, because of different RNGs

plan(sequential)
} ## if (requireNamespace("caret") && requireNamespace("randomForest"))

#' @importFrom utils globalVariables
globalVariables(".")
globalVariables(c("CALL", "OPTS", "EXPR", "NAME"))

#' Efficient partial substitution in expressions
#' 
#' @param expr A bquoted R expression
#' 
#' @param substitute If TRUE, `expr` is substituted.
#' 
#' @param tmpl A bquote template.
#' 
#' @param \ldots Named arguments.
#' 
#' @param envir The environment in which the bquoted values
#' should be resolved.
#' 
#' @return
#' `bquote_compile()` returns a compiled bquote template, where attribute
#'  `expression` holds the original expression `expr`.
#' 
#' `bquote_apply()` returns an expression.
#' 
#' @examples
#' tmpl <- bquote_compile(
#'   local({
#'     cl <- do.call(.(CALL), args = .(OPTS))
#'     oopts <- options(future.ClusterFuture.clusterEvalQ = "error")
#'     on.exit(options(oopts))
#'     .(EXPR)
#'   })
#' )
#' 
#' expr <- bquote_apply(tmpl, CALL = quote(foo::bar),
#'                      OPTS = list(a = 1, b = 2), EXPR = quote(1 + 2))
#' 
#' @noRd
bquote_compile <- function(expr, substitute = TRUE) {
  if (substitute) expr <- substitute(expr)
  
  tmpl <- list()
  
  unquote <- function(e, at = integer(0L)) {
    n <- length(e)
    if (n == 0L) return()

    if (is.pairlist(e)) {
      for (kk in 1:n) unquote(e[[kk]], at = c(at, kk))
      return()
    }

    if (!is.call(e)) return()
    
    ## .(<name>)?
    if (is.name(e[[1L]]) && as.character(e[[1]]) == ".") {
      ## Record location in expression tree
      entry <- list(
        expression = e[[2L]],
        at         = at
      )
      tmpl <<- c(tmpl, list(entry))
      return()
    }
  
    ## `{`, `+`, ...
    for (kk in 1:n) unquote(e[[kk]], at = c(at, kk))
  }

  dummy <- unquote(expr)
  attr(tmpl, "expression") <- expr
  tmpl
}


#' @noRd
bquote_apply <- function(tmpl, ..., envir = parent.frame()) {
  expr <- attr(tmpl, "expression")

  args <- list(...)
  if (length(args) > 0) {
    envir <- new.env(parent = envir)
    for (name in names(args)) {
      assign(name, args[[name]], envir = envir, inherits = FALSE)
    }
  }
  
  for (kk in seq_along(tmpl)) {
    entry <- tmpl[[kk]]
    value <- eval(entry[["expression"]], envir = envir)
    at <- entry[["at"]]
    
    ## Special case: Result becomes just a value
    nat <- length(at)
    if (nat == 0) return(value)

    ## Inject a NULL (needs special care) or a regular value?
    if (is.null(value)) {
      head <- if (nat == 1L) NULL else at[-nat]
      e <- if (is.null(head)) expr else expr[[head]]
      if (is.call(e)) {
        f <- as.list(e)
        f[at[nat]] <- list(NULL)
        e <- as.call(f)
      } else if (is.pairlist(e)) {
        e[at[nat]] <- list(NULL)
        e <- as.pairlist(e)
      } else {
        stop(sprintf("Unknown type of expression (please report to the maintainer): %s", sQuote(paste(deparse(e), collapse = "\\n"))))
      }
      if (is.null(head)) {
        expr <- e
      } else {
        expr[[head]] <- e
      }
    } else {
      expr[[at]] <- value
    }
  }

  expr
}

#' @tags pkg-doFuture
#' @tags %dofuture%
#' @tags %dopar%
#' @tags detritus-files
#' @tags sequential multisession cluster multicore

if (requireNamespace("foreach") && requireNamespace("doFuture")) {
library(futurize)
library(foreach)

doFuture::registerDoFuture()

strategies <- future:::supportedStrategies()

message("*** doFuture() - error handling w/ .errorhandling in c('stop', 'remove', 'pass') ...")

.options.future <- list(errors = "foreach")
for (strategy in strategies) {
  message(sprintf("- plan('%s') ...", strategy))
  plan(strategy)

  mu <- 1.0
  sigma <- 2.0

  message("Futureverse-style error handling ...")
  res <- tryCatch({
    foreach(i = 1:10) %do% {
      if (i %% 2 == 0) stop(sprintf("Index error ('stop'), because i = %d", i))
      list(i = i, value = dnorm(i, mean = mu, sd = sigma))
    } |> futurize_and_verify()
  }, FuturizeTestAssertionError = stop, error = identity)
  str(res)
  stopifnot(
    inherits(res, "error")
  )
  message("Futureverse-style error handling ... DONE")


  for (.errorhandling in c("stop", "remove", "pass")) {
    message(sprintf(".errorhandling = '%s' ...", .errorhandling))
    
    truth <- tryCatch({
      foreach(i = 1:10, .errorhandling = .errorhandling, .options.future = .options.future) %dopar% {
        if (i %% 2 == 0) stop(sprintf("Index error ('stop'), because i = %d", i))
        list(i = i, value = dnorm(i, mean = mu, sd = sigma))
      }
    }, FuturizeTestAssertionError = stop, error = identity)
    str(truth)

    res <- tryCatch({
      foreach(i = 1:10, .errorhandling = .errorhandling, .options.future = .options.future) %do% {
        if (i %% 2 == 0) stop(sprintf("Index error ('stop'), because i = %d", i))
        list(i = i, value = dnorm(i, mean = mu, sd = sigma))
      } |> futurize_and_verify()
    }, FuturizeTestAssertionError = stop, error = identity)
    str(res)
    stopifnot(
      length(res) == length(truth),
      identical(names(res), names(truth))
    )
    if (inherits(res, "condition")) {
    ## FIXME: HB 2025-08-19
##      stopifnot(identical(conditionMessage(res), conditionMessage(truth)))
    }
    
    if (.errorhandling == "stop") {
      stopifnot(
        inherits(res, "error"),
        grepl("Index error", conditionMessage(res))
    ## FIXME: HB 2025-08-19
#        identical(conditionMessage(res), conditionMessage(truth))
      )
    } else if (.errorhandling == "remove") {
      stopifnot(identical(res, truth))
    } else if (.errorhandling == "pass") {
      mapply(res, truth, FUN = function(r, t) {
        if (inherits(r, "error")) {
          r$call <- NULL
          t$call <- NULL
          stopifnot(identical(r, t))
        }
      })
    }

    message(sprintf(".errorhandling = '%s' ... done", .errorhandling))
  } ## for (.errorhandling ...)

  # Shutdown current plan
  plan(sequential)

  message(sprintf("- plan('%s') ... DONE", strategy))
} ## for (strategy ...)

message("*** doFuture() - error handling w/ .errorhandling in c('stop', 'remove', 'pass') ... DONE")


message("*** doFuture() - error handling w/ 'pass' ...")

for (strategy in strategies) {
  message(sprintf("- plan('%s') ...", strategy))
  plan(strategy)

  mu <- 1.0
  sigma <- 2.0
  res <- foreach(i = 1:10, .errorhandling = "pass", .options.future = .options.future) %do% {
    if (i %% 2 == 0) stop(sprintf("Index error ('pass'), because i = %d", i))
    list(i = i, value = dnorm(i, mean = mu, sd = sigma))
  } |> futurize_and_verify()
  str(res)
  stopifnot(
    is.list(res),
    length(res) == 10L
  )

  # Shutdown current plan
  plan(sequential)

  message(sprintf("- plan('%s') ... DONE", strategy))
} ## for (strategy ...)

message("*** doFuture() - error handling w/ 'pass' ... DONE")


message("*** doFuture() - error handling w/ 'remove' ...")

for (strategy in strategies) {
  message(sprintf("- plan('%s') ...", strategy))
  plan(strategy)

  mu <- 1.0
  sigma <- 2.0
  res <- foreach(i = 1:10, .errorhandling = "remove", .options.future = .options.future) %do% {
    if (i %% 2 == 0) stop(sprintf("Index error ('remove'), because i = %d", i))
    list(i = i, value = dnorm(i, mean = mu, sd = sigma))
  } |> futurize_and_verify()
  str(res)
  stopifnot(
    is.list(res),
    length(res) == 5L
  )

  # Shutdown current plan
  plan(sequential)
  
  message(sprintf("- plan('%s') ... DONE", strategy))
} ## for (strategy ...)

message("*** doFuture() - error handling w/ 'remove' ... DONE")

message("*** doFuture() - invalid accumulator ...")

boom <- function(...) stop("boom!")
res <- tryCatch({
  foreach(i = 1:3, .combine = boom) %do% { i } |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
print(res)
stopifnot(
  inherits(res, "error"),
  inherits(res, "FutureError")
)

message("*** doFuture() - invalid accumulator ... DONE")

plan(sequential)
} ## if (requireNamespace("foreach") && requireNamespace("doFuture"))

#' @tags pkg-furrr
#' @tags detritus-files

if (requireNamespace("purrr") && requireNamespace("furrr")) {
library(futurize)


library(purrr)

# ------------------------------------------------------------------------------
# map()

message("future_map() matches map() for simple cases")
stopifnot(identical(
  map(1:3, ~.x) |> futurize_and_verify(),
  map(1:3, ~.x)
))


message("names of `.x` are retained")
x <- c(a = 1, b = 2)
stopifnot(identical(names(map(x, ~1) |> futurize_and_verify()), c("a", "b")))


message("named empty input makes named empty output")
x <- set_names(list(), character())
stopifnot(identical(names(map(x, ~.x) |> futurize::futurize()), character()))


# ------------------------------------------------------------------------------
# atomic variants

message("future_map_dbl() works")
x <- c(1, 2, 3)

stopifnot(identical(
  map_dbl(x, ~.x) |> futurize_and_verify(),
  map_dbl(x, ~.x)
))


message("future_map_int() works")
x <- c(1L, 2L, 3L)

stopifnot(identical(
  map_int(x, ~.x) |> futurize_and_verify(),
  map_int(x, ~.x)
))


message("future_map_lgl() works")
x <- c(TRUE, FALSE, TRUE)

stopifnot(identical(
  map_lgl(x, ~.x) |> futurize_and_verify(),
  map_lgl(x, ~.x)
))


message("future_map_chr() works")
x <- c("a", "b", "c")

stopifnot(identical(
  map_chr(x, ~.x) |> futurize_and_verify(),
  map_chr(x, ~.x)
))


message("names of `.x` are retained")
x <- c(a = 1, b = 2)
stopifnot(identical(names(map_dbl(x, ~1) |> futurize_and_verify()), c("a", "b")))


# ------------------------------------------------------------------------------
# data frame variants

message("future_map_dfr() works")
x <- c("a", "b", "c")

stopifnot(identical(
  map_dfr(x, ~data.frame(x = .x)) |> futurize_and_verify(),
  map_dfr(x, ~data.frame(x = .x))
))


message("future_map_dfc() works")
x <- c("a", "b", "c")

stopifnot(identical(
  map_dfc(x, ~as.data.frame(set_names(list(1), .x))) |> futurize_and_verify(),
  map_dfc(x, ~as.data.frame(set_names(list(1), .x)))
))


# ------------------------------------------------------------------------------
# size

message("future_map() works with size zero input")
stopifnot(identical(map(list(), identity) |> futurize::futurize(), list()))


message("atomic variants work with size zero input")
stopifnot(identical(map_chr(list(), identity) |> futurize::futurize(), character()))
stopifnot(identical(map_dbl(list(), identity) |> futurize::futurize(), double()))
stopifnot(identical(map_int(list(), identity) |> futurize::futurize(), integer()))
stopifnot(identical(map_lgl(list(), identity) |> futurize::futurize(), logical()))


# ------------------------------------------------------------------------------
# at / if variants

message("future_map_at() works")
x <- list("a", "b", "c")

stopifnot(identical(
  map_at(x, 2, ~3) |> futurize_and_verify(),
  map_at(x, 2, ~3)
))


message("names of `.x` are retained")
x <- list(a = "a", b = "b", c = "c")
stopifnot(identical(names(map_at(x, 2, ~3) |> futurize_and_verify()), c("a", "b", "c")))


message("future_map_if() works")
x <- list("a", "b", "c")

stopifnot(identical(
  map_if(x, ~.x %in% c("a", "c"), ~3) |> futurize_and_verify(),
  map_if(x, ~.x %in% c("a", "c"), ~3)
))


message("names of `.x` are retained")
x <- list(a = "a", b = "b", c = "c")
stopifnot(identical(names(map_if(x, ~.x %in% c("a", "c"), ~3) |> futurize_and_verify()), c("a", "b", "c")))


message("`.else` can be used")
x <- list("a", "b", "c")

stopifnot(identical(
  map_if(x, ~.x %in% c("a", "c"), ~3, .else = ~-1) |> futurize_and_verify(),
  map_if(x, ~.x %in% c("a", "c"), ~3, .else = ~-1)
))


# ------------------------------------------------------------------------------
# Miscellaneous

message("Calling `~` from within `.f` works")
x <- list(
  list(a = 4, b = 6),
  list(c = 5, d = 7)
)

stopifnot(identical(map(x, ~map(.x, ~.x)) |> futurize_and_verify(), x))


message("Calling `~` from within `.f` inside a `mutate()` works (#7, #123)")
x <- list(
  list(a = 4, b = 6),
  list(c = 5, d = 7)
)

df <- dplyr::tibble(x = x)

stopifnot(identical(
  dplyr::mutate(df, x = map(x, ~map(.x, ~.x)) |> futurize_and_verify()),
  df
))


message("globals in `.x` are found (#16)")
fn <- function(x) sum(x, na.rm = TRUE)

x <- list(c(1, 2, NA), c(2, 3, 4))

fns1 <- map(x, ~purrr::partial(fn, x = .x))
fns2 <- map(x, ~function() fn(.x))

stopifnot(identical(map_dbl(fns1, ~.x()) |> futurize_and_verify(), c(3, 9)))
stopifnot(identical(map_dbl(fns2, ~.x()) |> futurize_and_verify(), c(3, 9)))


message("globals in `.x` are only exported to workers that use them")
plan(multisession, workers = 2)
on.exit(plan(sequential), add = TRUE)

# Use `local()` to ensure that the wrapper functions and the anonymous
# functions created with `~` don't pick up extra globals
my_wrapper1 <- local({
  my_mean1 <- function(x) mean(x, na.rm = TRUE)

  function(x) {
    my_mean1(x)
    exists("my_mean1")
  }
})

my_wrapper2 <- local({
  my_mean2 <- function(x) mean(x, na.rm = FALSE)

  function(x) {
    my_mean2(x)
    exists("my_mean1")
  }
})

x <- list(my_wrapper1, my_wrapper2)

stopifnot(identical(
  map_lgl(.x = x, .f = ~.x(c(1, NA))) |> futurize_and_verify(),
  c(TRUE, FALSE)
))


message("base package functions can be exported to workers (HenrikBengtsson/future#401)")
stopifnot(identical(map(1:2, identity) |> futurize_and_verify(), list(1L, 2L)))


message("`.f` globals are only looked up in the function env of `.f` (#153)")
fn <- function(x) {
  y
}

fn2 <- local({
  y <- -1

  function(x) {
    y
  }
})

wrapper <- function(f) {
  y <- 1
  map(1:2, f) |> futurize_and_verify()
}

res <- tryCatch(wrapper(fn), FuturizeTestAssertionError = stop, error = identity)
stopifnot(
  inherits(res, "error"),
  grepl("'y' not found", conditionMessage(res))
)
stopifnot(identical(wrapper(fn2), list(-1, -1)))


message("`...` globals/packages are found")
# We set the function environments to the global environment to ensure
# that they aren't set to something else while `test()` is running

fn <- function(x, fn_arg) {
  fn_arg()
}
environment(fn) <- .GlobalEnv

fn_arg_env <- new.env(parent = .GlobalEnv)
fn_arg_env$x <- 1

# This function is passed through `...`
fn_arg <- function() {
  x
}
environment(fn_arg) <- fn_arg_env

stopifnot(identical(
  map(1:2, fn, fn_arg = fn_arg) |> futurize_and_verify(),
  list(1, 1)
))


plan(sequential)
} ## if (requireNamespace("purrr") && requireNamespace("furrr"))

now <- function(x = Sys.time(), format = "[%H:%M:%OS3] ") {
  ## format(x, format = format) ## slower
  format(as.POSIXlt(x, tz = ""), format = format)
}

debug_indent <- local({
  prefix <- ""
  depth <- 0L
  symbols <- rep(c("|", ":", "."), times = 10L)
 
  function(delta = 0L) {
    if (delta == 0) return(prefix)
    if (delta > 0) {
      depth <<- depth + 1L
    } else if (delta < 0) {
      depth <<- depth - 1L
      if (depth < 0L) {
        calls <- paste(vapply(sys.calls(), FUN = deparse, FUN.VALUE = NA_character_), collapse = " -> ")
        warning(sprintf("[INTERNAL WARNING]: There appears to be one mdebug_pop() too many: %s", calls), call. = TRUE, immediate. = TRUE)
        depth <<- 0L
      }
    }
    prefix <<- if (depth == 0) "" else paste(paste(symbols[seq_len(depth)], " "), collapse = "")
  }
})

.debug <- new.env(parent = emptyenv())
.debug$stack <- list()

mdebug_push <- function(..., debug = isTRUE(getOption("futurize.debug"))) {
  msg <- mdebug(..., debug = debug)
  debug_indent(+1)
  .debug$stack <- c(.debug$stack, msg)
  invisible(msg)
}

mdebugf_push <- function(..., debug = isTRUE(getOption("futurize.debug"))) {
  msg <- mdebugf(..., debug = debug)
  debug_indent(+1)
  .debug$stack <- c(.debug$stack, msg)
  invisible(msg)
}

mdebug_pop <- function(..., debug = isTRUE(getOption("futurize.debug"))) {
  n <- length(.debug$stack)
  if (n == 0) stop("Called mdebug_pop() on an empty debug stack")
  msg <- .debug$stack[n]
  .debug$stack <- .debug$stack[-n]
  debug_indent(-1)
  mdebug(sprintf("%s done", msg), debug = debug)
}

mdebugf_pop <- function(..., debug = isTRUE(getOption("futurize.debug"))) {
  n <- length(.debug$stack)
  if (n == 0) stop("Called mdebug_pop() on an empty debug stack")
  msg <- .debug$stack[n]
  .debug$stack <- .debug$stack[-n]
  debug_indent(-1)
  mdebug(sprintf("%s done", msg), debug = debug)
}

mdebug <- function(..., prefix = now(), debug = isTRUE(getOption("futurize.debug"))) {
  prefix <- paste(prefix, debug_indent(), sep = "")
  msg <- paste(..., sep = "")
  message(sprintf("%s%s", prefix, msg))
  invisible(msg)
}

mdebugf <- function(..., appendLF = TRUE,
                    prefix = now(), debug = isTRUE(getOption("futurize.debug"))) {
  prefix <- paste(prefix, debug_indent(), sep = "")
  msg <- sprintf(...)
  message(sprintf("%s%s", prefix, msg), appendLF = appendLF)
  invisible(msg)
}

#' @importFrom utils capture.output
mprint <- function(..., appendLF = TRUE, prefix = now(), debug = isTRUE(getOption("futurize.debug"))) {
  prefix <- paste(prefix, debug_indent(), sep = "")
  message(paste(prefix, capture.output(print(...)), sep = "", collapse = "\n"), appendLF = appendLF)
}

#' @tags pkg-future.apply
#' @tags future_eapply
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)



message("*** future_eapply() ...")

message("- From example(eapply) ...")

for (strategy in supportedStrategies()) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)
  
  env <- new.env(hash = FALSE)
  env$a <- 1:10
  env$beta <- exp(-3:3)
  env$logic <- c(TRUE, FALSE, FALSE, TRUE)
  
  y0 <- unlist(eapply(env, mean, USE.NAMES = FALSE))
  y1 <- unlist(eapply(env, mean, USE.NAMES = FALSE) |> futurize_and_verify())
  stopifnot(all.equal(y1, y0))
  
  y0 <- eapply(env, quantile, probs = 1:3/4)
  y1 <- eapply(env, quantile, probs = 1:3/4) |> futurize_and_verify()
  stopifnot(all.equal(y1, y0))
  
  y0 <- eapply(env, quantile)
  y1 <- eapply(env, quantile) |> futurize_and_verify()
  stopifnot(all.equal(y1, y0))
  y2 <- eapply(env, "quantile") |> futurize_and_verify()
  stopifnot(all.equal(y2, y0))

  plan(sequential)
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
} ## for (strategy in ...) 

message("*** future_eapply() ... DONE")



} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags pkg-future.apply

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)



message("*** future_replicate() ...")

for (strategy in supportedStrategies()) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)
  
  y0 <- replicate(5L, sample(10L, size = 1L))
  y1 <- replicate(5L, sample(10L, size = 1L)) |> futurize_and_verify()
  stopifnot(length(y0) == length(y1))
  
  set.seed(0xBEEF)
  y1 <- replicate(5L, sample(10L, size = 1L)) |> futurize_and_verify()
  set.seed(0xBEEF)
  y2 <- replicate(5L, sample(10L, size = 1L)) |> futurize_and_verify()
  stopifnot(all.equal(y2, y1))
  
  y3 <- replicate(5L, sample(10L, size = 1L)) |> futurize_and_verify(seed = 0xBEEF)
  y4 <- replicate(5L, sample(10L, size = 1L)) |> futurize_and_verify(seed = 0xBEEF)
  stopifnot(all.equal(y4, y3))
  
  
  message("- example(replicate) ...")
  
  foo <- function(x = 1, y = 2) c(x, y)
  bar0 <- function(n, x) replicate(n, foo(x = x))
  y0 <- bar0(5, x = 3)
  
  bar1 <- function(n, x) replicate(n, foo(x = x)) |> futurize_and_verify()
  y1 <- bar1(5, x = 3)
  stopifnot(all.equal(y1, y0))
  
  plan(sequential)
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
} ## for (strategy in ...) 

message("*** future_replicate() ... DONE")



} ## if (requireNamespace("future.apply", quietly = TRUE))

# seriation::train(...) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), local({
#   options(future.disposable = structure(OPTS, dispose = FALSE))
#   on.exit(options(future.disposable = NULL))
#   seriation::seriate_best(...)
# }))
#
append_transpilers_for_seriation <- function() {
  transpilers <- make_package_transpilers("seriation", FUN = function(fcn, name) {
    if ("parallel" %in% names(formals(fcn))) {
      defaults <- list(
        seed = TRUE,
        label = sprintf("fz:seriation::%s-%%d", name)
      )
      list(
        label = sprintf("seriation::%s() ~> seriation::%s()", name, name),
        transpiler = make_futurize_for_doFuture(args = list(parallel = TRUE), defaults = defaults)
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("seriation", "doFuture")
}

#' @tags pkg-modelsummary
if (requireNamespace("modelsummary") && requireNamespace("future.apply")) {
  library(futurize)
  library(modelsummary)
  options(future.rng.onMisuse = "error")

  plan(multisession, workers = 2L)

  ## Fit two simple linear models
  fit1 <- lm(mpg ~ cyl, data = mtcars)
  fit2 <- lm(mpg ~ cyl + hp, data = mtcars)
  models <- list(Model1 = fit1, Model2 = fit2)

  ## -------------------------------------------------------
  ## modelsummary()
  ## -------------------------------------------------------

  ## Truth (sequential, with modelsummary_future = FALSE)
  oopts <- options(modelsummary_future = FALSE)
  result_truth <- modelsummary::modelsummary(
    models,
    output = "data.frame",
    fmt = 2
  )
  options(oopts)

  ## Futurized modelsummary()
  result <- modelsummary::modelsummary(
    models,
    output = "data.frame",
    fmt = 2
  ) |> futurize_and_verify()

  stopifnot(
    is.data.frame(result),
    nrow(result) > 0,
    all.equal(result, result_truth)
  )

  ## -------------------------------------------------------
  ## modelplot()
  ## -------------------------------------------------------

  ## Truth (sequential, with modelsummary_future = FALSE)
  oopts <- options(modelsummary_future = FALSE)
  result_plot_truth <- modelsummary::modelplot(models, draw = FALSE)
  options(oopts)

  ## Futurized modelplot()
  result_plot <- modelsummary::modelplot(models, draw = FALSE) |> futurize_and_verify()

  stopifnot(
    is.data.frame(result_plot),
    nrow(result_plot) > 0,
    all.equal(result_plot, result_plot_truth)
  )

  plan(sequential)
}

#' @tags pkg-purrr
if (requireNamespace("purrr") && requireNamespace("furrr")) {
library(futurize)
library(purrr)
options(future.rng.onMisuse = "error")

plan(multisession)

counters <- plan("backend")[["counters"]]
y <- map(1:3, function(x) { print(x) }) |> futurize(stdout = FALSE)
delta <- plan("backend")[["counters"]] - counters
cat(sprintf("Futures created: %d\n", delta[["created"]]))
stopifnot(delta[["created"]] > 0L)
print(y)


xs <- list(aa = 1, bb = 1:2, cc = 1:10, dd = 1:5, .ee = -6:6)
FUN_no_rng <- function(x, na.rm = TRUE) {
  a <- 1:5
  add <- NULL
  if (length(x) == 2) add <- list(C = 42)
  median(c(a, x), na.rm = na.rm)
}

FUN_rng <- function(x, na.rm = TRUE) {
  dummy <- sample.int(10L)
  a <- 1:5
  add <- NULL
  if (length(x) == 2) add <- list(C = 42)
  median(c(a, x), na.rm = na.rm)
}

es <- as.environment(xs)


exprs <- list(
      map = quote(map(xs, FUN) ),
  map_dbl = quote(map_dbl(xs, FUN) )
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))

  FUN <- FUN_no_rng
  truth <- eval(expr)
  counters <- plan("backend")[["counters"]]
  expr_f <- bquote(.(expr) |> futurize())
  res <- eval(expr_f)
  delta <- plan("backend")[["counters"]] - counters
  cat(sprintf("Futures created: %d\n", delta[["created"]]))
  stopifnot(delta[["created"]] > 0L)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> futurize(stdout = FALSE, conditions = character(0L)))
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(
    identical(out, character(0L)),
    identical(res2, res),
    identical(res2, truth)
  )

  expr_f3 <- bquote(.(expr) |> futurize(chunk_size = 1L))
  res3 <- eval(expr_f3)
  stopifnot(
    identical(res3, res),
    identical(res3, truth)
  )

  message("Test with RNG:")
  FUN <- FUN_rng
  expr_f4 <- bquote(.(expr) |> futurize(seed = TRUE))
  print(expr_f4)
  res4 <- local({
    opts <- options(future.rng.onMisuse = "error")
    on.exit(options(opts))
    eval(expr_f4)
  })
  stopifnot(
    identical(res4, res),
    identical(res4, truth)
  )
}

message("Special case: Zero futurize() options")
y <- map(1, identity) |> futurize(options = list())


plan(sequential)
} ## if (requireNamespace("purrr") && requireNamespace("furrr"))

.package <- new.env(parent = emptyenv())
.package[[".futurize"]] <- TRUE

## covr: skip=all
#' @importFrom utils packageVersion
.onLoad <- function(libname, pkgname) {
  ## R CMD build
  register_vignette_engine_during_build_only(pkgname)
  
  .package[["version"]] <- packageVersion(pkgname)

  update_package_option("futurize.debug", mode = "logical")
  debug <- isTRUE(getOption("futurize.debug"))

  if (debug) {
    envs <- Sys.getenv()
    envs <- envs[grep("R_FUTURIZE_", names(envs), fixed = TRUE)]
    envs <- sprintf("- %s=%s", names(envs), sQuote(envs))
    mdebug(paste(c("Futurize-specific environment variables:", envs), collapse = "\n"))
  }

  ## Set future options based on environment variables
  update_package_options(debug = debug)

  ## Register all futurize transpilers
  register_all_transpilers()
} ## .onLoad()

#' @tags pkg-DiceKriging
if (requireNamespace("DiceKriging") && requireNamespace("doFuture")) {
library(futurize)
library(DiceKriging)
options(future.rng.onMisuse = "ignore")

plan(multisession)

## -------------------------------------------------------------------
## km() - kriging model with multistart optimization
## -------------------------------------------------------------------
## Adopted from help("km", package = "DiceKriging")
set.seed(42)
n <- 20L
d <- 2L
design.fact <- expand.grid(seq(0, 1, length = 4), seq(0, 1, length = 4))
design.fact <- data.frame(design.fact)
names(design.fact) <- c("x1", "x2")
y <- apply(design.fact, 1, function(x) x[1]^2 + x[2]^2)
response <- data.frame(y = y)

## Sequential (truth)
set.seed(42)
m_truth <- km(~., design = design.fact, response = response,
              multistart = 4L)
cat(sprintf("km truth: range = [%s]\n",
            paste(round(coef(m_truth, "range"), 4), collapse = ", ")))

## Futurized
set.seed(42)
m_fz <- km(~., design = design.fact, response = response,
           multistart = 4L) |> futurize_and_verify()
cat(sprintf("km futurized: range = [%s]\n",
            paste(round(coef(m_fz, "range"), 4), collapse = ", ")))

stopifnot(all.equal(coef(m_truth, "range"), coef(m_fz, "range"), tolerance = 1e-3))

plan(sequential)
} ## if (requireNamespace("DiceKriging"))

#' @tags pkg-future.apply
#' @tags future_lapply
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)



message("*** future_lapply() and RNGs ...")

options(future.debug = FALSE)

message("* lapply(x, ..., future.seed = <invalid>) |> futurize() ...")

res <- tryCatch({
  y <- lapply(1:3, FUN = identity, future.seed = as.list(1:2) |> futurize_and_verify())
}, FuturizeTestAssertionError = stop, error = identity)
print(res)
stopifnot(inherits(res, "simpleError"))

res <- tryCatch({
  y <- lapply(1:3, FUN = identity, future.seed = list(1, 2, 3:4) |> futurize_and_verify())
}, FuturizeTestAssertionError = stop, error = identity)
print(res)
stopifnot(inherits(res, "simpleError"))

res <- tryCatch({
  y <- lapply(1:3, FUN = identity, future.seed = as.list(1:3) |> futurize_and_verify())
}, FuturizeTestAssertionError = stop, error = identity)
print(res)
stopifnot(inherits(res, "simpleError"))

seeds <- lapply(1:3, FUN = as_lecyer_cmrg_seed)
res <- tryCatch({
  y <- lapply(1:3, FUN = identity, future.seed = lapply(seeds, FUN = as.numeric) |> futurize_and_verify())
}, FuturizeTestAssertionError = stop, error = identity)
print(res)
stopifnot(inherits(res, "simpleError"))

seeds[[1]][1] <- seeds[[1]][1] + 1L
res <- tryCatch({
  y <- lapply(1:3, FUN = identity, future.seed = seeds) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
print(res)
stopifnot(inherits(res, "simpleError"))

message("* lapply(x, ..., future.seed = <invalid>) |> futurize() ... DONE")


## Iterate of the same set in all tests
x <- 1:5

message("* lapply(x, ..., future.seed = FALSE) |> futurize() ...")

y0 <- y0_nested <- seed00 <- NULL
for (cores in 1:availCores) {
  message(sprintf("  - Testing with %d cores ...", cores))
  options(mc.cores = cores)
  
  for (strategy in supportedStrategies(cores)) {
    message(sprintf("* plan('%s') ...", strategy))
    plan(strategy)
  
    set.seed(0xBEEF)
    seed0 <- get_random_seed()
    y <- lapply(x, FUN = function(i) i) |> futurize_and_verify(seed = FALSE)
    y <- unlist(y)
    seed <- get_random_seed()
    if (is.null(y0)) {
      y0 <- y
      seed00 <- seed
    }
    str(list(y = y))
    stopifnot(identical(seed, seed0), identical(seed, seed00))
    ## NOTE: We cannot guarantee the same random numbers, because
    ## future.seed = FALSE.
    
    message(sprintf("* plan('%s') ... DONE", strategy))
  }  ## for (strategy ...)
  message(sprintf("  - Testing with %d cores ... DONE", cores))
} ## for (core ...)

message("* lapply(x, ..., future.seed = FALSE) |> futurize() ... DONE")


seed_sets <- list(
  A = TRUE,
##  B = NA,
  C = 42L,
  D = as_lecyer_cmrg_seed(42L),
  E = list(),
  F = vector("list", length = length(x)),
  G = NULL
)

## Generate sequence of seeds of the current RNGkind()
## NOTE: This is NOT a good way to generate random seeds!!!
seeds <- lapply(seq_along(x), FUN = function(i) {
  set.seed(i)
  globalenv()$.Random.seed
})
seed_sets$E <- seeds

## Generate sequence of L'Ecyer CMRG seeds
seeds <- seed_sets$F
seeds[[1]] <- seed_sets$D
for (kk in 2:length(x)) seeds[[kk]] <- parallel::nextRNGStream(seeds[[kk - 1]])
seed_sets$F <- seeds
seed_sets$G <- seed_sets$A

rm(list = "seeds")

for (name in names(seed_sets)) {
  future.seed <- seed_sets[[name]]

  if (is.list(future.seed)) {
    label <- sprintf("<list of %d seeds each being a %d-int seed>",
                     length(future.seed), length(future.seed[[1]]))
  } else {
    label <- hpaste(future.seed)
  }
  message(sprintf("* lapply(x, ..., future.seed = %s) |> futurize() ...", label))
  
  set.seed(0xBEEF)
  y0 <- seed00 <- NULL

  for (cores in 1:availCores) {
    message(sprintf("  - Testing with %d cores ...", cores))
    options(mc.cores = cores)
  
    for (strategy in supportedStrategies(cores)) {
      message(sprintf("* plan('%s') ...", strategy))
      plan(strategy)
      
      set.seed(0xBEEF)
      seed0 <- get_random_seed()
      y <- lapply(x, FUN = function(i) {
        rnorm(1L)
      }) |> futurize_and_verify(seed = future.seed)
      y <- unlist(y)
      seed <- get_random_seed()
      if (is.null(y0)) {
        y0 <- y
        seed00 <- seed
      }
      str(list(y = y))
      stopifnot(!identical(seed, seed0), identical(seed, seed00),
                identical(y, y0))
  
      ## RNG-based results should also be identical regardless of
      ## load-balance scheduling.
      for (scheduling in list(FALSE, TRUE, 0, 0.5, 2.0, Inf)) {
        set.seed(0xBEEF)
        seed0 <- get_random_seed()
        y <- lapply(x, FUN = function(i) {
          rnorm(1L)
        }) |> futurize_and_verify(seed = future.seed, scheduling = scheduling)
        seed <- get_random_seed()
        y <- unlist(y)
        str(list(y = y))
        stopifnot(!identical(seed, seed0), identical(seed, seed00),
                  identical(y, y0))
      }
  
      ## Nested future_lapply():s
      for (scheduling in list(FALSE, TRUE)) {
        y <- lapply(x, FUN = function(i) {
          .seed <- globalenv()$.Random.seed
          
          z <- lapply(1:3, FUN = function(j) {
            list(j = j, seed = globalenv()$.Random.seed)
          }) |> futurize_and_verify(seed = .seed)
    
          ## Assert that all future seeds are unique
          seeds <- lapply(z, FUN = function(x) x$seed)
          for (kk in 2:length(seeds)) stopifnot(!all(seeds[[kk]] == seeds[[1]]))
          
          list(i = i, seed = .seed, sample = rnorm(1L), z = z)
        }) |> futurize_and_verify(seed = 42L, scheduling = scheduling)
  
        if (is.null(y0_nested)) y0_nested <- y
        str(list(y = y))
    
        ## Assert that all future seeds (also nested ones) are unique
        seeds <- Reduce(c, lapply(y, FUN = function(x) {
          c(list(seed = x$seed), lapply(x$z, FUN = function(x) x$seed))
        }))
        for (kk in 2:length(seeds)) stopifnot(!all(seeds[[kk]] == seeds[[1]]))
        
        stopifnot(identical(y, y0_nested))
      }
      
      message(sprintf("* plan('%s') ... DONE", strategy))
    } ## for (strategy ...)
    message(sprintf("  - Testing with %d cores ... DONE", cores))
  } ## for (cores ...)
  
  message(sprintf("* lapply(x, ..., future.seed = %s) |> futurize() ... DONE", label))

} ## for (name ...)


message("*** future_lapply() and RNGs ... DONE")



} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags skip_on_cran  ## (35s) to limit total check time
if (requireNamespace("scuttle") && requireNamespace("doFuture") && requireNamespace("DelayedArray")) {
library(futurize)
library(scuttle)

## Use a small block size to ensure multiple blocks and thus multiple futures
DelayedArray::setAutoBlockSize(10000)

plan(multisession)

## Create a simple SingleCellExperiment
set.seed(42)
sce <- mockSCE(ncells = 20L, ngenes = 70L, nspikes = 30L)

## ---------------------------------------------------------
## perFeatureQCMetrics()
## ---------------------------------------------------------
result_truth <- perFeatureQCMetrics(sce)

result <- perFeatureQCMetrics(sce) |> futurize_and_verify()
stopifnot(all.equal(result, result_truth))

plan(sequential)
} ## if (requireNamespace("scuttle") && ...)

#' @tags pkg-shapr
if (requireNamespace("shapr")) {
  library(futurize)
  library(shapr)
  options(future.rng.onMisuse = "error")

  plan(multisession)

  ## Simple linear model
  set.seed(42)
  x_train <- data.frame(x1 = rnorm(100), x2 = rnorm(100))
  y_train <- 2 * x_train$x1 + x_train$x2 + rnorm(100)
  model <- lm(y_train ~ x1 + x2, data = x_train)
  x_explain <- data.frame(x1 = rnorm(3), x2 = rnorm(3))
  phi0 <- mean(y_train)

  ## Compare Shapley values only (ignoring metadata like timing and paths)
  all_equal <- function(a, b, ...) {
    all.equal(a$shapley_values, b$shapley_values, ...)
  }

  ## -------------------------------------------------------
  ## explain()
  ## -------------------------------------------------------
  set.seed(42)
  result_truth <- shapr::explain(
    model = model,
    x_explain = x_explain,
    x_train = x_train,
    approach = "empirical",
    phi0 = phi0,
    verbose = NULL
  )
  print(result_truth)

  ## Futurized via explain()
  set.seed(42)
  result <- explain(
    model = model,
    x_explain = x_explain,
    x_train = x_train,
    approach = "empirical",
    phi0 = phi0,
    verbose = NULL
  ) |> futurize_and_verify()
  print(result)

  stopifnot(all_equal(result, result_truth))

  ## Futurized via shapr::explain()
  set.seed(42)
  result2 <- shapr::explain(
    model = model,
    x_explain = x_explain,
    x_train = x_train,
    approach = "empirical",
    phi0 = phi0,
    verbose = NULL
  ) |> futurize_and_verify()
  print(result2)

  stopifnot(all_equal(result2, result_truth))


  plan(sequential)
} ## if (requireNamespace("shapr"))

#' @tags pkg-doFuture
#' @tags %dofuture%
#' @tags sequential multisession cluster multicore
#' @tags detritus-files

if (requireNamespace("foreach") && requireNamespace("doFuture")) {
library(futurize)
library(foreach)

## WORKAROUND: To avoid `R CMD check --as-cran` on MS Windows triggering:
##
## * checking for detritus in the temp directory ... NOTE
##   Found the following files/directories:
##    'Rscript171866c62e'
##
## we can set cluster nodes to use a temporary folder *inside* the parent
## temporary folder. This will hide the above files from 'R CMD check'.
## See https://bugs.r-project.org/show_bug.cgi?id=18133 for more details.
if (.Platform$OS.type == "windows") Sys.setenv(TMPDIR = tempdir())

strategies <- future:::supportedStrategies()

options(future.debug = FALSE)
options(doFuture.debug = FALSE)

plan_nested <- "sequential"

message("*** doFuture - nested %dofuture% ...")

message("*** doFuture - nested %dofuture%  and tricky globals ...")

## This works ...
x <- foreach(j = 1) %do% { j } |> futurize()
str(x)
rm(list = "x")

## ... but this would give a "globals-not-found" error in
## doFuture (<= 0.4.0) because 'j' was interpreted as global variable
x <- foreach(i = 1) %do% {
  foreach(j = 1) %do% { j } |> futurize()
} |> futurize(packages = "foreach")
str(x)
rm(list = "x")

message("*** doFuture - nested %dofuture%  and tricky globals ... DONE")


for (strategy1 in strategies) {
  for (strategy2 in strategies) {
    message(sprintf("- plan(list('%s', '%s')) ...", strategy1, strategy2))
    plan(list(a = strategy1, b = strategy2))

    as <- 1:2
    bs <- 3:1

    stopifnot(!exists("a", inherits = FALSE), !exists("b", inherits = FALSE))

    message("foreach() - level 1 ...")
    x <- foreach(a = as) %do% {
      plan_list <- future::plan("next")
      stopifnot(inherits(plan_list, strategy2))
      plan_a <- future::plan("list")
      str(plan_a)
      stopifnot(inherits(plan_a[[1]], strategy2))

      message("foreach() - level 2 ...")
      y <- foreach(b = bs) %do% {
        plan_list <- future::plan("next")
        message(capture.output(print(plan_list)))

        stopifnot(inherits(plan_list, "future"))
        ## future.plan can be either a string or a future function
        default <- getOption("future.plan", "sequential")
        if (is.function(default)) default <- class(default)
        message("Option 'future.plan': ", sQuote(default))
        stopifnot(inherits(plan_list, default))

        plan_b <- future::plan("list")
        str(plan_b)
        stopifnot(
          inherits(plan_b[[1]], "future"),
          inherits(plan_b[[1]], default)
        )

        list(a = a, plan_a = plan_a,
             b = b, plan_b = plan_b)
      } |> futurize(globals = c("a", "plan_a"))
      message("foreach() - level 2 ... DONE")

      y
    } |> futurize(globals = c("bs", "strategy2"), packages = c("foreach", "futurize"))
    message("foreach() - level 1 ... DONE")

    local({
      stopifnot(length(x) == length(as))
      for (aa in seq_along(as)) {
        x_aa <- x[[aa]]
        stopifnot(length(x_aa) == length(bs))
        a <- as[aa]
        for (bb in seq_along(bs)) {
          x_aa_bb <- x_aa[[bb]]
          b <- bs[bb]
          stopifnot(
            length(x_aa_bb) == 4L,
            all(names(x_aa_bb) == c("a", "plan_a", "b", "plan_b")),
            x_aa_bb$a == a,
            x_aa_bb$b == b,
            inherits(x_aa_bb$plan_a[[1]], strategy2),
            inherits(x_aa_bb$plan_b[[1]], "future"),
            inherits(x_aa_bb$plan_b[[1]], plan_nested)
          )
        }
      }
    })

    ## Cleanup in order make sure none of these variables exist as
    ## proxies for missing globals of the name names
    rm(list = c("as", "bs", "x"))

    ## WORKAROUND: Shut down *nested* parallel workers as an attempt to avoid
    ## * checking for detritus in the temp directory ... NOTE
    ## from 'R CMD check --as-cran' when running on MS Windows. This looks
    ## like a bug in R, cf. https://bugs.r-project.org/show_bug.cgi?id=18133    
    message("- shut down nested workers")
    dummy <- foreach(ii = 1:nbrOfWorkers()) %do% plan("sequential") |> futurize()
    plan("sequential")
    
    message(sprintf("- plan(list('%s', '%s')) ... DONE", strategy1, strategy2))
  }
}

message("*** doFuture - nested %dofuture% ... DONE")

plan(sequential)
} ## if (requireNamespace("foreach") && requireNamespace("doFuture"))

#' @tags skip_on_cran  ## (45s) to limit total check time
if (requireNamespace("GenomicAlignments") && requireNamespace("Rsamtools") && requireNamespace("doFuture")) {
library(futurize)
library(GenomicAlignments)
library(Rsamtools)
options(future.rng.onMisuse = "error")

plan(multisession)

## ---------------------------------------------------------
## summarizeOverlaps() with BamFileList
## ---------------------------------------------------------
## summarizeOverlaps() parallelizes via bplapply() when the
## input is a BamFileList, distributing work across BAM files.

## Create test BAM files (copies of example BAM)
fls <- system.file("extdata", "ex1.bam", package = "Rsamtools")
fls_idx <- paste0(fls, ".bai")
tmp_dir <- tempdir()
bam_files <- character(3L)
for (i in seq_along(bam_files)) {
  dst <- file.path(tmp_dir, sprintf("sample%d.bam", i))
  file.copy(fls, dst, overwrite = TRUE)
  if (file.exists(fls_idx))
    file.copy(fls_idx, paste0(dst, ".bai"), overwrite = TRUE)
  bam_files[i] <- dst
}

bf <- BamFileList(bam_files)
features <- GRanges("seq1",
  IRanges(start = c(1, 100, 200), end = c(50, 150, 250))
)
names(features) <- paste0("feature", 1:3)

result_truth <- summarizeOverlaps(features, bf)
str(result_truth)

result <- summarizeOverlaps(features, bf) |> futurize_and_verify()
str(result)
stopifnot(all.equal(
  SummarizedExperiment::assay(result),
  SummarizedExperiment::assay(result_truth)
))

result2 <- GenomicAlignments::summarizeOverlaps(features, bf) |> futurize_and_verify()
stopifnot(all.equal(
  SummarizedExperiment::assay(result2),
  SummarizedExperiment::assay(result_truth)
))

## Cleanup
file.remove(bam_files)
file.remove(paste0(bam_files, ".bai"))

plan(sequential)
} ## if (requireNamespace("GenomicAlignments") && ...)

#' @tags pkg-future.apply
#' @tags future_lapply
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)


library(tools) ## toTitleCase()

options(future.debug = FALSE)
options(future.apply.debug = TRUE)

message("*** future_lapply() - globals ...")

plan(cluster, workers = "localhost")

a <- 1
b <- 2

globals_set <- list(
  A = FALSE,
  B = TRUE,
  C = c("a", "b"),
  D = list(a = 2, b = 3)
)

x <- list(1)
y_truth <- list(A = NULL, B = list(1), C = list(1), D = list(2))
str(y_truth)

for (name in names(globals_set)) {
  globals <- globals_set[[name]]
  message("Globals set ", sQuote(name))
  y <- tryCatch({
    lapply(x, FUN = function(x) {
      median(c(x, a, b))
    }) |> futurize_and_verify(globals = globals, packages = "utils")
  }, FuturizeTestAssertionError = stop, error = identity)
  print(y)
  if (! "covr" %in% loadedNamespaces()) {
    stopifnot((name == "A" && inherits(y, "error")) || 
               identical(y, y_truth[[name]]))
  }
}

message("*** future_lapply() - globals ... DONE")


message("*** future_lapply() - manual globals ...")

d <- 42
y <- lapply(1:2, FUN = function(x) { x * d }) |> futurize_and_verify(globals = structure(FALSE, add = "d"))
stopifnot(identical(y, list(42, 84)))

e <- 42
res <- tryCatch({
  lapply(1:2, FUN = function(x) { 2 * e }) |> futurize_and_verify(globals = structure(TRUE, ignore = "e"))
}, FuturizeTestAssertionError = stop, error = identity)
if (! "covr" %in% loadedNamespaces()) {
  stopifnot(inherits(res, "error"))
}

message("*** future_lapply() - manual globals ... DONE")



## Test adopted from http://stackoverflow.com/questions/42561088/nested-do-call-within-a-foreach-dopar-environment-cant-find-function-passed-w

message("*** future_lapply() - tricky globals ...")

my_add <- function(a, b) a + b

call_my_add <- function(a, b) {
  do.call(my_add, args = list(a = a, b = b))
}

call_my_add_caller <- function(a, b, FUN = call_my_add) {
  do.call(FUN, args = list(a = a, b = b))
}

main <- function(x = 1:2, caller = call_my_add_caller,
                 args = list(FUN = call_my_add)) {
  results <- lapply(x, FUN = function(i) {
    do.call(caller, args = c(list(a = i, b = i + 1L), args))
  }) |> futurize_and_verify()
  results
}

x <- list(list(1:2))
z_length <- lapply(x, FUN = do.call, what = length)
fun <- function(...) sum(...)
z_fun <- lapply(x, FUN = do.call, what = fun)

y0 <- NULL
for (strategy in supportedStrategies()) {
  plan(strategy)

  y <- main(1:3)
  if (is.null(y0)) y0 <- y
  stopifnot(identical(y, y0))

  message("- lapply(x, FUN = do.call, ...) |> futurize() ...")
  z <- lapply(x, FUN = do.call, what = length) |> futurize_and_verify()
  stopifnot(identical(z, z_length))
  z <- lapply(x, FUN = do.call, what = fun) |> futurize_and_verify()
  stopifnot(identical(z, z_fun))

  message("- lapply(x, ...) |> futurize() - passing arguments via '...' ...")
  ## typeof() == "list"
  obj <- data.frame(a = 1:2)
  stopifnot(typeof(obj) == "list")
  y <- lapply(1L, function(a, b) typeof(b), b = obj) |> futurize_and_verify()
  stopifnot(identical(y[[1]], typeof(obj)))

  ## typeof() == "environment"
  obj <- new.env()
  stopifnot(typeof(obj) == "environment")
  y <- lapply(1L, function(a, b) typeof(b), b = obj) |> futurize_and_verify()
  stopifnot(identical(y[[1]], typeof(obj)))

  ## typeof() == "S4"
  if (requireNamespace("methods")) {
    obj <- methods::getClass("MethodDefinition")
    stopifnot(typeof(obj) == "S4")
    y <- lapply(1L, function(a, b) typeof(b), b = obj) |> futurize_and_verify()
    stopifnot(identical(y[[1]], typeof(obj)))
  }

  message("- lapply(X, ...) |> futurize() - 'X' containing globals ...")
  ## From https://github.com/futureverse/future.apply/issues/12
  a <- 42
  b <- 21
  X <- list(
    function(b) 2 * a,
    function() b / 2,
    function() a + b,
    function() nchar(tools::toTitleCase("hello world"))
  )
  z0 <- lapply(X, FUN = function(f) f())
  str(z0)
  z1 <- lapply(X, FUN = function(f) f()) |> futurize_and_verify()
  str(z1)
  stopifnot(identical(z1, z0))

#  message("- lapply(x, ...) |> futurize() - passing '...' as a global ...")
#  ## https://github.com/futureverse/future/issues/417
#  fcn0 <- function(...) { lapply(1, FUN = function(x) list(...)) }
#  z0 <- fcn0(a = 1)
#  str(list(z0 = z0))
#  stopifnot(identical(z0, list(list(a = 1))))
#  fcn <- function(...) { lapply(1, FUN = function(x) list(...)) } |> futurize_and_verify()
#  z1 <- fcn(a = 1)
#  str(list(z1 = z1))
#  stopifnot(identical(z1, z0))

  ## https://github.com/futureverse/future.apply/issues/47
  message("- lapply(X, ...) |> futurize() - '{ a <- a + 1; a }' ...")
  a <- 1
  z0 <- lapply(1, function(ii) {
    a <- a + 1
    a
  })
  z1 <- lapply(1, function(ii) {
    a <- a + 1
    a
  }) |> futurize_and_verify()
  stopifnot(identical(z1, z0))

  ## https://github.com/futureverse/future.apply/issues/47
  message("- lapply(X, ...) |> futurize() - '{ a; a <- a + 1 }' ...")
  z2 <- tryCatch(lapply(1, function(ii) {
    a
    a <- a + 1
  }) |> futurize_and_verify(), FuturizeTestAssertionError = stop, error = identity)
  stopifnot(identical(z2, z0))

  ## https://github.com/futureverse/future.apply/issues/85
  message("- lapply(..., future.globals = <list>) |> futurize() ...")
  a <- 0
  y <- lapply(1, FUN = function(x) a) |> futurize_and_verify(globals = list(a = 42))
  str(y)
  if (! "covr" %in% loadedNamespaces()) {
    stopifnot(y[[1]] == 42)
  }
} ## for (strategy ...)

message("*** future_lapply() - tricky globals ... DONE")


message("*** future_lapply() - missing arguments ...")

## Here 'abc' becomes missing, i.e. missing(abc) is TRUE
foo <- function(x, abc) lapply(x, FUN = function(y) y) |> futurize_and_verify()
y <- foo(1:2)
stopifnot(identical(y, as.list(1:2)))

message("*** future_lapply() - missing arguments ... DONE")


message("*** future_lapply() - false positives ...")

## Here 'abc' becomes a promise, which fails to resolve
## iff 'xyz' does not exist. (Issue #161)
suppressWarnings(rm(list = "xyz"))
foo <- function(x, abc) lapply(x, FUN = function(y) y) |> futurize_and_verify()
y <- foo(1:2, abc = (xyz >= 3.14))
stopifnot(identical(y, as.list(1:2)))

message("*** future_lapply() - false positives ... DONE")


message("*** future_lapply() - too large ...")

oMaxSize <- getOption("future.globals.maxSize")
X <- replicate(10L, 1:100, simplify = FALSE)
FUN <- function(x) {
  getOption("future.globals.maxSize")
}

y0 <- lapply(X, FUN = FUN)
stopifnot(all(sapply(y0, FUN = identical, oMaxSize)))

sizes <- unclass(c(FUN = object.size(FUN), X = object.size(X)))
cat(sprintf("Baseline size of globals: %.2f KiB\n", sizes[["FUN"]] / 1024))

message("- true positive ...")
options(future.globals.maxSize = 1L)
res <- tryCatch({
  y <- lapply(X, FUN = FUN) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))
res <- NULL
options(future.globals.maxSize = oMaxSize)

maxSize <- getOption("future.globals.maxSize")
y <- lapply(X, FUN = FUN) |> futurize_and_verify()
str(y)
stopifnot(all(sapply(y, FUN = identical, oMaxSize)))

message("- approximately invariant to chunk size ...")
maxSize <- sizes[["FUN"]] + sizes[["X"]] / length(X)
if ("covr" %in% loadedNamespaces()) maxSize <- maxSize + 65e3
options(future.globals.maxSize = maxSize)

for (chunk_size in c(1L, 2L, 5L, 10L)) {
  y <- lapply(X, FUN = FUN) |> futurize_and_verify(chunk_size = chunk_size)
  str(y)
  stopifnot(all(unlist(y) == maxSize))
  cat(sprintf("maxSize = %g bytes\nfuture.globals.maxSize = %g bytes\n",
              maxSize, getOption("future.globals.maxSize")))
  stopifnot(getOption("future.globals.maxSize") == maxSize)
}
y <- NULL
options(future.globals.maxSize = oMaxSize)


message("*** future_lapply() - too large ... DONE")


message("*** future_lapply() - globals exceptions ...")

res <- tryCatch({
  y <- lapply(1, FUN = function(x) x) |> futurize_and_verify(globals = 42)
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))

res <- tryCatch({
  y <- lapply(1, FUN = function(x) x) |> futurize_and_verify(globals = list(1))
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))

res <- tryCatch({
  y <- lapply(1, FUN = function(x) x) |> futurize_and_verify(globals = "...future.FUN")
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))

res <- tryCatch({
  y <- lapply(1, FUN = function(x) x) |> futurize_and_verify(globals = "...future.FUN")
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))

...future.elements_ii <- 42L
X <- list(function() 2 * ...future.elements_ii)
res <- tryCatch({
  y <- lapply(X, FUN = function(f) f()) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
if (! "covr" %in% loadedNamespaces()) {
  stopifnot(inherits(res, "error"))
}

message("*** future_lapply() - globals exceptions ... DONE")



} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags pkg-furrr
#' @tags detritus-files
if (requireNamespace("purrr") && requireNamespace("furrr")) {
library(futurize)


library(purrr)

# ------------------------------------------------------------------------------
# pmap()

message("future_pmap() matches pmap() for simple cases")
stopifnot(identical(
  pmap(list(1:3, 4:6, 7:9), ~.x + .y + ..3) |> futurize_and_verify(),
  pmap(list(1:3, 4:6, 7:9), ~.x + .y + ..3)
))


message("names of `.x` are retained")
x <- c(a = 1, b = 2)
y <- c(c = 1, d = 2)
stopifnot(identical(names(pmap(list(x, y), ~1) |> futurize_and_verify()), c("a", "b")))


message("named empty input makes named empty output")
x <- set_names(list(), character())
stopifnot(identical(names(pmap(list(x, x), ~.x) |> futurize::futurize()), character()))


# ------------------------------------------------------------------------------
# atomic variants

message("future_pmap_dbl() works")
x <- c(1, 2, 3)
y <- c(4, 5, 6)

stopifnot(identical(
  pmap_dbl(list(x, y), ~.x + .y) |> futurize_and_verify(),
  pmap_dbl(list(x, y), ~.x + .y)
))


message("future_pmap_int() works")
x <- c(1L, 2L, 3L)
y <- c(4L, 5L, 6L)

stopifnot(identical(
  pmap_int(list(x, y), ~.x + .y) |> futurize_and_verify(),
  pmap_int(list(x, y), ~.x + .y)
))


message("future_pmap_lgl() works")
x <- c(TRUE, FALSE, TRUE)
y <- c(FALSE, TRUE, TRUE)

stopifnot(identical(
  pmap_lgl(list(x, y), ~.x || .y) |> futurize_and_verify(),
  pmap_lgl(list(x, y), ~.x || .y)
))


message("future_pmap_chr() works")
x <- c("a", "b", "c")
y <- c("d", "e", "f")

stopifnot(identical(
  pmap_chr(list(x, y), ~.y) |> futurize_and_verify(),
  pmap_chr(list(x, y), ~.y)
))


message("names of `.x` are retained")
x <- c(a = 1, b = 2)
y <- c(c = 1, d = 2)
stopifnot(identical(names(pmap_dbl(list(x, y), ~1) |> futurize_and_verify()), c("a", "b")))


# ------------------------------------------------------------------------------
# data frame variants

message("future_pmap_dfr() works")
x <- c("a", "b", "c")
y <- c("d", "e", "f")

stopifnot(identical(
  pmap_dfr(list(x, y), ~data.frame(x = .x, y = .y)) |> futurize_and_verify(),
  pmap_dfr(list(x, y), ~data.frame(x = .x, y = .y))
))


message("future_pmap_dfc() works")
x <- c("a", "b", "c")
y <- c("d", "e", "f")

stopifnot(identical(
  pmap_dfc(list(x, y), ~as.data.frame(set_names(list(.x), .y))) |> futurize_and_verify(),
  pmap_dfc(list(x, y), ~as.data.frame(set_names(list(.x), .y)))
))


# ------------------------------------------------------------------------------
# size

message("future_pmap() works with completely empty list")
stopifnot(identical(pmap(list(), identity) |> futurize::futurize(), list()))
stopifnot(identical(pmap_dbl(list(), identity) |> futurize::futurize(), double()))


message("future_pmap() works with size zero input")
stopifnot(identical(pmap(list(list(), list()), identity) |> futurize::futurize(), list()))


message("atomic variants work with size zero input")
stopifnot(identical(pmap_chr(list(list(), list()), identity) |> futurize::futurize(), character()))
stopifnot(identical(pmap_dbl(list(list(), list()), identity) |> futurize::futurize(), double()))
stopifnot(identical(pmap_int(list(list(), list()), identity) |> futurize::futurize(), integer()))
stopifnot(identical(pmap_lgl(list(list(), list()), identity) |> futurize::futurize(), logical()))


message("size one recycling works")
stopifnot(identical(
  pmap(list(1, 1:2), ~c(.x, .y)) |> futurize_and_verify(),
  list(c(1, 1), c(1, 2))
))

stopifnot(identical(
  pmap(list(1:2, 1), ~c(.x, .y)) |> futurize_and_verify(),
  list(c(1, 1), c(2, 1))
))

stopifnot(identical(
  pmap(list(integer(), 1), ~c(.x, .y)) |> futurize::futurize(),
  list()
))

stopifnot(identical(
  pmap(list(1, integer()), ~c(.x, .y)) |> futurize::futurize(),
  list()
))


message("generally can't recycle to size zero")
res <- tryCatch({
  pmap(list(1:2, integer()), ~c(.x, .y)) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(
  inherits(res, "error"),
  grepl("Can't recycle", conditionMessage(res))
)

res <- tryCatch({
  pmap(list(integer(), 1:2), ~c(.x, .y)) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(
  inherits(res, "error"),
  grepl("Can't recycle", conditionMessage(res)
))


# ------------------------------------------------------------------------------
# Miscellaneous

message("named arguments can be passed through")
vec_mean <- function(.x, .y, na.rm = FALSE) {
  mean(c(.x, .y), na.rm = na.rm)
}

x <- list(c(NA, 1), 1:2)

stopifnot(identical(
  pmap(x, vec_mean, na.rm = TRUE) |> futurize_and_verify(),
  list(1, 1.5)
))


message("arguments can be matched by name")
x <- list(x = c(1, 2), y = c(3, 5))

fn <- function(y, x) {
  y - x
}

stopifnot(identical(pmap_dbl(x, fn) |> futurize_and_verify(), c(2, 3)))


message("unused components can be absorbed")
x <- list(c(1, 2), c(3, 5))

fn1 <- function(x) {
  x
}
fn2 <- function(x, ...) {
  x
}

res <- tryCatch({
  pmap_dbl(x, fn1) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(
  inherits(res, "error")
)
stopifnot(identical(pmap_dbl(x, fn2) |> futurize_and_verify(), c(1, 2)))


message("globals in `.x` and `.y` are found (#16)")
fn1 <- function(x) sum(x, na.rm = TRUE)
fn2 <- function(x) sum(x, na.rm = FALSE)

x <- list(c(1, 2, NA), c(2, 3, 4))

fns1 <- map(x, ~purrr::partial(fn1, x = .x))
fns2 <- map(x, ~purrr::partial(fn2, x = .x))

stopifnot(identical(
  pmap(list(fns1, fns2), ~c(.x(), .y())) |> futurize_and_verify(),
  list(c(3, NA), c(9, 9))
))


message("globals in `.l` are only exported to workers that use them")
plan(multisession, workers = 2)
on.exit(plan(sequential), add = TRUE)

# Use `local()` to ensure that the wrapper functions and the anonymous
# functions created with `~` don't pick up extra globals
my_wrapper1 <- local({
  my_mean1 <- function(x) mean(x, na.rm = TRUE)

  function(x) {
    my_mean1(x)
    exists("my_mean1")
  }
})

my_wrapper2 <- local({
  my_mean2 <- function(x) mean(x, na.rm = FALSE)

  function(x) {
    my_mean2(x)
    exists("my_mean1")
  }
})

x <- list(my_wrapper1, my_wrapper2)

stopifnot(identical(
  pmap_lgl(list(x), .f = ~.x(c(1, NA))) |> futurize_and_verify(),
  c(TRUE, FALSE)
))

plan(sequential)
} ## if (requireNamespace("purrr") && requireNamespace("furrr"))

# scuttle::logNormCounts(sce) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
#   options(future.disposable = <future arguments>)
#   scuttle::logNormCounts(sce, BPPARAM = BiocParallel::DoparParam())
# })
#
append_transpilers_for_scuttle <- function() {
  ## These scuttle functions accept BPPARAM via '...' (S4 generics) or
  ## directly in their formals
  names <- c(
    "calculateAverage",
    "logNormCounts", "normalizeCounts",
    "perCellQCMetrics", "perFeatureQCMetrics",
    "addPerCellQCMetrics", "addPerFeatureQCMetrics",
    "addPerCellQC", "addPerFeatureQC",
    "numDetectedAcrossCells", "numDetectedAcrossFeatures",
    "sumCountsAcrossCells", "sumCountsAcrossFeatures",
    "summarizeAssayByGroup",
    "aggregateAcrossCells", "aggregateAcrossFeatures",
    "librarySizeFactors", "computeLibraryFactors",
    "geometricSizeFactors", "computeGeometricFactors",
    "medianSizeFactors", "computeMedianFactors",
    "pooledSizeFactors", "computePooledFactors",
    "fitLinearModel"
  )

  ns <- getNamespace("scuttle")
  transpilers <- list()
  for (name in names) {
    if (exists(name, mode = "function", envir = ns, inherits = TRUE)) {
      args <- list(BPPARAM = quote(BiocParallel::DoparParam()))
      transpilers[[name]] <- list(
        label = sprintf("scuttle::%s() ~> scuttle::%s(..., BPPARAM = BiocParallel::DoparParam())", name, name),
        transpiler = make_futurize_for_doFuture(
          args = args,
          defaults = list(label = sprintf("fz:scuttle::%s-%%d", name))
        )
      )
    }
  }

  transpilers <- list(transpilers)
  names(transpilers) <- "scuttle"

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("scuttle", "BiocParallel", "doFuture")
}

#' @tags pkg-future.apply
#' @tags future_sapply
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)

message("*** future_sapply() ...")

xs <- list(
  A = c(a = 1, b = 2, c = 3),
  B = c(a = 1:2, b = 2:3, c = 3:4),
  C = letters[1:3],
  D = structure(10 * 5:8, names = LETTERS[1:4])
)

FUNS <- list(
  a = identity,
  b = as.matrix,
  c = function(x, y = 2 * 1:5) outer(rep(x, length.out = 3L), y)
)

for (strategy in supportedStrategies()) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)
  
  for (x in xs) {
    FUNS_x <- FUNS
    if (!is.numeric(x)) FUNS_x[["c"]] <- NULL
    
    for (USE.NAMES in list(FALSE, TRUE)) {
      for (simplify in list(FALSE, TRUE, "array")) {
        for (FUN in FUNS_x) {
          y0 <- sapply(x, FUN = FUN,
                              USE.NAMES = USE.NAMES, simplify = simplify)
          y1 <- sapply(x, FUN = FUN,
                              USE.NAMES = USE.NAMES, simplify = simplify) |> futurize()
          str(list(y0 = y0, y1 = y1))
          stopifnot(identical(y1, y0))
          if (identical(simplify, FALSE)) {
             y2 <- lapply(x, FUN = FUN)
             str(list(y0 = y0, y2 = y2))
             stopifnot(identical(unname(y2), unname(y0)))
          }
        }
      }
    }
  }

  y0 <- sapply(1:3, FUN = "sqrt")
  y1 <- sapply(1:3, FUN = "sqrt") #|> futurize()
  stopifnot(identical(y1, y0))
  
  ## https://github.com/futureverse/future.apply/issues/61
#  compute <- function(a, x_vec) a + x_vec
#  call_compute <- function(..., x_vec = 1:2){
#    compute_with_dots <- function(x) compute(..., x_vec = x)
#    sapply(x_vec, FUN = compute_with_dots) |> futurize()
#  }
#  y <- call_compute(0L)
#  print(y)
#  stopifnot(identical(y, 1:2))
  
  plan(sequential)
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
} ## for (strategy in ...) 

message("*** future_sapply() ... DONE")

} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags pkg-future.apply
if (requireNamespace("future.apply")) {
library(futurize)
library(stats)
library(datasets)
options(future.rng.onMisuse = "error")
options(futurize.debug = TRUE)

plan(multisession)

y <- lapply(X = 1:3, FUN = function(x) { print(x) }) |> futurize_and_verify(stdout = FALSE)
print(y)

xs <- list(aa = 1, bb = 1:2, cc = 1:10, dd = 1:5, .ee = -6:6)

X <- EuStockMarkets[,1:2]
k <- kernel("daniell", m = 50L)
stopifnot(inherits(X, "matrix"), inherits(X, "ts"), inherits(k, "tskernel"))

FUN_no_rng <- function(x, na.rm = TRUE) {
  a <- 1:5
  add <- NULL
  if (length(x) == 2) add <- list(C = 42)
  median(c(a, x), na.rm = na.rm)
}

FUN_rng <- function(x, na.rm = TRUE) {
  dummy <- sample.int(10L)
  a <- 1:5
  add <- NULL
  if (length(x) == 2) add <- list(C = 42)
  median(c(a, x), na.rm = na.rm)
}

es <- as.environment(xs)

exprs <- list(
     lapply = quote( lapply(X = xs, FUN = FUN) ),
     lapply = quote( base::lapply(X = xs, FUN = FUN) ),
     sapply = quote( sapply(X = xs, FUN = FUN) ),
     sapply = quote( base::sapply(X = xs, FUN = FUN) ),
     sapply = quote( base::sapply(X = xs, FUN = FUN, simplify = FALSE) ),
     sapply = quote( base::sapply(X = xs, FUN = FUN, USE.NAMES = FALSE) ),
     vapply = quote( base::vapply(X = xs, FUN.VALUE = NA_real_, FUN = FUN) ),
     vapply = quote( base::vapply(X = xs, FUN.VALUE = NA_real_, FUN = FUN, USE.NAMES = FALSE) ),
     eapply = quote( base::eapply(env = es, FUN = FUN) ),
     eapply = quote( base::eapply(env = es, FUN = FUN, all.names = TRUE) ),
     eapply = quote( base::eapply(env = es, FUN = FUN, USE.NAMES = FALSE) ),
  replicate = quote( replicate(2, 42) ),
  replicate = quote( base::replicate(2, 42) ),
  kernapply = quote( kernapply(x = X, k = k) )
)

flavors <- c("add-on", "built-in")

for (flavor in flavors) {
  for (kk in seq_along(exprs)) {
    name <- names(exprs)[kk]
    expr <- exprs[[kk]]
    message()
    message(sprintf("=== %s ==========================", name))
    print(expr)
    message(sprintf("---------------------------------"))
    
    if (flavor == "built-in") {
      if (name %in% c("replicate", "kernapply")) {
        message(sprintf("Skipping %s() - not yet implemented for flavor = %s", name, sQuote(flavor)))
        next
      }
    }
  
    FUN <- FUN_no_rng
    truth <- eval(expr)
    named_truth <- !is.null(names(truth))
  
    ## SPECIAL CASE: eapply() does not guarantee the order. To compare results
    ## later, we sort the results by name if they exist, otherwise by value.
    ## From ?eapply: "Note that the order of the components is arbitrary
    ## for hashed environments."
    if (name == "eapply" && !named_truth) {
      truth <- truth[order(unlist(truth))]
    }
    
    FUN <- FUN_no_rng

    expr_f <- bquote(.(expr) |> futurize_and_verify(flavor = .(flavor)))

    res <- eval(expr_f)

    if (name == "eapply") {
      res <- res[if (named_truth) names(truth) else order(unlist(res))]
    }
    stopifnot(identical(res, truth))
  
    out <- utils::capture.output({
      expr_f2 <- bquote(.(expr) |> futurize_and_verify(stdout = FALSE, conditions = character(0L), flavor = .(flavor)))
      res2 <- eval(expr_f2)
    })
    print(out)
  
    if (name == "eapply") {
      res2 <- res2[if (named_truth) names(truth) else order(unlist(res2))]
    }
    stopifnot(
      identical(res2, truth),
      identical(res2, res),
      identical(out, character(0L))
    )
    
    expr_f3 <- bquote(.(expr) |> futurize_and_verify(chunk_size = 1L, flavor = .(flavor)))
    res3 <- eval(expr_f3)
    if (name == "eapply") {
      res3 <- res3[if (named_truth) names(truth) else order(unlist(res3))]
    }
    stopifnot(
      identical(res3, truth),
      identical(res3, res)
    )
    
    message("Test with RNG:")
    FUN <- FUN_rng
    expr_f4 <- bquote(.(expr) |> futurize_and_verify(seed = TRUE, flavor = .(flavor)))
    print(expr_f4)
    res4 <- local({
      opts <- options(future.rng.onMisuse = "error")
      on.exit(options(opts))
      eval(expr_f4)
    })
  
    if (name == "eapply") {
      res4 <- res4[if (named_truth) names(truth) else order(unlist(res4))]
    }
    stopifnot(
      identical(res4, truth),
      identical(res4, res)
    )
  } ## for (kk ...)
} ## for (flavor ...)

message("futurize() for replicate() should default to seed = TRUE")
y <- replicate(2, rnorm(1)) |> futurize_and_verify()

## Switch to 'sequential' already here to avoid detritus files on Windows
plan(sequential)

message("futurize(seed = FALSE) gives RNG error with replicate()")
y <- tryCatch(replicate(2, rnorm(1)) |> futurize_and_verify(seed = FALSE), FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(y, "error"))

message("Special case: Zero futurize() options")
y <- lapply(1, identity) |> futurize_and_verify(options = list())

} ## if (requireNamespace("future.apply"))

#' @tags skip_on_cran  ## (35s) to limit total check time
if (requireNamespace("scater") && requireNamespace("doFuture")) {
library(futurize)
library(scater)
options(future.rng.onMisuse = "error")

plan(multisession)

## Create a simple SingleCellExperiment
set.seed(42)
n_cells <- 100L
sce <- scuttle::mockSCE(ncells = n_cells, ngenes = 200L)
sce <- scuttle::logNormCounts(sce)

## Standardize PCA sign so that the first loading of each
## component is positive.  This makes results comparable
## regardless of the arbitrary sign chosen by the SVD solver.
standardize_pca_sign <- function(pca) {
  rotation <- attr(pca, "rotation")
  signs <- sign(rotation[1L, ])
  signs[signs == 0] <- 1
  pca[] <- sweep(pca, 2L, signs, FUN = `*`)
  attr(pca, "rotation") <- sweep(rotation, 2L, signs, FUN = `*`)
  pca
}

## ---------------------------------------------------------
## runPCA()
## ---------------------------------------------------------
set.seed(42)
result_truth <- runPCA(sce)
result_truth <- reducedDim(result_truth, "PCA")
result_truth <- standardize_pca_sign(result_truth)
str(result_truth)

set.seed(42)
result <- runPCA(sce) |> futurize::futurize() ## FIXME: futurize_and_verify()
result <- reducedDim(result, "PCA")
result <- standardize_pca_sign(result)
str(result)
stopifnot(isTRUE(all.equal(result, result_truth, tolerance = 1e-6)))

set.seed(42)
result2 <- scater::runPCA(sce) |> futurize::futurize() ## FIXME: futurize_and_verify()
result2 <- reducedDim(result2, "PCA")
result2 <- standardize_pca_sign(result2)
str(result2)
stopifnot(all.equal(result2, result_truth, tolerance = 1e-6))


## ---------------------------------------------------------
## getVarianceExplained()
## ---------------------------------------------------------
sce2 <- sce
sce2$group <- factor(rep(c("A", "B"), each = n_cells / 2L))

result_truth <- getVarianceExplained(sce2, variables = "group")

result <- getVarianceExplained(sce2, variables = "group") |> futurize_and_verify()
stopifnot(isTRUE(all.equal(result, result_truth, tolerance = 1e-6)))

plan(sequential)
} ## if (requireNamespace("scater") && ...)

#' @tags pkg-future.apply

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)

message("*** future_replicate() ...")

for (strategy in supportedStrategies()) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)
  
  y0 <- replicate(5L, sample(10L, size = 1L))
  y1 <- replicate(5L, sample(10L, size = 1L)) |> futurize()
  stopifnot(length(y0) == length(y1))
  
  set.seed(0xBEEF)
  y1 <- replicate(5L, sample(10L, size = 1L)) |> futurize()
  set.seed(0xBEEF)
  y2 <- replicate(5L, sample(10L, size = 1L)) |> futurize()
  stopifnot(all.equal(y2, y1))
  
  y3 <- replicate(5L, sample(10L, size = 1L)) |> futurize(seed = 0xBEEF)
  y4 <- replicate(5L, sample(10L, size = 1L)) |> futurize(seed = 0xBEEF)
  stopifnot(all.equal(y4, y3))
  
  
  message("- example(replicate) ...")
  
  foo <- function(x = 1, y = 2) c(x, y)
  bar0 <- function(n, x) replicate(n, foo(x = x))
  y0 <- bar0(5, x = 3)
  
  bar1 <- function(n, x) replicate(n, foo(x = x)) |> futurize()
  y1 <- bar1(5, x = 3)
  stopifnot(all.equal(y1, y0))
  
  plan(sequential)
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
} ## for (strategy in ...) 

message("*** future_replicate() ... DONE")



} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags pkg-future.apply
#' @tags future_eapply
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)

message("*** future_eapply() ...")

message("- From example(eapply) ...")

for (strategy in supportedStrategies()) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)
  
  env <- new.env(hash = FALSE)
  env$a <- 1:10
  env$beta <- exp(-3:3)
  env$logic <- c(TRUE, FALSE, FALSE, TRUE)
  
  y0 <- unlist(eapply(env, mean, USE.NAMES = FALSE))
  y1 <- unlist(eapply(env, mean, USE.NAMES = FALSE) |> futurize())
  stopifnot(all.equal(y1, y0))
  
  y0 <- eapply(env, quantile, probs = 1:3/4)
  y1 <- eapply(env, quantile, probs = 1:3/4) |> futurize()
  stopifnot(all.equal(y1, y0))
  
  y0 <- eapply(env, quantile)
  y1 <- eapply(env, quantile) |> futurize()
  stopifnot(all.equal(y1, y0))
  y2 <- eapply(env, "quantile") |> futurize()
  stopifnot(all.equal(y2, y0))

  plan(sequential)
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
} ## for (strategy in ...) 

message("*** future_eapply() ... DONE")



} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags pkg-TSP
if (requireNamespace("TSP") && requireNamespace("doFuture")) {
library(futurize)
library(TSP)
options(future.rng.onMisuse = "error")

plan(multisession)

data("USCA50")

RNGkind("L'Ecuyer-CMRG")

set.seed(42)
y_truth <- solve_TSP(USCA50, method = "nn", rep = 10L)
print(y_truth)

set.seed(42)
y <- solve_TSP(USCA50, method = "nn", rep = 10L) |> futurize_and_verify()
print(y)

set.seed(42)
y2 <- solve_TSP(USCA50, method = "nn", rep = 10L) |> futurize_and_verify()
print(y2)
stopifnot(all.equal(y2, y))

plan(sequential)

set.seed(42)
y3 <- solve_TSP(USCA50, method = "nn", rep = 10L) |> futurize_and_verify()
print(y3)
stopifnot(all.equal(y3, y))

} ## if (requireNamespace("TSP"))

# Used by transpilers for the future.apply, furrr packages
#'
#' @return
#' A named list of lists of transpilers, where the name of the list is the package name.
#' The list of transpilers is also a named list, where each element is a transpiler
#' and the corresponding name is the function transpiled.
#' A transpiler is a named list with elements:
#'
#'  * `label` - a character string describing the transpiler
#'
#'  * `transpiler` - a function that takes an R expression and
#'                   an optional argument `options`
#'
#' @keywords internals
#' @noRd
make_addon_transpilers <- function(from_package, to_package, make_options) {
  call_template <- as.call(list(as.symbol("::"), as.symbol(to_package), as.symbol("<place-holder>")))
  make_call <- function(name) {
    call <- call_template
    call[[3]] <- as.symbol(name)
    call
  }

  transpilers <- list()
  
  exports <- names(getNamespaceInfo(to_package, "exports"))
  names <- grep("^future_", exports, value = TRUE)
  for (name in names) {
    basename <- sub("^future_", "", name)

    defaults <- list(
      ## for future.apply
      future.label = sprintf("fz:%s::%s-%%d", from_package, basename),
      ## for furrr
      prefix       = sprintf("fz:%s::%s", from_package, basename)
    )

    transpiler <- eval(bquote(function(expr, options = NULL) {
      call <- make_call(.(name))
      fcn <- eval(call)
      expr[[1]] <- call
      parts <- c(as.list(expr), .(make_options)(options, fcn, defaults = .(defaults)))
      expr <- as.call(parts)
      expr
    }))
    body(transpiler) <- body(transpiler)

    transpilers[[basename]] <- list(
      label = sprintf("%s::%s() -> %s::%s()", from_package, basename, to_package, name),
      transpiler = transpiler
    )
  }

  transpilers <- list(transpilers)
  names(transpilers) <- from_package
  transpilers
} ## make_addon_transpilers()

## Travis CI specific: Explicit garbage collection because it
## looks like Travis CI might run out of memory during 'covr'
## testing and we now have so many tests. /HB 2017-01-11
if ("covr" %in% loadedNamespaces()) {
  res <- gc()
  testme <- as.environment("testme")
  if (testme[["debug"]]) print(res)
}

#' @tags pkg-strucchange
if (requireNamespace("strucchange") && requireNamespace("doFuture")) {
library(futurize)
library(strucchange)
options(future.rng.onMisuse = "error")

plan(multisession)

data("Nile")

bp_truth <- breakpoints(Nile ~ 1)
print(bp_truth)

bp <- breakpoints(Nile ~ 1) |> futurize_and_verify()
print(bp)

common <- setdiff(names(bp), c("RSS", "extract.breaks", "extend.RSS.table", "nobs", "call", "datatsp"))
stopifnot(all.equal(bp[common], bp_truth[common]))

plan(sequential)
} ## if (requireNamespace("strucchange"))

# rugarch::arfimacv(..., cluster = cl) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   oopts <- options(future.ClusterFuture.clusterEvalQ = "ignore")
#   on.exit(options(oopts))
#   rugarch::arfimacv(..., cluster = cl)
# })
#
append_transpilers_for_rugarch <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'rugarch' functions requires R (>= 4.4.0)", getRversion()))
  }

  template_ignore_clusterEvalQ <- bquote_compile(
    local({
      cl <- do.call(.(CALL), args = .(OPTS))
      ## rugarch uses clusterEvalQ() for library(rugarch), which
      ## is already taken care of by the future framework
      oopts <- options(future.ClusterFuture.clusterEvalQ = "ignore")
      on.exit(options(oopts))
      .(EXPR)
    })
  )

  transpilers <- make_package_transpilers("rugarch", FUN = function(fcn, name) {
    if ("cluster" %in% names(formals(fcn))) {
      transpiler <- make_futurize_for_makeClusterFuture(
        template = template_ignore_clusterEvalQ,
        args = list(
          cluster = quote(cl)
        ),
        defaults = list(
          label = sprintf("fz:rugarch::%s", name),
          seed = TRUE
        )
      )
      
      list(
        label = sprintf("rugarch::%s() ~> rugarch::%s(..., cluster = cl)", name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("rugarch", "future")
}

#' @tags pkg-future.apply
#' @tags future_mapply
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)



message("*** future_mapply() ...")

message("- Parallel RNG truth (for later)...")
plan(sequential)
y_rng_0 <- mapply(stats::runif, n = 1:4, max = 2:5,
                         MoreArgs = list(min = 1)) |> futurize_and_verify(seed = 0xBEEF)
print(y_rng_0)

for (strategy in supportedStrategies()) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)

  message("- From example(mapply) ...")
  y0 <- mapply(rep, 1:4, 4:1)
  y1 <- mapply(rep, 1:4, 4:1) |> futurize_and_verify()
  stopifnot(identical(y1, y0))
  y2 <- mapply("rep", 1:4, 4:1) |> futurize_and_verify()
  stopifnot(identical(y2, y0))
  
  y0 <- mapply(rep, times = 1:4, x = 4:1)
  y1 <- mapply(rep, times = 1:4, x = 4:1) |> futurize_and_verify()
  stopifnot(identical(y1, y0))
  
  y0 <- mapply(rep, times = 1:4, MoreArgs = list(x = 42))
  y1 <- mapply(rep, times = 1:4, MoreArgs = list(x = 42)) |> futurize_and_verify()
  stopifnot(identical(y1, y0))
  
  y0 <- mapply(function(x, y) seq_len(x) + y,
               c(a =  1, b = 2, c = 3),  # names from first
               c(A = 10, B = 0, C = -10))
  y1 <- mapply(function(x, y) seq_len(x) + y,
                      c(a =  1, b = 2, c = 3),  # names from first
                      c(A = 10, B = 0, C = -10)) |> futurize_and_verify()
  stopifnot(identical(y1, y0))

  message("- future_.mapply()")
  dots <- list(c(a =  1, b = 2, c = 3),  # names from first
               c(A = 10, B = 0, C = -10))
  y2 <- .mapply(function(x, y) seq_len(x) + y, dots = dots, MoreArgs = list())
  names(y0) <- NULL  ## .mapply() don't set names
  stopifnot(all.equal(y2, y0))
  y3 <- .mapply(function(x, y) seq_len(x) + y, dots = dots, MoreArgs = list()) |> futurize_and_verify()
  stopifnot(all.equal(y3, y2))

  word <- function(C, k) paste(rep.int(C, k), collapse = "")
  for (chunk_size in list(1L, structure(2L, ordering = "random"), structure(3L, ordering = 5:1))) {
    y0 <- mapply(word, LETTERS[1:5], 5:1, SIMPLIFY = FALSE)
    y1 <- mapply(word, LETTERS[1:5], 5:1, SIMPLIFY = FALSE) |> futurize_and_verify(chunk_size = chunk_size)
    stopifnot(identical(y1, y0))

    dots <- list(LETTERS[1:5], 5:1)
    MoreArgs <- list()
    y2 <- .mapply(word, dots = dots, MoreArgs = list())
    names(y0) <- NULL  ## .mapply() don't set names
    stopifnot(all.equal(y2, y0))
    y3 <- .mapply(word, dots = dots, MoreArgs = list()) |> futurize_and_verify()
    stopifnot(all.equal(y3, y2))
  }

  message("- Subsetting (Issue #17) ...")
  X <- as.Date("2018-06-01")
  y0 <- mapply(FUN = identity, X, SIMPLIFY = FALSE)
  y1 <- mapply(FUN = identity, X, SIMPLIFY = FALSE) |> futurize_and_verify()
  stopifnot(identical(y1, y0))

  dots <- list(X)
  y2 <- .mapply(FUN = identity, dots = dots, MoreArgs = MoreArgs)
  stopifnot(identical(y2, y0))
  y3 <- .mapply(FUN = identity, dots = dots, MoreArgs = MoreArgs) |> futurize_and_verify()
  stopifnot(identical(y3, y2))

  message("- Non-recycling of MoreArgs (Issue #51) ...")
  y0 <- base::mapply(
    function(x, y) y,
    x = 1:2, MoreArgs = list(y = 3:4)
  )
  y1 <- mapply(
    function(x, y) y,
    x = 1:2, MoreArgs = list(y = 3:4)
  ) |> futurize_and_verify(seed = FALSE)
  stopifnot(identical(y1, y0))
  y2 <- mapply(
    function(x, y) y,
    x = 1:2, MoreArgs = list(y = 3:4)
  ) |> futurize_and_verify(seed = TRUE)
  stopifnot(identical(y2, y0))

  dots <- list(x = 1:2)
  MoreArgs <- list(y = 3:4)
  y3 <- .mapply(function(x, y) y, dots = dots, MoreArgs = MoreArgs)
  y4 <- .mapply(function(x, y) y, dots = dots, MoreArgs = MoreArgs) |> futurize_and_verify()
  stopifnot(identical(y4, y3))

  message("- Recycle arguments to same length ...")
  y0 <- mapply(rep, 1:4, 2:1)
  y1 <- mapply(rep, 1:4, 2:1) |> futurize_and_verify()
  stopifnot(identical(y1, y0))

  
  message("- Parallel RNG ...")
  y_rng_1 <- mapply(stats::runif, n = 1:4, max = 2:5,
                           MoreArgs = list(min = 1)) |> futurize_and_verify(seed = 0xBEEF)
  print(y_rng_1)
  stopifnot(all.equal(y_rng_1, y_rng_0))

  message("- future_Map() ...")
  xs <- replicate(5, stats::runif(10), simplify = FALSE)
  ws <- replicate(5, stats::rpois(10, 5) + 1, simplify = FALSE)
  y0 <- Map(weighted.mean, xs, ws)
  y1 <- Map(stats::weighted.mean, xs, ws) |> futurize_and_verify()
  stopifnot(all.equal(y1, y0))

  message("- future_mapply() - 'max-or-0-if' recycling rule ...")
  ## R (>= 4.2.0): mapply() & Map() follow usual "max-or-0-if" recycling rule
  ## and keeps returning a named list in the "empty" case.
  
  truth <- list()
  if (getRversion() >= "4.2.0") {
    y0 <- mapply(`+`, 1:3, NULL)
    stopifnot(identical(y0, truth))
  }
  y <- mapply(`+`, 1:3, NULL) |> futurize::futurize()
  stopifnot(identical(y, truth))
  
  truth <- setNames(list(), character())
  if (getRversion() >= "4.2.0") {
    y0 <- mapply(paste, character(), NULL)
    stopifnot(identical(y0, truth))
  }
  y <- mapply(paste, character(), NULL) |> futurize::futurize()
  stopifnot(identical(y, truth))
  
  if (getRversion() >= "4.2.0") {
    y0 <- mapply(paste, character(), letters)
    stopifnot(identical(y0, truth))
  }
  y <- mapply(paste, character(), letters) |> futurize::futurize()
  stopifnot(identical(y, truth))
  
  if (getRversion() >= "4.2.0") {
    y0 <- mapply(paste, "A", character())
    stopifnot(identical(y0, truth))
  }
  y <- mapply(paste, "A", character()) |> futurize::futurize()
  stopifnot(identical(y, truth))
  
  if (getRversion() >= "4.2.0") {
    y0 <- mapply(paste, character(), letters) 
    stopifnot(identical(y0, truth))
  }
  y <- mapply(paste, character(), letters) |> futurize::futurize()
  stopifnot(identical(y, truth))
  
  if (getRversion() >= "4.2.0") {
    y0 <- mapply(paste, character(), NULL)
    stopifnot(identical(y0, truth))
  }
  y <- mapply(paste, character(), NULL) |> futurize::futurize()
  stopifnot(identical(y, truth))
  
  if (getRversion() >= "4.2.0") {
    y0 <- mapply(paste, character(), letters)
    stopifnot(identical(y0, truth))
  }
  y <- mapply(paste, character(), letters) |> futurize::futurize()
  stopifnot(identical(y, truth))
  
  if (getRversion() >= "4.2.0") {
    y0 <- mapply(paste, "A", character()) 
    stopifnot(identical(y0, truth))
  }
  y <- mapply(paste, "A", character()) |> futurize::futurize()
  stopifnot(identical(y, truth))
  
  if (getRversion() >= "4.2.0") {
    y0 <- mapply(paste, character(), letters) 
    stopifnot(identical(y0, truth))
  }
  y <- mapply(paste, character(), letters) |> futurize::futurize()
  stopifnot(identical(y, truth))

  ## Gives an error in R-devel (2021-11-26 r81252)
  if (getRversion() >= "4.2.0" && FALSE) {
    y0 <- mapply(paste, c(a = "A"), character())
    stopifnot(identical(y0, truth))
  }
  y <- mapply(paste, c(a = "A"), character()) |> futurize::futurize()
  stopifnot(identical(y, truth))

  ## R (>= 4.2.0): Map() now recycles similar to basic Ops:
  truth <- as.list(1 + 1:3)
  if (getRversion() >= "4.2.0") {
    y0 <- Map(`+`, 1, 1:3)
    stopifnot(identical(y0, truth))
  }
  y <- Map(`+`, 1, 1:3) |> futurize_and_verify()
  stopifnot(identical(y, truth))
  
  truth <- as.list(numeric() + 1:3)
  if (getRversion() >= "4.2.0") {
    y0 <- Map(`+`, numeric(), 1:3)
    stopifnot(identical(y0, truth))
  }
  y <- Map(`+`, numeric(), 1:3) |> futurize::futurize()
  stopifnot(identical(y, truth))

  message("- mapply(x, ...) |> futurize() where x[[i]] subsets via S3 method ...")
  x <- structure(list(a = 1, b = 2), class = "Foo")
  `[[.Foo` <- function(x, ...) 0
  registerS3method("[[", class = "Foo", method = `[[.Foo`)
  y0 <- mapply(x, FUN = identity)
  stopifnot(identical(y0, c(a = 0, b = 0)))
  y1 <- mapply(x, FUN = identity) |> futurize_and_verify()
  if (getOption("future.apply.chunkWith", "[[") == "[") {
    stopifnot(identical(y1, unlist(x)))
  } else {
    stopifnot(identical(y1, y0))
  }

  plan(sequential)
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
} ## for (strategy in ...) 

message("- Empty input [non parallel] ...")
y0 <- mapply(search)
y1 <- mapply(search) |> futurize::futurize()
stopifnot(identical(y1, y0))

y0 <- mapply(list, integer(0L))
y1 <- mapply(list, integer(0L)) |> futurize::futurize()
stopifnot(identical(y1, y0))

message("*** future_mapply() - special cases ...")

X <- list()
names(X) <- character(0L)

y <- mapply(FUN = identity, X) |> futurize::futurize()
stopifnot(length(y) == 0L, !is.null(names(y)), identical(y, X))

y <- mapply(FUN = identity, X, X) |> futurize::futurize()
stopifnot(length(y) == 0L, !is.null(names(y)), identical(y, X))

message("*** future_mapply() - special cases ... DONE")

message("*** future_mapply() ... DONE")

} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags pkg-tm
if (requireNamespace("tm") && getRversion() >= "4.4.0") {
library(futurize)
library(tm)
data(crude)

plan(multisession)

# -------------------------------------------------------------------------
# tm_map()
# -------------------------------------------------------------------------
## Use wrapper to apply character processing function
a0 <- tm_map(crude, content_transformer(tolower))
a1 <- tm_map(crude, content_transformer(tolower)) |> futurize_and_verify()
stopifnot(all.equal(a0, a1), identical(a0, a1))

## Generate a custom transformation function which takes the heading as new content
headings <- function(x) {
  PlainTextDocument(meta(x, "heading"),
                    id = meta(x, "id"),
                    language = meta(x, "language"))
}                    
b0 <- tm_map(crude, headings)
b1 <- tm_map(crude, headings) |> futurize_and_verify()
stopifnot(all.equal(a0, a1), identical(a0, a1))

plan(sequential)
} ## if (requireNamespace("tm"))

#' @tags pkg-future.apply
#' @tags future_vapply
#' @tags detritus-files
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)



message("*** future_vapply() ...")

for (strategy in supportedStrategies()) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)
  
  x <- NULL
  fun <- is.factor
  fun_name <- "is.factor"
  fun_value <- logical(1L)
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize::futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  y2 <- vapply(x, FUN = fun_name, FUN.VALUE = fun_value) |> futurize::futurize()
  str(y2)
  stopifnot(all.equal(y2, y0))

  x <- list()
  fun <- is.numeric
  fun_value <- logical(1L)
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize::futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  x <- integer()
  fun <- identity
  fun_value <- fun(integer(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize::futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  df <- data.frame(x = 1:10, y = letters[1:10], stringsAsFactors=FALSE)
  fun <- class
  fun_value <- character(1L)
  y0 <- vapply(df, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(df, FUN = fun, FUN.VALUE = fun_value) |> futurize_and_verify()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  x <- 1:10
  fun <- function(x) double(0L)
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize_and_verify()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  fun <- function(x) integer(0L)
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize_and_verify()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  fun <- sqrt
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize_and_verify()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  fun <- function(x) c(x, x^2)
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize_and_verify()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  fun <- function(x) matrix(x, nrow = 2L, ncol = 2L)
  fun_value <- fun(integer(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize_and_verify()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  fun <- function(x) matrix(x, nrow = 2L, ncol = 2L)
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize_and_verify()
  str(y1)
  stopifnot(all.equal(y1, y0))

  ## Ditto with dimnames on FUN.VALUE
  fun <- function(x) {
    matrix(x, nrow = 2L, ncol = 2L, dimnames = list(c("a", "b"), c("A", "B")))
  }
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize_and_verify()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  message("- From example(vapply) ...")
  x <- list(a = 1:10, beta = exp(-3:3), logic = c(TRUE, FALSE, FALSE, TRUE))
  y0 <- vapply(x, FUN = quantile, FUN.VALUE = double(5L))
  y1 <- vapply(x, FUN = quantile, FUN.VALUE = double(5L)) |> futurize_and_verify()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  i39 <- sapply(3:9, seq)
  ys0 <- sapply(i39, fivenum)
  ys1 <- sapply(i39, fivenum) |> futurize_and_verify()
  stopifnot(all.equal(ys1, ys0))
  
  yv0 <- vapply(i39, fivenum,
          c(Min. = 0, "1st Qu." = 0, Median = 0, "3rd Qu." = 0, Max. = 0))
  yv1 <- vapply(i39, fivenum,
          c(Min. = 0, "1st Qu." = 0, Median = 0, "3rd Qu." = 0, Max. = 0)) |> futurize_and_verify()
  str(yv1)
  stopifnot(all.equal(yv1, yv0))
  
  v <- structure(10*(5:8), names = LETTERS[1:4])
  f <- function(x, y) outer(rep(x, length.out = 3L), y)
  ys0 <- sapply(v, f, y = 2*(1:5), simplify = "array")
  ys1 <- sapply(v, f, y = 2*(1:5), simplify = "array") |> futurize_and_verify()
  stopifnot(all.equal(ys1, ys0))
  
  fv <- outer(1:3, 1:5)
  y <- 2*(1:5)
  yv0 <- vapply(v, f, fv, y = y)
  yv1 <- vapply(v, f, fv, y = y) |> futurize_and_verify()
  str(yv1)
  stopifnot(all.equal(yv1, yv0))
  
  y0 <- vapply(mtcars, FUN = is.numeric, FUN.VALUE = logical(1L))
  y1 <- vapply(mtcars, FUN = is.numeric, FUN.VALUE = logical(1L)) |> futurize_and_verify()
  str(y1)
  stopifnot(all.equal(y1, y0))

  message("- vapply(x, ...) |> futurize() where length(x) != length(as.list(x)) ...")
  x <- structure(list(a = 1, b = 2), class = "Foo")
  as.list.Foo <- function(x, ...) c(x, c = 3)
  y0 <- vapply(x, FUN = length, FUN.VALUE = -1L)
  y1 <- vapply(x, FUN = length, FUN.VALUE = -1L) |> futurize_and_verify()
  stopifnot(identical(y1, y0))

  message("- exceptions ...")
  res <- tryCatch({
    y0 <- vapply(1:3, FUN = identity, FUN.VALUE = c(3, 3))
  }, FuturizeTestAssertionError = stop, error = identity)
  stopifnot(inherits(res, "error"))
  res <- tryCatch({
    y1 <- vapply(1:3, FUN = identity, FUN.VALUE = c(3, 3)) |> futurize_and_verify()
  }, FuturizeTestAssertionError = stop, error = identity)
  stopifnot(inherits(res, "error"))
  
  plan(sequential)
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
} ## for (strategy in ...) 


message("*** future_vapply() ... DONE")

} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags skip_on_cran  ## to limit total check time
#' @tags pkg-Rsamtools
if (requireNamespace("Rsamtools") && requireNamespace("doFuture")) {
library(futurize)
library(Rsamtools)
options(future.rng.onMisuse = "error")

plan(multisession)

## ---------------------------------------------------------
## countBam() with BamViews
## ---------------------------------------------------------
## countBam() parallelizes via bplapply() when the input is
## a BamViews object, distributing work across BAM files.

## Create test BAM files (copies of example BAM)
fls <- system.file("extdata", "ex1.bam", package = "Rsamtools")
fls_idx <- paste0(fls, ".bai")
tmp_dir <- tempdir()
bam_files <- character(3L)
for (i in seq_along(bam_files)) {
  dst <- file.path(tmp_dir, sprintf("sample%d.bam", i))
  file.copy(fls, dst, overwrite = TRUE)
  if (file.exists(fls_idx))
    file.copy(fls_idx, paste0(dst, ".bai"), overwrite = TRUE)
  bam_files[i] <- dst
}

## Set up BamViews
bv <- BamViews(bam_files)

result_truth <- countBam(bv)
str(result_truth)

result <- countBam(bv) |> futurize_and_verify()
str(result)
stopifnot(all.equal(result, result_truth))


## ---------------------------------------------------------
## scanBam() with BamViews
## ---------------------------------------------------------
## scanBam() parallelizes via bplapply() when the input is
## a BamViews object, distributing work across BAM files.

result_truth <- scanBam(bv)
str(result_truth)

result <- scanBam(bv) |> futurize_and_verify()
str(result)
stopifnot(all.equal(result, result_truth))

## Cleanup
file.remove(bam_files)
file.remove(paste0(bam_files, ".bai"))

plan(sequential)
} ## if (requireNamespace("Rsamtools") && ...)

#' @tags detritus-files

if (requireNamespace("glmnet")) {
library(futurize)
library(glmnet)
options(future.rng.onMisuse = "error")

all_equal <- function(a, b, ...) {
  a$call <- NULL
  b$call <- NULL
  a$glmnet.fit$call <- NULL
  b$glmnet.fit$call <- NULL
  all.equal(a, b, ...)
}

plan(multisession)

## Adopted from example("cv.glmnet", package = "glmnet")
n <- 500L
p <- 20L
nzc <- trunc(p / 10)
x <- matrix(rnorm(n * p), n, p)
beta <- rnorm(nzc)
fx <- x[, seq(nzc)] %*% beta
eps <- rnorm(n) * 5
y <- drop(fx + eps)

set.seed(1011)
cv_truth <- cv.glmnet(x, y, nfolds = 10L)
print(cv_truth)

set.seed(1011)
cv <- cv.glmnet(x, y, nfolds = 10L) |> futurize_and_verify()
print(cv)

res <- all_equal(cv, cv_truth)
print(res)
stopifnot(res)

plan(sequential)
} ## if (requireNamespace("glmnet"))

library(futurize)
options(futurize.debug = TRUE)

message("*** Internals")

# --------------------------------------------------------------------
# Debug functions
# --------------------------------------------------------------------
message("debug_indent()")
oopts <- options(warn = 2L)
res <- tryCatch(futurize:::debug_indent(delta = -1L), error = identity)
print(res)
stopifnot(inherits(res, "error"))
options(oopts)


# --------------------------------------------------------------------
# .onLoad()
# --------------------------------------------------------------------
message(".onLoad()")
futurize:::.onLoad("futurize", "futurize")


# --------------------------------------------------------------------
# bquote_compile() and bquote_apply()
# --------------------------------------------------------------------
message("bquote_compile() and bquote_apply()")
bquote_compile <- futurize:::bquote_compile
bquote_apply <- futurize:::bquote_apply

message("- substitute NULL into second pairlist element")
## Substituting Y = NULL in function(a = .(X), b = .(Y)) { a + b } should
## set the second argument value to NULL. There used to be a bug that hard
## coded it to always update the first argument.
tmpl <- bquote_compile(function(a = .(X), b = .(Y)) { a + b })
expr <- bquote_apply(tmpl, X = 42, Y = NULL)
f <- formals(eval(expr))
stopifnot(identical(f$a, 42))
stopifnot(is.null(f$b))


# --------------------------------------------------------------------
# import_future() and import_from()
# --------------------------------------------------------------------
message("import_future() and import_from()")
import_from <- futurize:::import_from
import_future <- futurize:::import_future

fcn <- import_future("plan")
stopifnot(is.function(fcn))

fcn <- import_future("nonExistingFcn", default = identity)
stopifnot(identical(fcn, identity))

res <- tryCatch(
  import_future("nonExistingFcn"),
  error = identity
)
stopifnot(inherits(res, "error"))

fcn <- import_from("lapply", package = "base")
stopifnot(identical(fcn, base::lapply))

fcn <- import_from("nonExisting", package = "base", default = sum)
stopifnot(identical(fcn, sum))

res <- tryCatch(
  import_from("nonExisting", package = "base"),
  error = identity
)
stopifnot(inherits(res, "error"))

#' @tags pkg-metafor
if (requireNamespace("metafor") && getRversion() >= "4.4.0") {
library(futurize)
library(metafor)

plan(multisession)

## -------------------------------------------------------------------
## profile() - profile likelihood plots for rma objects
## -------------------------------------------------------------------
## Example from help("rma", package = "metafor")
dat <- escalc(measure = "RR", ai = tpos, bi = tneg,
              ci = cpos, di = cneg, data = dat.bcg)
fit <- rma(yi, vi, data = dat)

prof <- profile(fit) |> futurize_and_verify()
print(prof)

plan(sequential)
} ## if (requireNamespace("metafor"))

#' @tags pkg-future.apply
#' @tags future_by
#' @tags listenv
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)


library(listenv)

all_equal_but_call <- function(target, current, ...) {
  attr(target, "call") <- NULL
  attr(current, "call") <- NULL
  all.equal(target = target, current = current, ...)
}

message("*** future_by() ...")

## ---------------------------------------------------------
## by()
## ---------------------------------------------------------
if (require("datasets") && require("stats")) { ## warpbreaks & lm()
  ## Use a local variable to test that it is properly exported, because
  ## 'warpbreaks' is available in all R sessions
  data <- warpbreaks
  
  y0 <- by(data[, 1:2], INDICES = data[,"tension"], FUN = summary)
  y1 <- by(data[, 1], INDICES = data[, -1], FUN = summary, digits = 2L)
  y2 <- by(data, INDICES = data[,"tension"], FUN = function(x, ...) {
    lm(breaks ~ wool, data = x, ...)
  }, singular.ok = FALSE)
  
  ## now suppose we want to extract the coefficients by group
  tmp <- with(data, by(data, INDICES = tension, FUN = function(x) {
    lm(breaks ~ wool, data = x)
  }))
  y3 <- sapply(tmp, coef)

  ## Source: {r-source}/tests/reg-tests-1d.R
  by2 <- function(data, INDICES, FUN) {
    by(data, INDICES = INDICES, FUN = FUN)
  }
  future_by2 <- function(data, INDICES, FUN) {
    by(data, INDICES = INDICES, FUN = FUN) |> futurize_and_verify()
  }
  y4 <- by2(data, INDICES = data[,"tension"], FUN = summary)

  for (cores in 1:availCores) {
    message(sprintf("Testing with %d cores ...", cores))
    options(mc.cores = cores)
    strategies <- supportedStrategies(cores)
  
    for (strategy in supportedStrategies()) {
      message(sprintf("- plan('%s') ...", strategy))
      plan(strategy)
    
      y0f <- by(data[, 1:2], INDICES = data[,"tension"], FUN = summary) |> futurize_and_verify()
      stopifnot(all_equal_but_call(y0f, y0, check.attributes = FALSE))
      
      y1f <- by(data[, 1], INDICES = data[, -1], FUN = summary, digits = 2L) |> futurize_and_verify()
      stopifnot(all_equal_but_call(y1f, y1))
      
      y2f <- by(data, INDICES = data[,"tension"], FUN = function(x, ...) {
        lm(breaks ~ wool, data = x, ...)
      }, singular.ok = FALSE) |> futurize_and_verify()
      stopifnot(all_equal_but_call(y2f, y2))
      
      ## now suppose we want to extract the coefficients by group
      tmp <- with(data, by(data, INDICES = tension, FUN = function(x) {
        lm(breaks ~ wool, data = x)
      }) |> futurize_and_verify())
      y3f <- sapply(tmp, coef)
      stopifnot(all_equal_but_call(y3f, y3))
      
      y4f <- future_by2(data, INDICES = data[,"tension"], FUN = summary)
      stopifnot(all_equal_but_call(y4f, y4))

      ## Defunct /HB 2025-01-11
      res <- tryCatch({
        y4f2 <- future_by2(data, INDICES = data[,"tension"], FUN = "summary")
      }, FuturizeTestAssertionError = stop, error = identity)
      stopifnot(inherits(res, "error"))
      if (getRversion() >= "3.6.0") {
        stopifnot(inherits(res, "defunctError"))
      }
    } ## for (strategy ...)
    
    message(sprintf("Testing with %d cores ... DONE", cores))
  } ## for (cores ...)
} ## if (require("stats"))

message("*** future_by() ... DONE")

} ## if (requireNamespace("future.apply", quietly = TRUE))

#! /usr/bin/env Rscript
## This runs testme test script inst/testme/test-futurize-SingleCellExperiment.R
## Don't edit - it was autogenerated by inst/testme/deploy.R
futurize:::testme("futurize-SingleCellExperiment")

#' @tags pkg-mgcv
if (requireNamespace("mgcv") && getRversion() >= "4.4.0") {
library(futurize)
library(mgcv)
options(future.rng.onMisuse = "error")

plan(multisession)

dat <- gamSim(1, n = 500L, dist = "normal", scale = 20)
bs <- "cr"
k <- 12

b_truth <- bam(y ~ s(x0, bs = bs) + s(x1, bs = bs) + s(x2, bs = bs, k = k) + s(x3, bs = bs), data = dat)
print(b_truth)

b <- bam(y ~ s(x0, bs = bs) + s(x1, bs = bs) + s(x2, bs = bs, k = k) + s(x3, bs = bs), data = dat) |> futurize_and_verify()
print(b)

stopifnot(all.equal(summary(b), summary(b_truth)))

## predict.bam() via stats::predict() S3 generic dispatch
stopifnot(inherits(b_truth, "bam"))

## predict() for 'bam' forces sequential processing if number of rows
## is strictly less than 100 * number of parallel workers
nrows <- 100 * nbrOfWorkers()
newdat <- dat[1:nrows, ]
p_truth <- predict(b_truth, newdata = newdat)

p <- predict(b_truth, newdata = newdat) |> futurize_and_verify()
## NOTE: mgcv::predict.bam() returns an array when run sequentially,
## but a named numeric vector when run in parallel via parLapply
stopifnot(all.equal(as.numeric(p), as.numeric(p_truth)))

p2 <- stats::predict(b_truth, newdata = newdat) |> futurize_and_verify()
stopifnot(all.equal(as.numeric(p2), as.numeric(p_truth)))

plan(sequential)
} ## if (requireNamespace("mgcv"))

# strucchange::breakpoints(...) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), local({
#   options(future.disposable = structure(OPTS, dispose = FALSE))
#   on.exit(options(future.disposable = NULL))
#   strucchange::breakpoints(..., hpc = "foreach")
# }))
#
append_transpilers_for_strucchange <- function() {
  transpilers <- make_package_transpilers("strucchange", FUN = function(fcn, name) {
    if (name == "breakpoints.formula") {
      list(
        label = sprintf("strucchange::%s() ~> strucchange::%s()", name, name),
        transpiler = make_futurize_for_doFuture(
          args = list(hpc = "foreach"),
          defaults = list(label = sprintf("fz:strucchange::%s-%%d", name))
        )
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("strucchange", "doFuture")
}

#' List packages and functions supporting futurization
#'
#' @param package A package name.
#'
#' @return
#' A character vector of package or function names.
#' `futurize_supported_functions()` produces an error if packages 
#' required by the futurize transpiler are not installed.
#'
#' @examples
#' pkgs <- futurize_supported_packages()
#' pkgs
#'
#' if (requireNamespace("future.apply")) {
#'   fcns <- futurize_supported_functions("base")
#'   print(fcns)
#' }
#'
#' if (requireNamespace("doFuture")) {
#'   fcns <- futurize_supported_functions("foreach")
#'   print(fcns)
#' }
#'
#' @export
futurize_supported_packages <- function() {
  db <- transpiler_packages(classes = c("futurize::add-on"))
  sort(unique(db[["package"]]))
}


#' @rdname futurize_supported_packages
#' @export
futurize_supported_functions <- function(package) {
  stopifnot(is.character(package), length(package) == 1L, !is.na(package), nzchar(package))
  
  db <- transpilers_for_package(action = "list")
  classes <- c("futurize::add-on")
  if (!is.null(classes)) {
    db <- db[names(db) %in% classes]
  }
  classes <- names(db)

  packages <- package
  if (!requireNamespace(package, quietly = TRUE)) {
    stop("Package is not installed: ", sQuote(package))
  }
  
  ## Special cases
  if (package == "stats") packages <- c(packages, "base")
  fcns <- lapply(classes, function(class) {
    ## "Activate" packages
    void <- lapply(packages, function(pkg) {
      activators <- db[[class]][[pkg]]
      lapply(activators, FUN = function(activator) activator())
    })
    transpilers <- get_transpilers(class)
    transpilers <- transpilers[names(transpilers) == package]
    names <- lapply(transpilers, FUN = names)
    names <- unlist(names, use.names = FALSE)
    names <- unique(sort(names))
    names
  })
  fcns <- unlist(fcns, use.names = TRUE)

  ns <- getNamespace(package)
  keep <- vapply(fcns, FUN.VALUE = FALSE, FUN = function(fcn) {
    exists(fcn, mode = "function", envir = ns, inherits = FALSE)
  })
  fcns <- fcns[keep]

  if (length(fcns) == 0) {
    stop(sprintf("Package %s does not support futurization", sQuote(package)))
  }

  fcns
}

#' @tags skip_on_cran  ## (35s) to limit total check time
if (requireNamespace("SingleCellExperiment") && requireNamespace("scuttle") && requireNamespace("doFuture")) {
library(futurize)
library(SingleCellExperiment)
library(scuttle)

plan(multisession)

## Create a simple SingleCellExperiment with alternative experiments
set.seed(42)
n_genes <- 50L
n_cells <- 20L
counts <- matrix(
  rpois(n_genes * n_cells, lambda = 10),
  nrow = n_genes,
  ncol = n_cells
)
rownames(counts) <- paste0("gene", seq_len(n_genes))
colnames(counts) <- paste0("cell", seq_len(n_cells))

sce <- SingleCellExperiment(
  assays = list(counts = counts)
)

## Add an alternative experiment (e.g. spike-ins)
spike_counts <- matrix(
  rpois(10L * n_cells, lambda = 5),
  nrow = 10L,
  ncol = n_cells
)
rownames(spike_counts) <- paste0("spike", seq_len(10L))
colnames(spike_counts) <- paste0("cell", seq_len(n_cells))

altExp(sce, "spikes") <- SingleCellExperiment(
  assays = list(counts = spike_counts)
)


## ---------------------------------------------------------
## applySCE() with perCellQCMetrics
## ---------------------------------------------------------
result_truth <- applySCE(sce, perCellQCMetrics)

counters <- plan("backend")[["counters"]]
result <- applySCE(sce, perCellQCMetrics) |> futurize()
delta <- plan("backend")[["counters"]] - counters
cat(sprintf("Futures created: %d\n", delta[["created"]]))
stopifnot(delta[["created"]] > 0L)
stopifnot(all.equal(result, result_truth))

result2 <- SingleCellExperiment::applySCE(sce, perCellQCMetrics) |> futurize()
stopifnot(all.equal(result2, result_truth))


plan(sequential)
} ## if (requireNamespace("SingleCellExperiment") && ...)

# metafor::profile.rma.uni(...) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   metafor::profile.rma.uni(..., parallel = "snow", ncpus = 2L, cl = cl)
# })
#
append_transpilers_for_metafor <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'metafor' functions requires R (>= 4.4.0)", getRversion()))
  }

  transpilers <- make_package_transpilers("metafor", FUN = function(fcn, name) {
    if ("parallel" %in% names(formals(fcn))) {
      transpiler <- make_futurize_for_makeClusterFuture(
        args = list(
          parallel = "snow",
          ncpus = 2L,   ## only used for test ncpus > 1
          cl = quote(cl)
        ),
        defaults = list(label = sprintf("fz:metafor::%s", name))
      )
      
      list(
        label = sprintf("metafor::%s() ~> metafor::%s(..., parallel = \"snow\")", name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("metafor", "future")
}

#' @tags pkg-SimDesign
if (requireNamespace("SimDesign") && getRversion() >= "4.4.0") {
library(futurize)
library(SimDesign)

plan(multisession)

## -------------------------------------------------------------------
## runSimulation() - Monte Carlo simulation
## -------------------------------------------------------------------
## Adopted from help("runSimulation", package = "SimDesign")
Design <- createDesign(
  sample_size = c(30, 60),
  distribution = c("norm", "chi")
)

Generate <- function(condition, fixed_objects) {
  N <- condition$sample_size
  dist <- condition$distribution
  if (dist == "norm") {
    dat <- rnorm(N)
  } else if (dist == "chi") {
    dat <- rchisq(N, df = 5)
  }
  dat
}

Analyse <- function(condition, dat, fixed_objects) {
  ret <- mean(dat)
  c(mean_est = ret)
}

Summarise <- function(condition, results, fixed_objects) {
  obs_bias <- bias(results[, "mean_est"],
    parameter = ifelse(condition$distribution == "norm", 0, 5))
  obs_RMSE <- RMSE(results[, "mean_est"],
    parameter = ifelse(condition$distribution == "norm", 0, 5))
  c(bias = obs_bias, RMSE = obs_RMSE)
}

res <- runSimulation(
  design = Design,
  replications = 10,
  generate = Generate,
  analyse = Analyse,
  summarise = Summarise,
  save = FALSE
) |> futurize_and_verify()
print(res)

plan(sequential)
} ## if (requireNamespace("SimDesign"))

# crossmap::xmap_dbl(...) =>
#
# crossmap::future_xmap_dbl(..., .options = <future arguments>)
#
append_transpilers_for_crossmap <- function() {
  append_transpilers("futurize::add-on", make_addon_transpilers(
    "crossmap", "crossmap", make_options = make_options_for_furrr)
  )

  ## Return required packages
  c("crossmap")
}

#' @tags pkg-future.apply
#' @tags future_vapply
#' @tags detritus-files
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)

message("*** future_vapply() ...")

for (strategy in supportedStrategies()) {
  message(sprintf("*** strategy = %s ...", sQuote(strategy)))
  plan(strategy)
  
  x <- NULL
  fun <- is.factor
  fun_name <- "is.factor"
  fun_value <- logical(1L)
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  y2 <- vapply(x, FUN = fun_name, FUN.VALUE = fun_value) |> futurize()
  str(y2)
  stopifnot(all.equal(y2, y0))
  
  x <- list()
  fun <- is.numeric
  fun_value <- logical(1L)
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  x <- integer()
  fun <- identity
  fun_value <- fun(integer(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  
  df <- data.frame(x = 1:10, y = letters[1:10], stringsAsFactors=FALSE)
  fun <- class
  fun_value <- character(1L)
  y0 <- vapply(df, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(df, FUN = fun, FUN.VALUE = fun_value) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  x <- 1:10
  fun <- function(x) double(0L)
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  fun <- function(x) integer(0L)
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  fun <- sqrt
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  fun <- function(x) c(x, x^2)
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  fun <- function(x) matrix(x, nrow = 2L, ncol = 2L)
  fun_value <- fun(integer(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  fun <- function(x) matrix(x, nrow = 2L, ncol = 2L)
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))

  ## Ditto with dimnames on FUN.VALUE
  fun <- function(x) {
    matrix(x, nrow = 2L, ncol = 2L, dimnames = list(c("a", "b"), c("A", "B")))
  }
  fun_value <- fun(double(1L))
  y0 <- vapply(x, FUN = fun, FUN.VALUE = fun_value)
  str(y0)
  y1 <- vapply(x, FUN = fun, FUN.VALUE = fun_value) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  message("- From example(vapply) ...")
  x <- list(a = 1:10, beta = exp(-3:3), logic = c(TRUE, FALSE, FALSE, TRUE))
  y0 <- vapply(x, FUN = quantile, FUN.VALUE = double(5L))
  y1 <- vapply(x, FUN = quantile, FUN.VALUE = double(5L)) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))
  
  i39 <- sapply(3:9, seq)
  ys0 <- sapply(i39, fivenum)
  ys1 <- sapply(i39, fivenum) |> futurize()
  stopifnot(all.equal(ys1, ys0))
  
  yv0 <- vapply(i39, fivenum,
          c(Min. = 0, "1st Qu." = 0, Median = 0, "3rd Qu." = 0, Max. = 0))
  yv1 <- vapply(i39, fivenum,
          c(Min. = 0, "1st Qu." = 0, Median = 0, "3rd Qu." = 0, Max. = 0)) |> futurize()
  str(yv1)
  stopifnot(all.equal(yv1, yv0))
  
  v <- structure(10*(5:8), names = LETTERS[1:4])
  f <- function(x, y) outer(rep(x, length.out = 3L), y)
  ys0 <- sapply(v, f, y = 2*(1:5), simplify = "array")
  ys1 <- sapply(v, f, y = 2*(1:5), simplify = "array") |> futurize()
  stopifnot(all.equal(ys1, ys0))
  
  fv <- outer(1:3, 1:5)
  y <- 2*(1:5)
  yv0 <- vapply(v, f, fv, y = y)
  yv1 <- vapply(v, f, fv, y = y) |> futurize()
  str(yv1)
  stopifnot(all.equal(yv1, yv0))
  
  y0 <- vapply(mtcars, FUN = is.numeric, FUN.VALUE = logical(1L))
  y1 <- vapply(mtcars, FUN = is.numeric, FUN.VALUE = logical(1L)) |> futurize()
  str(y1)
  stopifnot(all.equal(y1, y0))

  message("- vapply(x, ...) |> futurize() where length(x) != length(as.list(x)) ...")
  x <- structure(list(a = 1, b = 2), class = "Foo")
  as.list.Foo <- function(x, ...) c(x, c = 3)
  y0 <- vapply(x, FUN = length, FUN.VALUE = -1L)
  y1 <- vapply(x, FUN = length, FUN.VALUE = -1L) |> futurize()
  stopifnot(identical(y1, y0))

  message("- exceptions ...")
  res <- tryCatch({
    y0 <- vapply(1:3, FUN = identity, FUN.VALUE = c(3, 3))
  }, error = identity)
  stopifnot(inherits(res, "error"))
  res <- tryCatch({
    y1 <- vapply(1:3, FUN = identity, FUN.VALUE = c(3, 3)) |> futurize()
  }, error = identity)
  stopifnot(inherits(res, "error"))
  
  plan(sequential)
  message(sprintf("*** strategy = %s ... done", sQuote(strategy)))
} ## for (strategy in ...) 


message("*** future_vapply() ... DONE")

} ## if (requireNamespace("future.apply", quietly = TRUE))

# DESeq2::DESeq(dds) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
#   options(future.disposable = <future arguments>)
#   DESeq2::DESeq(dds, parallel = TRUE, BPPARAM = BiocParallel::DoparParam())
# })
#
append_transpilers_for_DESeq2 <- function() {
  transpilers <- make_package_transpilers("DESeq2", FUN = function(fcn, name) {
    if ("BPPARAM" %in% names(formals(fcn))) {
      args <- list(BPPARAM = quote(BiocParallel::DoparParam()))
      if ("parallel" %in% names(formals(fcn))) {
        args <- c(list(parallel = TRUE), args)
      }
      
      list(
        label = sprintf("DESeq2::%s() ~> DESeq2::%s(..., parallel = TRUE, BPPARAM = BiocParallel::DoparParam())", name, name),
        transpiler = make_futurize_for_doFuture(
          args = args,
          defaults = list(label = sprintf("fz:DESeq2::%s-%%d", name))
        )
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("DESeq2", "BiocParallel", "doFuture")
}

#' @tags pkg-crossmap
if (requireNamespace("crossmap")) {
library(futurize)
library(crossmap)
options(future.rng.onMisuse = "error")

plan(multisession)

xs <- list(1:5, 1:5)

exprs <- list(
  xmap_dbl = quote( xmap_dbl(xs, ~ .y * .x) ),
  xmap_chr = quote( xmap_chr(xs, ~ paste(.y, "*", .x, "=", .y * .x)) )
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))

  truth <- eval(expr)
  expr_f <- bquote(.(expr) |> futurize_and_verify())

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }
}

plan(sequential)
} ## if (requireNamespace("crossmap"))

#' @tags pkg-parameters
if (requireNamespace("parameters") && getRversion() >= "4.4.0") {
library(futurize)
library(parameters)
options(future.rng.onMisuse = "error")

plan(multisession)

#------------------------------------------------------------------
# bootstrap_model()
#------------------------------------------------------------------
message("bootstrap_model() ...")

## Case A: lm model (uses bootstrap_model.default)
data(mtcars, package = "datasets")
model_lm <- lm(mpg ~ wt, data = mtcars)

set.seed(42)
fit_lm <- bootstrap_model(model_lm, iterations = 10L) |> futurize_and_verify()

print(head(fit_lm))
stopifnot(inherits(fit_lm, "bootstrap_model"))

## Case B: merMod model (uses bootstrap_model.merMod)
if (requireNamespace("lme4", quietly = TRUE)) {
  library(lme4)
  model_lmer <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)

  set.seed(42)
  fit_lmer <- bootstrap_model(model_lmer, iterations = 10L) |> futurize_and_verify()
  ## If it correctly uses ClusterFuture, delta[["created"]] should be > 0
  ## Actually, future::makeClusterFuture() creates a cluster that 
  ## bootstrap_model() then uses.
  print(head(fit_lmer))
  stopifnot(inherits(fit_lmer, "bootstrap_model"))
}

message("bootstrap_model() ... done")


#------------------------------------------------------------------
# bootstrap_parameters()
#------------------------------------------------------------------
message("bootstrap_parameters() ...")

set.seed(42)

params <- bootstrap_parameters(model_lm, iterations = 10L) |> futurize_and_verify()

print(params)
stopifnot(inherits(params, "parameters_model"))

message("bootstrap_parameters() ... done")

plan(sequential)
}

#' @tags pkg-rugarch
if (requireNamespace("rugarch") && getRversion() >= "4.4.0") {
library(futurize)
library(rugarch)
options(future.rng.onMisuse = "error")

plan(multisession)

#------------------------------------------------------------------
# ugarchroll()
#------------------------------------------------------------------
message("ugarchroll() ...")

data(sp500ret, package = "rugarch")
spec <- ugarchspec()

## Small example to keep it fast
sp500ret_small <- sp500ret[1:100, , drop = FALSE]
n_start <- 50

set.seed(42)
roll_truth <- ugarchroll(spec, sp500ret_small, n.start = n_start, refit.window = "moving", refit.every = 25)
print(roll_truth)

set.seed(42)
roll <- ugarchroll(spec, sp500ret_small, n.start = n_start, refit.window = "moving", refit.every = 25) |> futurize_and_verify()
print(roll)

## rugarch objects might have some environment or pointer differences, 
## but let's see if all.equal works.
roll@model$elapsed <- NULL
roll_truth@model$elapsed <- NULL
roll@model$con_Args$cluster <- NULL
roll_truth@model$con_Args$cluster <- NULL
stopifnot(all.equal(roll, roll_truth))

message("ugarchroll() ... done")

plan(sequential)
}

# caret::train(...) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
#   options(future.disposable = <future arguments>)
#   caret::train(...)
# })
#
append_transpilers_for_caret <- function() {
  transpilers <- make_package_transpilers("caret", FUN = function(fcn, name) {
    ## nnnControl()
    if ("allowParallel" %in% names(formals(fcn))) {
      if (name == "nearZeroVar") {
        list(
          label = sprintf("caret::%s() ~> caret::%s()", name, name),
          transpiler = make_futurize_for_doFuture(
            defaults = list(
              seed = TRUE,
              label = sprintf("fz:caret::%s-%%d", name)
            ),
            args = list(foreach = TRUE)
          )
        )
      }
    } else {
      ## nnn() -> nnnControl()
      nameControl <- sprintf("%sControl", name)
      if (exists(nameControl, mode = "function", envir = getNamespace("caret"), inherits = FALSE)) {
        fcnControl <- get(nameControl, mode = "function", envir = getNamespace("caret"), inherits = FALSE)
        if ("allowParallel" %in% names(formals(fcnControl))) {
          list(
            label = sprintf("caret::%s() ~> caret::%s()", name, name),
            transpiler = make_futurize_for_doFuture(
              defaults = list(
                seed = TRUE,
                label = sprintf("fz:caret::%s-%%d", name)
              )
            )
          )
        }
      }
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("caret", "doFuture")
}

# tm::tm_index(...) =>
#
# local({
#   old_engine <- tm::tm_parLapply_engine()
#   on.exit(tm::tm_parLapply_engine(old_engine))
#   tm::tm_parLapply_engine(
#     future::makeClusterFuture(<future arguments>)
#   )
#   tm::tm_index(...)
# })
#
append_transpilers_for_tm <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'tm' functions requires R (>= 4.4.0)", getRversion()))
  }

  template <- bquote_compile(
    local({
      old_engine <- tm::tm_parLapply_engine()
      oopts <- options(future.ClusterFuture.clusterEvalQ = "error")
      on.exit({
        options(oopts)
        tm::tm_parLapply_engine(old_engine)
      })
      tm::tm_parLapply_engine(
        do.call(.(CALL), args = .(OPTS))
      )
      .(EXPR)
    })
  )
  
  ## To please 'R CMD check' on R (< 4.4.0), where
  ## future::makeClusterFuture() is not available
  call <- as.call(lapply(c("::", "future", "makeClusterFuture"), as.name))

  template2 <- bquote_compile(function(expr, options = NULL) {
    defaults <- list(
      packages = "tm",
      label = sprintf("fz:tm::%s", .(NAME))
    )
    opts <- make_options_for_makeClusterFuture(options, defaults = defaults)
    bquote_apply(template,
      CALL = call,
      OPTS = opts,
      EXPR = expr
    )
  })

  transpilers <- make_package_transpilers("tm", FUN = function(fcn, name) {
    if (!name %in% c("tm_map", "tm_index", "TermDocumentMatrix")) return(NULL)
    transpiler <- eval(bquote_apply(template2, NAME = name))
    list(
      label = sprintf("tm::%s() ~> tm::%s()", name, name),
      transpiler = transpiler
    )
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("tm", "future")
}

library(futurize)

message("*** Futurize internals")
options(futurize.debug = TRUE)

message("descend_wrappers()")
try(futurize:::descend_wrappers(NULL, unwrap = list()))
try(futurize:::descend_wrappers(quote({ lapply(x, f) }), unwrap = list(`{`), debug = TRUE))
try(futurize:::descend_wrappers(quote(with(data, lapply(x, f))), unwrap = list(base::with), debug = TRUE))

message("register_all_transpilers()")
futurize:::register_all_transpilers()

message("register_vignette_engine_during_build_only()")
futurize:::register_vignette_engine_during_build_only("futurize")
Sys.setenv(R_CMD = "something")
futurize:::register_vignette_engine_during_build_only("futurize")

message("transpiler_packages()")
db <- futurize:::transpiler_packages()
print(db)
if (requireNamespace("future.apply", quietly = TRUE)) {
  y <- lapply(x, f) |> futurize(eval = FALSE)
  db <- futurize:::transpiler_packages()
  print(db)
  db <- futurize:::transpiler_packages(classes = c("futurize::add-on"))
  print(db)
}


message("list_transpilers()")
ts <- futurize:::list_transpilers(class = "non-existing")
str(ts)
ts <- futurize:::list_transpilers(class = "futurize::add-on")
str(ts)
ts <- futurize:::list_transpilers(pattern = ".*", class = "futurize::add-on")
str(ts)
if (requireNamespace("future.apply", quietly = TRUE)) {
  y <- lapply(x, f) |> futurize(eval = FALSE)
  ts <- futurize:::list_transpilers(pattern = ".*", class = "futurize::add-on")
  str(ts)
}


message("*** is_s3_generic()")
stopifnot(isTRUE(futurize:::is_s3_generic(print)))
stopifnot(isTRUE(futurize:::is_s3_generic(summary)))
stopifnot(isFALSE(futurize:::is_s3_generic(identity)))
stopifnot(isFALSE(futurize:::is_s3_generic(sum))) ## primitive
foo <- function() NULL ## A function with NULL body
stopifnot(
  is.null(body(foo)),
  isFALSE(futurize:::is_s3_generic(foo))
)


message("*** parse_call() - namespace-qualified calls")
## base::lapply - valid namespace-qualified call
info <- futurize:::parse_call(quote(base::lapply), envir = globalenv())
stopifnot(info$fcn_name == "lapply", info$ns_name == "base")

## Non-existing function in a namespace
res <- tryCatch(
  futurize:::parse_call(quote(base::nonExistingFcn), envir = globalenv()),
  error = identity
)
stopifnot(inherits(res, "error"))

## Non-symbol operator in call, e.g. function call as operator
res <- tryCatch(
  futurize:::parse_call(quote((function() NULL)()), envir = globalenv()),
  error = identity
)
stopifnot(inherits(res, "error"))

## Non-supported operator (not :: or :::)
res <- tryCatch(
  futurize:::parse_call(quote(base$lapply), envir = globalenv()),
  error = identity
)
stopifnot(inherits(res, "error"))

## Unknown function
res <- tryCatch(
  futurize:::parse_call(as.symbol("nonExistingFcn"), envir = globalenv()),
  error = identity
)
stopifnot(inherits(res, "error"))

## Unknown infix operator
res <- tryCatch(
  futurize:::parse_call(as.symbol("%nonExisting%"), envir = globalenv()),
  error = identity
)
stopifnot(inherits(res, "error"))

## Non-call, non-symbol expression
res <- tryCatch(
  futurize:::parse_call(42, envir = globalenv()),
  error = identity
)
stopifnot(inherits(res, "error"))

## Primitive function - namespace detection
info <- futurize:::parse_call(as.symbol("sum"), envir = globalenv())
stopifnot(info$fcn_name == "sum", info$ns_name == "base")

## Debug mode with namespace-qualified call
info <- futurize:::parse_call(quote(base::lapply), envir = globalenv(), debug = TRUE)
stopifnot(info$fcn_name == "lapply", info$ns_name == "base")

## Debug mode with unqualified call
info <- futurize:::parse_call(as.symbol("lapply"), envir = globalenv(), debug = TRUE)
stopifnot(info$fcn_name == "lapply", info$ns_name == "base")


message("*** find_s3_method()")
find_s3_method <- futurize:::find_s3_method

## print() on a data.frame
df <- data.frame(a = 1)
res <- find_s3_method(
  fcn = print, fcn_name = "print",
  call = quote(print(df)), envir = environment()
)
stopifnot(
  is.list(res),
  res$name == "print.data.frame",
  is.character(res$package), nzchar(res$package)
)

## Ditto with debug = TRUE
res <- find_s3_method(
  fcn = print, fcn_name = "print",
  call = quote(print(df)), envir = environment(),
  debug = TRUE
)
stopifnot(is.list(res), res$name == "print.data.frame")

## Function with no formals
res <- find_s3_method(
  fcn = function() NULL, fcn_name = "nofmls",
  call = quote(nofmls()), envir = environment()
)
stopifnot(is.null(res))

## First argument is "..." returns NULL
dotfcn <- function(...) UseMethod("dotfcn")
res <- find_s3_method(
  fcn = dotfcn, fcn_name = "dotfcn",
  call = quote(dotfcn(x)), envir = environment()
)
stopifnot(is.null(res))

## Dispatch argument not provided in call
res <- find_s3_method(
  fcn = print, fcn_name = "print",
  call = quote(print()), envir = environment()
)
stopifnot(is.null(res))

## Dispatch argument is a literal
res <- find_s3_method(
  fcn = print, fcn_name = "print",
  call = quote(print(42)), envir = environment()
)
stopifnot(is.null(res))

## Dispatch object cannot be evaluated
res <- find_s3_method(
  fcn = print, fcn_name = "print",
  call = quote(print(nonExistingVar)), envir = environment()
)
stopifnot(is.null(res))

## No S3 method registered for this generic + class
my_generic <- function(x) UseMethod("my_generic")
obj <- structure(1, class = "noMethodForThis")
res <- find_s3_method(
  fcn = my_generic, fcn_name = "my_generic",
  call = quote(my_generic(obj)), envir = environment()
)
stopifnot(is.null(res))


message("*** append_call_arguments()")
call <- quote(my_fcn(x, y))
call2 <- futurize:::append_call_arguments(call, z = 42, w = quote(1 + 2))
stopifnot(length(call2) == 5L)
stopifnot("z" %in% names(as.list(call2)))
stopifnot("w" %in% names(as.list(call2)))

call3 <- futurize:::append_call_arguments(call, .args = list(a = 1))
stopifnot("a" %in% names(as.list(call3)))


message("transpilers_for_package()")
ts <- futurize:::transpilers_for_package(type = "unknown", package = "base", fcn = lapply, debug = TRUE)
str(ts)
ts <- futurize:::transpilers_for_package(type = "unknown", package = "base", fcn = lapply, action = "get", debug = TRUE)
str(ts)
ts <- futurize:::transpilers_for_package(type = "unknown", package = "base", fcn = lapply, action = "list", debug = TRUE)
str(ts)
ts <- futurize:::transpilers_for_package(type = "unknown", package = "base", fcn = lapply, action = "reset", debug = TRUE)
str(ts)
res <- tryCatch(futurize:::transpilers_for_package(type = "unknown", package = "unknown", action = "make", debug = TRUE), error = identity)
str(res)

#' @tags pkg-Sim.DiffProc
if (requireNamespace("Sim.DiffProc") && getRversion() >= "4.4.0") {
library(futurize)
library(Sim.DiffProc)
options(future.rng.onMisuse = "error")

plan(multisession)

#------------------------------------------------------------------
# MCM.sde()
#------------------------------------------------------------------
message("MCM.sde() ...")

f <- expression(0)
g <- expression(1)
mod1d <- snssde1d(drift=f, diffusion=g, x0=1, M=10, N=100)
stat <- function(x, ...) mean(x)

res_truth <- MCM.sde(mod1d, statistic = stat, R = 10, time = 0.5)

res <- MCM.sde(mod1d, statistic = stat, R = 10, time = 0.5) |> futurize_and_verify()

## Compare structure/metadata since simulation results themselves are stochastic
stopifnot(
  identical(class(res), class(res_truth)),
  identical(names(res), names(res_truth)),
  identical(res$dim, res_truth$dim),
  identical(res$Class, res_truth$Class)
)

message("MCM.sde() ... done")

plan(sequential)
}

#' @noRd
make_options_for_future.apply <- local({
  get_defaults <- function(fcn) {
    defaults <- formals(fcn)
    names <- setdiff(names(defaults), "future.envir")
    keep <- grep("^future[.]", names, value = TRUE)
    defaults[keep]
  }
  
  defaults_base <- NULL

  function(options, fcn, defaults = NULL) {
    ## Nothing to do?
    if (length(options) == 0L && length(defaults) == 0L) return(options)

    if (is.null(defaults_base)) {
      defaults_base <<- get_defaults(future.apply::future_lapply)
    }

    ## Keep only 'future.apply' defaults
    if (length(defaults) > 0) {
      defaults <- defaults[grepl("^future[.]", names(defaults))]
    }
    
    ## Default future.* arguments
    defaults <- c(defaults_base, get_defaults(fcn), defaults)
    keep <- !duplicated(names(defaults), fromLast = TRUE)
    defaults <- defaults[keep]
   
    ## Specified future.* arguments
    specified <- attr(options, "specified")

    ## Remap chunk_size -> chunk.size
    if (length(specified) > 0) {
      specified <- sub("^chunk_size$", "chunk.size", specified)
      names(options) <- sub("^chunk_size$", "chunk.size", names(options))
    }

    ## Remap future options for future.apply
    specified <- sprintf("future.%s", specified)

    names <- setdiff(names(defaults), specified)
    names(options) <- sprintf("future.%s", names(options))
    for (name in names) options[[name]] <- defaults[[name]]

    ## Silently drop non-existing future options
    keep <- intersect(names(options), names(defaults))
    options <- options[keep]
    
    options
  }
})

#' @tags pkg-future.apply
#' @tags future_apply
#' @tags globals
#' @tags detritus-files
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

if (packageVersion("future") > "1.49.0" && isTRUE(getOption("future.globals.keepWhere"))) {

  library(futurize)


  options(future.debug = FALSE)
  
  foo <- function(..., FUN = function(...) list(...)) {
    args <- list(...)
    lapply(1L, FUN = function(x) {
      do.call(FUN, args = c(list(x), args))
    }) |> futurize_and_verify()
  }
  
  bar <- function(..., fun = function(...) list(...)) {
    lapply(1L, FUN = function(x) fun(x, ...)) |> futurize_and_verify()
  }
  
  yaa <- function(..., FUN = function(...) list(...)) {
    lapply(1L, FUN = function(x) {
      do.call(FUN, args = c(list(x), ...))
    }) |> futurize_and_verify()
  }
  
  for (strategy in supportedStrategies()) {
    message(sprintf("- plan('%s') ...", strategy))
    plan(strategy)
  
    x1 <- foo()
    y1 <- bar()
    stopifnot(identical(y1, x1))
    z1 <- yaa()
    stopifnot(identical(z1, x1))
  
    x2 <- foo(a = 2)
    y2 <- bar(a = 2)
    z2 <- yaa(a = 2)
    stopifnot(identical(y2, x2))
    stopifnot(identical(z2, x2))
    
    message(sprintf("- plan('%s') ... done", strategy))
  }

} ## if (packageVersion("future") > ...)

} ## if (requireNamespace("future.apply", quietly = TRUE))

#' Transpile an R expression
#'
#' @param expr An \R expression, typically a function call to transpile.
#' If FALSE, then the transpiler is disabled, and if TRUE, it is re-enabled.
#' If NA, then TRUE is returned if the transpiler is enabled, otherwise FALSE.
#'
#' @param options (optional) Named options for the transpilation.
#' 
#' @param when If TRUE (default), the expression is transpiled, otherwise not.
#'
#' @param eval If TRUE (default), the transpiled expression is evaluated,
#' otherwise it is returned.
#'
#' @param envir The environment where the expression should be evaluated.
#'
#' @param type Type of the transpiler to use.
#'
#' @param unwrap (optional) A list of functions that should be considered
#' wrapping functions that the transpiler should unwrap ("enter"). This
#' allows us to transpile expressions within `{ ... }` and `local( ... )`.
#'
#' @returns
#' Returns the value of the evaluated expression `expr` if `eval = TRUE`,
#' otherwise the transpiled expression.
#' If `expr` is NA, then TRUE is returned if the transpiler is enabled,
#' otherwise FALSE.
#'
#' @keywords internal
transpile <- local({
  .enabled <- list()
  
  function(expr, options = list(...), ..., when = TRUE, eval = TRUE, envir = parent.frame(), disable = FALSE, type = "built-in", what = "transpile", unwrap = list(base::`{`, base::`(`, base::`!`, base::local, base::with, base::I, base::identity, base::invisible, base::suppressMessages, base::suppressWarnings, base::suppressPackageStartupMessages), debug = FALSE) {
    if (debug) {
      mdebug_push("transpile() ...")
      on.exit(mdebug_pop())
    }
  
    stopifnot(
      is.logical(when), length(when) == 1L, !is.na(when)
    )
  
    ## Enable or disable transpiler, or query its state?
    enabled <- .enabled[[type]]
    if (is.null(enabled)) {
      enabled <- TRUE
      .enabled[[type]] <<- enabled
    }
    
    ## e.g. transpile(TRUE), transpile(FALSE), or transpile(NA)?
    if (is.logical(expr) && length(expr) == 1L) {
      if (is.na(expr)) return(enabled)
      old_enabled <- enabled
      .enabled[[type]] <<- expr
      return(invisible(old_enabled))
    }
  
    ## Don't transpile, i.e. evaluate as-is?
    if (!enabled || !when || disable) {
      if (eval) {
        if (debug) mdebug("Evaluate call expression")
        return(eval(expr, envir = envir))
      } else {
        if (debug) mdebug("Return call expression")
        return(expr)
      }
    }

    repeat {
      ## 1a. Get a matching transpiler
      transpiler <- get_transpiler(expr, envir = envir, type = type, what = what, unwrap = unwrap, debug = debug)
    
      transpile <- transpiler[["transpiler"]]
  
      ## 1b. If not a nested transpiler function, we're done here
      if (!inherits(transpile, "transpiler")) break
  
      ## 1c. Generate transpiled expression of nested transpiler
      expr <- local({
        if (debug) {
          mdebug_push("Apply nested transpiler ...")
          on.exit({
            mprint(expr)
            mdebug_pop()
          })
          mprint(expr)
        }
        parts <- as.list(expr)
        parts$eval <- FALSE
        expr2 <- as.call(parts)
        expr <- eval(expr2, envir = envir)
        expr
      })
    }
  
  
    ## 2. Transpile
    if (debug) {
      mdebug_push("Transpile call expression ...")
    }
  
    expr_transpiled <- transpile(expr, options = options)
    class(expr_transpiled) <- c("transpiled_call", class(expr_transpiled))
    
    if (debug) {
      mprint(expr_transpiled)
      mdebug_pop()
    }
  
  
    ## 3. Evaluate or return transpiled expression?
    if (eval) {
      if (debug) mdebug("Evaluate transpiled call expression")
      eval(expr_transpiled, envir = envir)
    } else {
      if (debug) mdebug("Return transpiled call expression")
      expr_transpiled
    }
  } ## transpile()
})
class(transpile) <- c("transpiler", class(transpile))


#' Finds transpiler for S3 method for S3 generic function call and object
#'
#' @inheritParams find_s3_method
#'
#' @return
#' A transpiler function, or NULL if none exists.
#'
#' @noRd
find_s3_method_transpiler <- function(fcn, fcn_name, call, envir, type, debug = FALSE) {
  method <- find_s3_method(fcn, fcn_name = fcn_name, call = call, envir = envir, debug = debug)
  if (is.null(method)) return(NULL)

  pkg <- method[["package"]]
  name <- method[["name"]]

  ## Look up registered transpiler for the package of the S3 method
  transpiler_sets <- get_transpilers(type)
  transpilers <- transpiler_sets[[pkg]]

  ## If non-existing, retry by first trying to register transpilers for the package
  if (is.null(transpilers)) {
    transpilers <- tryCatch({
      transpilers_for_package(type = type, package = pkg, action = "make", debug = debug)
      transpiler_sets <- get_transpilers(type)
      transpiler_sets[[pkg]]
    }, error = function(e) NULL)
  }

  ## No transpilers registered for this package?
  if (is.null(transpilers)) return(NULL)

  transpilers[[name]]
} ## find_s3_method_transpiler()


find_s4_method_transpiler <- function(fcn, fcn_name, call, envir, type, debug = FALSE) {
  method <- find_s4_method(fcn, fcn_name = fcn_name, call = call, envir = envir, debug = debug)
  if (is.null(method)) return(NULL)

  pkg <- method[["package"]]
  name <- method[["name"]]

  ## Look up registered transpiler for the package of the S4 method
  transpiler_sets <- get_transpilers(type)
  transpilers <- transpiler_sets[[pkg]]

  ## If non-existing, retry by first trying to register transpilers for the package
  if (is.null(transpilers)) {
    transpilers <- tryCatch({
      transpilers_for_package(type = type, package = pkg, action = "make", debug = debug)
      transpiler_sets <- get_transpilers(type)
      transpiler_sets[[pkg]]
    }, error = function(e) NULL)
  }

  ## No transpilers registered for this package?
  if (is.null(transpilers)) return(NULL)

  transpilers[[name]]
} ## find_s4_method_transpiler()


#' Get a registered transpiler for an R expression
#' 
#' @inheritParams transpile
#' @inheritParams parse_call
#' 
#' @param expr The R expression, which contains an R symbol or an R call
#' to be transpiled.
#'
#' @return
#' A transpiler, which is a named list with elements:
#'
#'  * `label` - a character string describing the transpiler
#'
#'  * `transpiler` - a function that takes an R expression and
#'                   an optional argument `options`
#'
#' @keywords internal
#' @noRd
get_transpiler <- function(expr, envir = parent.frame(), unwrap = list(), type, what, debug = FALSE) {
  if (debug) {
    mdebug_push("get_transpiler() ...")
    on.exit(mdebug_pop())
    mdebug_push("Finding call to be transpiled ...")
  }
  
  call_pos <- descend_wrappers(expr, envir = envir, unwrap = unwrap, what = what, debug = debug)

  call <- expr[[call_pos]]
  call_info <- parse_call(call, envir = envir, what = what, debug = debug)
  fcn <- call_info[["fcn"]]
  fcn_name <- call_info[["fcn_name"]]
  ns_name <- call_info[["ns_name"]]

  if (debug) {
    mdebugf("Position of call to be transpiled in expression: c(%s)", comma(call_pos))
    mprint(call)
    mdebug_pop()   
    mdebugf_push("Locating %s transpiler for %s::%s() of class %s ...", sQuote(type), ns_name, fcn_name, sQuote(class(fcn)[1]))
  }

  ## Special case: A nested transpiler function?
  if (inherits(fcn, "transpiler")) {
    if (debug) {
      mdebugf("Detected a nested transpiler function: %s::%s()", ns_name, fcn_name)
    }
    transpiler <- list(
      label      = fcn_name,
      transpiler = fcn
    )

    stopifnot(call_pos == 1L)
    return(transpiler)
  }

  transpiler_sets <- get_transpilers(type)
  transpilers <- transpiler_sets[[ns_name]]
  if (is.null(transpilers)) {
    if (!requireNamespace(ns_name, quietly = TRUE)) {
      info <- if (grepl("^%.*%$", fcn_name)) {
        sprintf("%s::`%s`", ns_name, fcn_name)
      } else {
        sprintf("%s::%s()", ns_name, fcn_name)
      }
      stop(sprintf("Please install %s in order to %s %s",
           sQuote(ns_name), what, info))
    }

    ## Get transpiler package addons
    ## tryCatch() is needed for cases where a package re-exports a generic
    ## from another package, e.g. scater::runPCA() is a re-export of
    ## BiocSingular::runPCA()
    req_pkgs <- tryCatch(
      transpilers_for_package(type = type, package = ns_name, action = "make", debug = debug),
      error = function(e) NULL
    )
    if (!is.null(req_pkgs)) {
      if (debug) {
        mdebugf("Required packages: [n=%d] %s", length(req_pkgs), commaq(req_pkgs))
      }

      okay <- vapply(req_pkgs, FUN.VALUE = NA, FUN = requireNamespace, quietly = TRUE)
      if (!all(okay)) {
        pkgs <- req_pkgs[!okay]
        info <- if (grepl("^%.*%$", fcn_name)) {
          sprintf("%s::`%s`", ns_name, fcn_name)
        } else {
          sprintf("%s::%s()", ns_name, fcn_name)
        }
        stop(sprintf("Please install %s in order to %s %s",
             commaq(pkgs), what, info))
      }
      transpiler_sets <- get_transpilers(type)
      transpilers <- transpiler_sets[[ns_name]]
    }
  }

  if (debug) {
    mdebugf("Namespaces registered with %s(): %s", what, commaq(names(transpiler_sets)))
  }
  
  ## Is there a registered transpiler for the function?
  if (is.null(transpilers) || ! fcn_name %in% names(transpilers)) {
    ## Fallback: S3 generic dispatching to a method in another package
    ## Note: 'call' is expr[[call_pos]] = the function head symbol, not the
    ## full call. The full call expression including arguments is needed for
    ## match.call(), so we reconstruct it here.
    full_call <- if (length(call_pos) == 1L) expr else expr[[call_pos[-length(call_pos)]]]
    if (is_s3_generic(fcn)) {
      transpiler <- find_s3_method_transpiler(fcn, fcn_name, full_call, type, envir = envir, debug = debug)
    } else if (inherits(fcn, "standardGeneric")) {
      transpiler <- find_s4_method_transpiler(fcn, fcn_name, full_call, type, envir = envir, debug = debug)
    } else {
      transpiler <- NULL
    }
    if (is.null(transpiler)) {
      if (is.null(transpilers)) {
        stop_with_version(sprintf("Function %s::%s() is not in one of the registered %s namespaces: %s", ns_name, fcn_name, what, commaq(names(transpiler_sets))))
      }
      stop_with_version(sprintf("Do not know how to %s function: %s()", what, deparse(call)))
    }
  } else {
    transpiler <- transpilers[[fcn_name]]
  }
  if (debug) {
    stopifnot(is.list(transpiler), "label" %in% names(transpiler), "transpiler" %in% names(transpiler))
    mdebugf("Transpiler description: %s", transpiler[["label"]])
    mdebug("Transpiler function:")
    mprint(transpiler[["transpiler"]])
  }

  if (length(call_pos) > 1L) {
    if (debug) {
      mdebug_push("Creating wrapper transpiler ...")
    }
    transpiler_inner <- transpiler[["transpiler"]]
    transpiler <- list(
      label      = sprintf("Apply transpiler to inner expression at c(%s)", comma(call_pos)),
      transpiler = function(expr, ...) {
        inner_pos <- call_pos[-length(call_pos)]         
        expr_inner <- expr[[inner_pos]]
        expr_inner <- transpiler_inner(expr_inner, ...)
        expr[[inner_pos]] <- expr_inner
        expr
      }
    )
    if (debug) {
      mprint(transpiler)
      mdebug_pop()
    }
  }

  if (debug) mdebugf_pop()

  transpiler
} ## get_transpiler()


.env <- new.env()
.env[["transpiler_db"]] <- list()

get_transpilers <- function(type) {
  transpiler_db <- .env[["transpiler_db"]]
  transpiler_db[[type]]
}

append_transpilers <- function(type, ...) {
  transpiler_db <- .env[["transpiler_db"]]
  transpilers <- transpiler_db[[type]]
  transpilers <- c(transpilers, ...)
  transpiler_db[[type]] <- transpilers
  .env[["transpiler_db"]] <- transpiler_db
}


list_transpilers <- function(pattern = NULL, class) {
  data <- list()
  transpiler_db <- .env[["transpiler_db"]]
  db <- transpiler_db[[class]]
  if (is.null(db)) db <- list()
  types <- names(db)
  if (!is.null(pattern)) {
    types <- grep(pattern, types, value = TRUE)
  }
  for (type in types) {
    transpilers <- db[[type]]
    fcns <- unique(names(transpilers))
    for (fcn in fcns) {
      idxs <- which(fcn == names(transpilers))
      if (length(idxs) == 1) {
        transpilers_fcn <- transpilers[idxs]
      } else {
        ## length(idxs) > 1 should not happen, but in case ...
        transpilers_fcn <- list()
        for (idx in idxs) {
          transpilers_fcn <- c(transpilers_fcn, transpilers[idx])
        }
        drop <- duplicated(names(transpilers_fcn), fromLast = TRUE)
        transpilers_fcn <- transpilers_fcn[!drop]
      }
      transpilers_fcn <- transpilers_fcn[order(names(transpilers_fcn))]
      names <- names(transpilers_fcn)
      labels <- vapply(transpilers_fcn, FUN = function(t) t$label, FUN.VALUE = "")
      dd <- data.frame(type = type, fcn = names, description = labels)
      data <- c(data, list(dd))
    }
  }
  data <- Reduce(rbind, data)
  rownames(data) <- NULL
  data
}


transpilers_for_package <- local({
  .db <- list()
  
  function(type = "default", package, fcn, action = c("add", "make", "get", "list", "reset"), debug = FALSE) {
    stopifnot(is.character(type), length(type) == 1L, !is.na(type))
    action <- match.arg(action, several.ok = FALSE)
    
    if (debug) {
      mdebugf_push("transpilers_for_package(action = %s, type = %s) ...", sQuote(action), sQuote(type))
      on.exit(mdebug_pop())
    }

    db <- .db[[type]]
    if (is.null(db)) db <- list()
    
    if (action == "add") {
      stopifnot(
        is.character(package), length(package) == 1L,
        is.function(fcn)
      )
      if (debug) {
        mdebugf(" - package: %s", sQuote(package))
      }
      fcns <- old_fcns <- db[[package]]
      fcns <- if (length(fcns) == 0) list(fcn) else c(fcns, list(fcn))
      db[[package]] <- fcns
      .db[[type]] <<- db
      invisible(old_fcns)
    } else if (action == "get") {
      if (debug) {
        mdebugf(" - package: %s", sQuote(package))
      }
      stopifnot(
        is.character(package), length(package) == 1L
      )
      if (debug) mdebugf(" - package: %s", sQuote(package))
      db[[package]]
    } else if (action == "make") {
      stopifnot(
        is.character(package), length(package) == 1L
      )
      if (debug) {
        mdebugf(" - package: %s", sQuote(package))
      }
      fcns <- db[[package]]
      if (debug) mprint(list(fcns = fcns))
      if (length(fcns) == 0L) {
        stop_with_version(sprintf("There are no factory functions for creating %s transpilers for package %s", sQuote(type), sQuote(package)))
      }
      req_pkgs <- lapply(fcns, FUN = function(fcn) fcn())
      req_pkgs <- unlist(req_pkgs, use.names = FALSE)
      req_pkgs <- sort(unique(req_pkgs))
      req_pkgs
    } else if (action == "list") {
      .db
    } else if (action == "reset") {
      old_db <- db
      db <- list()
      .db[[type]] <<- db
      invisible(old_db)
    }
  }
})


transpiler_packages <- function(classes = NULL) {
  db <- transpilers_for_package(action = "list")
  if (!is.null(classes)) {
    db <- db[names(db) %in% classes]
  }
  transpilers <- data.frame(class = character(0L), package = character(0L))
  for (class in names(db)) {
    set <- db[[class]]
    pkgs <- names(set)
    transpilers <- rbind(transpilers, data.frame(class = class, package = pkgs))
  }
  transpilers
}



#' @param package (character string) Package name.
#'
#' @param FUN A functions taking arguments `fcn` (a function),
#' `package` (character string), and `name` (character string).
#'
#' @param export (logical) If TRUE, exported functions are considered.
#'
#' @param s3methods (logical) If TRUE, registered S3 methods are considered.
#'
#' @return
#' A named list of lists of transpilers, where the names correspond
#' to function names of the package `package` and transpilers are lists.
#'
#' @noRd
make_package_transpilers <- function(package, FUN, exports = TRUE, s3methods = TRUE) {
  transpilers <- list()
  ns <- getNamespace(package)

  names <- character(0L)
  if (exports) {
    exports <- names(getNamespaceInfo(ns, "exports"))
    names <- c(names, exports)
  }
  if (s3methods) {
    s3methods <- getNamespaceInfo(ns, "S3methods")[,3]
    if (!is.character(s3methods)) {
      ## Note, although 's3methods' is typically a character vector, it
      ## might also be a list. For example, loadNamespace("strucchange")
      ## results in a character vector, but if we then load
      ## loadNamespace("partykit"), the registered S3 methods for
      ## 'strucchange' becomes a list with two function objects appended
      ## at the very end.
      s3methods <- s3methods[vapply(s3methods, FUN.VALUE = FALSE, FUN = is.character)]
      s3methods <- unlist(s3methods, use.names = FALSE)
    }
    names <- c(names, s3methods)
  }
  
  for (name in names) {
    if (exists(name, mode = "function", envir = ns, inherits = FALSE)) {
      fcn <- get(name, mode = "function", envir = ns, inherits = FALSE)
      transpilers[[name]] <- FUN(fcn, name = name)
    }
  }

  transpilers <- list(transpilers)
  names(transpilers) <- package

  transpilers
} ## make_package_transpilers()


#' @importFrom utils packageName
stop_with_version <- function(msg, ...) {
  msg <- sprintf("[%s %s] %s", packageName(), getNamespaceVersion(packageName()), msg)
  stop(msg, ...)
}

#' @tags pkg-furrr
#' @tags detritus-files

if (requireNamespace("purrr") && requireNamespace("furrr")) {
library(futurize)


library(purrr)

# ------------------------------------------------------------------------------
# map2()

message("future_map2() matches map2() for simple cases")
stopifnot(identical(
  map2(1:3, 4:6, ~.x + .y) |> futurize_and_verify(),
  map2(1:3, 4:6, ~.x + .y)
))


message("names of `.x` are retained")
x <- c(a = 1, b = 2)
y <- c(c = 1, d = 2)
stopifnot(identical(names(map2(x, y, ~1) |> futurize_and_verify()), c("a", "b")))


message("named empty input makes named empty output")
x <- set_names(list(), character())
stopifnot(identical(names(map2(x, x, ~.x) |> futurize::futurize()), character()))


# ------------------------------------------------------------------------------
# atomic variants

message("map2_dbl() |> futurize() works")
x <- c(1, 2, 3)
y <- c(4, 5, 6)

stopifnot(identical(
  map2_dbl(x, y, ~.x + .y) |> futurize_and_verify(),
  map2_dbl(x, y, ~.x + .y)
))


message("future_map2_int() works")
x <- c(1L, 2L, 3L)
y <- c(4L, 5L, 6L)

stopifnot(identical(
  map2_int(x, y, ~.x + .y) |> futurize_and_verify(),
  map2_int(x, y, ~.x + .y)
))


message("future_map2_lgl() works")
x <- c(TRUE, FALSE, TRUE)
y <- c(FALSE, TRUE, TRUE)

stopifnot(identical(
  map2_lgl(x, y, ~.x || .y) |> futurize_and_verify(),
  map2_lgl(x, y, ~.x || .y)
))


message("future_map2_chr() works")
x <- c("a", "b", "c")
y <- c("d", "e", "f")

stopifnot(identical(
  map2_chr(x, y, ~.y) |> futurize_and_verify(),
  map2_chr(x, y, ~.y)
))


message("names of `.x` are retained")
x <- c(a = 1, b = 2)
y <- c(c = 1, d = 2)
stopifnot(identical(names(map2_dbl(x, y, ~1) |> futurize_and_verify()), c("a", "b")))


# ------------------------------------------------------------------------------
# data frame variants

message("future_map2_dfr() works")
x <- c("a", "b", "c")
y <- c("d", "e", "f")

stopifnot(identical(
  map2_dfr(x, y, ~data.frame(x = .x, y = .y)) |> futurize_and_verify(),
  map2_dfr(x, y, ~data.frame(x = .x, y = .y))
))


message("future_map2_dfc() works")
x <- c("a", "b", "c")
y <- c("d", "e", "f")

stopifnot(identical(
  map2_dfc(x, y, ~as.data.frame(set_names(list(.x), .y))) |> futurize_and_verify(),
  map2_dfc(x, y, ~as.data.frame(set_names(list(.x), .y)))
))


# ------------------------------------------------------------------------------
# size

message("future_map2() works with size zero input")
stopifnot(identical(map2(list(), list(), identity) |> futurize::futurize(), list()))


message("atomic variants work with size zero input")
stopifnot(identical(map2_chr(list(), list(), identity) |> futurize::futurize(), character()))
stopifnot(identical(map2_dbl(list(), list(), identity) |> futurize::futurize(), double()))
stopifnot(identical(map2_int(list(), list(), identity) |> futurize::futurize(), integer()))
stopifnot(identical(map2_lgl(list(), list(), identity) |> futurize::futurize(), logical()))


message("size one recycling works")
stopifnot(identical(
  map2(1, 1:2, ~c(.x, .y)) |> futurize_and_verify(),
  list(c(1, 1), c(1, 2))
))

stopifnot(identical(
  map2(1:2, 1, ~c(.x, .y)) |> futurize_and_verify(),
  list(c(1, 1), c(2, 1))
))

stopifnot(identical(
  map2(integer(), 1, ~c(.x, .y)) |> futurize::futurize(),
  list()
))

stopifnot(identical(
  map2(1, integer(), ~c(.x, .y)) |> futurize::futurize(),
  list()
))


message("generally can't recycle to size zero")
res <- tryCatch({
  map2(1:2, integer(), ~c(.x, .y)) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(
  inherits(res, "error"),
  grepl("Can't recycle", conditionMessage(res))
)

res <- tryCatch({
  map2(integer(), 1:2, ~c(.x, .y)) |> futurize_and_verify()
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(
  inherits(res, "error"),
  grepl("Can't recycle", conditionMessage(res))
)


# ------------------------------------------------------------------------------
# Miscellaneous

message("globals in `.x` and `.y` are found (#16)")
fn1 <- function(x) sum(x, na.rm = TRUE)
fn2 <- function(x) sum(x, na.rm = FALSE)

x <- list(c(1, 2, NA), c(2, 3, 4))

fns1 <- map(x, ~purrr::partial(fn1, x = .x))
fns2 <- map(x, ~purrr::partial(fn2, x = .x))

stopifnot(identical(
  map2(fns1, fns2, ~c(.x(), .y())) |> futurize_and_verify(),
  list(c(3, NA), c(9, 9))
))


message("chunk balancing is correct after a recycle (#30)")
stopifnot(identical(
  map2(1, 1:4, ~c(.x, .y)) |> futurize_and_verify(),
  list(c(1, 1), c(1, 2), c(1, 3), c(1, 4))
))

plan(sequential)
} ## if (requireNamespace("purrr") && requireNamespace("furrr"))

# shapr::explain(...) =>
#
# local({
#   ## This will be automatically consumed and removed by 'future.apply'
#   options(future.disposable = structure(<future options>, dispose = FALSE))
#   on.exit(options(future.disposable = NULL))
#   shapr::explain(...)
# })
#
append_transpilers_for_shapr <- function() {
  transpilers <- make_package_transpilers("shapr", FUN = function(fcn, name) {
    if (name %in% c("explain", "explain_forecast")) {
      defaults <- list(
        future.seed = TRUE,
        future.label = sprintf("fz:shapr::%s-%%d", name)
      )
      list(
        label = sprintf("shapr::%s() ~> shapr::%s()", name, name),
        transpiler = make_futurize_for_future.apply(defaults = defaults)
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("shapr", "future.apply")
}

# SingleCellExperiment::applySCE(sce, FUN, ...) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
#   options(future.disposable = <future arguments>)
#   SingleCellExperiment::applySCE(sce, FUN, ..., BPPARAM = BiocParallel::DoparParam())
# })
#
append_transpilers_for_SingleCellExperiment <- function() {
  names <- c(
    "applySCE"
  )

  ns <- getNamespace("SingleCellExperiment")
  transpilers <- list()
  for (name in names) {
    if (exists(name, mode = "function", envir = ns, inherits = TRUE)) {
      args <- list(BPPARAM = quote(BiocParallel::DoparParam()))
      transpilers[[name]] <- list(
        label = sprintf("SingleCellExperiment::%s() ~> SingleCellExperiment::%s(..., BPPARAM = BiocParallel::DoparParam())", name, name),
        transpiler = make_futurize_for_doFuture(
          args = args,
          defaults = list(label = sprintf("fz:SingleCellExperiment::%s-%%d", name))
        )
      )
    }
  }

  transpilers <- list(transpilers)
  names(transpilers) <- "SingleCellExperiment"

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("SingleCellExperiment", "BiocParallel", "doFuture")
}

#' @tags pkg-lme4
if (requireNamespace("lme4") && getRversion() >= "4.4.0") {
library(futurize)
library(lme4)
options(future.rng.onMisuse = "error")

## Disable checks for stray assignments to the global environment(),
## because lme4::allFit() assigns a 'ctrl' variable when called.
## See https://github.com/lme4/lme4/issues/853 (aug 2025) for details.
options(future.globalenv.onMisuse = "warning")

all_equal <- function(a, b, ...) {
  ## Cannot use all.equal(a, b), because there is timing data
  a <- lapply(a, summary)
  b <- lapply(b, summary)
  all.equal(a, b, ...)
}

plan(multisession)

## Adopted from example("allFit", package = "lme4")
gm1 <- glmer(cbind(incidence, size - incidence) ~ period + (1 | herd),
             data = cbpp, family = binomial)

message("Ordinary processing:")
set.seed(42)
gm_all_truth <- allFit(gm1)
print(gm_all_truth)

message("Futurized processing:")
set.seed(42)
gm_all <- allFit(gm1) |> futurize_and_verify()
print(gm_all)

message("Comparing results:")
stopifnot(all_equal(gm_all, gm_all_truth))

if (utils::packageVersion("lme4") >= "2.0.1") {
  ## influence.merMod() via stats::influence() S3 generic dispatch
  ## Adopted from example("influence.merMod", package = "lme4")
  fm1 <- lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy)
  stopifnot(inherits(fm1, "merMod"))
  
  message("Ordinary processing:")
  inf_truth <- influence(fm1, groups = "Subject")
  
  message("Futurized processing:")
  inf <- influence(fm1, groups = "Subject") |> futurize_and_verify()
  
  message("Comparing results:")
  stopifnot(all.equal(inf, inf_truth))
}

plan(sequential)
} ## if (requireNamespace("lme4"))

# ez::ezBoot(...) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
#   options(future.disposable = <future arguments>)
#   ez::ezBoot(..., parallel = TRUE)
# })
#
# and similar for ezPerm(), ezPlot2()
#
append_transpilers_for_ez <- function() {
  transpilers <- make_package_transpilers("ez", FUN = function(fcn, name) {
    if ("parallel" %in% names(formals(fcn))) {
      defaults <- list(label = sprintf("fz:ez::%s-%%d", name))
      if (name %in% c("ezBoot", "ezPerm")) {
        defaults[["seed"]] <- TRUE
      }
      
      list(
        label = sprintf("ez::%s() ~> ez::%s(..., parallel = TRUE)", name, name),
        transpiler = make_futurize_for_doFuture(defaults = defaults, args = list(parallel = TRUE))
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("ez", "doFuture")
}

#' @tags pkg-DESeq2
#' @tags skip_on_cran  ## (35s) to limit total check time
if (requireNamespace("DESeq2") && requireNamespace("doFuture")) {
library(futurize)
library(DESeq2)
options(future.rng.onMisuse = "error")

plan(multisession)

## Create a simple DESeqDataSet
set.seed(42)
n_genes <- 30L
n_samples <- 4L
counts <- matrix(
  as.integer(runif(n_genes * n_samples, min = 0, max = 1000)),
  nrow = n_genes,
  ncol = n_samples
)
rownames(counts) <- paste0("gene", seq_len(n_genes))
colnames(counts) <- paste0("sample", seq_len(n_samples))

col_data <- data.frame(
  condition = factor(rep(c("control", "treated"), each = n_samples / 2L)),
  row.names = colnames(counts)
)

dds <- DESeqDataSetFromMatrix(
  countData = counts,
  colData = col_data,
  design = ~ condition
)

## ---------------------------------------------------------
## DESeq()
## ---------------------------------------------------------
set.seed(42)
result_truth <- DESeq(dds)
print(result_truth)

set.seed(42)
result <- DESeq(dds) |> futurize_and_verify()
print(result)
stopifnot(all.equal(results(result), results(result_truth)))

set.seed(42)
result2 <- DESeq2::DESeq(dds) |> futurize_and_verify()
stopifnot(all.equal(results(result2), results(result_truth)))

plan(sequential)
} ## if (requireNamespace("DESeq2") && ...)

#' @tags pkg-furrr
if (requireNamespace("purrr") && requireNamespace("furrr")) {
library(futurize)


library(purrr)

# ------------------------------------------------------------------------------
# future_modify()

message("future_modify() default method works")
stopifnot(identical(modify(list(1, 2), ~3) |> futurize_and_verify(), list(3, 3)))
stopifnot(identical(
  modify(data.frame(x = 1, y = 2), ~3) |> futurize_and_verify(),
  data.frame(x = 3, y = 3)
))


# TODO: Fix `NULL` behavior to match what is done here
# https://github.com/tidyverse/purrr/pull/754
message("modify() |> futurize() is not stable when returning `NULL`")
stopifnot(identical(
  modify(
    list(1, 2),
    ~if (.x == 1) {
      NULL
    } else {
      .x
    }
  ) |> futurize_and_verify(),
  list(2, 2)
))


message("future_modify() variants work")
stopifnot(identical(modify(c(1L, 2L, 3L), ~2L) |> futurize_and_verify(), rep(2L, 3)))
stopifnot(identical(modify(c(1, 2, 3), ~2) |> futurize_and_verify(), rep(2, 3)))
stopifnot(identical(modify(c("a", "b", "c"), toupper) |> futurize_and_verify(), c("A", "B", "C")))
stopifnot(identical(modify(c(TRUE, FALSE, TRUE), ~TRUE) |> futurize_and_verify(), rep(TRUE, 3)))


message("modify(<pairlist>) |> futurize() works")
x <- as.pairlist(list(1, 2))
stopifnot(identical(class(modify(x, ~.x) |> futurize_and_verify()), "pairlist"))


# ------------------------------------------------------------------------------
# future_modify_at()

message("future_modify_at() default works")
stopifnot(identical(modify_at(list(1, 2, 3), c(1, 3), ~5) |> futurize_and_verify(), list(5, 2, 5)))
stopifnot(identical(
  modify_at(data.frame(x = 1, y = 2), 2, ~3) |> futurize_and_verify(),
  data.frame(x = 1, y = 3)
))


message("future_modify_at() variants works")
stopifnot(identical(modify_at(c(1L, 2L, 3L), c(1, 3), ~5L) |> futurize_and_verify(), c(5L, 2L, 5L)))
stopifnot(identical(modify_at(c(1, 2, 3), c(1, 3), ~5) |> futurize_and_verify(), c(5, 2, 5)))
stopifnot(identical(
  modify_at(c("a", "b", "c"), c(1, 3), toupper) |> futurize_and_verify(),
  c("A", "b", "C")
))
stopifnot(identical(
  modify_at(c(TRUE, FALSE, TRUE), c(1, 3), ~NA) |> futurize_and_verify(),
  c(NA, FALSE, NA)
))


# ------------------------------------------------------------------------------
# future_modify_if()

message("future_modify_if() default works")
stopifnot(identical(modify_if(list(1, 2), ~.x == 1, ~3) |> futurize_and_verify(), list(3, 2)))
stopifnot(identical(
  modify_if(data.frame(x = 1, y = 2), ~.x == 1, ~3) |> futurize_and_verify(),
  data.frame(x = 3, y = 2)
))


message("future_modify_if() `.else` works for default")
stopifnot(identical(
  modify_if(list(1, 2, 1, 4), ~.x == 1, ~3, .else = ~4) |> futurize_and_verify(),
  list(3, 4, 3, 4)
))


message("future_modify_if() variants works")
stopifnot(identical(
  modify_if(c(1L, 2L, 3L), ~.x == 1L, ~2L, .else = ~3L) |> futurize_and_verify(),
  c(2L, 3L, 3L)
))
stopifnot(identical(
  modify_if(c(1, 2, 3), ~.x == 1, ~2, .else = ~3) |> futurize_and_verify(),
  c(2, 3, 3)
))
stopifnot(identical(
  modify_if(c("a", "b", "c"), ~.x == "a", toupper, .else = ~"d") |> futurize_and_verify(),
  c("A", "d", "d")
))
stopifnot(identical(
  modify_if(c(TRUE, FALSE, TRUE), ~.x == TRUE, ~TRUE, .else = ~NA) |> futurize_and_verify(),
  c(TRUE, NA, TRUE)
))

plan(sequential)
} ## if (requireNamespace("purrr") && requireNamespace("furrr"))

#' @tags pkg-stars
if (requireNamespace("stars")) {
  library(futurize)
  library(stars)

  plan(multisession)

  ## Create a small stars object for testing
  m <- matrix(1:20, nrow = 5, ncol = 4)
  s <- st_as_stars(m)

  ## -------------------------------------------------------
  ## st_apply()
  ## -------------------------------------------------------
  result_truth <- stars::st_apply(s, MARGIN = 1, FUN = mean)
  print(result_truth)

  ## Futurized via st_apply()
  result <- st_apply(s, MARGIN = 1, FUN = mean) |> futurize_and_verify()
  print(result)

  stopifnot(all.equal(result, result_truth))

  ## Futurized via stars::st_apply()
  result2 <- stars::st_apply(s, MARGIN = 1, FUN = mean) |> futurize_and_verify()
  print(result2)

  stopifnot(all.equal(result2, result_truth))

  ## Assert that future.globals.maxSize is restored
  oopts <- getOption("future.globals.maxSize")
  result3 <- stars::st_apply(s, MARGIN = 1, FUN = mean) |> futurize_and_verify()
  stopifnot(identical(getOption("future.globals.maxSize"), oopts))

  plan(sequential)
} ## if (requireNamespace("stars"))

#' @tags pkg-future.apply

if (requireNamespace("future.apply", quietly = TRUE)) {

if (require("datasets") && require("stats")) {
  library(futurize)


  library(datasets)
  
  plan(multisession)
  
  ## Adopted from example("kernapply", package = "stats")

  ## ------------------------------------------------------
  ## Test {future_}kernapply() for 'default'
  ## ------------------------------------------------------
  X <- EuStockMarkets[, 1:2]
  X <- unclass(X)
  stopifnot(inherits(X, "matrix"), !inherits(X, "ts"))

  k1 <- kernel("daniell", m = 50L)
  stopifnot(inherits(k1, "tskernel"))
  X1_truth <- kernapply(X, k = k1)
  str(X1_truth)
  X1 <- kernapply(X, k = k1) |> futurize_and_verify()
  str(X1)
  stopifnot(identical(X1, X1_truth))


  ## ------------------------------------------------------
  ## Test {future_}kernapply() for 'ts'
  ## ------------------------------------------------------
  X <- EuStockMarkets[, 1:2]
  stopifnot(inherits(X, "matrix"), inherits(X, "ts"))

  k1 <- kernel("daniell", m = 50L)
  stopifnot(inherits(k1, "tskernel"))
  X1_truth <- kernapply(X, k = k1)
  str(X1_truth)
  X1 <- kernapply(X, k = k1) |> futurize_and_verify()
  str(X1)
  stopifnot(identical(X1, X1_truth))


  plan(sequential)
}

} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @tags skip_on_cran  ## (20s) to limit total check time
if (requireNamespace("riskRegression") && requireNamespace("survival") && requireNamespace("doFuture")) {
library(futurize)
library(riskRegression)
library(survival)
options(future.rng.onMisuse = "error")

plan(multisession)

## -------------------------------------------------------------------
## Score() - bootstrap cross-validation
## -------------------------------------------------------------------
set.seed(42)
d <- sampleData(200, outcome = "competing.risks")
fit <- CSC(Hist(time, event) ~ X1 + X2 + X7 + X8, data = d)

set.seed(42)
sc_truth <- Score(list("CSC" = fit), data = d,
                  formula = Hist(time, event) ~ 1,
                  times = 5, B = 10, split.method = "bootcv",
                  seed = 42)
print(sc_truth)

set.seed(42)
sc <- Score(list("CSC" = fit), data = d,
            formula = Hist(time, event) ~ 1,
            times = 5, B = 10, split.method = "bootcv",
            seed = 42) |> futurize_and_verify()
print(sc)

plan(sequential)
} ## if (requireNamespace("riskRegression"))

# gamlss::gamlssCV(...) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   gamlss::gamlssCV(..., parallel = "snow", ncpus = 2L, cl = cl)
# })
#
append_transpilers_for_gamlss <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'gamlss' functions requires R (>= 4.4.0)", getRversion()))
  }

  ## Functions that do NOT respect a user-provided 'cl' argument:
  ## stepGAIC(), stepGAICAll.A(), stepGAICAll.B(), stepTGD(),
  ## stepTGDAll.A() always create their own cluster via
  ## parallel::makeForkCluster() regardless of the 'cl' value.
  ##
  ## Functions that use parallel::clusterEvalQ(), which is not
  ## supported: chooseDist(), chooseDistPred().
  skip <- c("stepGAIC", "stepGAICAll.A", "stepGAICAll.B",
            "stepTGD", "stepTGDAll.A",
            "chooseDist", "chooseDistPred")

  transpilers <- make_package_transpilers("gamlss", FUN = function(fcn, name) {
    if (name %in% skip) return(NULL)
    if (all(c("parallel", "ncpus", "cl") %in% names(formals(fcn)))) {
      transpiler <- make_futurize_for_makeClusterFuture(
        defaults = list(
          seed = (name == "gamlssCV"),
          packages = "gamlss",
          label = sprintf("fz:gamlss::%s", name)
        ),
        args = list(
          parallel = "snow",
          ncpus = 2L,   ## only used for test ncpus > 1
          cl = quote(cl)
        )
      )

      list(
        label = sprintf("gamlss::%s() ~> gamlss::%s(..., parallel = TRUE)", name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("gamlss", "future")
}

#' @tags pkg-furrr
if (requireNamespace("purrr") && requireNamespace("furrr")) {
library(futurize)


library(purrr)

message("imap functions work with unnamed input")

stopifnot(identical(imap(1:2, ~.y) |> futurize_and_verify(), list(1L, 2L)))
stopifnot(identical(imap_chr(1:2, ~as.character(.y)) |> futurize_and_verify(), c("1", "2")))
stopifnot(identical(imap_int(1:2, ~.y) |> futurize_and_verify(), c(1L, 2L)))
stopifnot(identical(imap_dbl(1:2, ~.y) |> futurize_and_verify(), c(1, 2)))
stopifnot(identical(imap_lgl(1:2, ~identical(.y, 1L)) |> futurize_and_verify(), c(TRUE, FALSE)))
stopifnot(identical(
  imap_dfr(1:2, ~data.frame(x = .y)) |> futurize_and_verify(),
  data.frame(x = c(1L, 2L))
))
stopifnot(identical(
  imap_dfc(1:2, ~vctrs::new_data_frame(set_names(list(1), .y))) |> futurize_and_verify(),
  vctrs::new_data_frame(list(`1` = 1, `2` = 1))
))


message("imap functions work with named input")
x <- set_names(1:2, c("x", "y"))
stopifnot(identical(imap(x, ~.y) |> futurize_and_verify(), list(x = "x", y = "y")))
stopifnot(identical(imap_chr(x, ~as.character(.y)) |> futurize_and_verify(), c(x = "x", y = "y")))
stopifnot(identical(
  imap_int(x, ~if (.y == "x") 1L else 2L) |> futurize_and_verify(),
  c(x = 1L, y = 2L)
))
stopifnot(identical(
  imap_dbl(x, ~if (.y == "x") 1 else 2) |> futurize_and_verify(),
  c(x = 1, y = 2)
))
stopifnot(identical(
  imap_lgl(x, ~if (.y == "x") TRUE else FALSE) |> futurize_and_verify(),
  c(x = TRUE, y = FALSE)
))
stopifnot(identical(
  imap_dfr(x, ~data.frame(x = .y)) |> futurize_and_verify(),
  data.frame(x = c("x", "y"))
))
stopifnot(identical(
  imap_dfc(x, ~vctrs::new_data_frame(set_names(list(1), .y))) |> futurize_and_verify(),
  vctrs::new_data_frame(list(x = 1, y = 1))
))
  
plan(sequential)
} ## if (requireNamespace("purrr") && requireNamespace("furrr"))

# plyr::llply(xs, fcn) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), 
#   plyr::llply(xs, fcn,
#     .parallel = TRUE,
#     .paropts = list(.options.future = <future arguments>)
#   )
# )
#
append_transpilers_for_plyr <- function() {
  template <- bquote_compile(
    with(doFuture::registerDoFuture(flavor = "%dofuture%"), (.(EXPR)))
  )

  template2 <- bquote_compile(function(expr, options = NULL) {
    defaults <- list(label = sprintf("fz:plyr::%s-%%d", .(NAME)))
    options <- make_options_for_doFuture(options, defaults = defaults, wrap = TRUE)
    expr <- append_call_arguments(expr,
      .parallel = TRUE,
      .paropts = options
    )
    bquote_apply(template, EXPR = expr)
  })

  transpilers <- make_package_transpilers("plyr", FUN = function(fcn, name) {
    if (".parallel" %in% names(formals(fcn))) {
      transpiler <- eval(bquote_apply(template2, NAME = name))
      list(
        label = sprintf("plyr::%s() ~> plyr::%s(..., parallel = TRUE)", name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("plyr", "doFuture")
}

#' @tags pkg-partykit
if (requireNamespace("partykit") && requireNamespace("future.apply")) {
library(futurize)
library(partykit)
options(future.rng.onMisuse = "error")

plan(multisession)

RNGkind("L'Ecuyer-CMRG")
set.seed(42)
cf_truth <- partykit::cforest(dist ~ speed, data = cars, ntree = 10L)
print(summary(cf_truth))

set.seed(42)
cf <- partykit::cforest(dist ~ speed, data = cars, ntree = 10L) |> futurize_and_verify()
print(summary(cf))

stopifnot(
  all.equal(summary(cf), summary(cf_truth))
)

## Prediction
nd_truth <- data.frame(speed = 4:25)
nd_truth$mean  <- predict(cf_truth, newdata = nd_truth, type = "response")

nd <- data.frame(speed = 4:25)
nd$mean  <- predict(cf, newdata = nd, type = "response")

stopifnot(
  all.equal(nd, nd_truth)
)

plan(sequential)
} ## if (requireNamespace("partykit"))

# BiocParallel::bplapply(xs, fcn) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
#   options(future.disposable = <future arguments>)
#   BiocParallel::bplapply(xs, fcn, BPPARAM = BiocParallel::DoparParam())
# })
#
append_transpilers_for_BiocParallel <- function() {
  template_stdout <- bquote_compile(
    with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
      ## This will be automatically removed by doFuture
      options(future.disposable = .(OPTS))
      .(EXPR)
    })
  )

  ## WORKAROUNDS for BiocParallel
  ## https://github.com/Bioconductor/BiocParallel/issues/276
  ## https://github.com/Bioconductor/BiocParallel/issues/277
  template_no_stdout <- bquote_compile(
    with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
      ## This will be automatically removed by doFuture
      options(future.disposable = .(OPTS))
      ...futurize.result <- tryCatch({
        ...futurize.nullcon <- file(nullfile(), open = "w")
        sink(file = ...futurize.nullcon, type = "output")
        sink(file = ...futurize.nullcon, type = "message")
        withVisible(.(EXPR))
      }, finally = {
        sink(NULL, type = "message")
        sink(NULL, type = "output")
        close(...futurize.nullcon)
      })
      if (...futurize.result[["visible"]]) {
        ...futurize.result[["value"]]
      } else {
        invisible(...futurize.result[["value"]])
      }
    })
  )

  template2 <- bquote_compile(function(expr, options = NULL) {
    stdout <- options[["stdout"]]
    template <- if (isFALSE(stdout)) template_no_stdout else template_stdout
    expr <- append_call_arguments(expr,
      BPPARAM = quote(BiocParallel::DoparParam())
    )
    defaults <- list(label = sprintf("fz:BiocParallel::%s-%%d", .(NAME)))
    opts <- make_options_for_doFuture(options, defaults = defaults, wrap = TRUE)
    ## Update 'OPTS'
    bquote_apply(template,
      OPTS = opts,
      EXPR = expr
    )
  })

  transpilers <- make_package_transpilers("BiocParallel", FUN = function(fcn, name) {
    ## Skip some BiocParallel functions
    if (name %in% c("bpvectorize", "register")) return(NULL)
    if ("BPPARAM" %in% names(formals(fcn))) {
      transpiler <- eval(bquote_apply(template2, NAME = name))
      list(
        label = sprintf("BiocParallel::%s() ~> BiocParallel::%s(..., BPPARAM = BiocParallel::DoparParam())", name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("BiocParallel", "doFuture")
}

#' @tags pkg-future.apply
#' @tags future_lapply
#' @tags listenv
#' @tags sequential multisession multicore

if (requireNamespace("future.apply", quietly = TRUE)) {

library(futurize)


library(listenv)

message("*** future_lapply() ...")

x_a <- list(a = "integer", b = "numeric", c = "character", c = "list")
str(list(x_a = x_a))
y_a <- lapply(x_a, FUN = base::vector, length = 2L)
str(list(y_a = y_a))

x_b <- list(a = c("hello", b = 1:100))
str(list(x_b = x_b))
y_b <- lapply(x_b, FUN = future:::hpaste, collapse = "; ", maxHead = 3L)
str(list(y_b = y_b))

x_c <- list()
y_c <- listenv()
y_c$A <- 3L
x_c$a <- y_c
y_c<- listenv()
y_c$A <- 3L
y_c$B <- c("hello", b = 1:100)
x_c$b <- y_c
print(x_c)
y_c <- lapply(x_c, FUN = listenv::mapping)
str(list(y_c = y_c))

for (cores in 1:availCores) {
  message(sprintf("Testing with %d cores ...", cores))
  options(mc.cores = cores)
  strategies <- supportedStrategies(cores)

  for (strategy in supportedStrategies()) {
    message(sprintf("- plan('%s') ...", strategy))
    plan(strategy)

    for (scheduling in list(FALSE, TRUE, structure(TRUE, ordering = "random"), structure(TRUE, ordering = function(n) rev(seq_len(n))))) {
      message("- lapply(x, FUN = vector, ...) |> futurize() ...")
      y <- lapply(x_a, FUN = vector, length = 2L) |> futurize_and_verify(scheduling = scheduling)
      str(list(y = y))
      stopifnot(identical(y, y_a))

      y <- lapply(x_a, FUN = "vector", length = 2L) |> futurize_and_verify(scheduling = scheduling)
      str(list(y = y))
      stopifnot(identical(y, y_a))

      message("- lapply(x, FUN = base::vector, ...) |> futurize() ...")
      y <- lapply(x_a, FUN = base::vector, length = 2L) |> futurize_and_verify(scheduling = scheduling)
      str(list(y = y))
      stopifnot(identical(y, y_a))

      message("- lapply(x, FUN = future:::hpaste, ...) |> futurize() ...")
      y <- lapply(x_b, FUN = future:::hpaste, collapse = "; ", maxHead = 3L) |> futurize_and_verify(scheduling = scheduling)
      str(list(y = y))
      stopifnot(identical(y, y_b))

      message("- lapply(x, FUN = listenv::listenv, ...) |> futurize() ...")
      y <- lapply(x_c, FUN = listenv::mapping) |> futurize_and_verify(scheduling = scheduling)
      str(list(y = y))
      stopifnot(identical(y, y_c))
    } ## for (scheduling ...)

    message("- lapply(x, FUN, ...) |> futurize() for large length(x) ...")
    a <- 3.14
    x_d <- 1:1e4
    y <- lapply(x_d, FUN = function(z) sqrt(z + a)) |> futurize_and_verify()
    y <- unlist(y, use.names = FALSE)
    stopifnot(all.equal(y, sqrt(x_d + a)))

    message("- lapply(x, FUN = table, ...) |> futurize() ...")
    x <- list(a = 1:4, b = 5:8)
    y0 <- lapply(x, FUN = table)
    y1 <- lapply(x, FUN = table) |> futurize_and_verify()
    stopifnot(all.equal(y1, y0, check.attributes = FALSE)) ## FIXME

    message("- lapply(x, ...) |> futurize() where length(x) != length(as.list(x)) ...")
    x <- structure(list(a = 1, b = 2), class = "Foo")
    as.list.Foo <- function(x, ...) c(x, c = 3)
    registerS3method("as.list", class = "Foo", method = as.list.Foo)
    y0 <- lapply(x, FUN = length)
    stopifnot(identical(y0, list(a = 1L, b = 1L, c = 1L)))
    y1 <- lapply(x, FUN = length) |> futurize_and_verify()
    stopifnot(identical(y1, y0))
    rm(list = "as.list.Foo")
    
    message("- lapply(x, ...) |> futurize() where x[[i]] subsets via S3 method ...")
    x <- structure(list(a = 1, b = 2), class = "Bar")
    `[[.Bar` <- function(x, ...) 0
    registerS3method("[[", class = "Bar", method = `[[.Bar`)
    y0 <- lapply(x, FUN = identity)
    stopifnot(identical(y0, list(a = 0, b = 0)))
    y1 <- lapply(x, FUN = identity) |> futurize_and_verify()
    if (getOption("future.apply.chunkWith", "[[") == "[") {
      stopifnot(identical(y1, unclass(x)))
    } else {
      stopifnot(identical(y1, y0))
    }
    rm(list = "[[.Foo")
} ## for (strategy ...)

  message(sprintf("Testing with %d cores ... DONE", cores))
} ## for (cores ...)


message("*** future_lapply() - special cases ...")

X <- list()
names(X) <- character(0L)
y <- lapply(X, FUN = identity) |> futurize::futurize()
stopifnot(length(y) == 0L, !is.null(names(y)), identical(y, X))

X <- character(0L)
y0 <- lapply(X, FUN = identity)
y <- lapply(X, FUN = identity) |> futurize::futurize()
stopifnot(identical(y, y0))

X <- character(0L)
names(X) <- character(0L)
y0 <- lapply(X, FUN = identity)
y <- lapply(X, FUN = identity) |> futurize::futurize()
stopifnot(identical(y, y0))

message("*** future_lapply() - special cases ... DONE")


message("*** future_lapply() - exceptions ...")

res <- tryCatch({
  lapply(1:3, FUN = identity) |> futurize_and_verify(chunk_size = structure(1L, ordering = "invalid"))
}, FuturizeTestAssertionError = stop, error = identity)
stopifnot(inherits(res, "error"))

message("*** future_lapply() - exceptions ... DONE")

message("*** future_lapply() ... DONE")

} ## if (requireNamespace("future.apply", quietly = TRUE))

#' @noRd
make_options_for_furrr <- local({
  defaults_base <- NULL
  
  function(options, fcn, defaults = NULL) {
    if (is.null(defaults_base)) {
      defaults_base <<- setdiff(names(formals(furrr::furrr_options)), "...")
    }

    names <- names(options)
    if ("label" %in% names) {
      names[names == "label"] <- "prefix"
      names(options) <- names
    }
    
    ## Set default 'prefix'?
    if (!"prefix" %in% names && "prefix" %in% names(defaults)) {
      options[["prefix"]] <- defaults[["prefix"]]
    }

    ## Silently drop unknown future options
    keep <- intersect(defaults_base, names(options))
    options <- options[keep]

    options <- do.call(furrr::furrr_options, args = options)
    options <- list(.options = options)
    
    options
  }
})

# foreach(...) %do% { ... } =>
#   foreach(..., .options.future = <future arguments>) %dofuture% { ... }
#'
# times(...) %do% { ... } =>
#   local({
#     oopts <- options(future.disposable = <future arguments>)
#     on.exit(options(oopts))
#     times(...) %dofuture% { ... }
#   })
#
append_transpilers_for_doFuture <- function() {
  template <- bquote_compile(local({
    oopts <- options(future.disposable = .(OPTS))
    on.exit(options(oopts))
    .(EXPR)
  }))
  
  transpiler <- function(expr, options = NULL) {
    ## Replace `%do%` with doFuture::`%dofuture%`
    expr[[1]] <- quote(doFuture::`%dofuture%`)
    call <- expr[[2]]
    fcn <- call[[1]]
    
    ## times()?
    if (identical(fcn, as.symbol("times")) ||
        identical(fcn, quote(foreach::times))) {
      ## Default to seed = TRUE
      if (!"seed" %in% attr(options, "specified")) {
        options[["seed"]] <- TRUE
      }
      if (!"label" %in% attr(options, "specified")) {
        options[["label"]] <- "fz:foreach::times-%d"
      }
      expr <- bquote_apply(template,
        OPTS = options,
        EXPR = expr
      )
    } else {
      if (identical(fcn, as.symbol("%:%")) ||
                  identical(fcn, quote(foreach::`%:%`))) {
        name <- "%:%"
        label <- "%%:%%"
      } else {
        name <- "foreach"
        label <- "foreach"
      }
      defaults <- list(label = sprintf("fz:foreach::%s-%%d", label))
      options <- make_options_for_doFuture(options, defaults = defaults, wrap = TRUE)
      if (identical(fcn, as.symbol("%:%")) ||
               identical(fcn, quote(foreach::`%:%`))) {
        idx_EXPR <- 2:3
      } else {
        idx_EXPR <- 2L
      }

      expr[[idx_EXPR]] <- append_call_arguments(expr[[idx_EXPR]],
        .args = options
      )
    }
    expr
  }

  transpilers <- list()
  transpilers[["%do%"]] <- list(
    label = "foreach::foreach() %do% { ... } -> foreach::foreach() %dofuture% { ... }",
    transpiler = transpiler
  )

  for (name in c("%dofuture%", "%dopar%")) {
    transpilers[[name]] <- list(
      label = sprintf("foreach::foreach() %s { ... } - not supported", name),
      transpiler = eval(bquote(function(...) {
        stop(sprintf("Cannot futurize foreach::foreach() %s { ... } - use %%do%% instead", .(name)))
      }))
    )
  }

  transpilers <- list(transpilers)
  names(transpilers) <- "foreach"

  append_transpilers("futurize::add-on", transpilers)
  
  ## Return required packages
  c("doFuture")
}

# DiceKriging::km(..., multistart = N) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
#   options(future.disposable = <future arguments>)
#   DiceKriging::km(..., multistart = N)
# })
#
# Note: The parallel foreach loop is triggered only when multistart > 1.
# The loop lives in the internal function kmEstimate(), which is called by km().
#
append_transpilers_for_DiceKriging <- function() {
  transpilers <- make_package_transpilers("DiceKriging", FUN = function(fcn, name) {
    ## Only km() triggers the internal foreach %dopar% loop
    if (name != "km") return(NULL)

    list(
      label = "DiceKriging::km() ~> DiceKriging::km() [via doFuture]",
      transpiler = make_futurize_for_doFuture(
        defaults = list(
          label = "fz:DiceKriging::km-%d"
        )
      )
    )
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("DiceKriging", "doFuture")
}

# pbapply::pblapply(xs, fcn, ...) =>
#
# local({
#   options(future.disposable = structure(OPTS, dispose = FALSE))
#   on.exit(options(future.disposable = NULL))
#   pbapply::pblapply(xs, fcn, ..., cl = "future")
# })
#
append_transpilers_for_pbapply <- function() {
  template <- bquote_compile(
   local({
      ## This will be automatically consumed and removed by 'future.apply'
      options(future.disposable = structure(.(OPTS), dispose = FALSE))
      on.exit(options(future.disposable = NULL))
      .(EXPR)
    })
  )

  template2 <- bquote_compile(function(expr, options = NULL) {
    ## Specified future.* arguments
    specified <- attr(options, "specified")

    names(options) <- sub("chunk_size", "chunk.size", names(options), fixed = TRUE)
     if (!"label" %in% specified && !"label" %in% names(options)) {
      options[["label"]] <- sprintf("fz:pbapply::%s-%%d", .(NAME))
    }

    ## pbapply does not handle stdout, messages, and warnings, so disable
    ## those by default
    if (!"stdout" %in% specified) {
      options$stdout <- FALSE
    }
    if (!"conditions" %in% specified) {
      options$conditions <- structure(options$conditions,
                               exclude = c("message", "warning"))
    }

    expr <- append_call_arguments(expr,
      cl = "future"
    )

    bquote_apply(template, OPTS = options, EXPR = expr)
  })

  transpilers <- make_package_transpilers("pbapply", FUN = function(fcn, name) {
    if ("cl" %in% names(formals(fcn))) {
      transpiler <- eval(bquote_apply(template2, NAME = name))

      list(
        label = sprintf("pbapply::%s() ~> pbapply::%s(..., cl = \"future\")", name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("pbapply")
}

## This runs 'testme' test inst/testme/test-<name>.R scripts
## Don't edit - it was autogenerated by inst/testme/deploy.R
testme <- function(name) {
  path <- system.file(package = 'futurize', 'testme', mustWork = TRUE)
  Sys.setenv(R_TESTME_PATH = path)
  Sys.setenv(R_TESTME_PACKAGE = 'futurize')
  Sys.setenv(R_TESTME_NAME = name)
  on.exit(Sys.unsetenv('R_TESTME_NAME'))
  source(file.path(path, 'run.R'))
}

# SimDesign::runSimulation(...) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   SimDesign::runSimulation(..., parallel = TRUE, ncores = 2L, cl = cl)
# })
#
append_transpilers_for_SimDesign <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'SimDesign' functions requires R (>= 4.4.0)", getRversion()))
  }

  transpilers <- make_package_transpilers("SimDesign", FUN = function(fcn, name) {
    if (all(c("parallel", "ncores", "cl") %in% names(formals(fcn)))) {
      transpiler <- make_futurize_for_makeClusterFuture(args = list(
        parallel = TRUE,
        ncores = 2L,   ## only used for test ncores > 1
        cl = quote(cl)
      ), defaults = list(label = sprintf("fz:SimDesign::%s", name)))

      list(
        label = sprintf("SimDesign::%s() ~> SimDesign::%s(..., parallel = TRUE)", name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("SimDesign", "future")
}

# pls::mvr(...) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   old_parallel <- pls::pls.options("parallel")
#   on.exit(pls::pls.options(old_parallel))
#   pls::pls.options(parallel = cl)
#   pls::mvr(...)
# })
#
append_transpilers_for_pls <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'pls' functions requires R (>= 4.4.0)", getRversion()))
  }

  template_pls <- bquote_compile(
    local({
      cl <- do.call(.(CALL), args = .(OPTS))
      old_parallel <- pls::pls.options("parallel")
      on.exit(pls::pls.options(old_parallel))
      pls::pls.options(parallel = cl)
      .(EXPR)
    })
  )

  transpilers <- make_package_transpilers("pls", FUN = function(fcn, name) {
    if (name %in% c("mvr", "plsr", "pcr", "cppls", "crossval")) {
      seed <- name == "crossval"
      transpiler <- make_futurize_for_makeClusterFuture(
        template = template_pls,
        defaults = list(
          seed = seed,
          label = sprintf("fz:pls::%s", name)
        )
      )

      list(
        label = sprintf("pls::%s() ~> pls.options(parallel = cl); pls::%s()",  name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("pls", "future")
}

# TSP::solve_TSP(...) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), local({
#   options(future.disposable = structure(OPTS, dispose = FALSE))
#   on.exit(options(future.disposable = NULL))
#   TSP::solve_TSP(...)
# }))
#
append_transpilers_for_TSP <- function() {
  transpilers <- make_package_transpilers("TSP", FUN = function(fcn, name) {
    if (name %in% c("solve_TSP")) {
      defaults <- list(
        seed = TRUE,
        label = sprintf("fz:TSP::%s-%%d", name)
      )
      list(
        label = sprintf("TSP::%s() ~> TSP::%s()", name, name),
        transpiler = make_futurize_for_doFuture(defaults = defaults)
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("TSP", "doFuture")
}

# GSVA::gsva(gsvaParam(expr, geneSets)) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
#   options(future.disposable = <future arguments>)
#   GSVA::gsva(gsvaParam(expr, geneSets), BPPARAM = BiocParallel::DoparParam())
# })
#
append_transpilers_for_GSVA <- function() {
  ns <- getNamespace("GSVA")

  ## gsva() in GSVA (< 2.4.2) and GSVA (>= 2.5.0 & < 2.5.7) relies on
  ## BiocParallel::bpiterate(), which does _not_ support DoparParam.
  ns_version <- getNamespaceVersion(ns)
  supports_gsva <- (ns_version >= "2.5.7") ||
                   (ns_version >= "2.4.2" && ns_version < "2.5.0")
                   
  ## These GSVA functions accept BPPARAM via '...' (S4 generics)
  names <- c(
    if (supports_gsva) "gsva",
    "gsvaRanks", "gsvaScores",
    "spatCor"
  )
  
  transpilers <- list()
  for (name in names) {
    if (exists(name, mode = "function", envir = ns, inherits = TRUE)) {
      args <- list(BPPARAM = quote(BiocParallel::DoparParam()))
      transpilers[[name]] <- list(
        label = sprintf("GSVA::%s() ~> GSVA::%s(..., BPPARAM = BiocParallel::DoparParam())", name, name),
        transpiler = make_futurize_for_doFuture(
          args = args,
          defaults = list(label = sprintf("fz:GSVA::%s-%%d", name))
        )
      )
    }
  }

  transpilers <- list(transpilers)
  names(transpilers) <- "GSVA"

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("GSVA", "BiocParallel", "doFuture")
}

register_vignette_engine_during_build_only <- function(pkgname) {
  ## HACK: Only register vignette engine 'selfonly' during R CMD build
  if (Sys.getenv("R_CMD") == "") return()

  tools::vignetteEngine("selfonly", package = pkgname, pattern = "[.]md$",
    weave = function(file, ...) {
      output <- sprintf("%s.html", tools::file_path_sans_ext(basename(file)))
      md <- readLines(file)

      title <- grep("%\\VignetteIndexEntry{", md, fixed = TRUE, value = TRUE)
      title <- gsub(".*[{](.*)[}].*", "\\1", title)

      ## Inject vignette title
      md <- c(sprintf("# %s\n\n", title), md)

      html <- commonmark::markdown_html(md,
                                        smart = FALSE,
                                        extensions = "table",
                                        normalize = FALSE)
      
      ## Embed images as <img src="data:image/png;base64...">
      mimes <- list(
        gif = "image/gif",
        jpg = "image/jpeg",
        png = "image/png",
        svg = "image/svg+xml",
        webp = "image/webp"
      )
      html <- unlist(strsplit(html, split = "\n", fixed = TRUE))
      for (ext in names(mimes)) {
        mime <- mimes[[ext]]
        pattern <- sprintf('(.*[ ]src=")([^"]+[.]%s)(".*)', ext)
        idxs <- grep(pattern, html)
        if (length(idxs) == 0) next
        if (!requireNamespace(stealth <- "base64enc", quietly = TRUE)) {
          stop("This vignette requires the ", sQuote(stealth), 
               " package because it contains a ", sQuote(toupper(ext)), 
               " image")
        }
        ns <- getNamespace(stealth)
        dataURI <- get("dataURI", mode = "function", envir = ns)
        for (idx in idxs) {
          file <- gsub(pattern, "\\2", html[idx])
          uri <- dataURI(file = file, mime = mime)
          html[idx] <- gsub(pattern, sprintf("\\1%s\\3", uri), html[idx])
        }
      }

      ## Inject HTML environment
      html <- c("<!DOCTYPE html>",
                "<html lang=\"en\">",
                "<head>",
                "<meta charset=\"UTF-8\">",
                sprintf("<title>%s</title>", title),
                "<style>",
                readLines("incl/extra.css", warn = FALSE),
                "</style>",
                "</head>",
                "<body>", html, "</body>",
                "</html>")

      writeLines(html, con = output)
      output
    },

    tangle = function(file, ...) {
      ## As of R 3.3.2, vignette engines must produce tangled output, but as
      ## long as it contains all comments then 'R CMD build' will drop it.
      output <- sprintf("%s.R", tools::file_path_sans_ext(basename(file)))
      cat(sprintf("### This is an R script tangled from %s\n",
                  sQuote(basename(file))), file = output)
      output
    }
  )
}

# base::lapply(xs, fcn, ...) =>
#
# future.apply::future_lapply(xs, fcn, ..., <future.* arguments>)
#
append_transpilers_for_future.apply <- function() {
  ## base::apply(), ...
  append_transpilers("futurize::add-on", make_addon_transpilers(
    "base", "future.apply", make_options = make_options_for_future.apply)
  )

  ## stats::kernapply()
  append_transpilers("futurize::add-on", make_addon_transpilers(
    "stats", "future.apply", make_options = make_options_for_future.apply)
  )

  ## Return required packages
  c("future.apply")
}

# Sim.DiffProc::MCM.sde(...) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   options(future.ClusterFuture.clusterEvalQ = "ignore")
#   on.exit(options(oopts))
#   Sim.DiffProc::MCM.sde(..., parallel = "snow", ncpus = 2L, cl = cl)
# })
#
append_transpilers_for_Sim.DiffProc <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'Sim.DiffProc' functions requires R (>= 4.4.0)", getRversion()))
  }

  template_ignore_clusterEvalQ <- bquote_compile(
    local({
      cl <- do.call(.(CALL), args = .(OPTS))
      oopts <- options(future.ClusterFuture.clusterEvalQ = "ignore")
      on.exit(options(oopts))
      .(EXPR)
    })
  )

  transpilers <- make_package_transpilers("Sim.DiffProc", FUN = function(fcn, name) {
    if ("cl" %in% names(formals(fcn))) {
      transpiler <- make_futurize_for_makeClusterFuture(
        template = template_ignore_clusterEvalQ,
        args = list(
          parallel = "snow",
          ncpus = 2L,   ## only used for test ncpus > 1
          cl = quote(cl)
        ), defaults = list(
          packages = "Sim.DiffProc",
          seed = TRUE,
          label = sprintf("fz:Sim.DiffProc::%s", name)
        )
      )

      list(
        label = sprintf("Sim.DiffProc::%s() ~> Sim.DiffProc::%s(..., parallel = TRUE)", name, name),
        transpiler = transpiler
      )
    }
  })

  ## Register both generic and method calls
  transpiler_generic <- make_futurize_for_makeClusterFuture(
    template = template_ignore_clusterEvalQ,
    args = list(
      parallel = "snow",
      ncpus = 2L,
      cl = quote(cl)
    ), defaults = list(
      packages = "Sim.DiffProc",
      seed = TRUE,
      label = "fz:Sim.DiffProc::MCM.sde"
    )
  )

  transpilers[["Sim.DiffProc"]][["MCM.sde"]] <- list(
    label = "Sim.DiffProc::MCM.sde() ~> Sim.DiffProc::MCM.sde(..., parallel = TRUE)",
    transpiler = transpiler_generic
  )

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("Sim.DiffProc", "future")
}

import_from <- function(name, mode = "function", default = NULL, package) {
  ns <- getNamespace(package)
  if (exists(name, mode = mode, envir = ns, inherits = FALSE)) {
    get(name, mode = mode, envir = ns, inherits = FALSE)
  } else if (!is.null(default)) {
    default
  } else {
    stop(sprintf("No such '%s' function: %s()", package, name))
  }
}

## The 'futurize' package don't really need to have a hard dependency on
## the 'future' package. However, we make 'futurize' attach 'future' for
## conveniency so that `plan()` is available.
#' @importFrom future plan
import_future <- function(name, default = NULL) {
  import_from(name, default = default, package = "future")
}

# glmmTMB::confint(...) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   glmmTMB::confint(..., parallel = "snow", ncpus = 2L, cl = cl)
# })
#
append_transpilers_for_glmmTMB <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'glmmTMB' functions requires R (>= 4.4.0)", getRversion()))
  }

  template <- bquote_compile(
    local({
      ## WORKAROUND: https://github.com/glmmTMB/glmmTMB/issues/1265
      base_attach <- base::attach # silence R CMD check
      base_attach(list(new_cl = FALSE), name = "glmmTMB:patch")
      on.exit(detach("glmmTMB:patch"))
      
      cl <- do.call(.(CALL), args = .(OPTS))
      oopts <- options(future.ClusterFuture.clusterEvalQ = "error")
      on.exit(options(oopts), add = TRUE)
      .(EXPR)
    })
  )

  transpilers <- make_package_transpilers("glmmTMB", FUN = function(fcn, name) {
    if (all(c("parallel", "ncpus", "cl") %in% names(formals(fcn)))) {
      ## Skip confint() for 'glmmTMB' until has been fixed per
      ## https://github.com/glmmTMB/glmmTMB/issues/1268
      if (name == "confint.glmmTMB") {
        return()
      }

      list(
        label = sprintf("glmmTMB::%s() ~> glmmTMB::%s(..., parallel = \"snow\")", name, name),
        transpiler = make_futurize_for_makeClusterFuture(
          args = list(
            parallel = "snow",
            ncpus = 2L,   ## only used for test ncpus > 1
            cl = quote(cl)
          ),
          defaults = list(label = sprintf("fz:glmmTMB::%s", name)),
          template = template
        )
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("glmmTMB", "future")
}

# stars::st_apply(X, MARGIN, FUN, ...) =>
#
# local({
#   ## This will be automatically consumed and removed by 'future.apply'
#   options(future.disposable = structure(OPTS, dispose = FALSE))
#   ## Protect against 'stars' crude setting of future.globals.maxSize
#   ...future_oopts <- options(future.globals.maxSize = getOption("future.globals.maxSize"))
#   on.exit({
#     options(...future_oopts)
#     options(future.disposable = NULL)
#   })
#   stars::st_apply(X, MARGIN, FUN, ..., FUTURE = TRUE)
# })
#
append_transpilers_for_stars <- function() {
  template <- bquote_compile(
   local({
      ## This will be automatically consumed and removed by 'future.apply'
      options(future.disposable = structure(.(OPTS), dispose = FALSE))
      ## Protect against 'stars' crude setting of future.globals.maxSize
      ...future_oopts <- options(future.globals.maxSize = getOption("future.globals.maxSize"))
      on.exit({
        options(...future_oopts)
        options(future.disposable = NULL)
      })
      .(EXPR)
    })
  )

  transpilers <- make_package_transpilers("stars", FUN = function(fcn, name) {
    if ("FUTURE" %in% names(formals(fcn))) {
      defaults <- list(
        future.label = sprintf("fz:stars::%s-%%d", name)
      )
      list(
        label = sprintf("stars::%s() ~> stars::%s(..., FUTURE = TRUE)", name, name),
        transpiler = make_futurize_for_future.apply(
          defaults = defaults,
          args = list(FUTURE = TRUE),
          template = template
        )
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("stars", "future.apply")
}

#' Options controlling resources, scheduling and evaluation of futures
#'
#' Some futures require specific **resources** and capabilities (e.g., RNG,
#' packages, and globals) in order to be evaluated. These can be declared
#' via this function.
#' In addition, this function can control how futures are **evaluated**,
#' including what is collected from futures (e.g., standard output and
#' conditions).
#' It can also control how future are **scheduled**, including how a set
#' of futures are partitioned (e.g. chunking).
#'
#' @inheritParams future::future
#'
#' @param scheduling (scheduling) Average number of futures ("chunks") per
#'        worker.
#'        If `0.0`, then a single future is used to process all elements.
#'        If `1.0` or `TRUE`, then one future per worker is used.
#'        If `2.0`, then each worker will process two futures
#'        (if there are enough elements).
#'        If `Inf` or `FALSE`, then one future per element
#'        is used.
#'        Only used if `chunk_size` is `NULL`.
#'
#' @param chunk_size (scheduling) The average number of elements per future
#'        ("chunk").
#'        If `Inf`, then all elements are processed in a single future.
#'        If `NULL`, then argument `scheduling` is used.
#
#' @param \ldots Additional named options.
#'
#' @return
#' A named list of future options.
#' Attribute `specified` is a character vector of future options
#' that were explicitly specified.
#'
#' @examples
#' # Default futurize options
#' str(futurize_options())
#'
#' @export
futurize_options <- function(seed = FALSE, globals = TRUE, packages = NULL, stdout = TRUE, conditions = "condition", scheduling = 1.0, chunk_size = NULL, ...) {
  args <- list(
     # Resources:
           seed = seed,
        globals = globals,
       packages = packages,
     # Evaluation:
         stdout = stdout,
     conditions = conditions,
     # Scheduling:
     scheduling = scheduling,
     chunk_size = chunk_size,
                  ...
  )
  specified <- character(0L)
  if (!missing(seed)) specified <- c(specified, "seed")
  if (!missing(globals)) specified <- c(specified, "globals")
  if (!missing(packages)) specified <- c(specified, "packages")
  if (!missing(stdout)) specified <- c(specified, "stdout")
  if (!missing(conditions)) specified <- c(specified, "conditions")
  if (!missing(scheduling)) specified <- c(specified, "scheduling")
  if (!missing(chunk_size)) specified <- c(specified, "chunk_size")
  specified <- c(specified, ...names())
  attr(args, "specified") <- specified
  args
}

#' Parse an R call to be transpiled
#'
#' @param call An R symbol or an R call.
#'
#' @param envir The environment from which to search for the function that
#' is called in the `call`.
#'
#' @param what A character string used in error messages describing what
#' type of transpiler is used.
#'
#' @param debug If TRUE, output debug information.
#'
#' @return
#' A named list with elements:
#'
#'  * `fcn`: the function called in the `call`
#'
#'  * `fcn_name`: the name of the function
#'
#'  * `ns_name`: the name of the namespace where the function lives
#'
#' @keywords internal
#' @noRd
parse_call <- function(call, envir = parent.frame(), what = "transpiler", debug = FALSE) {
  if (debug) {
    mdebug_push("parse_call() ...")
    on.exit(mdebug_pop())
  }
  
  if (is.symbol(call)) {
    ## Examples: lapply(...), map(...)
    namespace <- NULL
    fcn <- call
  } else if (is.call(call)) {
    stopifnot(length(call) == 3L)
    ## Examples: base::lapply(...), purrr::map(...)
    ## Not supported: base:::lapply(), baseenv()$lapply(...), ...
    op <- call[[1]]
    if (is.symbol(op)) {
      op_name <- as.character(op)
      if (op_name == "::") {
      } else if (op_name == ":::") {
        stop(sprintf("Do not know how to %s a private function: %s()", what, deparse(call)))
      } else {
        stop(sprintf("Do not know how to %s function %s()", what, deparse(call)))
      }
      namespace <- call[[2]]
      fcn <- call[[3]]
    } else {
      stop(sprintf("Do not know how to %s object of type %s: %s()", what, sQuote(typeof(op)), deparse(call)))
    }
  } else {
    stop(sprintf("Do not know how to %s expression: %s", what, as.character(call)))
  }

  ns_name <- as.character(namespace)
  fcn_name <- as.character(fcn)
  if (debug) {
    if (length(ns_name) == 1L) {
      msg <- sprintf("Function: %s::%s(...)", ns_name, fcn_name)
    } else {
      msg <- sprintf("Function: %s(...)", fcn_name)
    }
    mdebug(msg)
  }

  ## Does the function exist?
  if (is.null(namespace)) {
    if (debug) mdebug_push("Locate function ...")
    if (!exists(fcn_name, mode = "function", envir = envir, inherits = TRUE)) {
      name <- deparse(call)
      if (grepl("^%.*%$", name)) {
        stop(sprintf("Unknown infix operator: %s", name))
      } else {
        stop(sprintf("Unknown function: %s()", name))
      }
    }
    fcn <- get(fcn_name, mode = "function", envir = envir, inherits = TRUE)
    env <- environment(fcn)

    if (is.null(env) && is.primitive(fcn)) {
      env <- baseenv()
    } else if (inherits(fcn, "standardGeneric")) {
      env <- parent.env(env)
    }
    tenv <- env
    repeat {
      ns_name <- environmentName(tenv)
      if (nzchar(ns_name)) break
      tenv_next <- parent.env(tenv)
      if (identical(tenv_next, tenv)) {
        stop(sprintf("Internal error: infinite environment stack in parse_call(): %s", sQuote(ns_name)))
      }
      tenv <- tenv_next
    }
    
    stopifnot(nzchar(ns_name))
    if (debug) {
      mdebugf("Function located in: %s", sQuote(ns_name))
      mdebug_pop()
    }
  } else {
    ns <- getNamespace(ns_name)
    if (!exists(fcn_name, mode = "function", envir = ns, inherits = TRUE)) {
      stop(sprintf("No such function in %s: %s()", sQuote(ns_name), deparse(call)))
    }
    fcn <- get(fcn_name, mode = "function", envir = ns, inherits = TRUE)
  }

  list(fcn = fcn, ns_name = ns_name, fcn_name = fcn_name)
} ## parse_call()




#' Append arguments to a call.
#'
#' @param expr An \R call expression.
#'
#' @param \ldots,.args Named arguments to be appended to the call expression.
#'
#' @return
#' Return the expression with arguments appended.
#'
#' @examples
#' call <- quote(my_fcn(x, y))
#' print(call)
#' #> my_fcn(x, y)
#'
#' call2 <- append_call_arguments(call, z = 42, w = quote(1 + 2))
#' print(call2)
#' #> my_fcn(x, y, z = 42, w = 1 + 2)
#'
#' @noRd
append_call_arguments <- function(expr, ..., .args = list(...)) {
  ## Update 'EXPR'
  as.call(c(
    as.list(expr),
    .args
  ))
} ## append_call_arguments()

# lme4::allFit(...) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   lme4::allFit(..., parallel = "snow", ncpus = 2L, cl = cl)
# })
#
# lme4:::influence.merMod(...) [via stats::influence()] =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   stats::influence(..., parallel = "snow", ncpus = 2L, cl = cl)
# })
#
append_transpilers_for_lme4 <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'lme4' functions requires R (>= 4.4.0)", getRversion()))
  }

  transpilers <- make_package_transpilers("lme4", FUN = function(fcn, name) {
    if ("parallel" %in% names(formals(fcn))) {
      defaults <- list(label = sprintf("fz:lme4::%s", name))
      if (name %in% c("allFit", "influence.merMod")) {
        defaults$packages <- "lme4"
      }

      list(
        label = sprintf("lme4::%s() ~> lme4::%s(..., parallel = TRUE)", name, name),
        transpiler = make_futurize_for_makeClusterFuture(defaults = defaults, args = list(
          parallel = "snow",
          ncpus = 2L,   ## only used for test ncpus > 1
          cl = quote(cl)
        ))
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("lme4", "future")
}

descend_wrappers <- function(expr, envir = parent.frame(), unwrap, what = "unwrap", debug = FALSE) {
  ## Nothing to do?
  if (length(unwrap) == 0 || !is.call(expr)) return(1L)

  if (debug) {
    mdebug_push("descend_wrappers() ...")
    on.exit(mdebug_pop())
  }

  call <- expr[[1]]  ## e.g. {, local, lapply, ...
  if (debug) {
    mdebug("Call:")
    mprint(call)
  }

  call_info <- parse_call(call, envir = envir, what = what, debug = debug)
  fcn <- call_info[["fcn"]]
  fcn_name <- call_info[["fcn_name"]]
  ns_name <- call_info[["ns_name"]]

  for (wrapper in unwrap) {
    if (identical(fcn, wrapper)) {  ## a wrapper call?
      if (debug) {
        info <- switch(fcn_name,
          "{" = "{ ... }",
          "(" = "( ... )",
          sprintf("%s( ... )", fcn_name)
        )
        mdebugf("Wrapped in %s", info)
      }

      ## Which argument is the wrapped expression?
      index <- if (identical(fcn, base::`{`)) {
        length(expr)
      } else if (identical(fcn, base::with)) {
        ## with(data, expr) - the expression is the second argument (position 3)
        3L
      } else {
        2L
      }

      ## Safety check
      if (index > length(expr)) return(1L)

      return(c(index, descend_wrappers(expr[[index]], envir = envir, unwrap = unwrap, what = what, debug = debug)))
    } ## if (identical(fcn, wrapper)
  } ## for (wrapper in unwrap)

  return(1L)
} ## descend_wrappers()



#' @export
print.transpiled_call <- function(x, ...) {
  stopifnot(inherits(x, "call"))

  ...x_org... <- x

  ## Make sure attributes are displayed
  ## Whoa; this might be more convoluted than writing a custom
  ## print() method from scratch
  ...x... <- as.list(x)
  for (...kk... in seq_along(...x...)) {
    ...x...[[...kk...]] <- local({
      expr <- ...x...[[...kk...]]
      code <- c("substitute(", deparse(expr), ")")
      tryCatch({
        eval(parse(text = code), enclos = emptyenv())
      }, error = function(e) {
        ## Some deparsed expression cannot be parsed
        expr
      })
    })
  }
  x <- as.call(...x...)
  
  NextMethod()

  invisible(...x_org...)
}

# riskRegression::Score(...) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), local({
#   options(future.disposable = structure(OPTS, dispose = FALSE))
#   on.exit(options(future.disposable = NULL))
#   riskRegression::Score(..., parallel = "as.registered",
#                         progress.bar = NULL)
# }))
#
append_transpilers_for_riskRegression <- function() {
  template <- bquote_compile(
    with(doFuture::registerDoFuture(flavor = "%dofuture%"),
      local({
        options(future.disposable = structure(.(OPTS), dispose = FALSE))
        on.exit(options(future.disposable = NULL))
        .(EXPR)
      })
    )
  )

  template2 <- bquote_compile(function(expr, options = NULL) {
    defaults <- list(
      seed = TRUE,
      label = sprintf("fz:riskRegression::%s-%%d", .(NAME))
    )
    expr <- append_call_arguments(expr,
      parallel = "as.registered"
    )
    ## NOTE: We cannot use append_call_arguments() for progress.bar,
    ## because c() drops NULL elements
    expr["progress.bar"] <- list(NULL)

    opts <- make_options_for_doFuture(options, defaults = defaults, wrap = FALSE)
    bquote_apply(template,
      OPTS = opts,
      EXPR = expr
    )
  })

  transpilers <- make_package_transpilers("riskRegression", FUN = function(fcn, name) {
    if ("parallel" %in% names(formals(fcn))) {
      transpiler <- eval(bquote_apply(template2, NAME = name))
      list(
        label = sprintf("riskRegression::%s() ~> riskRegression::%s(..., parallel = \"as.registered\")", name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("riskRegression", "doFuture")
}

# pvclust::pvclust(..., parallel = FALSE) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   pvclust::pvclust(..., parallel = cl)
# })
#
append_transpilers_for_pvclust <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'pvclust' functions requires R (>= 4.4.0)", getRversion()))
  }

  template_ignore_clusterEvalQ <- bquote_compile(
    local({
      cl <- do.call(.(CALL), args = .(OPTS))
      oopts <- options(future.ClusterFuture.clusterEvalQ = "ignore")
      on.exit(options(oopts))
      .(EXPR)
    })
  )

  transpilers <- make_package_transpilers("pvclust", FUN = function(fcn, name) {
    if (name == "pvclust") {
      base_transpiler <- make_futurize_for_makeClusterFuture(
        template = template_ignore_clusterEvalQ,
        args = list(
          parallel = quote(cl)
        ), defaults = list(
          label = "fz:pvclust::pvclust",
          seed = TRUE
        )
      )

      transpiler <- eval(bquote(function(expr, options = NULL) {
        expr <- match.call(definition = .(fcn), call = expr)
        expr$parallel <- NULL
        .(base_transpiler)(expr, options = options)
      }))

      list(
        label = "pvclust::pvclust() ~> pvclust::pvclust(..., parallel = cl)",
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("pvclust", "future")
}

# kernelshap::kernelshap(...) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
#   options(future.disposable = <future arguments>)
#   kernelshap::kernelshap(..., parallel = TRUE)
# })
#
append_transpilers_for_kernelshap <- function() {
  transpilers <- make_package_transpilers("kernelshap", FUN = function(fcn, name) {
    if (name %in% c("kernelshap", "permshap")) {
      defaults <- list(
        seed = TRUE,
        label = sprintf("fz:kernelshap::%s-%%d", name)
      )
      list(
        label = sprintf("kernelshap::%s() ~> kernelshap::%s(..., parallel = TRUE)", name, name),
        transpiler = make_futurize_for_doFuture(defaults = defaults, args = list(parallel = TRUE))
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("kernelshap", "doFuture")
}

#! /usr/bin/env Rscript
## This runs testme test script inst/testme/test-future.apply-options,nested.R
## Don't edit - it was autogenerated by inst/testme/deploy.R
futurize:::testme("future.apply-options,nested")

# fwb::fwb(...) =>
#
# local({
#   fwb::fwb(..., cl = "future")
# })
#
append_transpilers_for_fwb <- function() {
  template <- bquote_compile(
   local({
      ## This will be automatically consumed and removed by 'future.apply',
      ## which pblapply() calls, which fwb() calls.
      options(future.disposable = structure(.(OPTS), dispose = FALSE))
      ...future_old_rngkind <- RNGkind("L'Ecuyer-CMRG")
      on.exit({
        do.call(RNGkind, args = as.list(...future_old_rngkind))
        options(future.disposable = NULL)
      })
      .(EXPR)
    })
  )
  
  template2 <- bquote_compile(function(expr, options = NULL) {
    ## Specified future.* arguments
    specified <- attr(options, "specified")

    names(options) <- sub("chunk_size", "chunk.size", names(options), fixed = TRUE)
    
    ## pbapply does not handle stdout, messages, and warnings, so disable
    ## those by default
    if (!"stdout" %in% specified) {
      options$stdout <- FALSE
    }
    if (!"conditions" %in% specified) {
      options$conditions <- structure(options$conditions,
                               exclude = c("message", "warning"))
    }

    defaults <- list(
      seed = TRUE,
      label = sprintf("fz:fwb::%s-%%d", .(NAME))
    )
    names <- setdiff(names(defaults), attr(options, "specified"))
    for (name in names) {
      if (name == "packages") {
        options[[name]] <- c(options[[name]], defaults[[name]])
      } else {
        options[[name]] <- defaults[[name]]
      }
    }
    
    expr <- append_call_arguments(expr,
      cl = "future"
    )
    
    bquote_apply(template,
      OPTS = options,
      EXPR = expr
    )
  })

  transpilers <- make_package_transpilers("fwb", FUN = function(fcn, name) {
    if ("cl" %in% names(formals(fcn))) {
      transpiler <- eval(bquote_apply(template2, NAME = name))

      list(
        label = sprintf("fwb::%s() ~> fwb::%s(..., cl = \"future\")", name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("fwb", "future")
}

#' @noRd
make_options_for_makeClusterFuture <- local({
  get_defaults <- function(fcn) {
    defaults <- names(formals(fcn))
    excl <- c("expr", "substitute", "envir", "earlySignal", "gc", "...")
    setdiff(defaults, excl)
  }
  
  defaults_base <- NULL

  function(options, defaults = NULL) {
    ## Nothing to do?
    if (length(options) == 0L && length(defaults) == 0L) return(options)

    if (is.null(defaults_base)) {
      defaults_base <<- get_defaults(future::future)
    }

    if (length(defaults) > 0) {
      names <- setdiff(names(defaults), attr(options, "specified"))
      for (name in names) {
        if (name == "packages") {
          options[[name]] <- c(options[[name]], defaults[[name]])
        } else {
          options[[name]] <- defaults[[name]]
        }
      }
    }

    options <- options[names(options) %in% defaults_base]

    options
  }
})

# Rsamtools::countBam(bamViews, ...) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
#   options(future.disposable = <future arguments>)
#   Rsamtools::countBam(bamViews, ..., BPPARAM = BiocParallel::DoparParam())
# })
#
append_transpilers_for_Rsamtools <- function() {
  ## These Rsamtools functions accept BPPARAM via '...' (S4 generics).
  ## Parallelization occurs when the input is a BamViews object,
  ## which distributes work across BAM files via bplapply().
  names <- c(
    "countBam",
    "scanBam"
  )

  ns <- getNamespace("Rsamtools")
  transpilers <- list()
  for (name in names) {
    if (exists(name, mode = "function", envir = ns, inherits = TRUE)) {
      args <- list(BPPARAM = quote(BiocParallel::DoparParam()))
      transpilers[[name]] <- list(
        label = sprintf("Rsamtools::%s() ~> Rsamtools::%s(..., BPPARAM = BiocParallel::DoparParam())", name, name),
        transpiler = make_futurize_for_doFuture(
          args = args,
          defaults = list(label = sprintf("fz:Rsamtools::%s-%%d", name))
        )
      )
    }
  }

  transpilers <- list(transpilers)
  names(transpilers) <- "Rsamtools"

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("Rsamtools", "BiocParallel", "doFuture")
}

#' Create a futurize transpiler based on makeClusterFuture()
#' 
#' @param expr An \R call expression.
#'
#' @param args Named arguments to be appended to the call expression.
#'
#' @param template An optional custom template.
#'
#' @returns
#' A transpiler function.
#'
#' @noRd
make_futurize_for_makeClusterFuture <- function(defaults = list(), args = list(), template = NULL) {
  if (is.null(template)) {
    template <- bquote_compile(
      local({
        cl <- do.call(.(CALL), args = .(OPTS))
        oopts <- options(future.ClusterFuture.clusterEvalQ = "error")
        on.exit(options(oopts))
        .(EXPR)
      })
    )
  }

  ## To please 'R CMD check' on R (< 4.4.0), where
  ## future::makeClusterFuture() is not available
  call <- as.call(lapply(c("::", "future", "makeClusterFuture"), as.name))

  function(expr, options = NULL) {
    ## If specified, assert that 'seed' is logical; numeric not supported
    if ("seed" %in% attr(options, "specified")) {
      seed_value <- options[["seed"]]
      if (is.numeric(seed_value)) {
        stop(sprintf("futurize() option 'seed' must be logical for this function call, not %s: %s", class(seed_value)[1], deparse(expr)))
      }
    }
    expr <- append_call_arguments(expr, .args = args)
    opts <- make_options_for_makeClusterFuture(options, defaults = defaults)
    bquote_apply(template,
      CALL = call,
      OPTS = opts,
      EXPR = expr
    )
  }
} ## make_futurize_for_makeClusterFuture()

# sandwich::vcovBS(...) =>
#
# sandwich::vcovBS(..., applyfun = function(...) {
#   future.apply::future_lapply(..., future.seed = TRUE)
# })
#
append_transpilers_for_sandwich <- function() {
  transpilers <- make_package_transpilers("sandwich", FUN = function(fcn, name) {
    if ("applyfun" %in% names(formals(fcn)) || name %in% c("vcovBS", "vcovJK")) {
      list(
        label = sprintf("sandwich::%s() ~> sandwich::%s()", name, name),
        transpiler = make_futurize_for_future.apply(
          defaults = list(
            future.seed = TRUE,
            future.label = sprintf("fz:sandwich::%s-%%d", name)
          ),
          args = list(
            applyfun = function(...) {
              future.apply::future_lapply(...)
            }
          )
        ) ## make_futurize_for_future.apply()
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("sandwich", "future.apply")
}

#' Finds registered S3 method for S3 generic function and object
#'
#' @param fcn,fcn_name An S3 generic function and its name.
#'
#' @param call The S3 function call, which includes the dispatch object.
#'
#' @param envir The environment in which the dispatch object should be
#' resolved.
#'
#' @param debug If TRUE, debug output is given.
#'
#' @return
#' Returns a named list of elements `package` and `name` if found,
#' otherwise NULL.
#'
#' @noRd
#' @importFrom utils getS3method
find_s3_method <- function(fcn, fcn_name, call, envir, debug = FALSE) {
  ## Get the name of the first argument, which is the S3 dispatch argument
  fmls <- formals(fcn)
  if (length(fmls) == 0L) return(NULL)

  ## FIXME: Here we assume we're dispatching on the first argument
  dispatch_arg_name <- names(fmls)[[1L]]
  if (dispatch_arg_name == "...") return(NULL) ## FIXME: Skip for now

  ## Use match.call() to correctly handle named and reordered arguments
  matched_call <- tryCatch(
    match.call(fcn, call = call),
    error = function(e) NULL
  )
  if (is.null(matched_call)) return(NULL)

  dispatch_expr <- matched_call[[dispatch_arg_name]]
  if (is.null(dispatch_expr)) return(NULL)
  if (!is.symbol(dispatch_expr) && !is.call(dispatch_expr)) return(NULL)

  ## Evaluate the dispatch argument to get its class
  dispatch_obj <- tryCatch(
    eval(dispatch_expr, envir = envir),
    error = function(e) NULL
  )
  if (is.null(dispatch_obj)) return(NULL)

  ## Use .class2() to get the full S3 dispatch chain, which includes inherited
  ## classes from S4/R5 hierarchies not visible in class() alone.
  ## Example: class(lmerMod_obj) = "lmerMod", but .class2() = c("lmerMod", "merMod")
  dispatch_classes <- .class2(dispatch_obj)

  ## Walk the class hierarchy to find a dispatched S3 method
  method <- NULL
  dispatch_class <- NULL
  for (cls in dispatch_classes) {
    m <- getS3method(fcn_name, cls, optional = TRUE)
    if (!is.null(m)) {
      method <- m
      dispatch_class <- cls
      break
    }
  }
  if (is.null(method)) return(NULL)

  ## Determine the package the method lives in
  method_env <- environment(method)
  if (is.null(method_env)) return(NULL)
  method_pkg <- environmentName(topenv(method_env))

  method_name <- paste0(fcn_name, ".", dispatch_class)

  if (debug) {
    mdebugf("S3 generic %s() dispatches to %s::%s() for class %s",
            fcn_name, method_pkg, method_name, sQuote(dispatch_class))
  }

  list(package = method_pkg, name = method_name)
} ## find_s3_method()

## This function registered functions that adds transpilers for specific
## packages, without loading those packages.
register_all_transpilers <- function() {
  ## Map-reduce packages (base-R)
  transpilers_for_package("futurize::add-on", package = "base",         append_transpilers_for_future.apply)
  transpilers_for_package("futurize::add-on", package = "stats",        append_transpilers_for_future.apply)

  ## Map-reduce packages (pbapply)
  transpilers_for_package("futurize::add-on", package = "pbapply",      append_transpilers_for_pbapply)

  ## Map-reduce packages (Tidyverse)
  transpilers_for_package("futurize::add-on", package = "purrr",        append_transpilers_for_furrr)
  transpilers_for_package("futurize::add-on", package = "crossmap",     append_transpilers_for_crossmap)

  ## Map-reduce packages (foreach)
  transpilers_for_package("futurize::add-on", package = "foreach",      append_transpilers_for_doFuture)
  transpilers_for_package("futurize::add-on", package = "plyr",         append_transpilers_for_plyr)
  transpilers_for_package("futurize::add-on", package = "BiocParallel", append_transpilers_for_BiocParallel)

  ## Domain-specific "recommended" packages
  transpilers_for_package("futurize::add-on", package = "boot",         append_transpilers_for_boot)
  transpilers_for_package("futurize::add-on", package = "mgcv",         append_transpilers_for_mgcv)

  ## Domain-specific CRAN packages
  transpilers_for_package("futurize::add-on", package = "caret",          append_transpilers_for_caret)
  transpilers_for_package("futurize::add-on", package = "DiceKriging",    append_transpilers_for_DiceKriging)
  transpilers_for_package("futurize::add-on", package = "ez",             append_transpilers_for_ez)
  transpilers_for_package("futurize::add-on", package = "fwb",            append_transpilers_for_fwb)
  transpilers_for_package("futurize::add-on", package = "gamlss",         append_transpilers_for_gamlss)
  transpilers_for_package("futurize::add-on", package = "glmnet",         append_transpilers_for_glmnet)
  transpilers_for_package("futurize::add-on", package = "kernelshap",     append_transpilers_for_kernelshap)
  transpilers_for_package("futurize::add-on", package = "glmmTMB",        append_transpilers_for_glmmTMB)
  transpilers_for_package("futurize::add-on", package = "lme4",           append_transpilers_for_lme4)
  transpilers_for_package("futurize::add-on", package = "metafor",        append_transpilers_for_metafor)
  transpilers_for_package("futurize::add-on", package = "modelsummary",   append_transpilers_for_modelsummary)
  transpilers_for_package("futurize::add-on", package = "partykit",       append_transpilers_for_partykit)
  transpilers_for_package("futurize::add-on", package = "parameters",     append_transpilers_for_parameters)
  transpilers_for_package("futurize::add-on", package = "pls",            append_transpilers_for_pls)
  transpilers_for_package("futurize::add-on", package = "pvclust",        append_transpilers_for_pvclust)
  transpilers_for_package("futurize::add-on", package = "riskRegression", append_transpilers_for_riskRegression)
  transpilers_for_package("futurize::add-on", package = "rugarch",        append_transpilers_for_rugarch)
  transpilers_for_package("futurize::add-on", package = "sandwich",       append_transpilers_for_sandwich)
  transpilers_for_package("futurize::add-on", package = "seriation",      append_transpilers_for_seriation)
  transpilers_for_package("futurize::add-on", package = "Sim.DiffProc",   append_transpilers_for_Sim.DiffProc)
  transpilers_for_package("futurize::add-on", package = "SimDesign",      append_transpilers_for_SimDesign)
  transpilers_for_package("futurize::add-on", package = "shapr",          append_transpilers_for_shapr)
  transpilers_for_package("futurize::add-on", package = "stars",          append_transpilers_for_stars)
  transpilers_for_package("futurize::add-on", package = "strucchange",    append_transpilers_for_strucchange)
  transpilers_for_package("futurize::add-on", package = "SuperLearner",   append_transpilers_for_SuperLearner)
  transpilers_for_package("futurize::add-on", package = "tm",             append_transpilers_for_tm)
  transpilers_for_package("futurize::add-on", package = "TSP",            append_transpilers_for_TSP)
  transpilers_for_package("futurize::add-on", package = "vegan",          append_transpilers_for_vegan)

  ## Domain-specific Bioconductor packages
  transpilers_for_package("futurize::add-on", package = "DESeq2",               append_transpilers_for_DESeq2)
  transpilers_for_package("futurize::add-on", package = "fgsea",                append_transpilers_for_fgsea)
  transpilers_for_package("futurize::add-on", package = "GenomicAlignments",    append_transpilers_for_GenomicAlignments)
  transpilers_for_package("futurize::add-on", package = "Rsamtools",            append_transpilers_for_Rsamtools)
  transpilers_for_package("futurize::add-on", package = "GSVA",                 append_transpilers_for_GSVA)
  transpilers_for_package("futurize::add-on", package = "scater",               append_transpilers_for_scater)
  transpilers_for_package("futurize::add-on", package = "scuttle",              append_transpilers_for_scuttle)
  transpilers_for_package("futurize::add-on", package = "SingleCellExperiment", append_transpilers_for_SingleCellExperiment)
  transpilers_for_package("futurize::add-on", package = "sva",                  append_transpilers_for_sva)
} ## register_all_transpilers()

#' Create a futurize transpiler based on 'future.apply'
#' 
#' @inheritParams make_futurize_for_makeClusterFuture
#'
#' @param fcn The \pkg{future.apply} function to generate options for.
#'
#' @returns
#' A transpiler function.
#'
#' @noRd
make_futurize_for_future.apply <- function(defaults = list(), args = list(), template = NULL, fcn = future.apply::future_lapply) {
  if (is.null(template)) {
    template <- bquote_compile(
     local({
        ## This will be automatically consumed and removed by 'future.apply'
        options(future.disposable = structure(.(OPTS), dispose = FALSE))
        on.exit(options(future.disposable = NULL))
        .(EXPR)
      })
    )
  }
  
  function(expr, options = NULL) {
    expr <- append_call_arguments(expr, .args = args)
    opts <- make_options_for_future.apply(options, defaults = defaults, fcn = fcn)
    names(opts) <- sub("^future[.]", "", names(opts))
    
    bquote_apply(template,
      OPTS = opts,
      EXPR = expr
    )
  }
} ## make_futurize_for_future.apply()

# SuperLearner::CV.SuperLearner(...) =>
#
# local({
#   cl <- future::makeClusterFuture(<future arguments>)
#   SuperLearner::CV.SuperLearner(..., parallel = cl)
# })
#
append_transpilers_for_SuperLearner <- function() {
  if (getRversion() < "4.4.0") {
    stop(sprintf("You are running R %s, but futurization of 'SuperLearner' functions requires R (>= 4.4.0)", getRversion()))
  }

  transpilers <- make_package_transpilers("SuperLearner", FUN = function(fcn, name) {
    if ("parallel" %in% names(formals(fcn)) && name == "CV.SuperLearner") {
      transpiler <- make_futurize_for_makeClusterFuture(
        args = list(
          parallel = quote(cl)
        ), defaults = list(
          packages = "SuperLearner",
          seed = TRUE,
          label = sprintf("fz:SuperLearner::%s", name)
        )
      )

      list(
        label = sprintf("SuperLearner::%s() ~> SuperLearner::%s(..., parallel = cl)", name, name),
        transpiler = transpiler
      )
    }
  })

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("SuperLearner", "future")
}

# purrr::map(...) =>
#
# furrr::future_map(..., .options = <future.* arguments>)
#
append_transpilers_for_furrr <- function() {
  append_transpilers("futurize::add-on", make_addon_transpilers(
    "purrr", "furrr", make_options = make_options_for_furrr)
  )

  ## Return required packages
  c("purrr", "furrr")
}

# scater::runPCA(sce) =>
#
# with(doFuture::registerDoFuture(flavor = "%dofuture%"), {
#   options(future.disposable = <future arguments>)
#   scater::runPCA(sce, BPPARAM = BiocParallel::DoparParam())
# })
#
append_transpilers_for_scater <- function() {
  ## These scater functions accept BPPARAM via '...' (S4 generics) or
  ## directly in their formals
  names <- c(
    "calculatePCA", "calculateTSNE", "calculateUMAP",
    "runPCA", "runTSNE", "runUMAP",
    "runColDataPCA",
    "nexprs",
    "getVarianceExplained",
    "plotRLE"
  )

  ns <- getNamespace("scater")
  transpilers <- list()
  for (name in names) {
    if (exists(name, mode = "function", envir = ns, inherits = TRUE)) {
      args <- list(BPPARAM = quote(BiocParallel::DoparParam()))
      transpilers[[name]] <- list(
        label = sprintf("scater::%s() ~> scater::%s(..., BPPARAM = BiocParallel::DoparParam())", name, name),
        transpiler = make_futurize_for_doFuture(
          args = args,
          defaults = list(label = sprintf("fz:scater::%s-%%d", name))
        )
      )
    }
  }

  transpilers <- list(transpilers)
  names(transpilers) <- "scater"

  append_transpilers("futurize::add-on", transpilers)

  ## Return required packages
  c("scater", "BiocParallel", "doFuture")
}

#' Finds S4 method package for S4 generic function and object
#'
#' @param fcn,fcn_name An S4 generic function and its name.
#'
#' @param call The S4 function call, which includes the dispatch object.
#'
#' @param envir The environment in which the dispatch object should be
#' resolved.
#'
#' @param debug If TRUE, debug output is given.
#'
#' @return
#' Returns a named list of elements `package` and `name` if found,
#' otherwise NULL.
#'
#' @noRd
find_s4_method <- function(fcn, fcn_name, call, envir, debug = FALSE) {
  if (!methods::isGeneric(fcn_name)) return(NULL)

  ## Get the name of the first argument (dispatch argument)
  fmls <- formals(fcn)
  if (length(fmls) == 0L) return(NULL)

  ## FIXME: Here we assume we're dispatching on the first argument
  dispatch_arg_name <- names(fmls)[[1L]]
  if (dispatch_arg_name == "...") return(NULL) ## FIXME: Skip for now

  ## Use match.call() to correctly handle named and reordered arguments
  matched_call <- tryCatch(
    match.call(fcn, call = call),
    error = function(e) NULL
  )
  if (is.null(matched_call)) return(NULL)

  dispatch_expr <- matched_call[[dispatch_arg_name]]
  if (is.null(dispatch_expr)) return(NULL)
  if (!is.symbol(dispatch_expr) && !is.call(dispatch_expr)) return(NULL)

  ## Evaluate the dispatch argument to get its class
  dispatch_obj <- tryCatch(
    eval(dispatch_expr, envir = envir),
    error = function(e) NULL
  )
  if (is.null(dispatch_obj)) return(NULL)

  dispatch_class <- class(dispatch_obj)[[1L]]

  ## Find the S4 method
  method <- tryCatch(
    methods::selectMethod(fcn_name, signature = dispatch_class),
    error = function(e) NULL
  )
  if (is.null(method)) return(NULL)

  ## Determine the package the method is defined in
  method_env <- environment(method)
  if (is.null(method_env)) return(NULL)
  method_pkg <- environmentName(topenv(method_env))

  if (debug) {
    mdebugf("S4 generic %s() dispatches to %s::%s() for class %s",
            fcn_name, method_pkg, fcn_name, sQuote(dispatch_class))
  }

  list(package = method_pkg, name = fcn_name)
} ## find_s4_method()

#' Create a futurize transpiler based on %dofuture%
#'
#' @inheritParams make_futurize_for_makeClusterFuture
#'
#' @returns
#' A transpiler function.
#'
#' @noRd
make_futurize_for_doFuture <- function(defaults = list(), args = list(), template = NULL) {
  if (is.null(template)) {
    template <- bquote_compile(
      with(doFuture::registerDoFuture(flavor = "%dofuture%"),
        local({
          options(future.disposable = structure(.(OPTS), dispose = FALSE))
          on.exit(options(future.disposable = NULL))
          .(EXPR)
        })
      )
    )
  }

  function(expr, options = NULL) {
    expr <- append_call_arguments(expr,
      .args = args
    )
   opts <- make_options_for_doFuture(options, defaults = defaults, wrap = FALSE)
    
    ## Update 'OPTS'
    bquote_apply(template,
      OPTS = opts,
      EXPR = expr
    )
  }
} ## make_futurize_for_doFuture()

<!--
%\VignetteIndexEntry{Parallelize 'ez' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{ez}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(ez)

data(ANT)
rt <- ezBoot(
  data = ANT,
  dv = rt,
  wid = subnum,
  within = .(cue, flank),
  between = group,
  iterations = 1e3
) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
**[ez]** functions such as `ezBoot()`, `ezPerm()`, and `ezPlot2()`.
The functions `ezBoot()`, `ezPerm()`, and `ezPlot2()` support parallel
evaluation via the `parallel` argument. By piping to `futurize()`, you
can leverage any future-based parallel backend for these computations.


## Example: Bootstrap resampling

The `ezBoot()` function computes bootstrap resampled predictions for
each cell in an experimental design. We can parallelize this as:

```r
library(futurize)
plan(multisession)
library(ez)

data(ANT)
rt <- ezBoot(
  data = ANT,
  dv = rt,
  wid = subnum,
  within = .(cue, flank),
  between = group,
  iterations = 1e3
) |> futurize()
```

This will distribute the bootstrap iterations across the available
parallel workers.


## Example: Permutation testing

The `ezPerm()` function performs a non-parametric factorial
permutation test, and can be parallelized as:

```r
library(futurize)
plan(multisession)
library(ez)
library(plyr)

data(ANT)
cell_stats <- ddply(
  .data = ANT,
  .variables = .(subnum, group, cue, flank),
  .fun = function(x) {
    data.frame(mrt = mean(x$rt[x$error == 0]))
  }
)
gmrt <- ddply(
  .data = cell_stats,
  .variables = .(subnum, group),
  .fun = function(x) {
    data.frame(mrt = mean(x$mrt))
  }
)

mean_rt_perm <- ezPerm(
  data = gmrt,
  dv = mrt,
  wid = subnum,
  between = group,
  perms = 1e3
) |> futurize()
```


# Supported Functions

The following **ez** functions are supported by `futurize()`:

* `ezBoot()` with `seed = TRUE` as the default
* `ezPerm()` with `seed = TRUE` as the default
* `ezPlot2()`


[ez]: https://cran.r-project.org/package=ez
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'purrr' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{purrr}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/purrr-logo.webp" alt="The 'purrr' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(purrr)

slow_fcn <- function(x) {
  message("x = ", x)
  Sys.sleep(0.1)  # emulate work
  x^2
}

xs <- 1:10
ys <- xs |> map(slow_fcn) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
**[purrr]** functions such as `map()`, `map_dbl()`, and `walk()`.

The **purrr** `map()` function is commonly used to apply a function to
the elements of a vector or a list. For example,

```r
library(purrr)
xs <- 1:1000
ys <- map(xs, slow_fcn)
```

or equivalently using pipe syntax

```r
xs <- 1:1000
ys <- xs |> map(slow_fcn)
```

Here `map()` evaluates sequentially, but we can easily make it
evaluate in parallel, by using:

```r
library(purrr)

library(futurize)
plan(multisession) ## parallelize on local machine

xs <- 1:1000
ys <- xs |> map(slow_fcn) |> futurize()
#> x = 1
#> x = 2
#> x = 3
#> ...
#> x = 10
```

Note how messages produced on parallel workers are relayed as-is back
to the main R session as they complete. Not only messages, but also
warnings and other types of conditions are relayed back as-is.
Likewise, standard output produced by `cat()`, `print()`, `str()`, and
so on is relayed in the same way. This is a unique feature of
Futureverse - other parallel frameworks in R, such as **parallel**,
**foreach** with **doParallel**, and **BiocParallel**, silently drop
standard output, messages, and warnings produced on workers. With
**futurize**, your code behaves the same whether it runs sequentially
or in parallel: nothing is lost in translation.

The built-in `multisession` backend parallelizes on your local
computer and it works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```

Another example is:

```r
library(purrr)
library(futurize)
plan(future.mirai::mirai_multisession)

ys <- 1:10 |>
        map(rnorm, n = 10) |> futurize(seed = TRUE) |>
        map_dbl(mean) |> futurize()
```


# Supported Functions

The `futurize()` function supports parallelization of the following **purrr** functions:

 * `map()`, `map_chr()`, `map_dbl()`, `map_int()`, `map_lgl()`, `map_dfr()`, `map_dfc()`, `walk()`
 * `map2()`, `map2_chr()`, `map2_dbl()`, `map2_int()`, `map2_lgl()`, `map2_dfr()`, `map2_dfc()`, `walk2()`
 * `pmap()`, `pmap_chr()`, `pmap_dbl()`, `pmap_int()`, `pmap_lgl()`, `pmap_dfr()`, `pmap_dfc()`, `pwalk()`
 * `imap()`, `imap_chr()`, `imap_dbl()`, `imap_int()`, `imap_lgl()`, `imap_dfr()`, `imap_dfc()`, `iwalk()`
 * `modify()`, `modify_if()`, `modify_at()`
 * `map_if()`, `map_at()`


# Progress Reporting via progressify

For progress reporting, please see the **[progressify]** package. It is
specially designed to work with the Futureverse ecosystem and provide
progress updates from parallelized computations in a near-live
fashion. See the `vignette("futurize-11-apply", package = "futurize")`
for more details and an example.


[progressify]: https://progressify.futureverse.org/
[purrr]: https://cran.r-project.org/package=purrr
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'BiocParallel' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{Bioconductor}
%\VignetteKeyword{BiocParallel}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/bioconductor-BiocParallel-logo.webp" alt="The Bioconductor 'BiocParallel' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

You can use **futurize** to make **[BiocParallel]** functions
parallelize via any of the [parallel backends] supported by
Futureverse, e.g.

```r
library(futurize)
plan(multisession)
library(BiocParallel)

slow_fcn <- function(x) {
  message("x = ", x)
  Sys.sleep(0.1)  # emulate work
  x^2
}

xs <- 1:10
ys <- bplapply(xs, slow_fcn) |> futurize()
```

# Introduction

This vignette demonstrates how to use this approach to parallelize
functions such as `bplapply()`, `bpmapply()`, and `bpvec()` in the
**BiocParallel** package. For example, consider the `bplapply()`
function. It works like base-R `lapply()`, but uses the
**BiocParallel** framework to process the tasks concurrently. It is
commonly used something like:

```r
library(BiocParallel)
xs <- 1:1000
ys <- bplapply(xs, slow_fcn)
```

The parallel backend is controlled by the `BiocParallel::register()`,
similar to how we use `future::plan()` in Futureverse. We can use
the **futurize** package to tell **BiocParallel** to hand over the
orchestration of parallel tasks to Futureverse. All we need to do is
to pass the expression to `futurize()` as in:

```r
library(BiocParallel)

library(futurize)
plan(multisession) ## parallelize on local machine

xs <- 1:1000
ys <- bplapply(xs, slow_fcn) |> futurize()
#> x = 1
#> x = 2
#> x = 3
#> ...
#> x = 10
```

Note how messages produced on parallel workers are relayed as-is back
to the main R session as they complete. Not only messages, but also
warnings and other types of conditions are relayed back as-is.
Likewise, standard output produced by `cat()`, `print()`, `str()`, and
so on is relayed in the same way. This is a unique feature of
Futureverse - other parallel frameworks in R, such as **parallel**,
**foreach** with **doParallel**, and **BiocParallel**, silently drop
standard output, messages, and warnings produced on workers. With
**futurize**, your code behaves the same whether it runs sequentially
or in parallel: nothing is lost in translation.

The built-in `multisession` backend parallelizes on your local
computer and it works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The `futurize()` function supports parallelization of all
**BiocParallel** functions that take argument
`BPPARAM`. Specifically,

 * `bplapply()` and `.bplapply_impl()`
 * `bpmapply()` and `.bpmapply_impl()`
 * `bpvec()`
 * `bpaggregate()`

The following functions are currently not supported:

 * `bpiterate()` - technically supported, but because this
   function does not support using `DoparParam()` with it, it
   effectively does not work with `futurize()`
 * `bpvectorize()`
 * `register()`


# Bioconductor packages using BiocParallel

Most Bioconductor packages that support parallelization do so via
**BiocParallel** internally. These packages typically expose a
`BPPARAM` argument in their functions, which controls the parallel
backend used. For example, `DESeq2::DESeq()` has a `BPPARAM` argument
that defaults to `BiocParallel::bpparam()`, which corresponds to the
currently registered **BiocParallel** backend. This means that, in
order to parallelize such a function, one can call
`BiocParallel::register()` to set a parallel backend, and then the
function will use it automatically.

However, not all packages default to `bpparam()`. For example,
`sva::ComBat()` defaults to `bpparam("SerialParam")`, which means it
always runs sequentially unless you explicitly pass a parallel
`BPPARAM` argument. Because of this, one cannot count on `bpparam()`
being the default everywhere - some functions require an explicit
`BPPARAM` to parallelize. With **futurize**, this is handled
automatically: `futurize()` injects the appropriate `BPPARAM` argument
regardless of what the default is, so that the parallel execution is
performed via the Futureverse, where the parallel backend is
controlled by `future::plan()`.


# Progress Reporting via progressify

For progress reporting, please see the **[progressify]** package. It is
specially designed to work with the Futureverse ecosystem and provide
progress updates from parallelized computations in a near-live
fashion. See the `vignette("futurize-11-apply", package = "futurize")`
for more details and an example.


[progressify]: https://progressify.futureverse.org/
[BiocParallel]: https://bioconductor.org/packages/BiocParallel/
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'Sim.DiffProc' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{Sim.DiffProc}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!

# TL;DR

```r
library(futurize)
plan(multisession)
library(Sim.DiffProc)

# Define 1D SDE model
f <- expression(0)
g <- expression(1)
mod1d <- snssde1d(drift = f, diffusion = g, x0 = 1, M = 10, N = 100)
stat <- function(x, ...) mean(x)

res <- MCM.sde(mod1d, statistic = stat, R = 10, time = 0.5) |> futurize()
```

# Introduction

This vignette demonstrates how to use this approach to parallelize
**[Sim.DiffProc]** functions such as `MCM.sde()`.

The **[Sim.DiffProc]** package provides a comprehensive framework for
numerical simulation and inference of Stochastic Differential
Equations (SDEs) in R. Because Monte Carlo simulation is highly
iterative, running multiple replications in parallel can significantly
reduce execution times.

## Example: Monte Carlo simulation of SDEs

The `MCM.sde()` function performs Monte Carlo simulations for SDEs.
For example:

```r
library(Sim.DiffProc)

f <- expression(0)
g <- expression(1)
mod1d <- snssde1d(drift = f, diffusion = g, x0 = 1, M = 10, N = 100)
stat <- function(x, ...) mean(x)

res <- MCM.sde(mod1d, statistic = stat, R = 10, time = 0.5)
```

Here `MCM.sde()` evaluates sequentially. To run in parallel,
pipe to `futurize()`:

```r
library(futurize)
library(Sim.DiffProc)

res <- MCM.sde(mod1d, statistic = stat, R = 10, time = 0.5) |> futurize()
```

This will distribute the Monte Carlo replications across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other parallel
backends] to choose from, including alternatives to parallelize
locally as well as distributed across remote machines, e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```

# Supported Functions

The following **Sim.DiffProc** functions are supported by `futurize()`:

* `MCM.sde()` with `seed = TRUE` as the default

[Sim.DiffProc]: https://cran.r-project.org/package=Sim.DiffProc
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'SuperLearner' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{SuperLearner}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!

# TL;DR

```r
library(futurize)
plan(multisession)
library(SuperLearner)

res <- CV.SuperLearner(Y = Y, X = X, SL.library = SL.library) |> futurize()
```

# Introduction

This vignette demonstrates how to use this approach to parallelize
**[SuperLearner]** functions such as `CV.SuperLearner()`.

The **[SuperLearner]** package provides a framework for ensemble
machine learning in R. The algorithm utilizes V-fold cross-validation
to combine multiple prediction algorithms into a single ensemble
predictor. Since cross-validation involves training many models
independently, it is a perfect candidate for parallelization.

## Example: Cross-Validated Super Learner

The `CV.SuperLearner()` function evaluates the cross-validated risk of
the Super Learner ensemble.  For example:

```r
library(SuperLearner)

n <- 100
p <- 5
X <- as.data.frame(matrix(rnorm(n * p), n, p))
Y <- X[, 1] + X[, 2] + rnorm(n)
SL.library <- c("SL.glm", "SL.mean")

res <- CV.SuperLearner(Y = Y, X = X, V = 10, SL.library = SL.library)
```

Here `CV.SuperLearner()` evaluates sequentially. To run in parallel,
pipe to `futurize()`:
    
```r
library(futurize)
library(SuperLearner)

res <- CV.SuperLearner(Y = Y, X = X, V = 10, SL.library = SL.library) |> futurize()
```

This will distribute the cross-validation fold evaluations across the
available parallel workers, given that we have set up parallel
workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other parallel
backends] to choose from, including alternatives to parallelize
locally as well as distributed across remote machines, e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```

# Supported Functions

The following **SuperLearner** functions are supported by `futurize()`:

* `CV.SuperLearner()` with `seed = TRUE` as the default

[SuperLearner]: https://cran.r-project.org/package=SuperLearner
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'modelsummary' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{modelsummary}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/modelsummary-logo.webp" alt="The 'modelsummary' hexlogo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(modelsummary)

fit1 <- lm(mpg ~ cyl, data = mtcars)
fit2 <- lm(mpg ~ cyl + hp, data = mtcars)
models <- list(Model1 = fit1, Model2 = fit2)

modelsummary(models) |> futurize()
```


# Introduction

The **[modelsummary]** package creates customizable tables and plots
to summarize statistical models side-by-side. For example,

```r
library(futurize)
plan(multisession)
library(modelsummary)

## fit multiple linear models
fit1 <- lm(mpg ~ cyl, data = mtcars)
fit2 <- lm(mpg ~ cyl + hp, data = mtcars)
models <- list(Model1 = fit1, Model2 = fit2)

## generate modelsummary table in parallel
tbl <- modelsummary(models, output = "data.frame") |> futurize()
print(tbl)
```

will parallelize model summary statistics extraction, given that we have
set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **modelsummary** functions are supported by `futurize()`:

* `modelsummary()` with `seed = TRUE` as the default
* `msummary()` with `seed = TRUE` as the default
* `modelplot()` with `seed = TRUE` as the default


[modelsummary]: https://cran.r-project.org/package=modelsummary
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'TSP' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{TSP}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/TSP-logo.webp" alt="The 'TSP' hexlogo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(TSP)

data("USCA50")
tour <- solve_TSP(USCA50, method = "nn", rep = 10L) |> futurize()
```

# Introduction

The **[TSP]** package provides algorithms for solving the traveling
salesperson problem (TSP).

## Example: 

Example adopted from `help("solve_TSP", package = "TSP")`:

```r
library(futurize)
plan(multisession)
library(TSP)

data("USCA50")
methods <- c(
  "identity", "random", "nearest_insertion", "cheapest_insertion",
  "farthest_insertion", "arbitrary_insertion", "nn", "repetitive_nn", 
  "two_opt", "sa"
)

## calculate tours - each tour in parallel
tours <- lapply(methods, FUN = function(m) {
  solve_TSP(USCA50, rep = 10L, method = m) |> futurize()
})
names(tours) <- methods
```

This will parallelize the computations, given that we have set up
parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **TSP** functions are supported by `futurize()`:

* `solve_TSP()` with `seed = TRUE` as the default


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `solve_TSP()`
using the **parallel** and **doParallel** packages directly, without
**futurize**:

```r
library(TSP)
library(parallel)
library(doParallel)

data("USCA50")

## Set up a PSOCK cluster and register it with foreach
ncpus <- 4L
cl <- makeCluster(ncpus)
registerDoParallel(cl)

## Solve the TSP in parallel via foreach
tour <- solve_TSP(USCA50, method = "nn", rep = 10L)

## Tear down the cluster
stopCluster(cl)
registerDoSEQ()  ## reset foreach to sequential
```

This requires you to manually create a cluster, register it with
**doParallel**, and remember to tear it down and reset the
**foreach** backend when done. If you forget to call
`stopCluster()`, or if your code errors out before reaching it, you
leak background R processes. You also have to decide upfront how
many CPUs to use and what cluster type to use. Switching to another
parallel backend, e.g. a Slurm cluster, would require a completely
different setup. With **futurize**, all of this is handled for you - just pipe
to `futurize()` and control the backend with `plan()`.


[TSP]: https://cran.r-project.org/package=TSP
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'lme4' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{lme4}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-lme4-logo.webp" alt="The 'lme4' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(lme4)

gm <- glmer(cbind(incidence, size - incidence) ~ period + (1 | herd),
             data = cbpp, family = binomial)
gm_all <- allFit(gm) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[lme4]**
functions such as `allFit()` and `bootMer()`.

The **[lme4]** package fits linear and generalized linear mixed-effects
models. Its `allFit()` function fits models using all available
optimizers to check for convergence issues, and `bootMer()` performs
parametric bootstrap inference. Both are excellent candidates for
parallelization.


## Example: Fitting with multiple optimizers

The `allFit()` function fits a model with each available optimizer,
which can be done in parallel:

```r
library(lme4)

## Fit a generalized linear mixed model
gm <- glmer(cbind(incidence, size - incidence) ~ period + (1 | herd),
            data = cbpp, family = binomial)

## Try all available optimizers
gm_all <- allFit(gm)
```

Here `allFit()` evaluates sequentially, but we can easily make it
evaluate in parallel by piping to `futurize()`:

```r
library(futurize)
library(lme4)

gm <- glmer(cbind(incidence, size - incidence) ~ period + (1 | herd),
            data = cbpp, family = binomial)
gm_all <- allFit(gm) |> futurize()
```

This will distribute the optimizer fits across the available parallel
workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

Unlike other parallel backends in R, `futurize()` relays standard
output, messages, and warnings produced by the parallel workers back
to your main R session. For instance, when running the above, you
will see the progress output from each optimizer as it completes:

```
> gm_all <- allFit(gm) |> futurize()
bobyqa : [OK]
Nelder_Mead : [OK]
nlminbwrap : [OK]
nmkbw : [OK]
optimx.L-BFGS-B : [OK]
nloptwrap.NLOPT_LN_NELDERMEAD : [OK]
nloptwrap.NLOPT_LN_BOBYQA : [OK]
```

This output originates from the parallel workers and is relayed to
your R session, so you get the same informative feedback as when
running sequentially.

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


## Example: Parametric bootstrap

The `bootMer()` function performs parametric bootstrap inference on
fitted models:

```r
library(futurize)
plan(multisession)
library(lme4)

## Fit a linear mixed model
fm <- lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy)

## Bootstrap the fixed-effect coefficients
boot_coef <- function(model) fixef(model)
b <- bootMer(fm, boot_coef, nsim = 100) |> futurize()
```


# Supported Functions

The following **lme4** functions are supported by `futurize()`:

* `allFit()`
* `bootMer()`
* `influence()` for 'merMod'
* `profile()` for 'merMod'


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `bootMer()`
using the **parallel** package directly, without **futurize**:

```r
library(lme4)
library(parallel)

## Fit a linear mixed model
fm <- lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy)

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Bootstrap the fixed-effect coefficients
boot_coef <- function(model) fixef(model)
b <- bootMer(fm, boot_coef, nsim = 100,
             parallel = "snow", ncpus = ncpus, cl = cl)

## Tear down the cluster
stopCluster(cl)
```

This requires you to manually create and manage the cluster
lifecycle. If you forget to call `stopCluster()`, or if your code
errors out before reaching it, you leak background R processes. You
also have to decide upfront how many CPUs to use and what cluster
type to use. Switching to another parallel backend, e.g. a Slurm
cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


[lme4]: https://cran.r-project.org/package=lme4
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'seriation' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{seriation}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/seriation-logo.webp" alt="The 'seriation' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(seriation)

o <- seriation::seriate_best(d_supreme) |> futurize()
```


# Introduction

The **[seriation]** package provides functions for ordering objects
using seriation, ordination techniques for reordering matrices,
dissimilarity matrices, and dendrograms.


## Example: Seriate best

Example adopted from `help("seriate_best", package = "seriation")`:

```r
library(futurize)
plan(multisession)
library(seriation)

data(SupremeCourt)
d_supreme <- as.dist(SupremeCourt)

o <- seriate_best(d_supreme, criterion = "AR_events") |> futurize()
print(o)
```

This will parallelize the computations, given that we have set up
parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **seriation** functions are supported by `futurize()`:

* `seriate_best()` with `seed = TRUE` as the default
* `seriate_rep()` with `seed = TRUE` as the default


[seriation]: https://cran.r-project.org/package=seriation
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'parameters' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{parameters}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/parameters-logo.webp" alt="The 'parameters' hexlogo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(parameters)

model <- lm(mpg ~ wt, data = mtcars)
fit <- bootstrap_model(model, iterations = 1000) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[parameters]**
functions, such as `bootstrap_model()` and `bootstrap_parameters()`.

The **[parameters]** package (part of the **easystats** ecosystem)
provides utilities for processing and summarizing statistical models.
The `bootstrap_model()` function generates a distribution of model
estimates by refitting the model multiple times using bootstrapped
samples. This process can be computationally demanding, especially for
complex models or a large number of iterations. Since each bootstrap
iteration is independent, it is a perfect candidate for
parallelization.


## Example: Bootstrapping a linear model

Consider a linear model where we want to obtain bootstrapped
estimates of the coefficients:

```r
library(parameters)

model <- lm(mpg ~ wt + cyl, data = mtcars)

## Generate 1000 bootstrap replicates (sequentially)
boot_dist <- bootstrap_model(model, iterations = 1000)
```

To parallelize this using **futurize**, simply pipe the call to
`futurize()`:

```r
library(futurize)
library(parameters)

model <- lm(mpg ~ wt + cyl, data = mtcars)

## Generate 1000 bootstrap replicates (in parallel)
boot_dist <- bootstrap_model(model, iterations = 1000) |> futurize()
```

This will distribute the bootstrap iterations across the available
parallel workers, given that we have set up a parallel backend, e.g.

```r
plan(multisession)
```


## Example: Bootstrapped parameters summary

The `bootstrap_parameters()` function is a higher-level wrapper that
calls `bootstrap_model()` and then summarizes the results. It can
also be parallelized in the same way:

```r
library(futurize)
plan(multisession)
library(parameters)

model <- lm(mpg ~ wt + cyl, data = mtcars)
boot_params <- bootstrap_parameters(model, iterations = 1000) |> futurize()
```


# Supported Functions

The following **parameters** functions are supported by `futurize()`:

* `bootstrap_model()` with `seed = TRUE` as the default
* `bootstrap_parameters()` with `seed = TRUE` as the default

# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `bootstrap_model()` using
the **parallel** package directly, without **futurize**:

```r
library(parameters)
library(parallel)

model <- lm(mpg ~ wt + cyl, data = mtcars)

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Run bootstrapping in parallel
boot_dist <- bootstrap_model(model, iterations = 1000, 
                             parallel = "snow", n_cpus = ncpus, 
                             cluster = cl)

## Tear down the cluster
stopCluster(cl)
```

With **futurize**, the cluster management is handled automatically.
You just control the backend with `plan()`.


[parameters]: https://cran.r-project.org/package=parameters
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'boot' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{boot}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-boot-logo.webp" alt="The 'boot' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(boot)

ratio <- function(pop, w) sum(w * pop$x) / sum(w * pop$u)
b <- boot(bigcity, statistic = ratio, R = 999, stype = "w") |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[boot]**
functions such as `boot()`, `censboot()`, and `tsboot()`.

The **[boot]** package is one of the "recommended" R packages, meaning
it is officially endorsed by the R Core Team, well maintained, and
installed by default with R. The package generates bootstrap samples
and provides statistical methods around them. Given the resampling
nature of bootstrapping, the algorithms are excellent candidates for
parallelization.


## Example: Bootstrap sampling

The core function `boot()` produces bootstrap samples of a statistic
applied to data. For example, consider the `bigcity` dataset, which
contains populations of 49 large U.S. cities in 1920 (`u`) and 1930
(`x`):

```r
library(boot)

## Draw 999 bootstrap samples of the population data. For each
## sample, calculate the ratio of mean-1930 over mean-1920 populations
ratio <- function(pop, w) sum(w * pop$x) / sum(w * pop$u)
b <- boot(bigcity, statistic = ratio, R = 999, stype = "w")
```

Here `boot()` evaluates sequentially, but we can easily make it
evaluate in parallel by piping to `futurize()`:

```r
library(futurize)
library(boot)

ratio <- function(pop, w) sum(w * pop$x) / sum(w * pop$u)
b <- boot(bigcity, statistic = ratio, R = 999, stype = "w") |> futurize()
```

This will distribute the 999 bootstrap samples across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


## Example: Time series bootstrap

The `tsboot()` function generates bootstrap samples from time series
data. For example, here we fit autoregressive models to bootstrap
replicates of the `lynx` time series:

```r
library(futurize)
plan(multisession)
library(boot)

## Fit AR models to bootstrap replicates of the lynx time series
lynx_fun <- function(tsb) {
    ar_fit <- ar(tsb, order.max = 25)
    c(ar_fit$order, mean(tsb), tsb)
}

lynx_boot <- tsboot(log(lynx), lynx_fun, R = 999, l = 20, sim = "geom") |> futurize()
```


## Example: Bootstrap for censored data

The `censboot()` function is used for bootstrap sampling of censored
data, which is common in survival analysis. For example, using the
`aml` data from the **[boot]** package:

```r
library(futurize)
plan(multisession)
library(boot)
library(survival)

data(aml, package = "boot")

aml_fun <- function(data) {
    surv <- survfit(Surv(time, cens) ~ group, data = data)
    out <- NULL
    st <- 1
    for (s in seq_along(surv$strata)) {
        inds <- st:(st + surv$strata[s] - 1)
        md <- min(surv$time[inds[1-surv$surv[inds] >= 0.5]])
        st <- st + surv$strata[s]
        out <- c(out, md)
    }
    out
}

aml_boot <- censboot(aml, aml_fun, R = 999, strata = aml$group) |> futurize()
```


# Supported Functions

The following **boot** functions are supported by `futurize()`:

* `boot()`
* `censboot()`
* `tsboot()`


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `boot()` using
the **parallel** package directly, without **futurize**:

```r
library(boot)
library(parallel)

ratio <- function(pop, w) sum(w * pop$x) / sum(w * pop$u)

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Run bootstrapping in parallel
b <- boot(bigcity, statistic = ratio, R = 999, stype = "w",
          parallel = "snow", ncpus = ncpus, cl = cl)

## Tear down the cluster
stopCluster(cl)
```

This requires you to manually create and manage the cluster
lifecycle. If you forget to call `stopCluster()`, or if your code
errors out before reaching it, you leak background R processes. You
also have to decide upfront how many CPUs to use and what cluster
type to use. Switching to another parallel backend, e.g. a Slurm
cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


[boot]: https://cran.r-project.org/package=boot
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'DESeq2' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{Bioconductor}
%\VignetteKeyword{DESeq2}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/DESeq2-logo.webp" alt="The 'DESeq2' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(DESeq2)

dds <- DESeqDataSetFromMatrix(countData, colData, design = ~ condition)
dds <- DESeq(dds) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
the **[DESeq2]** `DESeq()` function.

The **[DESeq2]** Bioconductor package provides methods to test for
differential expression in RNA-seq data. The main function `DESeq()`
runs a pipeline of gene-wise dispersion estimation, fitting, and
statistical testing, which can be parallelized across genes.


## Example: Running DESeq() in parallel

The `DESeq()` function performs the full differential expression
analysis:

```r
library(DESeq2)

# Simulate data
n_genes <- 100L
n_samples <- 8L
counts <- matrix(
  as.integer(runif(n_genes * n_samples, min = 0, max = 1000)),
  nrow = n_genes,
  ncol = n_samples,
  dimnames = list(
    paste0("gene", seq_len(n_genes)),
    paste0("sample", seq_len(n_samples))
  )
)
 
col_data <- data.frame(
  condition = factor(rep(c("control", "treated"), each = n_samples / 2L)),
  row.names = colnames(counts)
)

dds <- DESeqDataSetFromMatrix(
  countData = counts,
  colData = col_data,
  design = ~ condition
)

dds <- DESeq(dds)
res <- results(dds)
```

Here `DESeq()` runs sequentially, but we can easily make it run in
parallel by piping to `futurize()`:

```r
library(futurize)

dds <- DESeq(dds) |> futurize()
res <- results(dds)
```

This will distribute the work across the available parallel workers,
given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **DESeq2** functions are supported by `futurize()`:

* `DESeq()`
* `lfcShrink()`
* `results()`


[DESeq2]: https://bioconductor.org/packages/DESeq2/
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'Rsamtools' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{Bioconductor}
%\VignetteKeyword{Rsamtools}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/Rsamtools-logo.webp" alt="The 'Rsamtools' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(Rsamtools)

bv <- BamViews(bam_files)
counts <- countBam(bv) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
the **[Rsamtools]** functions.

The **[Rsamtools]** Bioconductor package provides an interface to
BAM (Binary Alignment Map) files and other high-throughput sequencing
data formats. Functions like `countBam()` and `scanBam()` can process
multiple BAM files in parallel when called with a `BamViews` object,
which distributes work across BAM files using `bplapply()`.


## Example: Counting reads across multiple BAM files in parallel

The `countBam()` function counts the number of records in BAM files.
When called with a `BamViews` object containing multiple BAM files,
the counting can be parallelized:

```r
library(Rsamtools)

bam_files <- c("sample1.bam", "sample2.bam", "sample3.bam")
bv <- BamViews(bam_files)

counts <- countBam(bv)
```

Here `countBam()` processes BAM files sequentially, but we can easily
make it process them in parallel by piping to `futurize()`:

```r
library(futurize)

counts <- countBam(bv) |> futurize()
```

This will distribute the BAM file processing across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **Rsamtools** functions are supported by `futurize()`:

* `countBam()`
* `scanBam()`


[Rsamtools]: https://bioconductor.org/packages/Rsamtools/
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'foreach' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{foreach}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-foreach-logo.webp" alt="The CRAN 'foreach' package">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(foreach)

slow_fcn <- function(x) {
  message("x = ", x)
  Sys.sleep(0.1)  # emulate work
  x^2
}

xs <- 1:10
ys <- foreach(x = xs) %do% slow_fcn(x) |> futurize()
```

# Introduction

This vignette demonstrates how to use this approach to parallelize
functions such as `foreach()` and `times()` of the
**[foreach]** package. For example, consider:

```r
library(foreach)
xs <- 1:1000
ys <- foreach(x = xs) %do% slow_fcn(x)
```

This `foreach()` construct is resolved sequentially. We can use the
**futurize** package to tell **foreach** to hand over the
orchestration of parallel tasks to futureverse. All we need to do is
to pass the expression to `futurize()` as in:

```r
library(foreach)

library(futurize)
plan(multisession) ## parallelize on local machine

xs <- 1:1000
ys <- foreach(x = xs) %do% slow_fcn(x) |> futurize()
#> x = 1
#> x = 2
#> x = 3
#> ...
#> x = 10
```

Note how messages produced on parallel workers are relayed as-is back
to the main R session as they complete. Not only messages, but also
warnings and other types of conditions are relayed back as-is.
Likewise, standard output produced by `cat()`, `print()`, `str()`, and
so on is relayed in the same way. This is a unique feature of
Futureverse - other parallel frameworks in R, such as **parallel**,
**foreach** with **doParallel**, and **BiocParallel**, silently drop
standard output, messages, and warnings produced on workers. With
**futurize**, your code behaves the same whether it runs sequentially
or in parallel: nothing is lost in translation.

The built-in `multisession` backend parallelizes on your local
computer and it works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```

Here is another example that parallelizes `times()` of the
**foreach** package via the futureverse ecosystem:


```r
library(foreach)
library(futurize)
ys <- times(10) %do% rnorm(3) |> futurize()
```


# Supported Functions

The `futurize()` function supports parallelization of the following
**foreach** functions:

 * `foreach(...) %do% { ... }`
 * `times(...) %do% { ... }` with `seed = TRUE` as the default


# Progress Reporting via progressify

For progress reporting, please see the **[progressify]** package. It is
specially designed to work with the Futureverse ecosystem and provide
progress updates from parallelized computations in a near-live
fashion. See the `vignette("futurize-11-apply", package = "futurize")`
for more details and an example.


[progressify]: https://progressify.futureverse.org/
[foreach]: https://cran.r-project.org/package=foreach
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'pbapply' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{pbapply}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-pbapply-logo.webp" alt="The 'pbapply' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(pbapply)

slow_fcn <- function(x) {
  message("x = ", x)
  Sys.sleep(0.1)  # emulate work
  sqrt(x)
}

xs <- 1:100
ys <- pblapply(xs, slow_fcn) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
**[pbapply]** functions such as `pblapply()`, `pbsapply()`, and
`pbvapply()`.

The **[pbapply]** package provides progress-bar versions of the
base-R `*apply()` family of functions. It supports parallel
processing via the `cl` argument, which accepts a PSOCK cluster
object or, when used with **futurize**, the string `"future"`.


## Example: Parallel lapply with progress bar

The `pblapply()` function works like `lapply()` but displays a
progress bar. For example:

```r
library(pbapply)

slow_fcn <- function(x) {
  Sys.sleep(0.1)  # emulate work
  sqrt(x)
}

## Apply a function to each element with a progress bar
xs <- 1:100
ys <- pblapply(xs, slow_fcn)
```

Here `pblapply()` evaluates sequentially, but we can easily make it
evaluate in parallel by piping to `futurize()`:

```r
library(pbapply)

library(futurize)
plan(multisession) ## parallelize on local machine

xs <- 1:100
ys <- pblapply(xs, slow_fcn) |> futurize()
```

Comment: The `message("x = ", x)` output is not relayed to the main R
session by design, because if it were, it would clutter up the
progress bar that **pbapply** renders, which is the whole purpose of
using **pbapply** in the first place.

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


## Example: Parallel sapply with progress bar

The `pbsapply()` function simplifies the result like `sapply()`:

```r
library(futurize)
plan(multisession)
library(pbapply)

xs <- 1:100
ys <- pbsapply(xs, slow_fcn) |> futurize()
```


# Supported Functions

The following **pbapply** functions are supported by `futurize()`:

* `pbapply()`
* `pbby()`
* `pbeapply()`
* `pblapply()`
* `pbreplicate()`
* `pbsapply()`
* `pbtapply()`
* `pbvapply()`
* `pbwalk()`


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `pblapply()`
using the **parallel** package directly, without **futurize**:

```r
library(pbapply)
library(parallel)

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Run pblapply in parallel
xs <- 1:100
ys <- pblapply(xs, slow_fcn, cl = cl)

## Tear down the cluster
stopCluster(cl)
```

This requires you to manually create and manage the cluster
lifecycle. If you forget to call `stopCluster()`, or if your code
errors out before reaching it, you leak background R processes. You
also have to decide upfront how many CPUs to use and what cluster
type to use. Switching to another parallel backend, e.g. a Slurm
cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


# Progress Reporting via progressify

An alternative to using **pbapply** for progress reporting is to use
the **[progressify]** package, which is specially designed to work with
the Futureverse ecosystem and provide progress updates from
parallelized computations in a near-live fashion. See the
`vignette("futurize-11-apply", package = "futurize")` for more
details.


[progressify]: https://progressify.futureverse.org/
[pbapply]: https://cran.r-project.org/package=pbapply
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'pls' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{pls}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(pls)

data(yarn)
m <- plsr(density ~ NIR, ncomp = 10, data = yarn, validation = "CV") |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[pls]**
functions such as `mvr()`, `plsr()`, `pcr()`, and `crossval()`.

The **[pls]** package provides Partial Least Squares Regression (PLSR)
and Principal Component Regression (PCR) methods. These methods
often use cross-validation (CV) to determine the number of components
to use, which can be computationally intensive and is an ideal
candidate for parallelization.


## Example: PLS Regression with Cross-Validation

The `plsr()` function is used to perform PLS regression. When
`validation = "CV"` is specified, it performs cross-validation.

```r
library(pls)
data(yarn)

## Sequential evaluation
m <- plsr(density ~ NIR, ncomp = 10, data = yarn, validation = "CV")
```

To make it evaluate in parallel, simply pipe the call to `futurize()`:

```r
library(futurize)
library(pls)
data(yarn)

## Parallel evaluation
m <- plsr(density ~ NIR, ncomp = 10, data = yarn, validation = "CV") |> futurize()
```

This will automatically use the parallel backend set by `plan()`, e.g.

```r
plan(multisession)
```


## Example: Stand-alone Cross-Validation

The `crossval()` function can be used to perform cross-validation on
an already fitted model:

```r
library(futurize)
plan(multisession)
library(pls)

data(yarn)
m1 <- plsr(density ~ NIR, ncomp = 10, data = yarn)

## Parallel cross-validation
m_cv <- crossval(m1, segments = 10) |> futurize()
```


# Supported Functions

The following **pls** functions are supported by `futurize()`:

* `mvr()`
* `plsr()`
* `pcr()`
* `cppls()`
* `crossval()` with `seed = TRUE` as the default


# Without futurize: Manual 'pls.options' setup

For comparison, here is what it takes to parallelize `pls` functions
using the **parallel** package directly, without **futurize**:

```r
library(pls)
library(parallel)

## Set up a cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Configure pls to use the cluster
old_opts <- pls.options(parallel = cl)

## Run regression with cross-validation
data(yarn)
m <- plsr(density ~ NIR, ncomp = 10, data = yarn, validation = "CV")

## Restore original options and stop the cluster
pls.options(old_opts)
stopCluster(cl)
```

This requires you to manually manage the cluster lifecycle and the
global `pls.options()`. With **futurize**, the cluster setup and
option management are handled automatically and localized to the
function call.


[pls]: https://cran.r-project.org/package=pls

<!--
%\VignetteIndexEntry{Parallelize 'metafor' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{metafor}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-metafor-logo.webp" alt="The CRAN 'metafor' package">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(metafor)

dat <- escalc(measure = "RR", ai = tpos, bi = tneg,
              ci = cpos, di = cneg, data = dat.bcg)
fit <- rma(yi, vi, data = dat)
pr <- profile(fit) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
**[metafor]** functions such as `profile()`, `rstudent()`,
`cooks.distance()`, and `dfbetas()`.

The **[metafor]** package provides a comprehensive collection of
functions for conducting meta-analyses in R. It supports fixed-effects,
random-effects, and mixed-effects (meta-regression) models and includes
functions for model diagnostics and profiling. Several of these
computations involve fitting the model repeatedly, making them
excellent candidates for parallelization.


## Example: Likelihood profile for a random-effects model

The `profile()` function computes the likelihood profile for model
parameters such as the variance component in a random-effects
meta-analysis. For example, using the built-in BCG vaccine dataset:

```r
library(metafor)

## Calculate log risk ratios and sampling variances
dat <- escalc(measure = "RR", ai = tpos, bi = tneg,
              ci = cpos, di = cneg, data = dat.bcg)

## Fit a random-effects model
fit <- rma(yi, vi, data = dat)

## Compute likelihood profile
pr <- profile(fit)
```

Here `profile()` is calculated sequentially. To calculate in
parallel, we can pipe to `futurize()`:

```r
library(futurize)
library(metafor)

dat <- escalc(measure = "RR", ai = tpos, bi = tneg,
              ci = cpos, di = cneg, data = dat.bcg)
fit <- rma(yi, vi, data = dat)
pr <- profile(fit) |> futurize()
```

This will distribute the profile computations across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **metafor** functions are supported by `futurize()`:

* `profile()` for 'rma.uni', 'rma.mv', 'rma.ls', and 'rma.uni.selmodel'
* `rstudent()` for 'rma.mv'
* `cooks.distance()` for 'rma.mv'
* `dfbetas()` for 'rma.mv'


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `profile()`
using the **parallel** package directly, without **futurize**:

```r
library(metafor)
library(parallel)

## Calculate log risk ratios and sampling variances
dat <- escalc(measure = "RR", ai = tpos, bi = tneg,
              ci = cpos, di = cneg, data = dat.bcg)

## Fit a random-effects model
fit <- rma(yi, vi, data = dat)

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Compute likelihood profile in parallel
pr <- profile(fit, parallel = "snow", ncpus = ncpus, cl = cl)

## Tear down the cluster
stopCluster(cl)
```

This requires you to manually create and manage the cluster
lifecycle. If you forget to call `stopCluster()`, or if your code
errors out before reaching it, you leak background R processes. You
also have to decide upfront how many CPUs to use and what cluster
type to use. Switching to another parallel backend, e.g. a Slurm
cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


[metafor]: https://cran.r-project.org/package=metafor
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'sva' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{Bioconductor}
%\VignetteKeyword{sva}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/sva-logo.webp" alt="The 'sva' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(sva)

adjusted <- ComBat(dat = dat, batch = batch) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
the **[sva]** functions.

The **[sva]** Bioconductor package provides functions for removing
batch effects and other unwanted variation in high-throughput
experiments. The `ComBat()` function is a widely used method for
batch effect correction using an empirical Bayes framework. It
supports parallelization via BiocParallel's BPPARAM argument.


## Example: Running ComBat() in parallel

The `ComBat()` function adjusts for known batch effects in
microarray or RNA-seq data:

```r
library(sva)

# Create example data with batch effect
set.seed(42)
n_genes <- 200L
n_samples <- 40L
dat <- matrix(rnorm(n_genes * n_samples), nrow = n_genes, ncol = n_samples)
rownames(dat) <- paste0("gene", seq_len(n_genes))
colnames(dat) <- paste0("sample", seq_len(n_samples))

batch <- rep(c(1, 2), each = n_samples / 2L)
dat[, batch == 2] <- dat[, batch == 2] + 2

adjusted <- ComBat(dat = dat, batch = batch)
```

Here `ComBat()` runs sequentially, but we can easily make it run in
parallel by piping to `futurize()`:

```r
library(futurize)

adjusted <- ComBat(dat = dat, batch = batch) |> futurize()
```

This will distribute the work across the available parallel workers,
given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


## Using ComBat() with a model matrix

You can also include a model matrix for biological covariates of
interest, which will be protected during batch correction:

```r
mod <- model.matrix(~ group)
adjusted <- ComBat(dat = dat, batch = batch, mod = mod) |> futurize()
```


# Supported Functions

The following **sva** functions are supported by `futurize()`:

* `ComBat()`
* `read.degradation.matrix()`


[sva]: https://bioconductor.org/packages/sva/
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'DiceKriging' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{DiceKriging}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(DiceKriging)

design <- expand.grid(x1 = seq(0, 1, length = 15), x2 = seq(0, 1, length = 15))
y <- apply(design, 1, function(x) x[1]^2 + x[2]^2)
m <- km(~., design = design, response = data.frame(y = y),
        multistart = 20) |> futurize()
```


# Introduction

This vignette demonstrates how to use **futurize** to parallelize
**[DiceKriging]** functions, specifically `km()`.
When fitting a kriging model via `km()`, the parameters of the
covariance function are estimated by maximum likelihood or
cross-validation. The optimization can be started from multiple
points (to avoid local optima), which can be done in parallel.


## Example: kriging model with multi-start optimization

Fitting a kriging model with a single starting point:

```r
library(DiceKriging)

design <- expand.grid(x1 = seq(0, 1, length = 15), x2 = seq(0, 1, length = 15))
y <- apply(design, MARGIN = 1, FUN = function(x) x[1]^2 + x[2]^2)
m <- km(~., design = design, response = data.frame(y = y))
```

To run multiple optimizer starts in parallel, set `multistart > 1`
and pipe to `futurize()`:

```r
library(futurize)
library(DiceKriging)

design <- expand.grid(x1 = seq(0, 1, length = 15), x2 = seq(0, 1, length = 15))
y <- apply(design, MARGIN = 1, FUN = function(x) x[1]^2 + x[2]^2)
m <- km(~., design = design, response = data.frame(y = y),
        multistart = 20) |> futurize()
```

This distributes the multi-start runs across the available parallel
workers, given that we have set up a parallel plan, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **DiceKriging** function is supported by `futurize()`:

* `km()`


[DiceKriging]: https://cran.r-project.org/package=DiceKriging
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'scuttle' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{Bioconductor}
%\VignetteKeyword{scuttle}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/bioconductor-scuttle-logo.webp" alt="The 'scuttle' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(scuttle)

qc <- perFeatureQCMetrics(sce) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
the **[scuttle]** functions.

The **[scuttle]** Bioconductor package provides basic utility
functions for single-cell RNA-seq data analysis, including quality
control, normalization, and aggregation, which can be parallelized
across cells or features.


## Example: Computing per-feature QC metrics in parallel

The `perFeatureQCMetrics()` function computes quality control metrics
for each feature (gene) in a `SingleCellExperiment` object:

```r
library(scuttle)

# Simulate data
sce <- mockSCE()

qc <- perFeatureQCMetrics(sce)
```

Here `perFeatureQCMetrics()` runs sequentially, but we can easily make
it run in parallel by piping to `futurize()`:

```r
library(futurize)

qc <- perFeatureQCMetrics(sce) |> futurize()
```

This will distribute the work across the available parallel workers,
given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **scuttle** functions are supported by `futurize()`:

* `calculateAverage()`
* `perFeatureQCMetrics()`
* `numDetectedAcrossFeatures()`
* `summarizeAssayByGroup()`
* `medianSizeFactors()`
* `computeMedianFactors()`
* `pooledSizeFactors()`
* `computePooledFactors()`
* `fitLinearModel()`


The following **scuttle** functions are deprecated in **scuttle** (>=
1.22) in favor of counter-part functions in Bioconductor package
**[scrapper]**. Support for `futurize()` of the these deprecated
functions remains, but will be phased out;

* `logNormCounts()`
* `normalizeCounts()`
* `perCellQCMetrics()`
* `addPerCellQCMetrics()`
* `addPerFeatureQCMetrics()`
* `addPerCellQC()`
* `addPerFeatureQC()`
* `numDetectedAcrossCells()`
* `sumCountsAcrossCells()`
* `sumCountsAcrossFeatures()`
* `aggregateAcrossCells()`
* `aggregateAcrossFeatures()`
* `librarySizeFactors()`
* `computeLibraryFactors()`
* `geometricSizeFactors()`
* `computeGeometricFactors()`

[scuttle]: https://bioconductor.org/packages/scuttle/
[scrapper]: https://bioconductor.org/packages/scrapper/
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'fgsea' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{Bioconductor}
%\VignetteKeyword{fgsea}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/bioconductor-fgsea-logo.webp" alt="The 'fgsea' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(fgsea)

res <- fgsea(pathways, stats) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
the **[fgsea]** functions.

The **[fgsea]** Bioconductor package implements fast preranked gene
set enrichment analysis (GSEA). The main functions `fgsea()`,
`fgseaMultilevel()`, and `fgseaSimple()` perform permutation-based
enrichment testing, which can be parallelized across gene sets.


## Example: Running fgseaSimple() in parallel

The `fgseaSimple()` function performs permutation-based gene set
enrichment analysis:

```r
library(fgsea)

# Create example data
set.seed(42)
n_genes <- 1000L
stats <- rnorm(n_genes)
names(stats) <- paste0("gene", seq_len(n_genes))

pathways <- list(
  pathway1 = paste0("gene", sample(n_genes, 50L)),
  pathway2 = paste0("gene", sample(n_genes, 100L)),
  pathway3 = paste0("gene", sample(n_genes, 150L))
)

res <- fgseaSimple(pathways, stats, nperm = 10000)
```

Here `fgseaSimple()` runs sequentially, but we can easily make it
run in parallel by piping to `futurize()`:

```r
library(futurize)

res <- fgseaSimple(pathways, stats, nperm = 10000) |> futurize()
```

This will distribute the work across the available parallel workers,
given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **fgsea** functions are supported by `futurize()`:

* `fgsea()`
* `fgseaMultilevel()`
* `fgseaSimple()`
* `fgseaLabel()`
* `geseca()`
* `gesecaSimple()`
* `collapsePathwaysGeseca()`


[fgsea]: https://bioconductor.org/packages/fgsea/
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'riskRegression' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{riskRegression}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-riskRegression-logo.webp" alt="The 'riskRegression' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(riskRegression)
library(survival)

set.seed(42)
d <- sampleData(200, outcome = "competing.risks")
fit <- CSC(Hist(time, event) ~ X1 + X2 + X7 + X8, data = d)
sc <- Score(list("CSC" = fit), data = d,
            formula = Hist(time, event) ~ 1,
            times = 5, B = 100, split.method = "bootcv") |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
**[riskRegression]** functions such as `Score()`.

The **[riskRegression]** package provides tools for risk regression
modeling and prediction in survival analysis with competing risks.
It supports fitting cause-specific Cox regression models, Fine-Gray
regression, and absolute risk regression models. The `Score()`
function performs bootstrap cross-validation for model evaluation,
which is an excellent candidate for parallelization.


## Example: Bootstrap cross-validation with Score()

The `Score()` function evaluates prediction models via bootstrap
cross-validation with metrics such as time-dependent AUC and Brier
scores:

```r
library(riskRegression)
library(survival)

set.seed(42)
d <- sampleData(200, outcome = "competing.risks")
fit <- CSC(Hist(time, event) ~ X1 + X2 + X7 + X8, data = d)

## Bootstrap cross-validation with 100 bootstrap samples
sc <- Score(list("CSC" = fit), data = d,
            formula = Hist(time, event) ~ 1,
            times = 5, B = 100, split.method = "bootcv")
```

Here `Score()` evaluates sequentially, but we can easily make it
evaluate in parallel by piping to `futurize()`:

```r
library(futurize)
library(riskRegression)
library(survival)

set.seed(42)
d <- sampleData(200, outcome = "competing.risks")
fit <- CSC(Hist(time, event) ~ X1 + X2 + X7 + X8, data = d)

sc <- Score(list("CSC" = fit), data = d,
            formula = Hist(time, event) ~ 1,
            times = 5, B = 100, split.method = "bootcv") |> futurize()
```

This will distribute the bootstrap samples across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **riskRegression** functions are supported by `futurize()`:

* `Score()` for 'list' with `seed = TRUE` as the default


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `Score()` using
the **parallel** and **doParallel** packages directly, without
**futurize**:

```r
library(riskRegression)
library(survival)
library(parallel)
library(doParallel)

set.seed(42)
d <- sampleData(200, outcome = "competing.risks")
fit <- CSC(Hist(time, event) ~ X1 + X2 + X7 + X8, data = d)

## Set up a PSOCK cluster and register it with foreach
ncpus <- 4L
cl <- makeCluster(ncpus)
registerDoParallel(cl)

## Bootstrap cross-validation in parallel via foreach
sc <- Score(list("CSC" = fit), data = d,
            formula = Hist(time, event) ~ 1,
            times = 5, B = 100, split.method = "bootcv",
            parallel = "as.registered")

## Tear down the cluster
stopCluster(cl)
registerDoSEQ()  ## reset foreach to sequential
```

This requires you to manually create a cluster, register it with
**doParallel**, and remember to tear it down and reset the
**foreach** backend when done. If you forget to call
`stopCluster()`, or if your code errors out before reaching it, you
leak background R processes. You also have to decide upfront how
many CPUs to use and what cluster type to use. Switching to another
parallel backend, e.g. a Slurm cluster, would require a completely
different setup. With **futurize**, all of this is handled for you - just pipe
to `futurize()` and control the backend with `plan()`.


[riskRegression]: https://cran.r-project.org/package=riskRegression
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'glmnet' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{glmnet}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/glmnet-logo.webp" alt="The 'glmnet' hexlogo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(glmnet)

n <- 1000
p <- 100
nzc <- trunc(p / 10)
x <- matrix(rnorm(n * p), n, p)
beta <- rnorm(nzc)
fx <- x[, seq_len(nzc)] %*% beta
eps <- rnorm(n) * 5
y <- drop(fx + eps)

cv <- cv.glmnet(x, y) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[glmnet]**
functions such as `cv.glmnet()`.

The **[glmnet]** package uses a highly optimized pathwise coordinate
descent algorithm to efficiently compute the entire regularization
path for penalized generalized linear models (Lasso, Ridge, Elastic
Net).  Its `cv.glmnet()` function performs cross-validation to select
the optimal regularization parameter, which is an excellent candidate
for parallelization.


## Example: Cross-validation for regularized regression

The `cv.glmnet()` function fits models across multiple folds and lambda
values. For example:

```r
library(glmnet)

## Generate simulated data
n <- 1000
p <- 100
nzc <- trunc(p / 10)
x <- matrix(rnorm(n * p), n, p)
beta <- rnorm(nzc)
fx <- x[, seq_len(nzc)] %*% beta
eps <- rnorm(n) * 5
y <- drop(fx + eps)

## Perform cross-validation to find optimal lambda
cv <- cv.glmnet(x, y)
```

Here `cv.glmnet()` evaluates sequentially, but we can easily make it
evaluate in parallel by piping to `futurize()`:

```r
library(futurize)
library(glmnet)

n <- 1000
p <- 100
nzc <- trunc(p / 10)
x <- matrix(rnorm(n * p), n, p)
beta <- rnorm(nzc)
fx <- x[, seq_len(nzc)] %*% beta
eps <- rnorm(n) * 5
y <- drop(fx + eps)

cv <- cv.glmnet(x, y) |> futurize()
```

This will distribute the cross-validation folds across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **glmnet** functions are supported by `futurize()`:

* `cv.glmnet()` with `seed = TRUE` as the default


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `cv.glmnet()`
using the **parallel** and **doParallel** packages directly, without
**futurize**:

```r
library(glmnet)
library(parallel)
library(doParallel)

## Generate simulated data
n <- 1000
p <- 100
nzc <- trunc(p / 10)
x <- matrix(rnorm(n * p), n, p)
beta <- rnorm(nzc)
fx <- x[, seq_len(nzc)] %*% beta
eps <- rnorm(n) * 5
y <- drop(fx + eps)

## Set up a PSOCK cluster and register it with foreach
ncpus <- 4L
cl <- makeCluster(ncpus)
registerDoParallel(cl)

## Perform cross-validation in parallel via foreach
cv <- cv.glmnet(x, y, parallel = TRUE)

## Tear down the cluster
stopCluster(cl)
registerDoSEQ()  ## reset foreach to sequential
```

This requires you to manually create a cluster, register it with
**doParallel**, and remember to tear it down and reset the
**foreach** backend when done. If you forget to call
`stopCluster()`, or if your code errors out before reaching it, you
leak background R processes. You also have to decide upfront how
many CPUs to use and what cluster type to use. Switching to another
parallel backend, e.g. a Slurm cluster, would require a completely
different setup. With **futurize**, all of this is handled for you - just pipe
to `futurize()` and control the backend with `plan()`.


[glmnet]: https://cran.r-project.org/package=glmnet
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'SimDesign' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{SimDesign}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-SimDesign-logo.webp" alt="The CRAN 'SimDesign' package">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(SimDesign)

res <- runSimulation(
  design = Design,
  replications = 1000,
  generate = Generate,
  analyse = Analyse,
  summarise = Summarise
) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
**[SimDesign]** functions such as `runSimulation()`.

The **[SimDesign]** package provides a comprehensive framework for
organizing Monte Carlo simulation experiments in R. It uses a
structured generate-analyse-summarise workflow for designing,
executing, and summarizing simulation studies. The replication-based
nature of simulations makes them excellent candidates for
parallelization.


## Example: Monte Carlo simulation

The `runSimulation()` function runs Monte Carlo simulations over a
design of experimental conditions. For example:

```r
library(SimDesign)

Design <- createDesign(
  sample_size = c(10, 20, 40),
  distribution = c("norm", "chi")
)

Generate <- function(condition, fixed_objects) {
  N <- condition$sample_size
  dist <- condition$distribution
  if (dist == "norm") rnorm(N) else rchisq(N, df = 5)
}

Analyse <- function(condition, dat, fixed_objects) {
  c(mean_est = mean(dat))
}

Summarise <- function(condition, results, fixed_objects) {
  obs_bias <- bias(results[, "mean_est"],
    parameter = ifelse(condition$distribution == "norm", 0, 5))
  obs_RMSE <- RMSE(results[, "mean_est"],
    parameter = ifelse(condition$distribution == "norm", 0, 5))
  c(bias = obs_bias, RMSE = obs_RMSE)
}

res <- runSimulation(
  design = Design,
  replications = 100,
  generate = Generate,
  analyse = Analyse,
  summarise = Summarise
)
```

Here `runSimulation()` evaluates sequentially. To run in parallel,
pipe to `futurize()`:

```r
library(futurize)
library(SimDesign)

res <- runSimulation(
  design = Design,
  replications = 100,
  generate = Generate,
  analyse = Analyse,
  summarise = Summarise
) |> futurize()
```

This will distribute the replications across the available parallel
workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **SimDesign** functions are supported by `futurize()`:

* `runSimulation()`
* `runArraySimulation()`


[SimDesign]: https://cran.r-project.org/package=SimDesign
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'vegan' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vegan}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-vegan-logo.webp" alt="The 'vegan' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(vegan)

data(dune)
data(dune.env)
dune.mrpp <- with(dune.env, { mrpp(dune, Management) }) |> futurize()
```

# Introduction

The **[vegan]** package provides methods for community and
vegetation ecologists. Some of the functions have built-in support
for parallelization, which **futurize** simplifies further.

## Example: MRPP

Example adopted from `help("mrpp", package = "vegan")`:

```r
library(futurize)
plan(multisession)
library(vegan)

data(dune)
data(dune.env)
dune.mrpp <- with(dune.env, { mrpp(dune, Management) }) |> futurize()
```

This will parallelize the computations, given that we have set up
parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


## Example: anova() for 'cca' objects

The `anova()` S3 method for 'cca' objects supports parallelization via
the `parallel` argument. With **futurize**, you can parallelize this
directly. Example adopted from `help("anova.cca", package = "vegan")`:

```r
library(futurize)
plan(multisession)
library(vegan)

data(dune)
data(dune.env)
ord <- cca(dune ~ A1 + Management, data = dune.env)
res <- anova(ord, permutations = 99) |> futurize()
```


# Supported Functions

The following **vegan** functions are supported by `futurize()`:

* `adonis()`
* `adonis2()`
* `anova()` for 'cca'
* `anosim()`
* `cascadeKM()` with `seed = TRUE` as the default
* `estaccumR()`
* `mantel()`
* `mantel.partial()`
* `metaMDSiter()`
* `mrpp()`
* `oecosimu()`
* `ordiareatest()`
* `permutest()` for 'betadisper', and 'cca'

# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `mrpp()` using
the **parallel** package directly, without **futurize**:

```r
library(vegan)
library(parallel)

data(dune)
data(dune.env)

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Run MRPP in parallel
dune.mrpp <- with(dune.env, {
  mrpp(dune, Management, parallel = cl)
})

## Tear down the cluster
stopCluster(cl)
```

This requires you to manually create and manage the cluster
lifecycle. If you forget to call `stopCluster()`, or if your code
errors out before reaching it, you leak background R processes. You
also have to decide upfront how many CPUs to use and what cluster
type to use. Switching to another parallel backend, e.g. a Slurm
cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


[vegan]: https://cran.r-project.org/package=vegan
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'plyr' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{plyr}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-plyr-logo.webp" alt="The 'plyr' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(plyr)
library(futurize)
plan(multisession)

slow_fcn <- function(x) {
  message("x = ", x)
  Sys.sleep(0.1)  # emulate work
  x^2
}

xs <- 1:10
ys <- llply(xs, slow_fcn) |> futurize()
```

# Introduction

This vignette demonstrates how to use this approach to parallelize
**[plyr]** functions such as `llply()`, `maply()`, and `ddply()`.

The **plyr** `llply()` function is commonly used to apply a function
to the elements of a list and return a list. For example,

```r
library(plyr)
xs <- 1:1000
ys <- llply(xs, slow_fcn)
```

Here `llply()` evaluates sequentially, but we can easily make it
evaluate in parallel, by using:

```r
library(plyr)

library(futurize)
plan(multisession) ## parallelize on local machine

xs <- 1:1000
ys <- xs |> llply(slow_fcn) |> futurize()
#> x = 1
#> x = 2
#> x = 3
#> ...
#> x = 10
```

Note how messages produced on parallel workers are relayed as-is back
to the main R session as they complete. Not only messages, but also
warnings and other types of conditions are relayed back as-is.
Likewise, standard output produced by `cat()`, `print()`, `str()`, and
so on is relayed in the same way. This is a unique feature of
Futureverse - other parallel frameworks in R, such as **parallel**,
**foreach** with **doParallel**, and **BiocParallel**, silently drop
standard output, messages, and warnings produced on workers. With
**futurize**, your code behaves the same whether it runs sequentially
or in parallel: nothing is lost in translation.

The built-in `multisession` backend parallelizes on your local
computer and it works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```

Another example is:

```r
library(plyr)
library(futurize)
plan(future.mirai::mirai_multisession)

ys <- llply(baseball, summary) |> futurize()
```


# Supported Functions

The `futurize()` function supports parallelization of the following
**plyr** functions:

* `a_ply()`, `aaply()`, `adply()`, `alply()`
* `d_ply()`, `daply()`, `ddply()`, `dlply()`
* `l_ply()`, `laply()`, `ldply()`, `llply()`
* `m_ply()`, `maply()`, `mdply()`, `mlply()`


# Progress Reporting via progressr

For progress reporting, please see the **[progressr]** package. It is
specially designed to work with the Futureverse ecosystem and provide
progress updates from parallelized computations in a near-live
fashion. See the `vignette("futurize-11-apply", package = "futurize")`
for more details and an example.


[plyr]: https://cran.r-project.org/package=plyr
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'sandwich' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{sandwich}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-sandwich-logo.webp" alt="The 'sandwich' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(sandwich)

fm <- lm(dist ~ speed, data = cars)
v <- vcovBS(fm) |> futurize()
```


# Introduction

The **[sandwich]** package provides model-agnostic robust covariance
matrix estimators.


## Example: Clustered bootstrap covariance matrix

Example adopted from `help("vcovBS", package = "sandwich")`:

```r
library(futurize)
plan(multisession)
library(sandwich)

## fit a simple linear model
fm <- lm(dist ~ speed, data = cars)

## bootstrap covariance matrix estimation in parallel
v <- vcovBS(fm, R = 250) |> futurize()

## summary of coefficients with robust standard errors
library(lmtest)
coeftest(fm, vcov = v)
```

This will parallelize the bootstrap replications, given that we have
set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **sandwich** functions are supported by `futurize()`:

* `vcovBS()` with `seed = TRUE` as the default
* `vcovJK()` with `seed = TRUE` as the default


# Without futurize: Manual setup

For comparison, here is what it takes to parallelize `vcovBS()`
using the **sandwich** package directly, without **futurize**:

```r
library(sandwich)
library(parallel)

## Fit a simple linear model
fm <- lm(dist ~ speed, data = cars)

## Bootstrap covariance matrix estimation in parallel using cores
v <- vcovBS(fm, R = 250, cores = 4L)
```

While **sandwich** has a built-in `cores` argument, it only supports
local multicore or PSOCK clusters depending on the OS. With
**futurize**, you can use any `future` backend, including remote
clusters and HPC environments, just by piping to `futurize()` and
controlling the backend with `plan()`.


[sandwich]: https://cran.r-project.org/package=sandwich
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'tm' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{tm}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-tm-logo.webp" alt="The 'tm' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(tm)

data("crude")
m <- tm_map(crude, content_transformer(tolower)) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[tm]**
functions such as `tm_map()`.

The **[tm]** package provides a variety of text-mining methods. The
`tm_map()` function applies transformations to a corpus of text
documents, and `TermDocumentMatrix()` constructs document-term matrices.
When working with large corpora, these operations benefit greatly from
parallelization.


## Example: Transforming a text corpus

The `tm_map()` function applies a transformation to each document in
a corpus:

```r
library(tm)

## Load the crude oil news corpus holding 20 documents
data("crude")

## Convert all text to lowercase
m <- tm_map(crude, content_transformer(tolower))
```

Here `tm_map()` evaluates sequentially, but we can easily make it
evaluate in parallel by piping to `futurize()`:

```r
library(tm)
library(futurize)
plan(multisession)

data("crude")
m <- tm_map(crude, content_transformer(tolower)) |> futurize()
```

This will distribute the document transformations across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **tm** functions are supported by `futurize()`:

* `tm_map()`
* `tm_index()`
* `TermDocumentMatrix()`


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `tm_map()`
using the **parallel** package directly, without **futurize**:

```r
library(tm)
library(parallel)

data("crude")

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Configure tm to use the cluster
old_engine <- tm_parLapply_engine()
tm_parLapply_engine(function(X, FUN, ...) parLapply(cl, X, FUN, ...))

## Transform the corpus in parallel
m <- tm_map(crude, content_transformer(tolower))

## Restore the old engine and tear down the cluster
tm_parLapply_engine(old_engine)
stopCluster(cl)
```

This requires you to manually create a cluster, configure **tm**'s
internal parallel engine, and remember to restore the engine and tear
down the cluster when done. If you forget to call `stopCluster()`,
or if your code errors out before reaching it, you leak background R
processes. You also have to decide upfront how many CPUs to use, what
cluster type to use. Switching to another parallel backend, e.g. a
Slurm cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


[tm]: https://cran.r-project.org/package=tm
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'pvclust' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{pvclust}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>+</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(pvclust)

data(mtcars, package = "datasets")
fit <- pvclust(mtcars, nboot = 1000) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[pvclust]**
functions, specifically `pvclust()`.

The **[pvclust]** package provides hierarchical clustering with
p-values (AU: Approximately Unbiased p-value, BP: Bootstrap
Probability) calculated via multiscale bootstrap resampling. This
method is computationally intensive because it requires repeating the
clustering process for many bootstrap replicates at different scales.
These calculations are naturally independent and thus excellent
candidates for parallelization.


## Example: Hierarchical clustering with p-values

The core function `pvclust()` performs multiscale bootstrap resampling
to assess the uncertainty in hierarchical cluster analysis. For
example, using the `mtcars` dataset:

```r
library(pvclust)

## Assess the uncertainty of hierarchical clustering of mtcars
## variables using 1000 bootstrap replicates
fit <- pvclust(mtcars, nboot = 1000)
```

Here `pvclust()` evaluates sequentially. We can easily make it
evaluate in parallel by piping to `futurize()`:

```r
library(futurize)
library(pvclust)

fit <- pvclust(mtcars, nboot = 1000) |> futurize()
```

This will distribute the bootstrap replications across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **pvclust** function is supported by `futurize()`:

* `pvclust()` with `seed = TRUE` as the default


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `pvclust()` using
the **parallel** package directly, without **futurize**:

```r
library(pvclust)
library(parallel)

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Run pvclust in parallel
fit <- pvclust(mtcars, nboot = 1000, parallel = cl)

## Tear down the cluster
stopCluster(cl)
```

This requires you to manually create and manage the cluster
lifecycle. If you forget to call `stopCluster()`, or if your code
errors out before reaching it, you leak background R processes. You
also have to decide upfront how many CPUs to use and what cluster
type to use. Switching to another parallel backend, e.g. a Slurm
cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


[pvclust]: https://cran.r-project.org/package=pvclust
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'fwb' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{fwb}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-fwb-logo.webp" alt="The 'fwb' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(fwb)

set.seed(123)
lm_fit <- lm(mpg ~ wt + am, data = mtcars)
b <- fwb(mtcars, statistic = function(data, w) {
  fit <- lm(mpg ~ wt + am, data = data, weights = w)
  coef(fit)
}, R = 999) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
**[fwb]** functions such as `fwb()` and `vcovFWB()`.

The **[fwb]** package implements the fractional weighted bootstrap
(also known as the Bayesian bootstrap). Rather than resampling units
to include in bootstrap samples, random weights are drawn and
applied to a weighted estimator. Given the resampling nature of
bootstrapping, the algorithm is an excellent candidate for
parallelization.


## Example: Fractional weighted bootstrap

The `fwb()` function produces fractional weighted bootstrap samples
of a statistic applied to data. For example, consider bootstrapping
a linear model on the `mtcars` dataset:

```r
library(fwb)

## Draw 999 bootstrap samples of the regression coefficients
set.seed(123)
b <- fwb(mtcars, statistic = function(data, w) {
  fit <- lm(mpg ~ wt + am, data = data, weights = w)
  coef(fit)
}, R = 999)
```

Here `fwb()` evaluates sequentially, but we can easily make it
evaluate in parallel by piping to `futurize()`:

```r
library(fwb)
library(futurize)

set.seed(123)
b <- fwb(mtcars, statistic = function(data, w) {
  fit <- lm(mpg ~ wt + am, data = data, weights = w)
  coef(fit)
}, R = 999) |> futurize()
```

This will distribute the 999 bootstrap samples across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


## Example: Bootstrap variance-covariance matrix

The `vcovFWB()` function computes a bootstrap variance-covariance
matrix for model coefficients:

```r
library(futurize)
plan(multisession)
library(fwb)

lm_fit <- lm(mpg ~ wt + am, data = mtcars)
V <- vcovFWB(lm_fit, R = 999) |> futurize()
```


# Supported Functions

The following **fwb** functions are supported by `futurize()`:

* `fwb()` with `seed = TRUE` as the default
* `vcovFWB()` with `seed = TRUE` as the default


[fwb]: https://cran.r-project.org/package=fwb
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'stars' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{stars}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(stars)

m <- matrix(1:20, nrow = 5, ncol = 4)
s <- st_as_stars(m)
res <- st_apply(s, MARGIN = 1, FUN = mean) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[stars]**
functions such as `st_apply()`.

The **[stars]** package provides a framework for "Spatiotemporal Arrays"
(raster and vector data cubes). It is a powerful tool for working with 
large-scale spatial and temporal data. Many operations in **stars**, 
particularly those involving applying functions across dimensions, 
can be computationally intensive and thus benefit significantly from 
parallelization.


## Example: Applying a function across dimensions

The `st_apply()` function applies a function to one or more dimensions 
of a `stars` object. By default, it runs sequentially. By piping the 
result to `futurize()`, we can easily enable parallel processing.

```r
library(futurize)
library(stars)

## Create a small stars object
m <- matrix(1:10000, nrow = 100, ncol = 100)
s <- st_as_stars(m)

## Calculate the mean across the first dimension
res <- st_apply(s, MARGIN = 1, FUN = mean) |> futurize()
```

When you pipe a `st_apply()` call to `futurize()`, it automatically 
configures the internal parallelization mechanism of the **stars** 
package to use the **future** framework. This ensures that the 
computation is distributed across the parallel workers defined by 
your current `plan()`.

For example, to parallelize on your local machine:

```r
plan(multisession)
```

The **futurize** package handles all the details of setting up the 
parallel environment, ensuring that global variables and packages 
are correctly exported to the workers, and that any output or 
conditions (like messages and warnings) are relayed back to your 
main R session.


# Supported Functions

The following **stars** functions are supported by `futurize()`:

* `st_apply()`


[stars]: https://cran.r-project.org/package=stars

<!--
%\VignetteIndexEntry{Parallelize 'glmmTMB' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{glmmTMB}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-glmmTMB-logo.webp" alt="The CRAN 'glmmTMB' package">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(glmmTMB)

m <- glmmTMB(count ~ mined + (1 | site), data = Salamanders, family = nbinom2)
pr <- profile(m) |> futurize()
```


# Introduction

This vignette demonstrates how to parallelize **[glmmTMB]** functions
such as `profile()` through `futurize()`.

The **[glmmTMB]** package fits generalized linear mixed models (GLMMs)
using Template Model Builder (TMB). Its `profile()` function computes
likelihood profiles for model parameters. These computations are
performed independently for each parameter, making them candidates for
parallelization.


## Example: Likelihood profile

The `profile()` function computes the likelihood profile for each
model parameter. For example, using the built-in `Salamanders` dataset
to model salamander counts:

```r
library(glmmTMB)

## Fit a negative binomial GLMM
m <- glmmTMB(count ~ mined + (1 | site), data = Salamanders, family = nbinom2)

## Compute likelihood profile
pr <- profile(m)
```

Here `profile()` is calculated sequentially. To calculate in
parallel, we can pipe to `futurize()`:

```r
library(futurize)
library(glmmTMB)

m <- glmmTMB(count ~ mined + (1 | site), data = Salamanders, family = nbinom2)
pr <- profile(m) |> futurize()
```

This will distribute the per-parameter profile computations across the
available parallel workers, given that we have set up parallel
workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other parallel
backends] to choose from, including alternatives to parallelize
locally as well as distributed across remote machines, e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **glmmTMB** functions are supported by `futurize()`:

* `profile()` for 'glmmTMB'


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `profile()`
using the **parallel** package directly, without **futurize**:

```r
library(glmmTMB)
library(parallel)

## Fit a negative binomial GLMM
m <- glmmTMB(count ~ mined + (1 | site), data = Salamanders, family = nbinom2)

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Compute likelihood profile in parallel
pr <- profile(m, parallel = "snow", ncpus = ncpus, cl = cl)

## Tear down the cluster
stopCluster(cl)
```

This requires you to manually create and manage the cluster
lifecycle. If you forget to call `stopCluster()`, or if your code
errors out before reaching it, you leak background R processes. You
also have to decide upfront how many CPUs to use and what cluster
type to use. Switching to another parallel backend, e.g. a Slurm
cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


[glmmTMB]: https://cran.r-project.org/package=glmmTMB
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'GSVA' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{Bioconductor}
%\VignetteKeyword{GSVA}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/GSVA-logo.webp" alt="The 'GSVA' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(GSVA)

param <- gsvaParam(expr, geneSets)
es <- gsva(param) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
the **[GSVA]** functions.

The **[GSVA]** Bioconductor package implements gene set variation
analysis, a non-parametric, unsupervised method for estimating
variation of gene set enrichment through the samples of an expression
data set. The main function `gsva()` computes enrichment scores for
each gene set and sample, which can be parallelized across gene sets.


## Example: Running gsva() in parallel

The `gsva()` function computes gene set enrichment scores using
different methods depending on the parameter object passed to it:

```r
library(GSVA)

# Create example data
set.seed(42)
n_genes <- 200L
n_samples <- 120L
expr <- matrix(rnorm(n_genes * n_samples), nrow = n_genes, ncol = n_samples)
rownames(expr) <- paste0("gene", seq_len(n_genes))
colnames(expr) <- paste0("sample", seq_len(n_samples))

geneSets <- list(
  geneSet1 = paste0("gene", sample(n_genes, 30L)),
  geneSet2 = paste0("gene", sample(n_genes, 50L)),
  geneSet3 = paste0("gene", sample(n_genes, 40L))
)

param <- gsvaParam(expr, geneSets)
es <- gsva(param)
```

Here `gsva()` runs sequentially, but we can easily make it run in
parallel by piping to `futurize()`:

```r
library(futurize)

es <- gsva(param) |> futurize()
```

This will distribute the work across the available parallel workers,
given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


## Other enrichment methods

GSVA supports multiple enrichment methods through different parameter
objects. All of them can be parallelized with `futurize()`:

```r
## ssGSEA method
es <- gsva(ssgseaParam(expr, geneSets)) |> futurize()

## PLAGE method
es <- gsva(plageParam(expr, geneSets)) |> futurize()

## Combined z-score method
es <- gsva(zscoreParam(expr, geneSets)) |> futurize()
```


# Supported Functions

The following **GSVA** functions are supported by `futurize()`:

* `gsva()` - requires **GSVA** (>= 2.4.2 or >= 2.5.7)
* `gsvaRanks()`
* `gsvaScores()`
* `spatCor()`


[GSVA]: https://bioconductor.org/packages/GSVA/
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'SingleCellExperiment' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{Bioconductor}
%\VignetteKeyword{SingleCellExperiment}
%\VignetteKeyword{scuttle}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/SingleCellExperiment-logo.webp" alt="The 'SingleCellExperiment' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(SingleCellExperiment)
library(scuttle)

result <- applySCE(sce, perFeatureQCMetrics) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
the **[SingleCellExperiment]** functions.

The **[SingleCellExperiment]** Bioconductor package defines the
`SingleCellExperiment` class for storing single-cell genomics data,
including alternative experiments (e.g. spike-in transcripts, antibody
tags). The `applySCE()` function applies a given function to the main
experiment and each alternative experiment, passing additional
arguments such as `BPPARAM` via `...` to enable parallelization of
the applied function.


## Example: Computing per-feature QC metrics in parallel

The `applySCE()` function applies a function across the main
experiment and its alternative experiments:

```r
library(SingleCellExperiment)
library(scuttle)

# Simulate data
sce <- mockSCE()

result <- applySCE(sce, perFeatureQCMetrics)
```

Here `applySCE()` runs `perFeatureQCMetrics()` from the **[scuttle]**
package sequentially on each experiment, but we can easily make it run
in parallel by piping to `futurize()`:

```r
library(futurize)

result <- applySCE(sce, perFeatureQCMetrics) |> futurize()
```

It is actually not `SingleCellExperiment::applySCE()` that
orchestrates the parallelization, but `scuttle::perFeatureQCMetrics()`,
which hands of the parallelization to **[BiocParallel]** which in turn
hands it of to futureverse.

The above will distribute the work across the available parallel
workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **SingleCellExperiment** functions are supported by `futurize()`:

* `applySCE()`


[BiocParallel]: https://bioconductor.org/packages/BiocParallel/
[SingleCellExperiment]: https://bioconductor.org/packages/SingleCellExperiment/
[scuttle]: https://bioconductor.org/packages/scuttle/
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'rugarch' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{rugarch}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(rugarch)

data(sp500ret, package = "rugarch")
spec <- ugarchspec()
roll <- ugarchroll(spec, sp500ret, n.start = 1000, refit.window = "moving", refit.every = 100) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[rugarch]**
functions such as `ugarchroll()`, `ugarchdistribution()`, and `ugarchboot()`.

The **[rugarch]** package provides a comprehensive set of methods for
Generalized Autoregressive Conditional Heteroskedasticity (GARCH)
modeling. Many of its functions, especially those involving
rolling estimation or bootstrapping, are computationally intensive and
benefit greatly from parallelization.


## Example: Rolling GARCH estimation

The `ugarchroll()` function performs rolling estimation and
forecasting. This can be time-consuming as it involves multiple fits
of the GARCH model.

```r
library(futurize)
plan(multisession)
library(rugarch)

data(sp500ret, package = "rugarch")
spec <- ugarchspec()

## Perform rolling estimation
roll <- ugarchroll(spec, sp500ret, n.start = 1000, 
                   refit.window = "moving", refit.every = 100) |> futurize()
```


## Example: GARCH parameter distribution

The `ugarchdistribution()` function simulates and estimates the
parameter distribution of a GARCH model.

```r
library(futurize)
plan(multisession)
library(rugarch)

data(sp500ret, package = "rugarch")
spec <- ugarchspec()
fit <- ugarchfit(spec, sp500ret)

## Estimate parameter distribution
dist <- ugarchdistribution(fit, n.sim = 100, n.hist = 10) |> futurize()
```


# Supported Functions

The following **rugarch** functions are supported by `futurize()`:

* `arfimacv()` with `seed = TRUE` as the default
* `arfimadistribution()` with `seed = TRUE` as the default
* `arfimaroll()` with `seed = TRUE` as the default
* `autoarfima()` with `seed = TRUE` as the default
* `multifilter()` with `seed = TRUE` as the default
* `multifit()` with `seed = TRUE` as the default
* `multiforecast()` with `seed = TRUE` as the default
* `ugarchboot()` with `seed = TRUE` as the default
* `ugarchdistribution()` with `seed = TRUE` as the default
* `ugarchroll()` with `seed = TRUE` as the default


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `ugarchroll()` using
the **parallel** package directly, without **futurize**:

```r
library(rugarch)
library(parallel)

data(sp500ret, package = "rugarch")
spec <- ugarchspec()

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Run rolling estimation in parallel
roll <- ugarchroll(spec, sp500ret, n.start = 1000, 
                   refit.window = "moving", refit.every = 100,
                   cluster = cl)

## Tear down the cluster
stopCluster(cl)
```

This requires you to manually create and manage the cluster
lifecycle. If you forget to call `stopCluster()`, or if your code
errors out before reaching it, you leak background R processes. You
also have to decide upfront how many CPUs to use and what cluster
type to use. Switching to another parallel backend, e.g. a Slurm
cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


[rugarch]: https://cran.r-project.org/package=rugarch

<!--
%\VignetteIndexEntry{Parallelize 'kernelshap' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{kernelshap}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/kernelshap-logo.webp" alt="The 'kernelshap' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(kernelshap)

ks <- kernelshap(
  model, X = x_explain, bg_X = bg_X
) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
**[kernelshap]** functions such as `kernelshap()` and `permshap()`.

The **[kernelshap]** package provides efficient implementations of
Kernel SHAP and permutation SHAP for explaining predictions from any
machine learning model. These functions iterate over observations to
compute Shapley value estimates, making the computation an excellent
candidate for parallelization.


## Example: Computing Kernel SHAP values in parallel

The `kernelshap()` function computes Kernel SHAP values for a set of
observations. For example, using a simple linear model:

```r
library(kernelshap)

## Fit a model
x_train <- data.frame(x1 = rnorm(100), x2 = rnorm(100))
y_train <- 2 * x_train$x1 + x_train$x2 + rnorm(100)
model <- lm(y ~ ., data = cbind(y = y_train, x_train))

## Compute Kernel SHAP values
x_explain <- x_train[1:5, ]
bg_X <- x_train[1:20, ]
ks <- kernelshap(model, X = x_explain, bg_X = bg_X)
```

Here `kernelshap()` processes observations sequentially, but we can
easily make it process them in parallel by piping to `futurize()`:

```r
library(futurize)
library(kernelshap)

ks <- kernelshap(
  model, X = x_explain, bg_X = bg_X
) |> futurize()
```

This will distribute the observation-level computations across the
available parallel workers, given that we have set up parallel
workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


## Example: Computing permutation SHAP values in parallel

The `permshap()` function works the same way:

```r
library(futurize)
library(kernelshap)

ps <- permshap(
  model, X = x_explain, bg_X = bg_X
) |> futurize()
```


# Supported Functions

The following **kernelshap** functions are supported by `futurize()`:

* `kernelshap()` with `seed = TRUE` as the default
* `permshap()` with `seed = TRUE` as the default


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `kernelshap()`
using the **parallel** and **doParallel** packages directly, without
**futurize**:

```r
library(kernelshap)
library(parallel)
library(doParallel)

## Fit a model
x_train <- data.frame(x1 = rnorm(100), x2 = rnorm(100))
y_train <- 2 * x_train$x1 + x_train$x2 + rnorm(100)
model <- lm(y ~ ., data = cbind(y = y_train, x_train))

x_explain <- x_train[1:5, ]
bg_X <- x_train[1:20, ]

## Set up a PSOCK cluster and register it with foreach
ncpus <- 4L
cl <- makeCluster(ncpus)
registerDoParallel(cl)

## Compute Kernel SHAP values in parallel via foreach
ks <- kernelshap(model, X = x_explain, bg_X = bg_X, parallel = TRUE)

## Tear down the cluster
stopCluster(cl)
registerDoSEQ()  ## reset foreach to sequential
```

This requires you to manually create a cluster, register it with
**doParallel**, and remember to tear it down and reset the
**foreach** backend when done. If you forget to call
`stopCluster()`, or if your code errors out before reaching it, you
leak background R processes. You also have to decide upfront how
many CPUs to use and what cluster type to use. Switching to another
parallel backend, e.g. a Slurm cluster, would require a completely
different setup. With **futurize**, all of this is handled for you - just pipe
to `futurize()` and control the backend with `plan()`.


[kernelshap]: https://cran.r-project.org/package=kernelshap
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'crossmap' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{crossmap}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/crossmap-logo.webp" alt="The 'crossmap' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!

# TL;DR

```r
library(futurize)
plan(multisession)
library(crossmap)

xs <- list(1:5, 1:5)
ys <- xmap(xs, ~ .y * .x) |> futurize()
```


# Introduction

The **[crossmap]** package adds to the **[purrr]**-set of functions. For example, `xmap()` can apply a function to every combination of elements in a list, e.g.

```r
library(crossmap)

# Multiply the 15 combinations of values in 1:3 and -2:2
xs <- list(1:3, -2:2)
ys <- xmap(xs, function(x, y) x * y) |> futurize()
```

Here `xmap()` evaluates sequentially over each combination of (.y, .x)
elements. The **crossmap** package provides its own future-counterpart functions, e.g. there is a `future_xmap()` that mimics `xmap()`. The `futurize()` function transpiles `xmap()` into `future_xmap()`, meaning you can do:

```r
library(futurize)

# Multiply the 15 combinations of values in 1:3 and -2:2
xs <- list(1:3, -2:2)
ys <- xmap(xs, function(x, y) x * y) |> futurize()
```

to process this `xmap()` call concurrently, which allows you to
execute it on a set of parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and it works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The `futurize()` function supports parallelization of the following **crossmap** functions:

* `imap_vec()`, `map_vec()`, `map2_vec()`, `pmap_vec()`, `xmap_vec()`
* `xmap()`
* `xmap_chr()`, `xmap_dbl()`, `xmap_int()`, `xmap_lgl()`
* `xmap_dfc()`, `xmap_dfr()`
* `xmap_mat()`, `xmap_arr()`
* `xwalk()`


[crossmap]: https://cran.r-project.org/package=crossmap
[purrr]: https://cran.r-project.org/package=purrr
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'shapr' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{shapr}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-shapr-logo.webp" alt="The 'shapr' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(shapr)

result <- explain(
  model = model,
  x_explain = x_explain,
  x_train = x_train,
  approach = "empirical",
  phi0 = mean(y_train)
) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
**[shapr]** functions such as `explain()`.

The **[shapr]** package implements dependence-aware Shapley values for
explaining predictions from machine learning models. Its `explain()`
function computes Shapley value estimates by evaluating conditional
expectations across multiple coalitions of features, making the
computation an excellent candidate for parallelization.


## Example: Computing Shapley values in parallel

The `explain()` function computes Shapley values for a set of
observations. For example, using a simple linear model:

```r
library(shapr)

## Fit a model
x_train <- data.frame(x1 = rnorm(100), x2 = rnorm(100))
y_train <- 2 * x_train$x1 + x_train$x2 + rnorm(100)
model <- lm(y_train ~ x1 + x2, data = x_train)

## Explain predictions
x_explain <- data.frame(x1 = rnorm(5), x2 = rnorm(5))
result <- explain(
  model = model,
  x_explain = x_explain,
  x_train = x_train,
  approach = "empirical",
  phi0 = mean(y_train)
)
```

Here `explain()` evaluates the coalitions sequentially, but we can
easily make it evaluate them in parallel by piping to `futurize()`:

```r
library(futurize)
library(shapr)

result <- explain(
  model = model,
  x_explain = x_explain,
  x_train = x_train,
  approach = "empirical",
  phi0 = mean(y_train)
) |> futurize()
```

This will distribute the coalition computations across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **shapr** functions are supported by `futurize()`:

* `explain()` with `seed = TRUE` as the default
* `explain_forecast()` with `seed = TRUE` as the default


[shapr]: https://cran.r-project.org/package=shapr
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'gamlss' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{gamlss}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-gamlss-logo.webp" alt="The 'gamlss' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(gamlss)

data(abdom, package = "gamlss.data")
cv <- gamlssCV(y ~ pb(x), data = abdom, K.fold = 10) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[gamlss]**
functions such as `gamlssCV()`, `add1All()`, `drop1All()`, `add1TGD()`,
and `drop1TGD()`.

The **[gamlss]** package implements Generalized Additive Models for
Location, Scale, and Shape (GAMLSS). GAMLSS models extend
traditional generalized additive models (GAMs) by allowing all
parameters of a distribution — not just the mean — to be modeled as
functions of explanatory variables. The package provides tools for
model fitting, selection, and diagnostics.

Several gamlss functions support parallel evaluation via the
`parallel`, `ncpus`, and `cl` arguments. By piping to `futurize()`,
you can leverage any future-based parallel backend for these
computations.


## Example: k-fold cross-validation

The `gamlssCV()` function performs k-fold cross-validation for model
selection. This is computationally intensive and benefits greatly
from parallelization:

```r
library(gamlss)

data(abdom, package = "gamlss.data")
cv <- gamlssCV(y ~ pb(x), data = abdom, K.fold = 10)
```

Here `gamlssCV()` evaluates each fold sequentially, but we can
easily make it evaluate in parallel by piping to `futurize()`:

```r
library(futurize)
library(gamlss)

data(abdom, package = "gamlss.data")
cv <- gamlssCV(y ~ pb(x), data = abdom, K.fold = 10) |> futurize()
```

This will distribute the cross-validation folds across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

Unlike other parallel backends in R, `futurize()` relays standard
output, messages, and warnings produced by the parallel workers back
to your main R session. For instance, when running the above, you
will see the progress output from each optimizer as it completes:

```
> cv <- gamlssCV(y ~ pb(x), data = abdom, K.fold = 10) |> futurize()
fold 1
fold 2
fold 3
fold 4
fold 5
fold 6
fold 7
fold 8
fold 9
fold 10
> 
```

This output originates from the parallel workers and is relayed to
your R session, so you get the same informative feedback as when
running sequentially.


The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


## Example: Drop terms from a model

The `drop1All()` function evaluates the effect of dropping each
term from a fitted GAMLSS model, which can be parallelized:

```r
library(futurize)
plan(multisession)
library(gamlss)

data(abdom, package = "gamlss.data")
m <- gamlss(y ~ pb(x) + x, data = abdom)
d <- drop1All(m) |> futurize()
```


# Supported Functions

The following **gamlss** functions are supported by `futurize()`:

* `add1All()`
* `add1TGD()`
* `drop1All()`
* `drop1TGD()`
* `gamlssCV()` with `seed = TRUE` as the default


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `gamlssCV()`
using the **parallel** package directly, without **futurize**:

```r
library(gamlss)
library(parallel)

data(abdom, package = "gamlss.data")

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)
clusterEvalQ(cl, library(gamlss))

## Perform k-fold cross-validation in parallel
cv <- gamlssCV(y ~ pb(x), data = abdom, K.fold = 10,
               parallel = "snow", ncpus = ncpus, cl = cl)

## Tear down the cluster
stopCluster(cl)
```

This requires you to manually create and manage the cluster
lifecycle. If you forget to call `stopCluster()`, or if your code
errors out before reaching it, you leak background R processes. You
also have to decide upfront how many CPUs to use and what cluster
type to use. Switching to another parallel backend, e.g. a Slurm
cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


[gamlss]: https://cran.r-project.org/package=gamlss
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'mgcv' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{mgcv}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-mgcv-logo.webp" alt="The 'mgcv' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(mgcv)

## Adopted from example("bam", package = "mgcv")
dat <- gamSim(1, n = 25000, dist = "normal", scale = 20)
bs <- "cr"
k <- 12

b <- bam(y ~ s(x0, bs = bs) + s(x1, bs = bs) + s(x2, bs = bs, k = k) +
             s(x3, bs = bs), data = dat) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[mgcv]**
functions such as `bam()`.

The **[mgcv]** package is one of the "recommended" packages in R. It
provides methods for fitting Generalized Additive Models (GAMs). The
`bam()` function can be used to fit GAMs for massive datasets
("Big Additive Models") with many thousands of observations, making
it an excellent candidate for parallelization.


## Example: Fitting a Big Additive Model

The `bam()` function supports parallel processing by setting up a
**parallel** cluster and passing it as argument `cluster`. This is
abstracted away by **futurize**:

```r
library(mgcv)

## Adopted from example("bam", package = "mgcv")
dat <- gamSim(1, n = 25000, dist = "normal", scale = 20)
bs <- "cr"
k <- 12

b <- bam(y ~ s(x0, bs = bs) + s(x1, bs = bs) + s(x2, bs = bs, k = k) +
             s(x3, bs = bs), data = dat) |> futurize()
```

This will distribute the calculations across the available parallel
workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **mgcv** functions are supported by `futurize()`:

* `bam()`
* `predict()` for 'bam'


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `bam()` using
the **parallel** package directly, without **futurize**:

```r
library(mgcv)
library(parallel)

## Adopted from example("bam", package = "mgcv")
dat <- gamSim(1, n = 25000, dist = "normal", scale = 20)
bs <- "cr"
k <- 12

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Fit the model in parallel
b <- bam(y ~ s(x0, bs = bs) + s(x1, bs = bs) + s(x2, bs = bs, k = k) +
             s(x3, bs = bs), data = dat, cluster = cl)

## Tear down the cluster
stopCluster(cl)
```

This requires you to manually create and manage the cluster
lifecycle. If you forget to call `stopCluster()`, or if your code
errors out before reaching it, you leak background R processes. You
also have to decide upfront how many CPUs to use and what cluster
type to use. Switching to another parallel backend, e.g. a Slurm
cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


[mgcv]: https://cran.r-project.org/package=mgcv
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'strucchange' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{strucchange}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-strucchange-logo.webp" alt="The 'strucchange' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(strucchange)

data("Nile")
bp.nile <- breakpoints(Nile ~ 1) |> futurize()
```


# Introduction

The **[strucchange]** package provides the `breakpoints()` function
for estimating one or more change points in a data trace,
e.g. in time-series data.


## Example: Finding breakpoints in time-series data

```r
library(futurize)
plan(multisession)
library(strucchange)

## UK Seatbelt data: a SARIMA(1,0,0)(1,0,0)_12 model
## (fitted by OLS) is used and reveals (at least) two
## breakpoints - one in 1973 associated with the oil crisis and
## one in 1983 due to the introduction of compulsory
## wearing of seatbelts in the UK.
data("UKDriverDeaths")

seatbelt <- log10(UKDriverDeaths)
seatbelt <- cbind(seatbelt, lag(seatbelt, k = -1), lag(seatbelt, k = -12))
colnames(seatbelt) <- c("y", "ylag1", "ylag12")
seatbelt <- window(seatbelt, start = c(1970, 1), end = c(1984, 12))
plot(seatbelt[,"y"], ylab = expression(log[10](casualties)))

## testing
re.seat <- efp(y ~ ylag1 + ylag12, data = seatbelt, type = "RE")
plot(re.seat)

## dating
bp.seat <- breakpoints(y ~ ylag1 + ylag12, data = seatbelt, h = 0.1) |> futurize()
lines(bp.seat, breaks = 2)
```

This will parallelize the dynamic programming algorithm for
computing the optimal breakpoints, given that we have set up
parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **strucchange** functions are supported by `futurize()`:

* `breakpoints()` for 'formula'


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize
`breakpoints()` using the **parallel** and **doParallel** packages
directly, without **futurize**:

```r
library(strucchange)
library(parallel)
library(doParallel)

data("Nile")

## Set up a PSOCK cluster and register it with foreach
ncpus <- 4L
cl <- makeCluster(ncpus)
registerDoParallel(cl)

## Find breakpoints in parallel via foreach
bp.nile <- breakpoints(Nile ~ 1, hpc = "foreach")

## Tear down the cluster
stopCluster(cl)
registerDoSEQ()  ## reset foreach to sequential
```

This requires you to manually create a cluster, register it with
**doParallel**, and remember to tear it down and reset the
**foreach** backend when done. If you forget to call
`stopCluster()`, or if your code errors out before reaching it, you
leak background R processes. You also have to decide upfront how
many CPUs to use and what cluster type to use. Switching to another
parallel backend, e.g. a Slurm cluster, would require a completely
different setup. With **futurize**, all of this is handled for you - just pipe
to `futurize()` and control the backend with `plan()`.


[strucchange]: https://cran.r-project.org/package=strucchange
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'caret' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{caret}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-caret-logo.webp" alt="The 'caret' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(caret)

ctrl <- trainControl(method = "cv", number = 10)
model <- train(Species ~ ., data = iris, method = "rf", trControl = ctrl) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize **[caret]**
functions such as `train()`.

The **[caret]** package provides a rich set of machine-learning tools
with a unified API. The `train()` function fits models using
cross-validation or bootstrap resampling, making it an excellent
candidate for parallelization.


## Example: Training a random forest with cross-validation

The `train()` function fits models across multiple resampling
iterations:

```r
library(caret)

## Set up 10-fold cross-validation
ctrl <- trainControl(method = "cv", number = 10)

## Train a random forest model
model <- train(Species ~ ., data = iris, method = "rf", trControl = ctrl)
```

Here `train()` evaluates sequentially, but we can easily make it
evaluate in parallel by piping to `futurize()`:

```r
library(futurize)
library(caret)

ctrl <- trainControl(method = "cv", number = 10)
model <- train(Species ~ ., data = iris, method = "rf", trControl = ctrl) |> futurize()
```

This will distribute the cross-validation folds across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **caret** functions are supported by `futurize()`:

* `bag()` with `seed = TRUE` as the default
* `gafs()` with `seed = TRUE` as the default
* `nearZeroVar()` with `seed = TRUE` as the default
* `rfe()` with `seed = TRUE` as the default
* `safs()` with `seed = TRUE` as the default
* `sbf()` with `seed = TRUE` as the default
* `train()` with `seed = TRUE` as the default


[caret]: https://cran.r-project.org/package=caret
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'scater' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{Bioconductor}
%\VignetteKeyword{scater}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/scater-logo.webp" alt="The 'scater' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(scater)

sce <- scuttle::logNormCounts(sce)
sce <- runPCA(sce) |> futurize()
sce <- runUMAP(sce) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
the **[scater]** functions.

The **[scater]** Bioconductor package provides tools for
single-cell RNA-seq data analysis, including dimensionality
reduction methods such as PCA, t-SNE, and UMAP, which can be
parallelized across cells.


## Example: Running PCA in parallel

The `runPCA()` function performs PCA on a `SingleCellExperiment`
object:

```r
library(scater)

# Simulate data
sce <- scuttle::mockSCE()
sce <- scuttle::logNormCounts(sce)

sce <- runPCA(sce)
```

Here `runPCA()` runs sequentially, but we can easily make it run in
parallel by piping to `futurize()`:

```r
library(futurize)

sce <- runPCA(sce) |> futurize()
```

This will distribute the work across the available parallel workers,
given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **scater** functions are supported by `futurize()`:

* `calculatePCA()`
* `calculateTSNE()`
* `calculateUMAP()`
* `runPCA()`
* `runTSNE()`
* `runUMAP()`
* `runColDataPCA()`
* `nexprs()`
* `getVarianceExplained()`
* `plotRLE()`


[scater]: https://bioconductor.org/packages/scater/
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize 'partykit' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{partykit}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/cran-partykit-logo.webp" alt="The 'partykit' image">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(partykit)

cf <- partykit::cforest(dist ~ speed, data = cars) |> futurize()
```


# Introduction

The **[partykit]** package provides a toolkit for recursive
partitioning.


## Example: Conditional random forests inference

Example adopted from `help("cforest", package = "partykit")`:

```r
library(futurize)
plan(multisession)
library(partykit)

## basic example: conditional inference forest for cars data
cf <- cforest(dist ~ speed, data = cars) |> futurize()

## prediction of fitted mean and visualization
nd <- data.frame(speed = 4:25)
nd$mean  <- predict(cf, newdata = nd, type = "response")
plot(dist ~ speed, data = cars)
lines(mean ~ speed, data = nd)
```

This will parallelize the computations of the variable selection
criterion, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **partykit** functions are supported by `futurize()`:

* `cforest()` with `seed = TRUE` as the default
* `ctree_control()` with `seed = TRUE` as the default
* `mob_control()` with `seed = TRUE` as the default
* `varimp()` for `cforest` with `seed = TRUE` as the default


# Without futurize: Manual PSOCK cluster setup

For comparison, here is what it takes to parallelize `cforest()`
using the **parallel** package directly, without **futurize**:

```r
library(partykit)
library(parallel)

## Set up a PSOCK cluster
ncpus <- 4L
cl <- makeCluster(ncpus)

## Fit a conditional inference forest in parallel
cf <- cforest(dist ~ speed, data = cars,
              applyfun = function(X, FUN, ...) parLapply(cl, X, FUN, ...))

## Tear down the cluster
stopCluster(cl)
```

This requires you to manually create and manage the cluster
lifecycle. If you forget to call `stopCluster()`, or if your code
errors out before reaching it, you leak background R processes. You
also have to decide upfront how many CPUs to use and what cluster
type to use. Switching to another parallel backend, e.g. a Slurm
cluster, would require a completely different setup. With
**futurize**, all of this is handled for you - just pipe to
`futurize()` and control the backend with `plan()`.


[partykit]: https://cran.r-project.org/package=partykit
[other parallel backends]: https://www.futureverse.org/backends.html

<!--
%\VignetteIndexEntry{Parallelize base-R apply functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/r-base-logo.webp" alt="The base-R logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)

slow_fcn <- function(x) {
  message("x = ", x)
  Sys.sleep(0.1)  # emulate work
  x^2
}

xs <- 1:10
ys <- lapply(xs, slow_fcn) |> futurize()
```

# Introduction

This vignette demonstrates how to use this approach to parallelize
functions such as `lapply()`, `tapply()`, `apply()`, and `replicate()`
in the **base** package, and `kernapply()` in the **stats**
package. For example, consider the base R `lapply()` function, which
is commonly used to apply a function to the elements of a vector or a
list, as in:

```r
xs <- 1:1000
ys <- lapply(xs, slow_fcn)
```

Here `lapply()` evaluates sequentially, but we can easily make it
evaluate in parallel, by using:

```r
library(futurize)
plan(multisession) ## parallelize on local machine

ys <- lapply(xs, slow_fcn) |> futurize()
#> x = 1
#> x = 2
#> x = 3
#> ...
#> x = 10
```

Note how messages produced on parallel workers are relayed as-is back
to the main R session as they complete. Not only messages, but also
warnings and other types of conditions are relayed back as-is.
Likewise, standard output produced by `cat()`, `print()`, `str()`, and
so on is relayed in the same way. This is a unique feature of
Futureverse - other parallel frameworks in R, such as **parallel**,
**foreach** with **doParallel**, and **BiocParallel**, silently drop
standard output, messages, and warnings produced on workers. With
**futurize**, your code behaves the same whether it runs sequentially
or in parallel: nothing is lost in translation.

The built-in `multisession` backend parallelizes on your local
computer and it works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


## Kernel smoothing

```r
library(futurize)
plan(multisession)

library(stats)

xs <- datasets::EuStockMarkets
k50 <- kernel("daniell", 50)
xs_smooth <- kernapply(xs, k = k50) |> futurize()
```



# Supported Functions

The `futurize()` function supports parallelization of the common base
R functions. The following **base** package functions are supported:

 * `lapply()`, `vapply()`, `sapply()`, `tapply()`
 * `mapply()`, `.mapply()`, `Map()`
 * `eapply()`
 * `apply()`
 * `replicate()` with `seed = TRUE` as the default
 * `by()`
 * `Filter()`

The `rapply()` function is not supported by `futurize()`.

The following **stats** package function is also supported:

 * `kernapply()`


# Progress Reporting via progressify

The **[progressify]** package is specially designed to work with the
Futureverse ecosystem. With **progressify**, progress can be reported
from parallelized computations in a near-live fashion. Progress
updates are propagated from the workers back to the main process,
where they are relayed to provide feedback during long-running
computations. This works because progress is signaled as R
conditions that the **future** package and most future backends
relay instantly. What is unique about **progressify** is that it
adds progress reporting to your existing code without you having
to change your code.

For example:

```r
library(futurize)
plan(multisession)
library(progressify)

xs <- 1:100
ys <- lapply(xs, slow_fcn) |> progressify() |> futurize()
```

Note how `futurize()` unwraps the expression - it descends
through `progressify()` to identify the `lapply()` call to be
futurized. Using the default progress handler, the above output and
progress reporting will appear as:

```plain
x = 1
x = 2
...
x = 20
  |=====                    |  20%
```



# Known issues

The **[BiocGenerics]** package defines generic functions `lapply()`,
`sapply()`, `mapply()`, and `tapply()`. These S4 generic functions
override the non-generic, counterpart functions in the **base**
package, which are only used as a fallback if there is no matching
method. For example, in a vanilla R session we have that both of the
following calls are identical:

```r
y_0 <- lapply(1:3, sqrt)
y_1 <- base::lapply(1:3, sqrt)
```

However, if we attach the **BiocGenerics** package, we have that the
following two calls are identical:

```r
library(BiocGenerics)
y_2 <- lapply(1:3, sqrt)
y_3 <- BiocGenerics::lapply(1:3, sqrt)
```

The reason is that `lapply()` here is no longer `base::lapply()`, but
the one defined by **BiocGenerics**, which masks the one in
**base**. We can see this with:

```r
find("lapply")
#> [1] "package:BiocGenerics" "package:base" 
```

This matters in the context of **futurize**. In a vanilla R session,

```r
y <- lapply(1:3, sqrt) |> futurize()
```

is identical to

```r
y <- base::lapply(1:3, sqrt) |> futurize()
```

However, with **BiocGenerics** attached, it is instead identical to:

```r
y <- BiocGenerics::lapply(1:3, sqrt) |> futurize()
```

which results in:

```
Error in transpilers_for_package(type = type, package = ns_name, action = "make",  : 
  There are no factory functions for creating 'futurize::add-on' transpilers for package 'BiocGenerics'
```

The solution is to specify that it is the **base** version we wish to futurize, i.e.

```r
y <- base::lapply(1:3, sqrt) |> futurize()
```


[progressify]: https://progressify.futureverse.org/
[other parallel backends]: https://www.futureverse.org/backends.html
[BiocGenerics]: https://www.bioconductor.org/packages/BiocGenerics/

<!--
%\VignetteIndexEntry{Parallelize 'GenomicAlignments' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{Bioconductor}
%\VignetteKeyword{GenomicAlignments}
%\VignetteKeyword{vignette}
%\VignetteKeyword{futurize}
%\VignetteEngine{futurize::selfonly}
-->

<div class="logos">
<img src="../man/figures/bioconductor-GenomicAlignments-logo.webp" alt="The 'GenomicAlignments' logo">
<span>+</span>
<img src="../man/figures/futurize-logo.webp" alt="The 'futurize' hexlogo">
<span>=</span>
<img src="../man/figures/future-logo.webp" alt="The 'future' logo">
</div>

The **futurize** package allows you to easily turn sequential code
into parallel code by piping the sequential code to the `futurize()`
function. Easy!


# TL;DR

```r
library(futurize)
plan(multisession)
library(GenomicAlignments)

se <- summarizeOverlaps(features, bam_files) |> futurize()
```


# Introduction

This vignette demonstrates how to use this approach to parallelize
the **[GenomicAlignments]** functions.

The **[GenomicAlignments]** Bioconductor package provides efficient
representation and manipulation of short genomic alignments. The
`summarizeOverlaps()` function counts the number of reads that map to
each feature (e.g. gene or exon) from one or more BAM files. When
called with a `BamFileList`, the work is distributed across BAM files
using `bplapply()`, which can be parallelized.


## Example: Running summarizeOverlaps() in parallel

The `summarizeOverlaps()` function counts reads overlapping genomic
features across multiple BAM files:

```r
library(GenomicAlignments)
library(Rsamtools)

bam_files <- BamFileList(c("sample1.bam", "sample2.bam", "sample3.bam"))
features <- GRanges("chr1",
  IRanges(start = c(1, 1000, 2000), end = c(500, 1500, 2500))
)

se <- summarizeOverlaps(features, bam_files)
```

Here `summarizeOverlaps()` processes BAM files sequentially, but we
can easily make it process them in parallel by piping to `futurize()`:

```r
library(futurize)

se <- summarizeOverlaps(features, bam_files) |> futurize()
```

This will distribute the BAM file processing across the available
parallel workers, given that we have set up parallel workers, e.g.

```r
plan(multisession)
```

The built-in `multisession` backend parallelizes on your local
computer and works on all operating systems. There are [other
parallel backends] to choose from, including alternatives to
parallelize locally as well as distributed across remote machines,
e.g.

```r
plan(future.mirai::mirai_multisession)
```

and

```r
plan(future.batchtools::batchtools_slurm)
```


# Supported Functions

The following **GenomicAlignments** functions are supported by `futurize()`:

* `summarizeOverlaps()`


[GenomicAlignments]: https://bioconductor.org/packages/GenomicAlignments/
[other parallel backends]: https://www.futureverse.org/backends.html

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/futurize_options.R
\name{futurize_options}
\alias{futurize_options}
\title{Options controlling resources, scheduling and evaluation of futures}
\usage{
futurize_options(
  seed = FALSE,
  globals = TRUE,
  packages = NULL,
  stdout = TRUE,
  conditions = "condition",
  scheduling = 1,
  chunk_size = NULL,
  ...
)
}
\arguments{
\item{seed}{(resource; optional) If TRUE, the random seed, that is, the
state of the random number generator (RNG) will be set such that
statistically sound random numbers are produced (also during
parallelization).
If FALSE (default), it is assumed that the future expression neither
needs nor uses random number generation.
To use a fixed random seed, specify a L'Ecuyer-CMRG seed (seven integers)
or a regular RNG seed (a single integer).  If the latter, then a
L'Ecuyer-CMRG seed will be automatically created based on the given seed.
Furthermore, if FALSE, then the future will be monitored to make sure it
does not use random numbers.  If it does and depending on the value of
option \link[future]{future.rng.onMisuse}, the check is
ignored, an informative warning, or error will be produced.
If \code{seed} is NULL, then the effect is as with \code{seed = FALSE}
but without the RNG check being performed.}

\item{globals}{(resource; optional) a logical, a character vector, or
a named list to control how globals are handled.
For details, see section 'Globals used by future expressions'
in the help for \code{\link[future:future]{future()}}.}

\item{packages}{(resource; optional) a character vector specifying
packages to be attached in the \R environment evaluating the future,
\emph{in addition to packages required by global variables} specified or
identified via argument \code{globals}.}

\item{stdout}{(evaluation) If TRUE (default), then the standard output is
captured, and re-outputted when \code{value()} is called.
If FALSE, any output is silenced (by sinking it to the null device as
it is outputted).
Using \code{stdout = structure(TRUE, drop = TRUE)} causes the captured
standard output to be dropped from the future object as soon as it has
been relayed. This can help decrease the overall memory consumed by
captured output across futures.
Using \code{stdout = NA} fully avoids intercepting the standard output;
behavior of such unhandled standard output depends on the future backend.}

\item{conditions}{(evaluation) A character string of condition classes to
be captured and relayed.  The default is to relay all conditions,
including messages and warnings.  To drop all conditions, use
\code{conditions = character(0)}.
Errors are always relayed.
Attribute \code{exclude} can be used to ignore specific classes, e.g.
\code{conditions = structure("condition", exclude = "message")} will capture
all \code{condition} classes except those that inherit from the \code{message} class.
Using \code{conditions = structure(..., drop = TRUE)} causes any captured
conditions to be dropped from the future object as soon as they have
been relayed, e.g. by \code{value(f)}. This can help decrease the overall
memory consumed by captured conditions across futures.
Using \code{conditions = NULL} (not recommended) avoids intercepting conditions,
except from errors; behavior of such unhandled conditions depends on the
future backend and the environment from which R runs.}

\item{scheduling}{(scheduling) Average number of futures ("chunks") per
worker.
If \code{0.0}, then a single future is used to process all elements.
If \code{1.0} or \code{TRUE}, then one future per worker is used.
If \code{2.0}, then each worker will process two futures
(if there are enough elements).
If \code{Inf} or \code{FALSE}, then one future per element
is used.
Only used if \code{chunk_size} is \code{NULL}.}

\item{chunk_size}{(scheduling) The average number of elements per future
("chunk").
If \code{Inf}, then all elements are processed in a single future.
If \code{NULL}, then argument \code{scheduling} is used.}

\item{\ldots}{Additional named options.}
}
\value{
A named list of future options.
Attribute \code{specified} is a character vector of future options
that were explicitly specified.
}
\description{
Some futures require specific \strong{resources} and capabilities (e.g., RNG,
packages, and globals) in order to be evaluated. These can be declared
via this function.
In addition, this function can control how futures are \strong{evaluated},
including what is collected from futures (e.g., standard output and
conditions).
It can also control how future are \strong{scheduled}, including how a set
of futures are partitioned (e.g. chunking).
}
\examples{
# Default futurize options
str(futurize_options())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/futurize_supported_tools.R
\name{futurize_supported_packages}
\alias{futurize_supported_packages}
\alias{futurize_supported_functions}
\title{List packages and functions supporting futurization}
\usage{
futurize_supported_packages()

futurize_supported_functions(package)
}
\arguments{
\item{package}{A package name.}
}
\value{
A character vector of package or function names.
\code{futurize_supported_functions()} produces an error if packages
required by the futurize transpiler are not installed.
}
\description{
List packages and functions supporting futurization
}
\examples{
pkgs <- futurize_supported_packages()
pkgs

if (requireNamespace("future.apply")) {
  fcns <- futurize_supported_functions("base")
  print(fcns)
}

if (requireNamespace("doFuture")) {
  fcns <- futurize_supported_functions("foreach")
  print(fcns)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/futurize.R
\name{futurize}
\alias{futurize}
\alias{fz}
\title{Turn common R function calls into concurrent calls for parallel evaluation}
\usage{
futurize(
  expr,
  substitute = TRUE,
  options = futurize_options(...),
  ...,
  when = TRUE,
  eval = TRUE,
  envir = parent.frame()
)
}
\arguments{
\item{expr}{An \R expression, typically a function call to futurize.
If FALSE, then futurization is disabled, and if TRUE, it is
re-enabled.}

\item{substitute}{If TRUE, argument \code{expr} is
\code{\link[base]{substitute}()}:d, otherwise not.}

\item{options, \ldots}{Named options, passed to \code{\link[=futurize_options]{futurize_options()}},
controlling how futures are resolved.}

\item{when}{If TRUE (default), the expression is futurized, otherwise not.}

\item{eval}{If TRUE (default), the futurized expression is evaluated,
otherwise it is returned.}

\item{envir}{The \link{environment} from where global objects should be
identified.}
}
\value{
Returns the value of the evaluated expression \code{expr}.

If \code{expr} is TRUE or FALSE, then a logical is returned indicating
whether futurization was previously enabled or disabled.
}
\description{
\if{html}{\figure{futurize-logo.png}{options: style='float: right;' alt='logo' width='120'}}
}
\section{Expression unwrapping}{

The transpilation mechanism includes logic to "unwrap" expressions
enclosed in constructs such as \code{!}, \code{{ }}, \verb{( )}, \code{local()}, \code{I()},
\code{identity()}, \code{invisible()}, \code{suppressMessages()}, \code{suppressWarnings()},
and \code{with()}. The transpiler descends through wrapping
constructs until it finds a transpilable expression, avoiding the
need to place \code{futurize()} inside such constructs. This allows for
patterns like:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{y <- \{
  lapply(xs, fcn)
\} |> suppressMessages() |> futurize()
}\if{html}{\out{</div>}}

avoiding having to write:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{y <- \{
  lapply(xs, fcn) |> futurize()
\} |> suppressMessages()
}\if{html}{\out{</div>}}
}

\section{Conditional futurization}{

It is possible to control whether futurization should take place at
run-time. For example,

\if{html}{\out{<div class="sourceCode r">}}\preformatted{y <- lapply(xs, fun) |> futurize(when = \{ length(xs) >= 10 \})
}\if{html}{\out{</div>}}

will be futurized, unless \code{length(xs)} is less than ten, in which case it is
evaluated as:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{y <- lapply(xs, fun)
}\if{html}{\out{</div>}}
}

\section{Disable and re-enable all futurization}{

It is possible to globally disable the effect of all \code{futurize()} calls
by calling \code{futurize(FALSE)}. The effect is as if \code{futurize()} was never
applied. For example,

\if{html}{\out{<div class="sourceCode r">}}\preformatted{futurize(FALSE)
y <- lapply(xs, fun) |> futurize()
}\if{html}{\out{</div>}}

evaluates as:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{y <- lapply(xs, fun)
}\if{html}{\out{</div>}}

To re-enable futurization, call \code{futurize(TRUE)}.
Please note that it is only the end-user that may control whether
futurization should be disabled and enabled. A package must \emph{never}
disable or enable futurization.
}

\examples{
xs <- list(1, 1:2, 1:2, 1:5)

# ------------------------------------------
# Base R apply functions
# ------------------------------------------
# Sequential lapply()
y <- lapply(X = xs, FUN = function(x) {
  sum(x)
})
   
# Parallel version
y <- lapply(X = xs, FUN = function(x) {
  sum(x)
}) |> futurize()
str(y)


# ------------------------------------------
# purrr map-reduce functions with pipes
# ------------------------------------------
if (require("purrr") && requireNamespace("furrr", quietly = TRUE)) {

# Sequential map()
y <- xs |> map(sum)
   
# Parallel version
y <- xs |> map(sum) |> futurize()
str(y)

} ## if (require ...)


# ------------------------------------------
# foreach map-reduce functions
# ------------------------------------------
if (require("foreach") && requireNamespace("doFuture", quietly = TRUE)) {

# Sequential foreach()
y <- foreach(x = xs) \%do\% {
  sum(x)
}
   
# Parallel version
y <- foreach(x = xs) \%do\% {
  sum(x)
} |> futurize()
str(y)


# Sequential times()
y <- times(3) \%do\% rnorm(1)
str(y)
   
# Parallel version
y <- times(3) \%do\% rnorm(1) |> futurize()
str(y)

} ## if (require ...)


# ------------------------------------------
# plyr map-reduce functions
# ------------------------------------------
if (require("plyr") && requireNamespace("doFuture", quietly = TRUE)) {

# Sequential llply()
y <- llply(xs, sum)
   
# Parallel version
y <- llply(xs, sum) |> futurize()
str(y)

} ## if (require ...)
}
\seealso{
To see which CRAN and Bioconductor packages are supported, use
\code{\link[=futurize_supported_packages]{futurize_supported_packages()}}.
To see which functions a specific package supports, use
\code{\link[=futurize_supported_functions]{futurize_supported_functions()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-options.R
\name{zzz-futurize.options}
\alias{zzz-futurize.options}
\alias{futurize.options}
\alias{futurize.debug}
\alias{futurize.enable}
\alias{R_FUTURIZE_DEBUG}
\alias{R_FUTURIZE_ENABLE}
\title{Options used by futurize}
\description{
Below are the \R options and environment variables that are used by the
\pkg{futurize} package and packages enhancing it.\cr
\cr
\emph{WARNING: Note that the names and the default values of these options may
change in future versions of the package.  Please use with care until
further notice.}
}
\section{Packages must not change futurize options}{


Just like for other R options, as a package developer you must \emph{not} change
any of the below \verb{futurize.*} options.  Only the end-user should set these.
If you find yourself having to tweak one of the options, make sure to
undo your changes immediately afterward.
}

\section{Options for debugging}{

\describe{
\item{\option{futurize.debug}:}{(logical) If \code{TRUE}, extensive debug messages are generated. (Default: \code{FALSE})}
}
}

\section{Options for controlling futurization}{

\describe{
\item{\option{futurize.enable}:}{(logical) If \code{TRUE} (default),
\code{futurize()} transpilation will be applied, otherwise not.}
}
}

\section{Environment variables that set R options}{

All of the above \R \option{futurize.*} options can be set by corresponding
environment variable \env{R_FUTURIZE_*} \emph{when the \pkg{futurize} package is
loaded}. This means that those environment variables must be set before
the \pkg{futurize} package is loaded in order to have an effect.
For example, if \code{R_FUTURIZE_DEBUG=true}, then option
\option{futurize.debug} is set to \code{TRUE} (logical).
}

\seealso{
To set \R options or environment variables when \R starts (even before the \pkg{futurize} package is loaded), see the \link[base]{Startup} help page.  The \href{https://cran.r-project.org/package=startup}{\pkg{startup}} package provides a friendly mechanism for configuring \R's startup process.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transpile.R
\name{transpile}
\alias{transpile}
\title{Transpile an R expression}
\usage{
transpile(
  expr,
  options = list(...),
  ...,
  when = TRUE,
  eval = TRUE,
  envir = parent.frame(),
  disable = FALSE,
  type = "built-in",
  what = "transpile",
  unwrap = list(base::`{`, base::`(`, base::`!`, base::local, base::with, base::I,
    base::identity, base::invisible, base::suppressMessages, base::suppressWarnings,
    base::suppressPackageStartupMessages),
  debug = FALSE
)
}
\arguments{
\item{expr}{An \R expression, typically a function call to transpile.
If FALSE, then the transpiler is disabled, and if TRUE, it is re-enabled.
If NA, then TRUE is returned if the transpiler is enabled, otherwise FALSE.}

\item{options}{(optional) Named options for the transpilation.}

\item{when}{If TRUE (default), the expression is transpiled, otherwise not.}

\item{eval}{If TRUE (default), the transpiled expression is evaluated,
otherwise it is returned.}

\item{envir}{The environment where the expression should be evaluated.}

\item{type}{Type of the transpiler to use.}

\item{unwrap}{(optional) A list of functions that should be considered
wrapping functions that the transpiler should unwrap ("enter"). This
allows us to transpile expressions within \code{{ ... }} and \code{local( ... )}.}
}
\value{
Returns the value of the evaluated expression \code{expr} if \code{eval = TRUE},
otherwise the transpiled expression.
If \code{expr} is NA, then TRUE is returned if the transpiler is enabled,
otherwise FALSE.
}
\description{
Transpile an R expression
}
\keyword{internal}
